/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../util/iutil.h"
#include "./ialgorithm.h"

#include <fstream>

bool IAlgorithm::isMinimalBasis(const IMonom& m, const ICoeff &c) const {
  bool r = true;
  ISetT::ConstIterator j = mSetT->begin();
  if (mSetT->coeffInterface()->type() == ICoeffInterface::GmpZZ) {
    while(j) {
      if ((*j)->lm().divisibility(m) && (*j)->poly().lc().cmpNorm(c) >= 0) {
	r = false;
	break;
      }
      ++j;
    }
  }
  else {
  /*  if (j && (*j)->lm().compare(m) > 0)
      return false;*/
    while(j /*&& (*j)->lm().compare(m) > 0*/) {
      if ((*j)->lm().divisibility(m)) {
	r = false;
	break;
      }
  /*    if ((*j)->lm().dependVar() == m.dependVar() &&  (*j)->lm().degree() > m.degree()) {
	r = false;
	break;
      }*/
      ++j;
    }
  }
  return r;
}

static int compare(const IPoly &a, const IPoly &b) {
  IPoly::ConstIterator &ia(*a.begin()),
                       &ib(*b.begin());
  int cmp = 0;
  while(ia && ib) {
    cmp = ia.monom().compare(ib.monom());
    if (cmp != 0)
      return cmp;
    ++ia;
    ++ib;
  }
  if (ia)
    cmp = 1;
  else
    cmp = -1;
  return cmp;
}

static void countsDegree(const IWrap& wrap, int var, int dimVar, int* res) {
  if (var == dimVar)
    ++*res;
  else {
    int deg = wrap.build(var);
    int d = 0;
    do {
      countsDegree(wrap, var+1, dimVar, res + d);
      ++d;
    } while(d < deg);
  }
}

void IAlgorithm::buildStatistics() {
  mStatistics.mOutOfTime = !mStatistics.mTimer.earlier_than_alarm();
  if (mDivision == NULL) {
    mStatistics.mLengthGB = 1;
    mStatistics.mLengthIB = 1;
  }
  else {
    ISetT::ConstIterator i = mSetT->begin();
    mStatistics.mLengthGB = 0;
    mStatistics.mLengthIB = 0;
    while(i) {
      if ((*i)->poly().lc().isNegative())
        (*i)->poly().minus();

      if (!(*i)->isProlong())
        ++mStatistics.mLengthGB;
      ++mStatistics.mLengthIB;

      ++i;
    }
    if (mDivision->coeffInterface()->type() != ICoeffInterface::GmpZZ) {
      if (mDivision->type() != IDivisionInterface::JanetLike) {
	IMonomInterface *monomInterface = mSetT->monomInterface();
	int dimVar = monomInterface->dimIndepend();
	mStatistics.mHilbertPoly.setDim(dimVar);
	mStatistics.mHilbertPoly.binomial(dimVar, dimVar);
	if (monomInterface->dimDepend())
	  mStatistics.mHilbertPoly.mult(monomInterface->dimDepend());
	IHilbertPoly tmp;
	tmp.setDim(dimVar);
	i = mSetT->begin();
	while(i) {
	  int m = (*i)->multi();
	  int degree = (*i)->lm().degree();
	  tmp.binomial(m - degree, m);
	  mStatistics.mHilbertPoly.sub(tmp);
	  ++i;
	}
      }
      else if (mDivision->type() == IDivisionInterface::JanetLike) {
	IMonomInterface *monomInterface = mSetT->monomInterface();
	int dimVar = monomInterface->dimIndepend();
	mStatistics.mHilbertPoly.setDim(dimVar);
	mStatistics.mHilbertPoly.binomial(dimVar, dimVar);
	if (monomInterface->dimDepend())
	  mStatistics.mHilbertPoly.mult(monomInterface->dimDepend());
	IHilbertPoly tmp;
	tmp.setDim(dimVar);
	i = mSetT->begin();
	while(i) {
	  int m = (*i)->multi();
	  int degree = (*i)->lm().degree();

	  tmp.binomial(m - degree, m);
	  mStatistics.mHilbertPoly.sub(tmp);

	  int deg = m - dimVar;
	  for(int var=0; var < dimVar; var++)
	    deg += (*i)->build(var);
	  IASSERT(deg >= 0);
	  if (deg > 0) {
	    ++deg;
	    int *res = new int[deg];
	    for(int d=0; d < deg; d++)
	      res[d] = 0;
	    countsDegree(**i, 0, dimVar, res);
	    for(int d=1; d < deg; d++) {
	      if (res[d] > 0) {
		tmp.binomial(m - degree - d, m);
		tmp.mult(res[d]);
		mStatistics.mHilbertPoly.sub(tmp);
	      }
	    }
	    delete[] res;
	  }

	  ++i;
	}
      }
    }
  }
}

IAlgorithm::IAlgorithm(ISetQ *setQ, IDivision *division, long time_bound, bool quiet, unsigned QlengthBound, unsigned MovedBound, unsigned DegreeBound):
    mStatistics(),
    mSetT(new ISetT(setQ)),
    mDivision(division),
    mQuiet(quiet),
    mQlengthBound(QlengthBound),
    mMovedBound(MovedBound),
    mDegreeBound(DegreeBound) {
  mStatistics.mTimer.setAlarm(time_bound);
}

IAlgorithm::~IAlgorithm() {
  delete mSetT;
  delete mDivision;
}

void IAlgorithm::read(std::istream &in) {
  mSetT->read(in);
  ISetT::ConstIterator j = mSetT->begin();
  while(j) {
    mDivision->insert(*j);
    ++j;
  }
  mStatistics.read(in);
}

bool IAlgorithmTQ::reduce(IWrap &wrap) {
  bool r = !wrap.isZero();
  if (r) {
    IWrap *d = mDivision->find(wrap.lm());
    if (d && wrap.coeffInterface()->type() == ICoeffInterface::GmpZZ
        && d->poly().lc().divisibilityProper(wrap.poly().lc()))
      d = NULL;
    if (d) {
      if (wrap.coeffInterface()->type() != ICoeffInterface::GmpZZ &&
          wrap.criterion(*d, *mSetT)) {
        ++mStatistics.mCriterion;
        r = false;
      }
      else {
        if (wrap.coeffInterface()->type() != ICoeffInterface::GmpZZ ||
            d->poly().lc().divisibility(wrap.poly().lc()))
          d->setEquivalent(wrap);
        wrap.updateProlong();
        ICoeff *multiplier = mDivision->wrapInterface()->coeffInterface()->create();
        IASSERT(multiplier->isZero());
        mStatistics.mReduction += mDivision->nf(*multiplier, wrap.poly(), d->poly(), *mSetT->setQ());
        delete multiplier;
        if (wrap.poly().isZero()) {
          ++mStatistics.mSpoly;
          r = false;
        }
        else {
          if (wrap.coeffInterface()->type() != ICoeffInterface::ModularShort)
            wrap.poly().pp();
          wrap.update();
        }
      }
    }
  }
  return r;
}

bool IAlgorithmTQ::reduceQ() {
  ISetQ::Iterator i(mSetT->setQ()->begin());
  int number = 0;
  while(i) {
    ISetQ::ConstIterator j(i.get());
    if (!mQuiet)
      std::cout << ++number << " " << std::flush;
    if (reduce(**j)) {
      if (mSetT->coeffInterface()->systemType()->rightPart() > 0 &&
	  (*j)->lm().dependVar() >= mSetT->coeffInterface()->systemType()->rightPart()) {
        if ((mSetT->monomInterface()->order() & IMonomInterface::DegRevLex) != 0)
          delete *j;
        // needs to be adapted for other cases
        // (e.g. problem for lex order: if *j is deleted, a not yet written out
        // prolongation of it will produce a segmentation fault later...)
      }
      else
	if (mDegreeBound == 0) {
          mSetT->setQ()->insert(j);
          break;
	}
	else {
	  (*j)->updateProlong();
	  if ((*j)->poly().degree() <= mDegreeBound) {
            mSetT->setQ()->insert(j);
	    break;
	  }
	  else {
            if ((mSetT->monomInterface()->order() & IMonomInterface::DegRevLex) != 0)
              delete *j;
            // see comment above
          }
	}
    }
    else {
      if ((mSetT->monomInterface()->order() & IMonomInterface::DegRevLex) != 0)
        delete *j;
      // see comment above
    }
  }
  if (!mQuiet)
    std::cout << std::endl;
  return !mSetT->setQ()->isEmpty();
}

/* experimental
bool IAlgorithmTQ::reduceQ() {
  bool r = false;
  ISetQ::Iterator i(mSetT->setQ()->begin());
  int number = 0;
  int pos = 0;

  if (mSetT->coeffInterface()->type() == ICoeffInterface::GmpZZ) {
    IPoly *pr = NULL;
    while(i) {
      if (reduceZZ(**i, pr)) {
	if ((mSetT->monomInterface()->order() & IMonomInterface::DegRevLexDeg) != 0) {
	  IASSERT(mDegreeBound == 0);
	  r = true;
	  int inspos = mSetT->setQ()->insert(i.get());
	  if (inspos == pos) {
	    ++i;
	    ++pos;
	  }
	  else if (inspos < pos)
	    ++pos;
          if (pr) {
	    inspos = mSetT->setQ()->insert(pr);
	    if (inspos == pos) {
	      ++i;
	      ++pos;
	    }
	    else if (inspos < pos)
	      ++pos;
	    pr = NULL;
	  }
	}
	else {
	  if (pr) {
	    mSetT->setQ()->insert(pr);
	    pr = NULL;
	  }
	  if (mDegreeBound == 0) {
	    r = true;
	    break;
	  }
	  else {
	    (*i)->updateProlong();
	    if ((*i)->poly().degree() <= mDegreeBound) {
	      r = true;
	      break;
	    }
	  }
	  ++i;
	}
      }
      else {
	if (!mQuiet)
	  std::cout << ++number << " " << std::flush;
	i.del();
      }
    }
  }
  else {
    while(i) {
      if (reduce(**i)) {
	if ((mSetT->monomInterface()->order() & IMonomInterface::DegRevLexDeg) != 0) {
	  IASSERT(mDegreeBound == 0);
	  r = true;
	  int inspos = mSetT->setQ()->insert(i.get());
	  if (inspos == pos) {
	    ++i;
	    ++pos;
	  }
	  else if (inspos < pos)
	    ++pos;
	}
	else {
	  if (mDegreeBound == 0) {
	    r = true;
	    break;
	  }
	  else {
	    (*i)->updateProlong();
	    if ((*i)->poly().degree() <= mDegreeBound) {
	      r = true;
	      break;
	    }
	  }
	  ++i;
	}
      }
      else {
	if (!mQuiet)
	  std::cout << ++number << " " << std::flush;
	i.del();
      }
    }
  }
  if (!mQuiet)
    std::cout << std::endl;
  return r;
}
*/

/* old
bool IAlgorithmTQ::reduceQ() {
  bool r = false;
  ISetQ::Iterator i(mSetT->setQ()->begin());
  int number = 0;
  int pos = 0;
  while(i) {
    if (reduce(**i)) {
      if ((mSetT->monomInterface()->order() & IMonomInterface::DegRevLexDeg) != 0) {
	IASSERT(mDegreeBound == 0);
	r = true;
	int inspos = mSetT->setQ()->insert(i.get());
	if (inspos == pos) {
	  ++i;
	  ++pos;
	}
	else if (inspos < pos)
	  ++pos;
      }
      else {
	if (mDegreeBound == 0) {
	  r = true;
	  break;
	}
	else {
	  (*i)->updateProlong();
	  if ((*i)->poly().degree() <= mDegreeBound) {
	    r = true;
	    break;
	  }
	}
        ++i;
      }
    }
    else {
      if (!mQuiet)
	std::cout << ++number << " " << std::flush;
      i.del();
    }
  }
  if (!mQuiet)
    std::cout << std::endl;
  return r;
}
*/

/* even older
bool IAlgorithmTQ::reduceQ() {
  ISetQ::Iterator i(mSetT->setQ()->begin());
  int number = 0;
  while(i)
    if (mDegreeBound == 0 ||
	(!(*i)->isZero() && (*i)->poly().degree() <= mDegreeBound)) {
      if (!mQuiet)
        std::cout << ++number << " " << std::flush;
      if (reduce(**i)) {
	r = true;
	break;
      }
      else
	i.del();
    }
    else
      ++i;
  if (!mQuiet)
    std::cout << std::endl;
  return r;
}
*/

/* old version
bool IAlgorithmTQ::reduceQ() {
  ISetQ::Iterator i(mSetT->setQ()->begin());
  while(i) {
    if (reduce(**i))
      break;
    else
      i.del();
  }
  return !mSetT->setQ()->isEmpty();
}
*/

IAlgorithmTQ::IAlgorithmTQ(ISetQ *setQ, IDivision *division, long time_bound, bool quiet, unsigned QlengthBound, unsigned MovedBound, unsigned DegreeBound):
    IAlgorithm(setQ, division, time_bound, quiet, QlengthBound, MovedBound, DegreeBound), mNumEq(0) {
}

IAlgorithmTQ::~IAlgorithmTQ() {}

void IAlgorithmTQ::build() {
  mStatistics.mTimer.cont();

  ICoeff *multiplier = mDivision->wrapInterface()->coeffInterface()->create();
  IASSERT(multiplier->isZero());

  unsigned step = 0;
  while (mStatistics.mTimer.earlier_than_alarm() && reduceQ()) {
    ++step;
    ISetQ::ConstIterator i = mSetT->setQ()->getMin(mNumEq);
    IASSERT(mDivision->find((*i)->lm()) == NULL || mDivision->coeffInterface()->type() == ICoeffInterface::GmpZZ);

    (*i)->updateProlong();
    // even if mSetT->coeffInterface()->isField() is true, call pp
    // in order to make (*i)->poly() monic!
    (*i)->poly().pp();

    int r = mDivision->nfTail(*multiplier, (*i)->poly());
    if (r) {
      mStatistics.mReduction += r;
      (*i)->poly().pp();
    }
    if ((*i)->lm().degree() == 0 && (*i)->poly().lc().isUnit() &&
        (*i)->lm().dependVar() < 0) {
      mStatistics.mTimer.stop();
      delete mDivision;
      mDivision = NULL;
      IWrapInterface* wrapInterface = mSetT->wrapInterface();
      delete mSetT;
      mSetT = new ISetT(wrapInterface);
      break;
    }
    else if (mSetT->isEmpty() ||
             (*i)->isProlong() ||
             isMinimalBasis((*i)->lm(), (*i)->poly().lc())) {
      mDivision->insert(*i, *mSetT->setQ());

      ISetT::Iterator j = mSetT->begin();
      while(j && (*j)->lm().compare((*i)->lm()) > 0) {
	r = mDivision->nfTail(*multiplier, (*j)->poly());
	if (r) {
	  mStatistics.mReduction += r;
	  (*j)->poly().pp();
	}
        ++j;
      }

      j.insert(i);
      ++j;
      while(j) {
	r = mDivision->nfTail(*multiplier, (*j)->poly());
	if (r) {
	  mStatistics.mReduction += r;
	  (*j)->poly().pp();
	}
        ++j;
      }
      IASSERTVALID(*mSetT);
    }
    else {
      ISetT::Iterator j = mSetT->begin();
      int dependVar = (*i)->lm().dependVar();
      if (mSetT->coeffInterface()->type() == ICoeffInterface::GmpZZ) {
         while(j) {
          if ((*j)->lm().dependVar() != dependVar)
            ++j;
          else {
            int cmp=(*j)->lm().compare((*i)->lm());
            if (cmp > 0 || (cmp == 0 && (*j)->poly().lc().cmpNorm((*i)->poly().lc()) > 0))
              mSetT->setQ()->insert(j.get());
            else
              ++j;
          }
        }
      }
      else {
         while(j) {
          if ((*j)->lm().dependVar() == dependVar && (*j)->lm().compare((*i)->lm()) > 0)
            mSetT->setQ()->insert(j.get());
          else
            ++j;
        }
      }
      mDivision->clear();

      j = mSetT->begin();
      while(j) {
        mDivision->insert(*j);
        ++j;
      }

      j = mSetT->begin();
      while(j) {
        mDivision->update(*j, *mSetT->setQ());
        ++j;
      }

      IASSERT(!mDivision->find((*i)->lm()) ||
        mSetT->coeffInterface()->type() == ICoeffInterface::GmpZZ);
      //if (mSetT->coeffInterface()->type() == ICoeffInterface::GmpZZ &&
      //  mDivision->find((*i)->lm(), (*i)->poly().lc()) != NULL) {
      //  mSetT->setQ()->insert(i);
      //}
      //else {
	mDivision->insert(*i, *mSetT->setQ());

	j = mSetT->begin();
	while(j && (*j)->lm().compare((*i)->lm()) > 0) {
	  r = mDivision->nfTail(*multiplier, (*j)->poly());
	  if (r) {
	    mStatistics.mReduction += r;
	    (*j)->poly().pp();
	  }
	  ++j;
	}
	j.insert(i);
	++j;
	while(j) {
	  r = mDivision->nfTail(*multiplier, (*j)->poly());
	  if (r) {
	    mStatistics.mReduction += r;
	    (*j)->poly().pp();
	  }
	  ++j;
	}
      //}
      IASSERTVALID(*mSetT);
    }

    mStatistics.mTimer.stop();
    unsigned Tlength = mSetT->length();
    unsigned Qlength = mSetT->setQ()->length();
    mStatistics.append(mStatistics.mTimer.userTime(), Tlength, Qlength, mNumEq);
    if (!mQuiet)
      fprintf(stdout, "%d %d %d %d\n", step, Tlength, Qlength, mNumEq);

    mStatistics.mTimer.cont();

    if (mQlengthBound > 0 && mMovedBound > 0 &&
      step - Tlength > mMovedBound && Qlength > mQlengthBound) {
      mStatistics.mComputationInterrupted = true;
      break;
    }

    /*
    ISetT::ConstIterator j1 = mSetT->begin();
    std::cout << "T:\n";
    while(j1) {
      std::cout << "[";
      for (int k1 = 0; k1 < (*j1)->lm().dimIndepend(); k1++) {
	if (k1 > 0) std::cout << " ";
	std::cout << (*j1)->lm()[k1];
      }
      std::cout << "]\n";

      //std::cout << "[";
      //for (int k2 = 0; k2 < (*j1)->lm().dimIndepend(); k2++) {
        //if (k2 > 0) std::cout << " ";
        //std::cout << (*j1)->ancestor()[k2];
      //}
      //std::cout << "]" << std::endl;

      std::cout << "[";
      for (int k3 = 0; k3 < (*j1)->lm().dimIndepend(); k3++) {
	if (k3 > 0) std::cout << " ";
	std::cout << (*j1)->build(k3);
      }
      std::cout << "]\n";

      std::cout << (*j1)->poly() << std::endl;
      std::cout << ",\n";
      ++j1;
    }

    ISetQ::ConstIterator j2 = mSetT->setQ()->begin();
    std::cout << "Q:\n";
    while(j2) {
      std::cout << "[";
      for (int k4 = 0; k4 < (*j2)->lm().dimIndepend(); k4++) {
	if (k4 > 0) std::cout << " ";
	std::cout << (*j2)->lm()[k4];
      }
      std::cout << "]\n";

      //std::cout << "[";
      //for (int k4 = 0; k4 < (*j2)->lm().dimIndepend(); k4++) {
	//if (k4 > 0) std::cout << " ";
	//std::cout << (*j2)->ancestor()[k4];
      //}
      //std::cout << "]" << std::endl;

      std::cout << (*j2)->poly() << std::endl;
      std::cout << ",\n";
      ++j2;
    }
    */
  }
  mStatistics.mTimer.stop();
  ISetT::ConstIterator j = mSetT->begin();
  while(j) {
    if ((*j)->poly().lc().isNegative())
      (*j)->poly().minus();
    ++j;
  }
  j = mSetT->begin();
  while(j) {
    mDivision->nfTailCanonical(*multiplier, (*j)->poly());
    ++j;
  }
  delete multiplier;
  buildStatistics();
}

bool IAlgorithmTQDegree::reduceQ() {
  ISetQ::Iterator i(mSetT->setQ()->begin());
  int deg = -1;
  ISetWrap stack;
  ISetWrap::Iterator iStack(stack.begin());
  int number = 0;
  while(i && stack.isEmpty()) {
    if (i) {
      (*i)->updateProlong();
      deg = (*i)->poly().degree();
    }
    while(i) {
      if (!mQuiet)
        std::cout << ++number << " " << std::flush;
      ISetQ::ConstIterator j(i.get());
      if (reduce(**j)) {
	if (mSetT->coeffInterface()->systemType()->rightPart() > 0 &&
	    (*j)->lm().dependVar() >= mSetT->coeffInterface()->systemType()->rightPart()) {
	  delete *j;
        }
        else
	  if (mDegreeBound == 0)
	    iStack.insert(j);
	  else {
	    (*j)->updateProlong();
	    if ((*j)->poly().degree() <= mDegreeBound) {
  /*            if (deg < (*j)->poly().degree())
		break;*/
	      iStack.insert(j);
	    }
	    else
	      delete *j;
	  }
      }
      else
	delete *j;
      if (i && !(*i)->isZero() && deg < (*i)->poly().degree())
        break;
    }
  }
  while(iStack) {
    mSetT->setQ()->insert(iStack.get());
  }
  if (!mQuiet)
    std::cout << std::endl;
  return !mSetT->setQ()->isEmpty();
}

IAlgorithmTQDegree::IAlgorithmTQDegree(ISetQ *setQ, IDivision *division, long time_bound, bool quiet, unsigned QlengthBound, unsigned MovedBound, unsigned DegreeBound):
    IAlgorithmTQ(setQ, division, time_bound, quiet, QlengthBound, MovedBound, DegreeBound) {
  if (!(mSetT->monomInterface()->order() & IMonomInterface::DegRevLex))
    IERROR("monomInterface is not DegRevLex order");
}

IAlgorithmTQDegree::~IAlgorithmTQDegree() {
}

bool IAlgorithmTQBlockHigh::reduceQBlock(ISetWrap &block) {
  IASSERT(block.isEmpty());
  ISetQ::Iterator i(mSetT->setQ()->begin());
  int deg = -1;
  int minDeg = -1;

  ISetWrap::Iterator j(block.begin());
  mNumEq = 0;
  while(i && block.isEmpty()) {
    if (i)
      deg = (*i)->poly().degree();
    while(i) {
      if (!(*i)->isZero() && deg < (*i)->poly().degree())
        break;
      if (!reduce(**i))
        i.del();
      else {
        if (minDeg < 0 ||  minDeg > (*i)->poly().degree())
          minDeg = (*i)->poly().degree();
        j.insert(i.get());
      }
    }
  }

  if (block.isEmpty())
    return false;

  ICoeff *multiplier = mDivision->wrapInterface()->coeffInterface()->create();
  IASSERT(multiplier->isZero());
  int r;
  j = block.begin();
  while(j) {
    (*j)->updateProlong();
    r = mDivision->nfTail(*multiplier, (*j)->poly());
    if (r) {
      mStatistics.mReduction += r;
      (*j)->poly().pp();
    }
    ++j;
  }
//   j = block.begin();
//   while(j) {
//     if (minDeg < (*j)->poly().degree()) {
//       mSetT->setQ()->insert(j.get());
//     }
//     else {
//       (*j)->updateProlong();
//       r = mDivision->nfTail(*multiplier, (*j)->poly());
//       if (r) {
//         mStatistics.mReduction += r
//         (*j)->poly().pp();
//       }
//       ++j;
//     }
//   }

  mNumEq = 1;
  int numEq;
  ISetWrap stack;
  ISetWrap::Iterator k(stack.begin());
  do {
    numEq = 1;
    j = block.begin();
    ISetWrap::Iterator r = j;
    int normr = (*r)->poly().norm();
    int lenr = (*r)->poly().length();
    ++j;
    while(j) {
      int cmp = (*j)->lm().compare((*r)->lm());
      if (cmp == 0)
        cmp = -compare((*j)->poly(), (*r)->poly());
      if (cmp > 0) {
        numEq = 1;
        normr = (*j)->poly().norm();
        lenr = (*j)->poly().length();
        r = j;
      }
      else if (cmp == 0) {
        ++numEq;
        int normj = (*j)->poly().norm();
        if (normj < normr) {
          normr = normj;
          lenr = (*j)->poly().length();
          r = j;
        }
        else if (normj == normr) {
          int lenj = (*j)->poly().length();
          if (lenj < lenr) {
            lenr = lenj;
            normr = (*j)->poly().norm();
            r = j;
          }
        }
      }
      ++j;
    }
    stack.begin().insert(r.get());
    if (numEq > mNumEq)
      mNumEq = numEq;

    IWrap *wrap = *stack.begin();
/*    mStatistics.mTimer.stop();
    std::cerr << wrap->lm() << std::endl;
    mStatistics.mTimer.cont();*/
    if (wrap->poly().degree() < minDeg)
      minDeg = wrap->poly().degree();

    j = block.begin();
    while(j) {
      int cmp = (*j)->lm().compare(wrap->lm());
      if (cmp != 0) {
        ++j;
      }
      else {
        if ((*j)->criterion(*wrap, *mSetT)) {
          ++mStatistics.mCriterion;
          j.del();
        }
        else {
          wrap->setEquivalent(*(*j));
          (*j)->poly().reduction(*multiplier, wrap->poly());
          ++mStatistics.mReduction;
          if ((*j)->poly().isZero()) {
            ++mStatistics.mSpoly;
            j.del();
          }
          else {
            (*j)->poly().pp();
            (*j)->update();
            ++j;
          }
        }
      }
    }
  } while(!block.isEmpty());

  k = stack.begin();
  j = block.begin();
  while(k) {
    if ((*k)->poly().degree() > minDeg)
      mSetT->setQ()->insert(k.get());
    else
      j.insert(k.get());
  }

  delete multiplier;
  return true;
}

IAlgorithmTQBlockHigh::IAlgorithmTQBlockHigh(ISetQ *setQ, IDivision *division, long time_bound, bool quiet, unsigned QlengthBound, unsigned MovedBound, unsigned DegreeBound):
    IAlgorithmTQ(setQ, division, time_bound, quiet, QlengthBound, MovedBound, DegreeBound) {
  if (!(mSetT->monomInterface()->order() & IMonomInterface::DegRevLex))
    IERROR("monomInterface is not DegRevLex order");
}

IAlgorithmTQBlockHigh::~IAlgorithmTQBlockHigh() {
}

void IAlgorithmTQBlockHigh::build() {
  mStatistics.mTimer.cont();
  unsigned step = 0;
  unsigned Tlength = 0, Qlength = 0, Blength = 0;
  int r;
  ICoeff *multiplier = mDivision->wrapInterface()->coeffInterface()->create();
  IASSERT(multiplier->isZero());

  ISetWrap block;
  while (mStatistics.mTimer.earlier_than_alarm() && reduceQBlock(block)) {
    ++step;

    mStatistics.mTimer.stop();
    Tlength = mSetT->length();
    Qlength = mSetT->setQ()->length();
    mStatistics.append(mStatistics.mTimer.userTime(), Tlength, Qlength, mNumEq);
    if (!mQuiet)
      fprintf(stdout, "%d %d %d %d\n", step, Tlength, Qlength, mNumEq);
    mStatistics.mTimer.cont();

    int path = 0;
    ISetWrap::ConstIterator i = block.begin();
    while(i) {
      if ((*i)->lm().degree() == 0 && (*i)->poly().lc().isUnit() &&
	  (*i)->lm().dependVar() < 0) {
        path = -1;
        break;
      }
      if (!mSetT->isEmpty() &&
          !(*i)->isProlong() &&
          !isMinimalBasis((*i)->lm(), (*i)->poly().lc())) {
        path = 1;
        break;
      }
      ++i;
    }

    if (path == -1) {
      mStatistics.mTimer.stop();
      delete mDivision;
      mDivision = NULL;
      IWrapInterface* wrapInterface = mSetT->wrapInterface();
      delete mSetT;
      mSetT = new ISetT(wrapInterface);
      break;
    }
    else {
      if (path == 1) {
        ISetT::Iterator j = mSetT->begin();

        int degree = (*block.begin())->poly().degree();
        while(j && (*j)->poly().degree() > degree) {
          mSetT->setQ()->insert(j.get());
        }
        mDivision->clear();

        j = mSetT->begin();
        while(j) {
          mDivision->insert(*j);
          ++j;
        }

        j = mSetT->begin();
        while(j) {
          mDivision->update(*j, *mSetT->setQ());
          ++j;
        }
      }

      i = block.begin();
      while(i) {
        IASSERT(mDivision->find((*i)->lm()) == NULL || mDivision->coeffInterface()->type() == ICoeffInterface::GmpZZ);
        mDivision->insert(*i, *mSetT->setQ());
        ++i;
      }

      i = block.begin();
      while(i) {
	r = mDivision->nfTail(*multiplier, (*i)->poly());
	if (r) {
	  mStatistics.mReduction += r;
          (*i)->poly().pp();
	}
        ++i;
      }

      ISetT::Iterator j = mSetT->begin();
      ISetWrap::Iterator k = block.begin();
      while(j && k) {
	r = mDivision->nfTail(*multiplier, (*j)->poly());
	if (r) {
	  mStatistics.mReduction += r;
          (*j)->poly().pp();
	}
        if ((*j)->lm().compare((*k)->lm()) < 0)
          j.insert(k.get());
        ++j;
      }
      while(k)
        j.insert(k.get());
    }

    if (mQlengthBound > 0 && mMovedBound > 0 &&
      step - Tlength > mMovedBound && Qlength > mQlengthBound) {
      mStatistics.mComputationInterrupted = true;
      break;
    }
  }
  mStatistics.mTimer.stop();
  ISetT::ConstIterator j = mSetT->begin();
  while(j) {
    if ((*j)->poly().lc().isNegative())
      (*j)->poly().minus();
    ++j;
  }
  j = mSetT->begin();
  while(j) {
    mDivision->nfTailCanonical(*multiplier, (*j)->poly());
    ++j;
  }
  delete multiplier;
  buildStatistics();
}

int redTail(ICoeff &multiplier, IPoly &poly, IPoly &red) {
  IASSERT(!poly.isZero());
  int r = 0;
  IPoly::Iterator &i(*poly.begin());
  ++i;
  while(i) {
    int cmp = i.monom().compare(red.lm());
    if (cmp > 0)
      ++i;
    else {
      if (cmp == 0) {
        r = 1;
        poly.redTail(multiplier, i, red);
        poly.pp();
      }
      break;
    }
  }
  return r;
}

// int redTail(ICoeff &multiplier, IPoly &poly, IPoly &red) {
//   IASSERT(!poly.isZero());
//   int r = 0;
//   IPoly::Iterator &i(*poly.begin());
//   ++i;
//   while(i) {
//     int cmp = i.monom().compare(red.lm());
//     if (cmp < 0)
//       break;
//     if (i.monom().divisibility(red.lm())) {
//       poly.redTail(multiplier, i, red);
//       ++r;
//     }
//     else
//       ++i;
//   }
//   if (r)
//     poly.pp();
//   return r;
// }

bool IAlgorithmTQBlockLow::reduceQBlock(ISetWrap &block) {
  IASSERT(block.isEmpty());
  ISetQ::Iterator i(mSetT->setQ()->begin());
  int deg = -1;
  int r;

  ISetWrap::Iterator j(block.begin());
  mNumEq = 0;
  while(i && block.isEmpty()) {
    if (i) {
      deg = (*i)->poly().degree();
    }
    while(i) {
      if (!(*i)->isZero() && deg < (*i)->poly().degree())
        break;
      if (!reduce(**i))
        i.del();
      else {
/*        if (deg < (*i)->poly().degree())
          break;*/
        j.insert(i.get());
      }
    }
  }
  if (block.isEmpty())
    return false;

  ICoeff *multiplier = mDivision->wrapInterface()->coeffInterface()->create();
  IASSERT(multiplier->isZero());
  j = block.begin();
  while(j) {
    (*j)->updateProlong();
    r = mDivision->nfTail(*multiplier, (*j)->poly());
    if (r) {
      mStatistics.mReduction += r;
      (*j)->poly().pp();
    }
    ++j;
  }
//   j = block.begin();
//   while(j) {
//     if (deg < (*j)->poly().degree()) {
//       mSetT->setQ()->insert(j.get());
//     }
//     else {
//       (*j)->updateProlong();
//       r = mDivision->nfTail(*multiplier, (*j)->poly());
//       if (r) {
//         mStatistics.mReduction += r;
//         (*j)->poly().pp();
//       }
//       ++j;
//     }
//   }

//   j = block.begin();
//   int sumMonom = 0;
//   while(j) {
//     sumMonom += (*j)->poly().length();
//     ++j;
//   }
//   std::cerr << "length() = "<< block.length() << " sumMonom = " << sumMonom << std::endl;

  mNumEq = 1;
  int numEq;
  ISetWrap stack;
  ISetWrap::Iterator k;
  do {
    numEq = 1;
    j = block.begin();
    ISetWrap::Iterator r = j;
    int normr = (*r)->poly().norm();
    int lenr = (*r)->poly().length();
    ++j;
    while(j) {
//       int cmp = (*j)->lm().compare((*r)->lm());
      int cmp = compare((*j)->poly(), (*r)->poly());
      if (cmp < 0) {
        numEq = 1;
        normr = (*j)->poly().norm();
        lenr = (*j)->poly().length();
        r = j;
      }
      else if (cmp == 0) {
        ++numEq;
        int normj = (*j)->poly().norm();
        if (normj < normr) {
          normr = normj;
          lenr = (*j)->poly().length();
          r = j;
        }
        else if (normj == normr) {
          int lenj = (*j)->poly().length();
          if (lenj < lenr) {
            lenr = lenj;
            normr = (*j)->poly().norm();
            r = j;
          }
        }
      }
      ++j;
    }
    if (numEq > mNumEq)
      mNumEq = numEq;

//     if (deg < (*r)->poly().degree()) {
//       j = block.begin();
//       while(j)
//         mSetT->setQ()->insert(j.get());
//       break;
//     }

    stack.begin().insert(r.get());
    IWrap *wrap = *stack.begin();
/*    mStatistics.mTimer.stop();
    std::cerr << wrap->lm() << std::endl;
    mStatistics.mTimer.cont();*/
    if (wrap->poly().degree() < deg)
      deg = wrap->poly().degree();

    k = stack.begin();
    ++k;
    j = block.begin();
    while(k && (*k)->lm().compare(wrap->lm()) > 0) {
      j.insert(k.get());
    }

    j = block.begin();
    while(j) {
      int cmp = (*j)->lm().compare(wrap->lm());
      if (cmp != 0) {
        if (cmp > 0)
          mStatistics.mReduction += redTail(*multiplier, (*j)->poly(), wrap->poly());
        ++j;
      }
/*    j = block.begin();
    while(j) {
      if (!(*j)->lm().divisibility(wrap->lm())) {
        if ((*j)->lm().compare(wrap->lm()) > 0)
          mStatistics.mReduction += redTail(*multiplier, (*j)->poly(), wrap->poly());
        ++j;
      }*/
      else {
        if ((*j)->criterion(*wrap, *mSetT)) {
          ++mStatistics.mCriterion;
          j.del();
        }
        else {
          wrap->setEquivalent(*(*j));
          (*j)->poly().reduction(*multiplier, wrap->poly());
          ++mStatistics.mReduction;
          if ((*j)->poly().isZero()) {
            ++mStatistics.mSpoly;
            j.del();
          }
          else {
            (*j)->poly().pp();
            (*j)->update();
            if ((*j)->poly().degree() < deg) {
              deg = (*j)->poly().degree();
            }
            ++j;
          }
        }
      }
    }
  } while(!block.isEmpty());

  k = stack.begin();
  j = block.begin();
  while(k) {
    if ((*k)->poly().degree() > deg)
      mSetT->setQ()->insert(k.get());
    else
      j.insert(k.get());
  }

  delete multiplier;
  return true;
}

IAlgorithmTQBlockLow::IAlgorithmTQBlockLow(ISetQ *setQ, IDivision *division, long time_bound, bool quiet, unsigned QlengthBound, unsigned MovedBound, unsigned DegreeBound):
    IAlgorithmTQBlockHigh(setQ, division, time_bound, quiet, QlengthBound, MovedBound, DegreeBound) {
  if (!(mSetT->monomInterface()->order() & IMonomInterface::DegRevLex))
    IERROR("monomInterface is not DegRevLex order");
}

IAlgorithmTQBlockLow::~IAlgorithmTQBlockLow() {
}

bool IAlgorithmTQBlockNorm::reduceQBlock(ISetWrap &block) {
  IASSERT(block.isEmpty());
  ISetQ::Iterator i(mSetT->setQ()->begin());
  int deg = -1;
  int r;

  ISetWrap::Iterator j(block.begin());
  mNumEq = 0;
  while(i && block.isEmpty()) {
    if (i) {
      deg = (*i)->poly().degree();
    }
    while(i) {
      if (!(*i)->isZero() && deg < (*i)->poly().degree())
        break;
      if (!reduce(**i))
        i.del();
      else {
/*        if (deg < (*i)->poly().degree())
          break;*/
        j.insert(i.get());
      }
    }
  }
  if (block.isEmpty())
    return false;

  ICoeff *multiplier = mDivision->wrapInterface()->coeffInterface()->create();
  IASSERT(multiplier->isZero());
  j = block.begin();
  while(j) {
    (*j)->updateProlong();
    r = mDivision->nfTail(*multiplier, (*j)->poly());
    if (r) {
      mStatistics.mReduction += r;
      (*j)->poly().pp();
    }
    ++j;
  }
//   j = block.begin();
//   while(j) {
//     if (deg < (*j)->poly().degree()) {
//       mSetT->setQ()->insert(j.get());
//     }
//     else {
//       (*j)->updateProlong();
//       r = mDivision->nfTail(*multiplier, (*j)->poly());
//       if (r) {
//         mStatistics.mReduction += r;
//         (*j)->poly().pp();
//       }
//       ++j;
//     }
//   }

//   j = block.begin();
//   int sumMonom = 0;
//   while(j) {
//     sumMonom += (*j)->poly().length();
//     ++j;
//   }
//   std::cerr << "length() = "<< block.length() << " sumMonom = " << sumMonom << std::endl;

  mNumEq = 1;
  int numEq;
  ISetWrap stack;
  ISetWrap::Iterator k;
  do {
    numEq = 1;
    j = block.begin();
    ISetWrap::Iterator r = j;
    int normr = (*r)->poly().norm();
    int lenr = (*r)->poly().length();
    ++j;
    while(j) {

      int normj = (*j)->poly().norm();
      if (normj < normr) {
        numEq = 1;
        normr = (*j)->poly().norm();
        lenr = (*j)->poly().length();
        r = j;
      }
      else if (normj == normr) {
        int cmp = (*j)->lm().compare((*r)->lm());
        if (cmp < 0) {
          numEq = 1;
          normr = (*j)->poly().norm();
          lenr = (*j)->poly().length();
          r = j;
        }
        else if (cmp == 0) {
          ++numEq;
          int normj = (*j)->poly().norm();
          if (normj < normr) {
            normr = normj;
            lenr = (*j)->poly().length();
            r = j;
          }
        }
      }

/*      int cmp = (*j)->lm().compare((*r)->lm());
      if (cmp < 0) {
        numEq = 1;
        normr = (*j)->poly().norm();
        lenr = (*j)->poly().length();
        r = j;
      }
      else if (cmp == 0) {
        ++numEq;
        int normj = (*j)->poly().norm();
        if (normj < normr) {
          normr = normj;
          lenr = (*j)->poly().length();
          r = j;
        }
        else if (normj == normr) {
          int lenj = (*j)->poly().length();
          if (lenj < lenr) {
            lenr = lenj;
            normr = (*j)->poly().norm();
            r = j;
          }
        }
      }*/
      ++j;
    }
    if (numEq > mNumEq)
      mNumEq = numEq;

//     if (deg < (*r)->poly().degree()) {
//       j = block.begin();
//       while(j)
//         mSetT->setQ()->insert(j.get());
//       break;
//     }

    stack.begin().insert(r.get());
    IWrap *wrap = *stack.begin();
/*    mStatistics.mTimer.stop();
    std::cerr << wrap->lm() << std::endl;
    mStatistics.mTimer.cont();*/
    if (wrap->poly().degree() < deg)
      deg = wrap->poly().degree();

//     k = stack.begin();
//     ++k;
//     while(k && (*k)->lm().compare(wrap->lm()) > 0) {
//       if ((*k)->lm().divisibility(wrap->lm()))
//         block.begin().insert(k.get());
//       else
//         ++k;
//     }

    j = block.begin();
    while(j) {
      int cmp = (*j)->lm().compare(wrap->lm());
      if (cmp != 0) {
        if (cmp > 0)
          mStatistics.mReduction += redTail(*multiplier, (*j)->poly(), wrap->poly());
        ++j;
      }
/*    j = block.begin();
    while(j) {
      if (!(*j)->lm().divisibility(wrap->lm())) {
        if ((*j)->lm().compare(wrap->lm()) > 0)
          mStatistics.mReduction += redTail(*multiplier, (*j)->poly(), wrap->poly());
        ++j;
      }*/
      else {
        if ((*j)->criterion(*wrap, *mSetT)) {
          ++mStatistics.mCriterion;
          j.del();
        }
        else {
          wrap->setEquivalent(*(*j));
          (*j)->poly().reduction(*multiplier, wrap->poly());
          ++mStatistics.mReduction;
          if ((*j)->poly().isZero()) {
            ++mStatistics.mSpoly;
            j.del();
          }
          else {
            (*j)->poly().pp();
            (*j)->update();
            if ((*j)->poly().degree() < deg) {
              deg = (*j)->poly().degree();
            }
            ++j;
          }
        }
      }
    }
  } while(!block.isEmpty());

  k = stack.begin();
  j = block.begin();
  while(k) {
    if ((*k)->poly().degree() > deg)
      mSetT->setQ()->insert(k.get());
    else
      j.insert(k.get());
  }

  delete multiplier;
  return true;
}

IAlgorithmTQBlockNorm::IAlgorithmTQBlockNorm(ISetQ *setQ, IDivision *division, long time_bound, bool quiet, unsigned QlengthBound, unsigned MovedBound, unsigned DegreeBound):
    IAlgorithmTQBlockHigh(setQ, division, time_bound, quiet, QlengthBound, MovedBound, DegreeBound) {
  if (!(mSetT->monomInterface()->order() & IMonomInterface::DegRevLex))
    IERROR("monomInterface is not DegRevLex order");
}

IAlgorithmTQBlockNorm::~IAlgorithmTQBlockNorm() {
}
