/***************************************************************************
 *   Copyright (C) 2006 by Daniel Robertz                                  *
 *   daniel@momo.math.rwth-aachen.de                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./ialgext.h"

IAlgExtInterface::IAlgExtInterface(IVariables *parameter, const IOneParameterPoly &minPoly):
    ICoeffInterface(),
    mMinPoly(minPoly) {
  IASSERT(parameter);
  mParameters = parameter;

  mType = ICoeffInterface::AlgebraicFieldExtension;
}

ICoeff* IAlgExtInterface::create() {
  return new IAlgExt(this);
}

ICoeff* IAlgExtInterface::copy(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  return new IAlgExt(a);
}

IAlgExt::IAlgExt(IAlgExtInterface* r):
   ICoeff(r),
   mPoly(r->mMinPoly.coeffInterface()) {
}

IAlgExt::IAlgExt(const ICoeff &a):
   ICoeff(a.coeffInterface()),
   mPoly(((const IAlgExt&)a).mPoly) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  IASSERTVALID(*this);
}

IAlgExt::~IAlgExt() {
}

void IAlgExt::set(const ICoeff& a) {
//std::cout << "IAlgExt::set (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  mPoly = ((const IAlgExt&)a).mPoly;
  IASSERTVALID(*this);
//std::cout << "IAlgExt::set (out)" << std::endl;
}

void IAlgExt::setZero() {
//std::cout << "IAlgExt::setZero (in)" << std::endl;
  mPoly.setZero();
//std::cout << "IAlgExt::setZero (out)" << std::endl;
}

void IAlgExt::setOne() {
//std::cout << "IAlgExt::setOne (in)" << std::endl;
  mPoly.setOne();
//std::cout << "IAlgExt::setOne (out)" << std::endl;
}

bool IAlgExt::isNegative() const {
//std::cout << "IAlgExt::isNegative (in)" << std::endl;
  return mPoly.isNegative();
//std::cout << "IAlgExt::isNegative (out)" << std::endl;
}

bool IAlgExt::isZero() const {
//std::cout << "IAlgExt::isZero (in)" << std::endl;
  return mPoly.isZero();
//std::cout << "IAlgExt::isZero (out)" << std::endl;
}

bool IAlgExt::isOne() const {
//std::cout << "IAlgExt::isOne (in)" << std::endl;
  return mPoly.isOne();
//std::cout << "IAlgExt::isOne (out)" << std::endl;
}

bool IAlgExt::isComposite() const {
//std::cout << "IAlgExt::isComposite (in)" << std::endl;
  return mPoly.isComposite();
//std::cout << "IAlgExt::isComposite (out)" << std::endl;
}

bool IAlgExt::isUnit() const {
//std::cout << "IAlgExt::isUnit (in)" << std::endl;
  // expecting mMinPoly to be irreducible
  return !mPoly.isZero();
//std::cout << "IAlgExt::isUnit (out)" << std::endl;
}

int IAlgExt::norm() const {
//std::cout << "IAlgExt::norm (in)" << std::endl;
  int r = 0;
  if (!mPoly.isZero())
    r = mPoly.degree();
  return r;
//std::cout << "IAlgExt::norm (out)" << std::endl;
}

void IAlgExt::minus() {
//std::cout << "IAlgExt::minus (in)" << std::endl;
  mPoly.minus();
//std::cout << "IAlgExt::minus (out)" << std::endl;
}

void IAlgExt::abs() {
}

void IAlgExt::add(const ICoeff& a) {
//std::cout << "IAlgExt::add (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  mPoly.add(((const IAlgExt&)a).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
//std::cout << "IAlgExt::add (out)" << std::endl;
}

void IAlgExt::sub(const ICoeff& a) {
//std::cout << "IAlgExt::sub (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  mPoly.sub(((const IAlgExt&)a).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
//std::cout << "IAlgExt::sub (out)" << std::endl;
}

void IAlgExt::mult(const ICoeff& a) {
//std::cout << "IAlgExt::mult (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  mPoly.mult(((const IAlgExt&)a).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
//std::cout << "IAlgExt::mult (out)" << std::endl;
}

void IAlgExt::diff(int , unsigned ) {
  mPoly.setZero();
}

void IAlgExt::shift(int , unsigned ) {
}

void IAlgExt::pow(unsigned deg) {
//std::cout << "IAlgExt::pow (in)" << std::endl;
  mPoly.pow(deg);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
//std::cout << "IAlgExt::pow (out)" << std::endl;
}

void IAlgExt::inverse() {
//std::cout << "IAlgExt::inverse (in)" << std::endl;
  IASSERT(!isZero());
  IASSERT(!minPoly().degree() == 0);
  IASSERT(mPoly.degree() <= minPoly().degree());

  if (mPoly.degree() == 0) {
    IOneParameterPoly::ConstIterator i(mPoly.begin());
    (*i).mCoeff->inverse();
  }
  else {
    IOneParameterPoly tmp(minPoly());
    IOneParameterPoly c1(minPoly().coeffInterface()), c2(minPoly().coeffInterface());
    c2.setOne();

    do {
      tmp.rem_ext(mPoly, c1, c2);

      if (tmp.isZero())
        IERROR("expecting irreducible minimal polynomial");
      mPoly.swap(tmp);
      c1.swap(c2);
    } while (mPoly.degree() > 0);
    c2.div(*(*mPoly.begin()).mCoeff);
    mPoly.swap(c2);
  }
//std::cout << "IAlgExt::inverse (out)" << std::endl;
}

void IAlgExt::div(const ICoeff& a) {
//std::cout << "IAlgExt::div (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  ICoeff* tmp = coeffInterface()->copy(a);
  tmp->inverse();
  mPoly.mult(((const IAlgExt&)*tmp).mPoly);
  delete tmp;
  IASSERTVALID(*this);
//std::cout << "IAlgExt::div (out)" << std::endl;
}

void IAlgExt::div(const ICoeff& a, const ICoeff& b) {
//std::cout << "IAlgExt::div (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  IASSERT(b.type() == ICoeffInterface::AlgebraicFieldExtension);
  ICoeff* tmp = coeffInterface()->copy(b);
  tmp->inverse();
  mPoly.mult(((const IAlgExt&)a).mPoly, ((const IAlgExt&)*tmp).mPoly);
  delete tmp;
  IASSERTVALID(*this);
//std::cout << "IAlgExt::div (out)" << std::endl;
}

void IAlgExt::gcd(const ICoeff& a) {
//std::cout << "IAlgExt::gcd (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  mPoly.gcd(((const IAlgExt&)a).mPoly);
  IASSERTVALID(*this);
//std::cout << "IAlgExt::gcd (out)" << std::endl;
}

void IAlgExt::gcd(const ICoeff& a, const ICoeff& b) {
//std::cout << "IAlgExt::gcd (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  IASSERT(b.type() == ICoeffInterface::AlgebraicFieldExtension);
  mPoly.gcd(((const IAlgExt&)a).mPoly, ((const IAlgExt&)b).mPoly);
  IASSERTVALID(*this);
//std::cout << "IAlgExt::gcd (out)" << std::endl;
}

void IAlgExt::addmul(const ICoeff& a, const ICoeff& b) {
//std::cout << "IAlgExt::addmul (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtension);
  IASSERT(b.type() == ICoeffInterface::AlgebraicFieldExtension);
  mPoly.addmul(((const IAlgExt&)a).mPoly, ((const IAlgExt&)b).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
//std::cout << "IAlgExt::addmul (out)" << std::endl;
}

void IAlgExt::read(std::istream &in) {
  mPoly.read(in);
  IASSERTVALID(*this);
}

void IAlgExt::write(std::ostream &out) const {
  mPoly.write(out);
}

void IAlgExt::input(std::istream& in, bool expanded) {
  std::streampos posbeg = in.tellg();
  int var = parameters()->read(in);
  if (var == 0) {
    mPoly.setOne();
    mPoly.mult(1);
    IASSERTVALID(*this);
  }
  else {
    ICoeff* mp = mPoly.coeffInterface()->create();
    in >> *mp;
    if (in.fail()) {
      in.clear();
      in.setstate(std::ios::failbit);
      in.seekg(posbeg);
    }
    else {
      mPoly.setZero();
      if (!mp->isZero()) {
        mPoly.setOne();
        mPoly.mult(*mp);
      }
      IASSERTVALID(*this);
    }
    delete mp;
  }
}

void IAlgExt::output(std::ostream& out) const {
  mPoly.output(out, parameters()->variable(0));
}

void IAlgExt::assertValid(const char* fileName, int fileLine) const {
  mPoly.assertValid(fileName, fileLine);
}

IAllocator IAlgExt::sAllocator(sizeof(IAlgExt));
