/***************************************************************************
 *   Copyright (C) 2006 by Daniel Robertz                                  *
 *   daniel@momo.math.rwth-aachen.de                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./ialgextgmpq.h"

IAlgExtGmpQInterface::IAlgExtGmpQInterface(IVariables *parameter, const IAlgExtGmpQAux &minPoly):
    ICoeffInterface(),
    mMinPoly(minPoly) {
  IASSERT(parameter);
  mParameters = parameter;

  mType = ICoeffInterface::AlgebraicFieldExtensionGmpQ;
}

ICoeff* IAlgExtGmpQInterface::create() {
  return new IAlgExtGmpQ(this);
}

ICoeff* IAlgExtGmpQInterface::copy(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  return new IAlgExtGmpQ(a);
}

IAlgExtGmpQ::IAlgExtGmpQ(IAlgExtGmpQInterface* r):
   ICoeff(r),
   mPoly() {
}

IAlgExtGmpQ::IAlgExtGmpQ(const ICoeff &a):
   ICoeff(a.coeffInterface()),
   mPoly(((const IAlgExtGmpQ&)a).mPoly) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  IASSERTVALID(*this);
}

IAlgExtGmpQ::~IAlgExtGmpQ() {
}

void IAlgExtGmpQ::set(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  mPoly = ((const IAlgExtGmpQ&)a).mPoly;
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::setZero() {
  mPoly.setZero();
}

void IAlgExtGmpQ::setOne() {
  mPoly.setOne();
}

bool IAlgExtGmpQ::isNegative() const {
  return mPoly.isNegative();
}

bool IAlgExtGmpQ::isZero() const {
  return mPoly.isZero();
}

bool IAlgExtGmpQ::isOne() const {
  return mPoly.isOne();
}

bool IAlgExtGmpQ::isComposite() const {
  return mPoly.isComposite();
}

bool IAlgExtGmpQ::isUnit() const {
  // expecting mMinPoly to be irreducible
  return !mPoly.isZero();
}

int IAlgExtGmpQ::norm() const {
  int r = 0;
  if (!mPoly.isZero())
    r = mPoly.degree();
  return r;
}

void IAlgExtGmpQ::minus() {
  mPoly.minus();
}

void IAlgExtGmpQ::abs() {
}

void IAlgExtGmpQ::add(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  mPoly.add(((const IAlgExtGmpQ&)a).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::sub(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  mPoly.sub(((const IAlgExtGmpQ&)a).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::mult(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  mPoly.mult(((const IAlgExtGmpQ&)a).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::diff(int , unsigned ) {
  mPoly.setZero();
}

void IAlgExtGmpQ::shift(int , unsigned ) {
}

void IAlgExtGmpQ::pow(unsigned deg) {
  mPoly.pow(deg);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::inverse() {
  IASSERT(!isZero());
  IASSERT(!minPoly().degree() == 0);
  IASSERT(mPoly.degree() <= minPoly().degree());

  if (mPoly.degree() == 0) {
    IAlgExtGmpQAux::ConstIterator i(mPoly.mHead.begin());
    mpq_inv(&(*i).mCoeff, &(*i).mCoeff);
  }
  else {
    IAlgExtGmpQAux tmp(minPoly());
    IAlgExtGmpQAux c1, c2;
    c2.setOne();

    do {
      tmp.rem_ext(mPoly, c1, c2);
      if (tmp.isZero())
        IERROR("expecting irreducible minimal polynomial");
      mPoly.mHead.swap(tmp.mHead);
      c1.mHead.swap(c2.mHead);
    } while (mPoly.degree() > 0);
    c2.div((*mPoly.mHead.begin()).mCoeff);
    mPoly.mHead.swap(c2.mHead);
  }
}

void IAlgExtGmpQ::div(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);

  ICoeff* tmp = coeffInterface()->copy(a);
  tmp->inverse();

  mPoly.mult(((const IAlgExtGmpQ&)*tmp).mPoly);
  delete tmp;
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::div(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  IASSERT(b.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);

  ICoeff* tmp = coeffInterface()->copy(b);
  tmp->inverse();
  mPoly.mult(((const IAlgExtGmpQ&)a).mPoly, ((const IAlgExtGmpQ&)*tmp).mPoly);
  delete tmp;
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::gcd(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  mPoly.gcd(((const IAlgExtGmpQ&)a).mPoly);
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::gcd(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  IASSERT(b.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  mPoly.gcd(((const IAlgExtGmpQ&)a).mPoly, ((const IAlgExtGmpQ&)b).mPoly);
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::addmul(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  IASSERT(b.type() == ICoeffInterface::AlgebraicFieldExtensionGmpQ);
  mPoly.addmul(((const IAlgExtGmpQ&)a).mPoly, ((const IAlgExtGmpQ&)b).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::read(std::istream &in) {
  mPoly.read(in);
  IASSERTVALID(*this);
}

void IAlgExtGmpQ::write(std::ostream &out) const {
  mPoly.write(out);
}

void IAlgExtGmpQ::input(std::istream& in, bool expanded) {
  std::streampos posbeg = in.tellg();
  int var = parameters()->read(in);
  if (var == 0) {
    mPoly.setOne();
    (*mPoly.begin()).mDegree = 1;
    IASSERTVALID(*this);
  }
  else {
    MP_RAT mp;
    mpq_init(&mp);
    in >> (&mp);
    mpq_canonicalize(&mp);
    if (in.fail()) {
      in.clear();
      in.setstate(std::ios::failbit);
      in.seekg(posbeg);
    }
    else {
      IAlgExtGmpQAux::Iterator i(mPoly.begin());
      i.clear();
      if (mpq_sgn(&mp) != 0) {
        i.insert();
        mpq_set(&(*i).mCoeff, &mp);
      }
      IASSERTVALID(*this);
    }
    mpq_clear(&mp);
  }
}

void IAlgExtGmpQ::output(std::ostream& out) const {
  if (isZero())
    out << '0';
  else {
    MP_RAT mp;
    mpq_init(&mp);
    IAlgExtGmpQAux::ConstIterator i(mPoly.begin());
    if (mpq_cmp_ui(&(*i).mCoeff, 1ul, 1ul) == 0)
      if ((*i).mDegree == 0)
        out << '1';
      else if ((*i).mDegree == 1)
        out << parameters()->variable(0);
      else
        out << parameters()->variable(0) << '^' << (*i).mDegree;
    else
      if ((*i).mDegree == 0)
        out << (&(*i).mCoeff);
      else if ((*i).mDegree == 1)
        out << (&(*i).mCoeff) << '*' << parameters()->variable(0);
      else
        out << (&(*i).mCoeff) << '*' << parameters()->variable(0) << '^' << (*i).mDegree;
    ++i;
    while(i) {
      if (mpq_sgn(&(*i).mCoeff) > 0) {
        out << " + ";
        if (mpq_cmp_ui(&(*i).mCoeff, 1ul, 1ul) == 0)
          if ((*i).mDegree == 0)
            out << '1';
          else if ((*i).mDegree == 1)
            out << parameters()->variable(0);
          else
            out << parameters()->variable(0) << '^' << (*i).mDegree;
        else
          if ((*i).mDegree == 0)
            out << (&(*i).mCoeff);
          else if ((*i).mDegree == 1)
            out << (&(*i).mCoeff) << '*' << parameters()->variable(0);
          else
            out << (&(*i).mCoeff) << '*' << parameters()->variable(0) << '^' << (*i).mDegree;
      }
      else {
        out << " - ";
        if (mpq_cmp_si(&(*i).mCoeff, -1, 1ul) == 0)
          if ((*i).mDegree == 0)
            out << '1';
          else if ((*i).mDegree == 1)
            out << parameters()->variable(0);
          else
            out << parameters()->variable(0) << '^' << (*i).mDegree;
        else {
          mpq_neg(&mp, &(*i).mCoeff);
          if ((*i).mDegree == 0)
            out << (&mp);
          else if ((*i).mDegree == 1)
            out << (&mp) << '*' << parameters()->variable(0);
          else
            out << (&mp) << '*' << parameters()->variable(0) << '^' << (*i).mDegree;
        }
      }
      ++i;
    }
    mpq_clear(&mp);
  }
}

void IAlgExtGmpQ::assertValid(const char* fileName, int fileLine) const {
  mPoly.assertValid(fileName, fileLine);
}

IAllocator IAlgExtGmpQ::sAllocator(sizeof(IAlgExtGmpQ));
