/***************************************************************************
 *   Copyright (C) 2006 by Daniel Robertz                                  *
 *   daniel@momo.math.rwth-aachen.de                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IALGEXTGMPQAUX_H
#define IALGEXTGMPQAUX_H

#include <iostream>
#include <gmp.h>

#include "../util/iutil.h"
#include "../util/ilist.h"
#include "../util/ivariables.h"

class IAlgExtGmpQ;

class IAlgExtGmpQAux {
  friend class IAlgExtGmpQ;

public:
  struct Term {
    unsigned  mDegree;
    MP_RAT  mCoeff;

    Term(): mDegree(0) { mpq_init(&mCoeff); }
    Term(const Term& a): mDegree(a.mDegree) {
      mpq_init(&mCoeff);
      mpq_set(&mCoeff, &a.mCoeff);
    }
    ~Term() { mpq_clear(&mCoeff); }

    void operator=(const Term& a) {
      mDegree = a.mDegree;
      mpq_set(&mCoeff, &a.mCoeff);
    }
  };

protected:
  IList<Term>  mHead;

  void add(IAlgExtGmpQAux& a);
  void sub(IAlgExtGmpQAux& a);
  void mult(const MP_RAT& coeff);
  void mult(unsigned degree, const MP_RAT& coeff);
  void div(const MP_RAT& coeff);
  void div(unsigned degree,  const MP_RAT& coeff);
  void rem(const IAlgExtGmpQAux& a);
  void rem_ext(const IAlgExtGmpQAux& a, IAlgExtGmpQAux& c1, IAlgExtGmpQAux& c2);

  void equality(std::istream& in, const IVariables &parameters);
  void additive(std::istream& in, const IVariables &parameters);
  void multiplicative(std::istream& in, const IVariables &parameters);
  void unary(std::istream& in, const IVariables &parameters);
  void power(std::istream& in, const IVariables &parameters);
  void bracket(std::istream& in, const IVariables &parameters);

public:
  typedef IList<Term>::Iterator  Iterator;
  typedef IList<Term>::ConstIterator  ConstIterator;
  IAlgExtGmpQAux():
      mHead() {
  }
  IAlgExtGmpQAux(const IAlgExtGmpQAux& a):
      mHead(a.mHead) {
  }

  ~IAlgExtGmpQAux() {}

  void operator=(const IAlgExtGmpQAux& a) {
    mHead = a.mHead;
  }
  void setZero() { mHead.clear(); }
  void setOne() {
    mHead.clear();
    Iterator i(mHead.begin());
    i.insert();
    mpq_set_ui(&(*i).mCoeff, 1ul, 1ul);
  }

  bool isZero() const { return mHead.isEmpty(); }
  bool isOne() const {
    ConstIterator i(mHead.begin());
    return i && (*i).mDegree == 0 && mpq_cmp_ui(&(*i).mCoeff, 1ul, 1ul) == 0;
  }
  bool isNegative() const {
    ConstIterator i(mHead.begin());
    return i && mpq_sgn(&(*i).mCoeff) < 0;
  }
  bool isComposite() const {
    ConstIterator i(mHead.begin());
    return mHead.length() > 1 || (i && mpq_sgn(&(*i).mCoeff) < 0);
  }
  unsigned degree() const {
    IASSERT(!isZero());
    return (*mHead.begin()).mDegree;
  }

  void minus();
  void add(const IAlgExtGmpQAux& a);
  void sub(const IAlgExtGmpQAux& a);
  void mult(const IAlgExtGmpQAux& a);
  void mult(const IAlgExtGmpQAux& a, const IAlgExtGmpQAux& b);
  void pow(unsigned deg);
  void div(const IAlgExtGmpQAux& a);
  void div(const IAlgExtGmpQAux& a, const IAlgExtGmpQAux& b) {
    *this = a;
    div(b);
  }
  void gcd(const IAlgExtGmpQAux& a);
  void gcd(const IAlgExtGmpQAux& a, const IAlgExtGmpQAux& b) {
    *this = a;
    gcd(b);
  }
  void addmul(const IAlgExtGmpQAux& a, const IAlgExtGmpQAux& b) {
    IAlgExtGmpQAux tmp;
    tmp.mult(a, b);
    add(tmp);
  }

  ConstIterator begin() const { return mHead.begin(); }
  Iterator begin() { return mHead.begin(); }

  void read(std::istream &in);
  void write(std::ostream &out) const;

  void input(std::istream& in, const IVariables &parameters);
  friend std::ostream& operator<<(std::ostream& out, const IAlgExtGmpQAux& a);

  void assertValid(const char* fileName, int fileLine) const;
};

#endif // IALGEXTGMPQAUX_H
