/***************************************************************************
 *   Copyright (C) 2006 by Daniel Robertz                                  *
 *   daniel@momo.math.rwth-aachen.de                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./ialgextmodularshort.h"
#include "../util/imodular.h"

IAlgExtModularShortInterface::IAlgExtModularShortInterface(IVariables *parameter, const IAlgExtModularShortAux &minPoly):
    ICoeffInterface(),
    mMinPoly(minPoly) {
  IASSERT(parameter);
  mParameters = parameter;

  mType = ICoeffInterface::AlgebraicFieldExtensionModularShort;
}

ICoeff* IAlgExtModularShortInterface::create() {
  return new IAlgExtModularShort(this);
}

ICoeff* IAlgExtModularShortInterface::copy(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  return new IAlgExtModularShort(a);
}

IAlgExtModularShort::IAlgExtModularShort(IAlgExtModularShortInterface* r):
   ICoeff(r),
   mPoly(r->mMinPoly.mPrime) {
}

IAlgExtModularShort::IAlgExtModularShort(const ICoeff &a):
   ICoeff(a.coeffInterface()),
   mPoly(((const IAlgExtModularShort&)a).mPoly) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  IASSERTVALID(*this);
}

IAlgExtModularShort::~IAlgExtModularShort() {
}

void IAlgExtModularShort::set(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  mPoly = ((const IAlgExtModularShort&)a).mPoly;
  IASSERTVALID(*this);
}

void IAlgExtModularShort::setZero() {
  mPoly.setZero();
}

void IAlgExtModularShort::setOne() {
  mPoly.setOne();
}

bool IAlgExtModularShort::isNegative() const {
  return mPoly.isNegative();
}

bool IAlgExtModularShort::isZero() const {
  return mPoly.isZero();
}

bool IAlgExtModularShort::isOne() const {
  return mPoly.isOne();
}

bool IAlgExtModularShort::isComposite() const {
  return mPoly.isComposite();
}

bool IAlgExtModularShort::isUnit() const {
  // expecting mMinPoly to be irreducible
  return !mPoly.isZero();
}

int IAlgExtModularShort::norm() const {
  int r = 0;
  if (!mPoly.isZero())
    r = mPoly.degree();
  return r;
}

void IAlgExtModularShort::minus() {
  mPoly.minus();
}

void IAlgExtModularShort::abs() {
}

void IAlgExtModularShort::add(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  mPoly.add(((const IAlgExtModularShort&)a).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtModularShort::sub(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  mPoly.sub(((const IAlgExtModularShort&)a).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtModularShort::mult(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  mPoly.mult(((const IAlgExtModularShort&)a).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtModularShort::diff(int , unsigned ) {
  mPoly.setZero();
}

void IAlgExtModularShort::shift(int , unsigned ) {
}

void IAlgExtModularShort::pow(unsigned deg) {
  mPoly.pow(deg);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtModularShort::inverse() {
  IASSERT(!isZero());
  IASSERT(!minPoly().degree() == 0);
  IASSERT(mPoly.degree() <= minPoly().degree());

  if (mPoly.degree() == 0) {
    IAlgExtModularShortAux::ConstIterator i(mPoly.mHead.begin());
    (*i).mCoeff = iInverse((*i).mCoeff, mPoly.mPrime);
  }
  else {
    IAlgExtModularShortAux tmp(minPoly());
    IAlgExtModularShortAux c1(mPoly.mPrime), c2(mPoly.mPrime);
    c2.setOne();

    do {
      tmp.rem_ext(mPoly, c1, c2);
      if (tmp.isZero())
        IERROR("expecting irreducible minimal polynomial");
      mPoly.mHead.swap(tmp.mHead);
      c1.mHead.swap(c2.mHead);
    } while (mPoly.degree() > 0);
    c2.div((*mPoly.mHead.begin()).mCoeff);
    mPoly.mHead.swap(c2.mHead);
  }
}

void IAlgExtModularShort::div(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);

  ICoeff* tmp = coeffInterface()->copy(a);
  tmp->inverse();
  mPoly.mult(((const IAlgExtModularShort&)*tmp).mPoly);
  delete tmp;
  IASSERTVALID(*this);
}

void IAlgExtModularShort::div(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  IASSERT(b.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);

  ICoeff* tmp = coeffInterface()->copy(b);
  tmp->inverse();
  mPoly.mult(((const IAlgExtModularShort&)a).mPoly, ((const IAlgExtModularShort&)*tmp).mPoly);
  delete tmp;
  IASSERTVALID(*this);
}

void IAlgExtModularShort::gcd(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  mPoly.gcd(((const IAlgExtModularShort&)a).mPoly);
  IASSERTVALID(*this);
}

void IAlgExtModularShort::gcd(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  IASSERT(b.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  mPoly.gcd(((const IAlgExtModularShort&)a).mPoly, ((const IAlgExtModularShort&)b).mPoly);
  IASSERTVALID(*this);
}

void IAlgExtModularShort::addmul(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  IASSERT(b.type() == ICoeffInterface::AlgebraicFieldExtensionModularShort);
  mPoly.addmul(((const IAlgExtModularShort&)a).mPoly, ((const IAlgExtModularShort&)b).mPoly);
  mPoly.rem(minPoly());
  IASSERTVALID(*this);
}

void IAlgExtModularShort::read(std::istream &in) {
  mPoly.read(in);
  IASSERTVALID(*this);
}

void IAlgExtModularShort::write(std::ostream &out) const {
  mPoly.write(out);
}

void IAlgExtModularShort::input(std::istream& in, bool expanded) {
  std::streampos posbeg = in.tellg();
  int var = parameters()->read(in);
  if (var == 0) {
    mPoly.setOne();
    (*mPoly.begin()).mDegree = 1;
    IASSERTVALID(*this);
  }
  else {
    int mp;
    in >> mp;
    if (mp < 0) {
      int factor = -(mp / mPoly.mPrime) + 1;
      mp += factor * mPoly.mPrime;
    }
    mp %= mPoly.mPrime;
    if (in.fail()) {
      in.clear();
      in.setstate(std::ios::failbit);
      in.seekg(posbeg);
    }
    else {
      IAlgExtModularShortAux::Iterator i(mPoly.begin());
      i.clear();
      if (mp != 0) {
        i.insert();
        (*i).mCoeff = mp;
      }
      IASSERTVALID(*this);
    }
  }
}

void IAlgExtModularShort::output(std::ostream& out) const {
  if (isZero())
    out << '0';
  else {
    IAlgExtModularShortAux::ConstIterator i(mPoly.begin());
    if ((*i).mCoeff == 1)
      if ((*i).mDegree == 0)
        out << '1';
      else if ((*i).mDegree == 1)
        out << parameters()->variable(0);
      else
        out << parameters()->variable(0) << '^' << (*i).mDegree;
    else
      if ((*i).mDegree == 0)
        out << (*i).mCoeff;
      else if ((*i).mDegree == 1)
        out << (*i).mCoeff << '*' << parameters()->variable(0);
      else
        out << (*i).mCoeff << '*' << parameters()->variable(0) << '^' << (*i).mDegree;
    ++i;
    while(i) {
      out << " + ";
      if ((*i).mCoeff == 1)
        if ((*i).mDegree == 0)
          out << '1';
        else if ((*i).mDegree == 1)
          out << parameters()->variable(0);
        else
          out << parameters()->variable(0) << '^' << (*i).mDegree;
      else
        if ((*i).mDegree == 0)
          out << (*i).mCoeff;
        else if ((*i).mDegree == 1)
          out << (*i).mCoeff << '*' << parameters()->variable(0);
        else
          out << (*i).mCoeff << '*' << parameters()->variable(0) << '^' << (*i).mDegree;
      ++i;
    }
  }
}

void IAlgExtModularShort::assertValid(const char* fileName, int fileLine) const {
  mPoly.assertValid(fileName, fileLine);
}

IAllocator IAlgExtModularShort::sAllocator(sizeof(IAlgExtModularShort));
