/***************************************************************************
 *   Copyright (C) 2006 by Daniel Robertz                                  *
 *   daniel@momo.math.rwth-aachen.de                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IALGEXTMODULARSHORTAUX_H
#define IALGEXTMODULARSHORTAUX_H

#include <iostream>

#include "ioneparametergmpzaux.h"

#include "../util/iutil.h"
#include "../util/ilist.h"
#include "../util/ivariables.h"

#ifdef TEST
class IAlgExtModularShortAux;
#include "../test/iginvtest.h"
#include "../test/ifactorizationzassenhaustest.h"
#endif

class IAlgExtModularShort;
class IFactorizationZassenhaus;

class IAlgExtModularShortAux {
  friend class IAlgExtModularShort;
  friend class IFactorizationZassenhaus;

#ifdef TEST
  friend class IGinvTest;
  friend class IFactorizationZassenhausTest;
#endif

public:
  struct Term {
    unsigned  mDegree;
    int mCoeff;

    Term(): mDegree(0) { }
    Term(const Term& a): mDegree(a.mDegree) {
      mCoeff = a.mCoeff;
    }
    ~Term() { }

    void operator=(const Term& a) {
      mDegree = a.mDegree;
      mCoeff = a.mCoeff;
    }
  };

protected:
  int          mPrime;
  IList<Term>  mHead;

  void equality(std::istream& in, const IVariables &parameters);
  void additive(std::istream& in, const IVariables &parameters);
  void multiplicative(std::istream& in, const IVariables &parameters);
  void unary(std::istream& in, const IVariables &parameters);
  void power(std::istream& in, const IVariables &parameters);
  void bracket(std::istream& in, const IVariables &parameters);

public:
  void mult(const int coeff);
  void mult(unsigned degree, const int coeff);
  void div(const int coeff);
  void div(unsigned degree,  const int coeff);
  void quo(const IAlgExtModularShortAux& a, const IAlgExtModularShortAux& b);
  void rem(const IAlgExtModularShortAux& a);
  void rem_ext(const IAlgExtModularShortAux& a, IAlgExtModularShortAux& c1, IAlgExtModularShortAux& c2);

  typedef IList<Term>::Iterator  Iterator;
  typedef IList<Term>::ConstIterator  ConstIterator;

  IAlgExtModularShortAux(int prime):
      mPrime(prime),
      mHead() {
  }
  IAlgExtModularShortAux(const IAlgExtModularShortAux& a):
      mPrime(a.mPrime),
      mHead(a.mHead) {
  }

  ~IAlgExtModularShortAux() {}

  void operator=(const IAlgExtModularShortAux& a);

  void setZero() { mHead.clear(); }
  void setOne() {
    mHead.clear();
    Iterator i(mHead.begin());
    i.insert();
    (*i).mCoeff = 1;
  }

  bool isZero() const { return mHead.isEmpty(); }
  bool isOne() const {
    ConstIterator i(mHead.begin());
    return i && (*i).mDegree == 0 && (*i).mCoeff == 1;
  }
  bool isNegative() const {
    return false;
  }
  bool isComposite() const {
    ConstIterator i(mHead.begin());
    return mHead.length() > 1;
  }
  unsigned degree() const {
    IASSERT(!isZero());
    return (*mHead.begin()).mDegree;
  }
  
  int lc() const {
    if (isZero())
      return 0;
    return (*mHead.begin()).mCoeff;
  }
  
  int prime() const {
    return mPrime;
  }
  
  void swap(IAlgExtModularShortAux& a) {
    mHead.swap(a.mHead);
  }

  void minus();
  void add(const IAlgExtModularShortAux& a);
  void sub(const IAlgExtModularShortAux& a);
  void sub(const IAlgExtModularShortAux& a, unsigned deg, int coeff);
  void mult(const IAlgExtModularShortAux& a);
  void mult(const IAlgExtModularShortAux& a, const IAlgExtModularShortAux& b);
  void diff();
  void pow(unsigned deg);
  void div(const IAlgExtModularShortAux& a);
  void div(const IAlgExtModularShortAux& a, const IAlgExtModularShortAux& b) {
    *this = a;
    div(b);
  }
  void gcd(const IAlgExtModularShortAux& a);
  void gcd(const IAlgExtModularShortAux& a, const IAlgExtModularShortAux& b) {
    *this = a;
    gcd(b);
  }
  void addmul(const IAlgExtModularShortAux& a, const IAlgExtModularShortAux& b) {
    IAlgExtModularShortAux tmp(mPrime);
    tmp.mult(a, b);
    add(tmp);
  }
  
  void mod(const IOneParameterGmpZAux& a);
  void convert(IOneParameterGmpZAux& a) const;

  ConstIterator begin() const { return mHead.begin(); }
  Iterator begin() { return mHead.begin(); }

  void read(std::istream &in);
  void write(std::ostream &out) const;

  void input(std::istream& in, const IVariables &parameters);
  friend std::ostream& operator<<(std::ostream& out, const IAlgExtModularShortAux& a);

  void assertValid(const char* fileName, int fileLine) const;
};

#endif // IALGEXTMODULARSHORTAUX_H
