/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <cstdarg>
#include <sstream>

#include <gmp.h>

#include "../util/iutil.h"
#include "../monom/imonom.h"
#include "./icoeff.h"
#include "./igmpz.h"
#include "./igmpzz.h"
#include "./igmpq.h"
#include "./imodularshort.h"
#include "./ioneparametergmpz.h"
#include "./ioneparametermodularshort.h"
#include "./ialgextgmpq.h"
#include "./ialgext.h"
#include "./ialgextmodularshort.h"
#include "./itwoparametermodularshort.h"
#include "./inparametergmpz.h"
#include "./inparameterfield.h"


static void* (*imallocGmp)(size_t) = imalloc;

static void* ireallocGmp(void *ptr, size_t, size_t n) {
  return irealloc(ptr, n);
}

static void ifreeGmp(void *ptr, size_t) {
  ifree(ptr);
}

IDenomList::IDenomList(): mDenomList() {
}

IDenomList::~IDenomList() {
  IList<ICoeff*>::Iterator i = mDenomList.begin();
  while (i) {
    delete *i;
    i.del();
  }
}

void IDenomList::addDenom(const ICoeff& a) {
  IList<ICoeff*>::Iterator i = mDenomList.begin();
  i.insert();
  *i = a.coeffInterface()->copy(a);
}

ICoeffInterface::ICoeffInterface():
    mType(ICoeffInterface::Indefinitely),
    mSystemType(),
    mParameters(NULL) {
  mp_set_memory_functions(imallocGmp, ireallocGmp, ifreeGmp);
}

ICoeffInterface* ICoeffInterface::create(ICoeffInterface::Type type, const ISystemType *systemType, ...) {
  if (type == ICoeffInterface::Indefinitely)
    IERROR("error in ICoeffInterface: type == ICoeffInterface::Indefinitely");
  if (systemType == NULL || systemType->type() == ISystemType::Indefinitely)
    IERROR("error in ICoeffInterface: systemType == NULL || systemType->type() == ISystemType::Indefinitely");

  va_list marker;
  va_start(marker, systemType);

  ICoeffInterface* r=NULL;
  switch(type) {
  case ICoeffInterface::GmpZ: {
    static IGmpZInterface *gmpZ = NULL;
    if (gmpZ == NULL)
      gmpZ = new IGmpZInterface;
    r = gmpZ;
    }
    break;
  case ICoeffInterface::GmpZZ: {
    static IGmpZZInterface *gmpZZ = NULL;
    if (gmpZZ == NULL)
      gmpZZ = new IGmpZZInterface;
    r = gmpZZ;
    }
    break;
  case ICoeffInterface::GmpQ: {
    static IGmpQInterface *gmpQ = NULL;
    if (gmpQ == NULL)
      gmpQ = new IGmpQInterface;
    r = gmpQ;
    }
    break;
  case ICoeffInterface::ModularShort:
    r = new IModularShortInterface(va_arg(marker, int));
    break;
  case ICoeffInterface::ReducibleModuloBasis:
    break;
  case ICoeffInterface::OneParameterGmpZ: {
      IVariables* parameter = va_arg(marker, IVariables*);
      r = new IOneParameterGmpZInterface(parameter);
    }
    break;
  case ICoeffInterface::OneParameterModularShort: {
      IVariables* parameter = va_arg(marker, IVariables*);
      int modularShort = va_arg(marker, int);
      r = new IOneParameterModularShortInterface(parameter, modularShort);
    }
    break;
  case ICoeffInterface::AlgebraicFieldExtensionGmpQ: {
      IVariables* parameter = va_arg(marker, IVariables*);
      const char **strings = va_arg(marker, const char**);
      IVariables* tmp = new IVariables();
      tmp->add(parameter->variable(0));
      IAlgExtGmpQAux minPoly;
      std::istringstream in(strings[0]);
      minPoly.input(in, *tmp);
      in >> std::ws;
      if (!in.eof())
        IERROR("error parsing minimal polynomial");
      r = new IAlgExtGmpQInterface(tmp, minPoly);

      for (int i = 1; i < parameter->dim(); i++) {
        IVariables* tmp = new IVariables();
        tmp->add(parameter->variable(i));
        IOneParameterPoly minPoly(r);
        std::istringstream in(strings[i]);
        minPoly.input(in, *tmp);
        in >> std::ws;
        if (!in.eof())
          IERROR("error parsing minimal polynomial");
        r = new IAlgExtInterface(tmp, minPoly);
      }
    }
    break;
  case ICoeffInterface::AlgebraicFieldExtensionNParameter: {
      IVariables* fieldParameter = va_arg(marker, IVariables*);
      IVariables* parameter = va_arg(marker, IVariables*);
      const char **strings = va_arg(marker, const char**);
      
      r = create(ICoeffInterface::NParameterField, systemType, fieldParameter);
      
      for (int i = 0; i < parameter->dim(); i++) {
        IVariables* tmp = new IVariables();
        tmp->add(parameter->variable(i));
        IOneParameterPoly minPoly(r);
        std::istringstream in(strings[i]);
        minPoly.input(in, *tmp);
        in >> std::ws;
        if (!in.eof())
          IERROR("error parsing minimal polynomial");
        r = new IAlgExtInterface(tmp, minPoly);
      }
    }
    break;
  case ICoeffInterface::AlgebraicFieldExtensionModularShort: {
      IVariables* parameter = va_arg(marker, IVariables*);
      const char **strings = va_arg(marker, const char**);
      int modularShort = va_arg(marker, int);
      IVariables* tmp = new IVariables();
      tmp->add(parameter->variable(0));
      IAlgExtModularShortAux minPoly(modularShort);
      std::istringstream in(strings[0]);
      minPoly.input(in, *tmp);
      in >> std::ws;
      if (!in.eof())
        IERROR("error parsing minimal polynomial");
      r = new IAlgExtModularShortInterface(tmp, minPoly);

      for (int i = 1; i < parameter->dim(); i++) {
        IVariables* tmp = new IVariables();
        tmp->add(parameter->variable(i));
        IOneParameterPoly minPoly(r);
        std::istringstream in(strings[i]);
        minPoly.input(in, *tmp);
        in >> std::ws;
        if (!in.eof())
          IERROR("error parsing minimal polynomial");
        r = new IAlgExtInterface(tmp, minPoly);
      }
    }
    break;
  case ICoeffInterface::TwoParameterGmpZ:
    break;
  case ICoeffInterface::TwoParameterModularShort: {
      IVariables* parameter = va_arg(marker, IVariables*);
      int modularShort = va_arg(marker, int);
      r = new ITwoParameterModularShortInterface(parameter, modularShort);
    }
    break;
  case ICoeffInterface::NParameterGmpZ: {
      IVariables* parameter = va_arg(marker, IVariables*);
      IMonomInterface* mi = IMonomInterface::create(IMonomInterface::Lex, systemType, parameter);
      r = new INParameterGmpZInterface(mi);
    }
    break;
  case ICoeffInterface::NParameterField: {
      IVariables* parameter = va_arg(marker, IVariables*);
      IMonomInterface* mi = IMonomInterface::create(IMonomInterface::Lex, systemType, parameter);
      r = new INParameterFieldInterface(mi);
    }
    break;
  
  case ICoeffInterface::NParameterModularShort:
    break;
  }
  va_end(marker);

  if (r)
    r->mSystemType = systemType;

  return r;
}

template <>
IAllocator IList<ICoeff*>::Node::sAllocator(sizeof(IList<ICoeff*>::Node));
