/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ICOEFF_H
#define ICOEFF_H

#include <iostream>

#include "../util/ivariables.h"
#include "../system/isystemtype.h"

class ICoeff;

class IDenomList {
protected:
  IList<ICoeff*> mDenomList;

public:
  IDenomList();
  ~IDenomList();

  IList<ICoeff*>::ConstIterator begin() const { return mDenomList.begin(); }
  int length() const { return mDenomList.length(); }
  void addDenom(const ICoeff& a);
};

class ICoeffInterface {
  friend class ICoeff;

public:
  enum Type {Indefinitely                         = 0x00000000,
             Field                                = 0x00000001,
             Pseudo                               = 0x00000002,
             Ring                                 = 0x00000004,
             GmpZ                                 = 0x00000008,
             GmpZZ                                = 0x00000008 | Ring,
             GmpQ                                 = 0x00000010 | Field,
             ModularShort                         = 0x00000020 | Field,
             ReducibleModuloBasis                 = 0x00000040,
             OneParameterGmpZ                     = 0x00000080 | Pseudo,
             OneParameterModularShort             = 0x00000100,
             AlgebraicFieldExtensionGmpQ          = 0x00000200 | Field,
             TwoParameterGmpZ                     = 0x00000400 | Pseudo,
             TwoParameterModularShort             = 0x00000800,
             NParameterGmpZ                       = 0x00001000 | Pseudo,
             NParameterModularShort               = 0x00002000,
             AlgebraicFieldExtensionModularShort  = 0x00004000 | Field,
             AlgebraicFieldExtension              = 0x00008000 | Field,
             NParameterField                      = 0x00010000 | Field,
             AlgebraicFieldExtensionNParameter    = 0x00020000 | Field,
  };

protected:
  Type               mType;
  const ISystemType *mSystemType;
  IVariables        *mParameters;

  IDenomList         mDenomList;

  ICoeffInterface();

public:
  static ICoeffInterface* create(Type type, const ISystemType *systemType,...);
  virtual ~ICoeffInterface() {}

  Type type() const { return mType; }
  bool isField() const { return (mType & Field) != 0; }
  bool isPseudo() const { return (mType & Pseudo) != 0; }
  const ISystemType* systemType() const { return mSystemType; }
  const IVariables* parameters() const { return mParameters; }

  virtual ICoeff* create() = 0;
  virtual ICoeff* copy(const ICoeff& a) = 0;

  IList<ICoeff*>::ConstIterator denomListbegin() const { return mDenomList.begin(); }
  int lengthDenom() const { return mDenomList.length(); }
  void addDenom(const ICoeff& a) { mDenomList.addDenom(a); }
};

class ICoeff {
protected:
  ICoeffInterface* const mRealization;

  ICoeff(ICoeffInterface* r): mRealization(r) { IASSERT(mRealization); }
  ICoeff(const ICoeff& a): mRealization(a.mRealization) {}

public:
  typedef ICoeffInterface::Type Type;

  virtual ~ICoeff() {}

  ICoeffInterface* coeffInterface() const { return mRealization; }

  ICoeff* copy() const { return mRealization->copy(*this); }

  Type type() const { return mRealization->type(); }
  bool isField() const { return mRealization->isField(); }
  bool isPseudo() const { return mRealization->isPseudo(); }
  const IVariables* parameters() const { return mRealization->parameters(); }

  virtual void set(const ICoeff& a) = 0;
  void operator=(const ICoeff& a) { set(a); }
  virtual void setZero() = 0;
  virtual void setOne() = 0;

  virtual bool isNegative() const = 0;
  virtual bool isZero() const = 0;
  virtual bool isOne() const = 0;
  virtual bool isComposite() const = 0;
  virtual bool isUnit() const = 0;
  virtual int norm() const = 0;
  virtual int compare(const ICoeff& a) const {
    return 0;
  }
  virtual int cmpNorm(const ICoeff& a) const {
    return 0;
  }

  virtual void minus() = 0;
  virtual void abs() = 0;
  virtual void add(const ICoeff& a) = 0;
  virtual void sub(const ICoeff& a) = 0;
  virtual void mult(const ICoeff& a) = 0;
  virtual void diff(int var, unsigned deg) = 0;
  virtual void shift(int var, unsigned deg) = 0;
  virtual void pow(unsigned deg) = 0;
  virtual void inverse() = 0;
  virtual bool divisibility(const ICoeff& a) const {
    return true;
  }
  virtual bool divisibilityProper(const ICoeff& a) const {
    return true;
  }
  virtual void div(const ICoeff& a) = 0;
  virtual void div(const ICoeff& a, const ICoeff& b) = 0;
  virtual void quotRem(const ICoeff& a, const ICoeff& b, ICoeff& r) {}
  virtual void gcd(const ICoeff& a) = 0;
  virtual void gcd(const ICoeff& a, const ICoeff& b) = 0;
  virtual void gcdExt(const ICoeff& a, const ICoeff& b, ICoeff& s, ICoeff& t) {}
  virtual void addmul(const ICoeff& a, const ICoeff& b) = 0;
  virtual void addmul(const ICoeff& a, unsigned long b) {}

  virtual void read(std::istream &in) = 0;
  virtual void write(std::ostream &out) const = 0;

  virtual void input(std::istream& in, bool expanded = false) = 0;
  virtual void output(std::ostream& out) const = 0;
  friend std::istream& operator>>(std::istream& in, ICoeff& a) {
    a.input(in);
    return in;
  }
  friend std::ostream& operator<<(std::ostream& out, const ICoeff& a) {
    a.output(out);
    return out;
  }
};

#endif // ICOEFF_H
