/***************************************************************************
 *   Copyright (C) 2008 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../util/iprime.h" 
#include "./ifactorizationz.h" 
#include "./ifactorizationzassenhaus.h" 
#include "./inparametergmpzaux.h" 

/* implementation of Cohen, Algorithm 3.5.7
 */
void IFactorizationZ::factor(const IOneParameterGmpZAux& a,
    IList<IOneParameterGmpZAux*>& factors,
    IList<unsigned>& multiplicities) {

  if (a.isZero())
    return;

  IOneParameterGmpZAux tmpa(a);

  // STEP 0: Remove possible trivial factor
  IOneParameterGmpZAux::ConstIterator ita = tmpa.begin();
  while (ita.isNext())
    ++ita;
  
  if ((*ita).mDegree) {
    int deg = (*ita).mDegree;
    
    IList<IOneParameterGmpZAux*>::Iterator itf = factors.begin();
    itf.insert();
    (*itf) = new IOneParameterGmpZAux();
    (*itf)->setOne();
    (*(*itf)->mHead.begin()).mDegree = 1;
    
    IList<unsigned>::Iterator itm = multiplicities.begin();
    itm.insert();
    (*itm) = deg;
    
    ita = tmpa.begin();
    while (ita) {
      (*ita).mDegree -= deg;
      ++ita;
    }
  }
  
  // STEP 1: Reduce to squarefree and primitive
  IOneParameterGmpZAux u(a);

  IOneParameterGmpZAux ad(a);
  ad.diff();
  
  // use the nparameter gcd algorithm
  INParameterGmpZAux tmpu(u);
  INParameterGmpZAux tmpud(ad);
  
  INParameterGmpZAux gcd(tmpu.interface());
  gcd.gcd(tmpu, tmpud);
  
  ad = IOneParameterGmpZAux(gcd);
  u.div(ad);
  
  // STEP 2: Find a squarefree factorization mod p
  IPrime prime;
//  prime.nextPrimeGreater(2);
  
  IFactorizationZassenhaus alg;
  MP_INT lcu;
  u.lc(lcu);
  
  if (mpz_cmp_ui(&lcu, 0ul) < 0) {
    mpz_abs(&lcu, &lcu);
    u.minus();
  }
  
  ModularShort primes[20];
  IList<unsigned> degrees[20];
  
  int i = 0;
  while (i < 20) {
    primes[i] = prime.prime();
    
    if (!mpz_divisible_ui_p(&lcu, primes[i])) {
      IAlgExtModularShortAux up(primes[i]);
      up.mod(u);

      IAlgExtModularShortAux upd(up);
      upd.diff();
      
      upd.gcd(up);

      if (upd.isOne()) {
        std::cout << "using prime " << primes[i] << std::endl;
        alg.degrees(up, degrees[i]);
        ++i;
      }
    }

    prime.next();
  }
  
  std::cout << "found the following possible degrees:" << std::endl;
  for (int j = 0; j < 20; j++) {
    std::cout << j << ": ";
    bool* str = createBinaryString(degrees[j], u.degree());
    for (int i = 0; i <= u.degree(); i++)
      std::cout << str[i];
    std::cout << std::endl;
  }

  ModularShort p;
  IList<IAlgExtModularShortAux*> factorsp;
  IList<unsigned> multiplicitiesp;

  do {
    p = prime.prime();
    
    if (!mpz_divisible_ui_p(&lcu, p)) {
      IAlgExtModularShortAux up(p);
      up.mod(u);

      IAlgExtModularShortAux upd(up);
      upd.diff();
      
      upd.gcd(up);

      if (upd.isOne()) {
        std::cout << "using prime " << p << std::endl;
        alg.factor(up, factorsp, multiplicitiesp);
      }
    }

    prime.next();
  } while (factorsp.isEmpty());
  mpz_clear(&lcu);

  std::cout << "step 3" << std::endl;
  // STEP 3: Find bound
  unsigned e = getExponent(u, p);
  
  std::cout << "step 4" << std::endl;
  // STEP 4: Lift factorization
  std::vector<IOneParameterGmpZAux*> lifts;
  lifts.reserve(factorsp.length());
  lift(lifts, factorsp, u, p, e);
  
  std::cout << "step 5" << std::endl;
  // STEP 5: Try combination
  MP_INT q;
  mpz_init(&q);
  mpz_ui_pow_ui(&q, p, e);
  
  IList<IOneParameterGmpZAux*>::Iterator itf = factors.begin();
  IList<unsigned>::Iterator itm = multiplicities.begin();
  
  for (int d = 1; d <= lifts.size()/2; d++) {
    std::cout << "combinations of length " << d << std::endl;
    unsigned* comb = initCombination(d, lifts.size());
    
    while (d <= lifts.size()/2 && isValidCombination(comb, d, lifts.size())) {
//      std::cout << "trying combination " << std::flush;
//      for (int i = 0; i < d; i++)
//        std::cout << comb[i] << ", ";
//      std::cout << std::endl;
      IOneParameterGmpZAux tmp;
      int multiplicity = testFactor(lifts, tmp, u, tmpa, comb, d, q);
      
      if (multiplicity) {
        std::cout << "found one!" << std::endl;
        itf.insert();
        (*itf) = new IOneParameterGmpZAux();
        (*itf)->swap(tmp);
        ++itf;
        
        itm.insert();
        (*itm) = multiplicity;
        ++itm;
        
        updateCombination(comb, d);
      }
      else {
        nextCombination(comb, d, lifts.size());
      }
    }
    
    delete[] comb;
  }
  
  if (!u.isOne()) {
    u.pp();
    itf.insert();
    (*itf) = new IOneParameterGmpZAux();
    (*itf)->swap(u);

    itm.insert();
    (*itm) = 0;

    while (tmpa.divtry(*(*itf)))
      ++(*itm);
  }

  mpz_clear(&q);
}

bool* IFactorizationZ::createBinaryString(IList<unsigned>& degrees, 
    int n) {
  
  bool* str = new bool[n+1];
  str[0] = true;
  
  for (IList<unsigned>::ConstIterator it = degrees.begin(); it; ++it) {
    for (int i = n; i >= 0; i--) {
      str[i] |= str[i-(*it)];
    }
  }
  
  return str;
}

void IFactorizationZ::createBinaryString(bool* str1, bool* str2, int n) {
  for (int i = 0; i <= n; i++) {
    str1[i] &= str2[i];
  }
}


unsigned* IFactorizationZ::initCombination(int k, int n) {
  unsigned* result = new unsigned[n];
  unsigned tmp = 0;
  for (int i = 0; i < k; i++) {
    result[i] = tmp++;
  }
  
  return result;
}

void IFactorizationZ::updateCombination(unsigned* comb, int k) {
  for (int i = 1; i < k; i++) {
    comb[i] = comb[i-1] + 1;
  }
}

void IFactorizationZ::nextCombination(unsigned* comb, int k, int n) {
  for (int i = k-1; i >= 0; i--) {
    if (comb[i] < n-(k-i)) {
      comb[i]++;
      for (int j = i+1; j < k; j++) {
        comb[j] = comb[j-1] + 1;
      }
      
      return;
    }
  }
  
  // make combination invalid
  comb[0] = n;
}

bool IFactorizationZ::isValidCombination(unsigned* comb, int k, int n) {
  if (comb[0] >= n)
    return false;
  
  for (int i = 1; i < k; i++)
    if (comb[i] <= comb[i-1])
      return false;
  
  return comb[k-1] < n;
}

/* checks, whether the product of the k elements of lifts divides u.
 * In this case, f will be set to this factor, and the number of times
 * that f divides a will be returned. Furthermore the elements will be
 * removed from lifts, and f will be factored out of u and a.
 */ 
int IFactorizationZ::testFactor(std::vector<IOneParameterGmpZAux*> lifts, 
    IOneParameterGmpZAux& f, 
    IOneParameterGmpZAux& u, 
    IOneParameterGmpZAux& a, 
    unsigned elements[], 
    int k,
    MP_INT& q) {

  MP_INT lcu;
  MP_INT ccu;
  u.lc(lcu);
  u.cc(ccu);
  
  f = *lifts[elements[0]];
  for (int i = 1; i < k; i++) {
    f.mult(*lifts[elements[i]]);
    f.mod(q);
  }
  f.mult(lcu);
  f.mod(q);
  setRange(f, q);
  
  MP_INT ccf;
  f.cc(ccf);
  mpz_mul(&ccu, &ccu, &lcu);
  
  if (!mpz_divisible_p(&ccu, &ccf))
    return 0;
  
  IOneParameterGmpZAux tmp(u);
  tmp.mult(lcu);
  
  if (!tmp.divtry(f))
    return 0;
  
  f.pp();
  
  u.div(f);
  
  int result = 0;
  while (a.divtry(f)) {
    result++;
  }
  
  return result;
}

/* sets the range of the coefficients of a to [-q/2, q/2)
 */
void IFactorizationZ::setRange(IOneParameterGmpZAux& a, MP_INT& q) {
  MP_INT q2;
  mpz_init(&q2);
  mpz_cdiv_q_ui(&q2, &q, 2ul);
  
  IOneParameterGmpZAux::Iterator it = a.begin();
  while (it) {
    if (mpz_cmp(&(*it).mCoeff, &q2) >= 0)
      mpz_sub(&(*it).mCoeff, &(*it).mCoeff, &q);
    ++it;
  }
}

/* implementation of Cohen, Algorithm 3.5.5
 * here q = p^e and r = p is a prime.
 */
void IFactorizationZ::henselLift(IOneParameterGmpZAux & a,
    IOneParameterGmpZAux & b,
    const IOneParameterGmpZAux & c,
    const IAlgExtModularShortAux & u,
    const IAlgExtModularShortAux & v,
    ModularShort p,
    unsigned e) {

  IAlgExtModularShortAux f(p);

  MP_INT q;
  mpz_init(&q);
  mpz_ui_pow_ui(&q, p, e);

  getPoly(f, a, b, c, q);

  IAlgExtModularShortAux tmpv(v);
  IAlgExtModularShortAux ap(p);
  ap.mod(a);

  tmpv.mult(f);

  IAlgExtModularShortAux t(p);
  t.quo(tmpv, ap);

  // calculate tmpl as a lift of vf - at
  IAlgExtModularShortAux tmp(ap);
  tmp.mult(t);
  tmpv.sub(tmp);

  IOneParameterGmpZAux tmpl;
  tmpv.convert(tmpl);

  tmpl.mult(q);
  a.add(tmpl);

  // calculate tmpl as a lift of uf + bt
  IAlgExtModularShortAux tmpu(u);
  tmpu.mult(f);

  tmp.mod(b);
  tmp.mult(t);
  tmpu.add(tmp);

  tmpu.convert(tmpl);
  tmpl.mult(q);
  b.add(tmpl);

  mpz_clear(&q);
}

void IFactorizationZ::lift(std::vector<IOneParameterGmpZAux*>& result,
    IList<IAlgExtModularShortAux*>& factors,
    IOneParameterGmpZAux& u,
    ModularShort p,
    unsigned e) {

//  IList<IAlgExtModularShortAux*>::ConstIterator ifactors = factors.begin();
//
//  int i = 0;
//
//  IAlgExtModularShortAux tmp1(*(*ifactors));
//  ++ifactors;
//  IAlgExtModularShortAux tmp2(p);
//
//  IOneParameterGmpZAux tmp3;
  
  IAlgExtModularShortAux** array = new IAlgExtModularShortAux*[factors.length()];
  int n = 0;
  for (IList<IAlgExtModularShortAux*>::ConstIterator it = factors.begin();
      it; ++it)
    array[n++] = (*it);
    

  MP_INT lcu;
  u.lc(lcu);
  ModularShort lcup= mpz_fdiv_ui(&lcu, p);
  mpz_clear(&lcu);

  std::cout << "n = " << n << std::endl;
  for (int i = 0; i < n; i++) {
    IAlgExtModularShortAux prod(p);
    prod.setOne();

    for (int j = 0; j < n; j++) {
      if (j != i)
        prod.mult(*array[j]);
    }
    prod.mult(lcup);

    IOneParameterGmpZAux* lifti = new IOneParameterGmpZAux();
    IOneParameterGmpZAux ignore;

    liftPoly(*lifti, ignore, u, *array[i], prod, p, e);

    result.push_back(lifti);
  }
}

/* for a*b = c mod p, calculate the lifts x and y mod p^e.
 */
void IFactorizationZ::liftPoly(IOneParameterGmpZAux& x,
    IOneParameterGmpZAux& y,
    IOneParameterGmpZAux& c,
    IAlgExtModularShortAux& a,
    IAlgExtModularShortAux& b,
    ModularShort p,
    unsigned e) {

  IAlgExtModularShortAux u(p);
  IAlgExtModularShortAux v(p);

  std::cout << "xea ... " << std::flush;
  xea(a, b, u, v);
  std::cout << "done" << std::endl;

  a.convert(x);
  b.convert(y);

  std::cout << "e = " << e << std::endl;
  for (unsigned i = 1; i < e; i++) {
    henselLift(x, y, c, u, v, p, i);
  }
}

/* sets s and t such that s*a + t*b = gcd(a, b)
 */
void IFactorizationZ::xea(const IAlgExtModularShortAux& a,
    const IAlgExtModularShortAux& b,
    IAlgExtModularShortAux& s,
    IAlgExtModularShortAux& t) {

  ModularShort p = a.prime();

  IAlgExtModularShortAux c(a);
  c.div(c.lc());
  IAlgExtModularShortAux d(b);
  d.div(d.lc());

  IAlgExtModularShortAux c1(p);
  IAlgExtModularShortAux c2(p);
  IAlgExtModularShortAux d1(p);
  IAlgExtModularShortAux d2(p);
  c1.setOne();
  d2.setOne();

  IAlgExtModularShortAux q(p);
  IAlgExtModularShortAux r(p);
  IAlgExtModularShortAux r1(p);
  IAlgExtModularShortAux r2(p);

  while (!d.isZero()) {
    polydiv(c, d, q, r);
    r1 = q;
    r1.mult(d1);
    r1.minus();
    r1.add(c1);

    r2.swap(q);
    r2.mult(d2);
    r2.minus();
    r2.add(c2);

    c.swap(d);
    c1.swap(d1);
    c2.swap(d2);

    d.swap(r);
    d1.swap(r1);
    d2.swap(r2);
  }

  ModularShort temp = c.lc();
  temp *= a.lc();
  temp %= p;
  temp = iInverse(temp, p);

  s.swap(c1);
  s.mult(temp);

  temp = c.lc();
  temp *= b.lc();
  temp %= p;
  temp = iInverse(temp, p);

  t.swap(c2);
  t.mult(temp);
}

/* computes q and r such that a = qb + r, deg(r) < deg(b);
 */
void IFactorizationZ::polydiv(const IAlgExtModularShortAux& a,
    const IAlgExtModularShortAux& b,
    IAlgExtModularShortAux& q,
    IAlgExtModularShortAux& r) {

  ModularShort p = a.prime();

  ModularShort lc = iInverse(b.lc(), p);

  r= a;

  q.setZero();

  int m2;
  ModularShort k2;
  IAlgExtModularShortAux::Iterator iq(q.begin());

  while (!r.isZero() && r.degree() >= b.degree()) {
    IAlgExtModularShortAux::Iterator i1(r.begin());
    IAlgExtModularShortAux::ConstIterator i2(b.begin());

    m2 = (*i1).mDegree - (*i2).mDegree;
    k2 = ((*i1).mCoeff * lc) % p;
    k2 = p - k2;

    i1.del();
    ++i2;

    while (i1 && i2) {
      if ((*i1).mDegree > (*i2).mDegree + m2) {
        ++i1;
      } else if ((*i1).mDegree < (*i2).mDegree + m2) {
        i1.insert();
        (*i1).mDegree = (*i2).mDegree + m2;
        (*i1).mCoeff = ((*i2).mCoeff * k2) %p;

        ++i1;
        ++i2;
      } else {
        (*i1).mCoeff += (*i2).mCoeff * k2;
        (*i1).mCoeff %= p;
        if ((*i1).mCoeff == 0)
          i1.del();
        else
          ++i1;
        ++i2;
      }
    }
    while (i2) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree + m2;
      (*i1).mCoeff = ((*i2).mCoeff * k2) %p;
      ++i1;
      ++i2;
    }
    iq.insert();
    (*iq).mDegree = m2;
    (*iq).mCoeff = p - k2;
    ++iq;
  }
}

/* calculates f = (c - ab)/q
 */
void IFactorizationZ::getPoly(IAlgExtModularShortAux & f,
    const IOneParameterGmpZAux & a,
    const IOneParameterGmpZAux & b,
    const IOneParameterGmpZAux & c,
    MP_INT& q) {

  IOneParameterGmpZAux tmp(a);
  tmp.mult(b);
  tmp.minus();
  tmp.add(c);
  tmp.div(q);

  f.mod(tmp);
}

void IFactorizationZ::mignotte(MP_INT& b, const IOneParameterGmpZAux& a) {
  MP_INT norma;
  mpz_init_set_ui(&norma, 0ul);
  mpz_init_set_ui(&b, 0ul);

  IOneParameterGmpZAux::ConstIterator it = a.begin();
  while (it) {
    mpz_addmul(&norma, &(*it).mCoeff, &(*it).mCoeff);

    ++it;
  }
  mpz_sqrt(&norma, &norma);
  mpz_add_ui(&norma, &norma, 1ul);

  MP_INT lca;
  a.lc(lca);

  int n = a.degree()/2;

  MP_INT tmp;
  MP_INT tmp2;
  mpz_init(&tmp);
  mpz_init(&tmp2);
  for (int j = 0; j <= n; j++) {
    mpz_bin_uiui(&tmp2, n-1, j);
    mpz_mul(&tmp2, &tmp2, &norma);
    mpz_bin_uiui(&tmp, n-1, j-1);
    mpz_addmul(&tmp2, &tmp, &lca);

    if (mpz_cmp(&tmp2, &b)> 0)
      mpz_set(&b, &tmp2);
  }

  mpz_clear(&lca);
  mpz_clear(&norma);
  mpz_clear(&tmp2);
  mpz_clear(&tmp);
}

unsigned IFactorizationZ::getExponent(const IOneParameterGmpZAux& u,
    ModularShort p) {

  MP_INT b;
  mignotte(b, u);
  MP_INT lcu;
  u.lc(lcu);
  mpz_mul(&b, &b, &lcu);
  mpz_add(&b, &b, &b);
  mpz_abs(&b, &b);

  MP_INT tmp;
  mpz_init_set_ui(&tmp, p);

  unsigned result = 1;
  while (mpz_cmp(&tmp, &b) <= 0) {
    mpz_mul_ui(&tmp, &tmp, p);
    ++result;
  }

  mpz_clear(&b);
  mpz_clear(&lcu);
  mpz_clear(&tmp);

  return result;
}
