/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./ifactorizationzassenhaus.h" 
#include "../util/imodular.h" 

void IFactorizationZassenhaus::randomPolynomial(
    IAlgExtModularShortAux& a, 
    unsigned n) {
  
  IASSERT(n > 1);
  unsigned deg = iRandom(n);

  a.setZero();

  while (a.isZero() || a.degree() == 0) {
    IAlgExtModularShortAux::Iterator it = a.begin();

    for (int i = 0; i <= deg; i++) {
      it.insert();
      (*it).mDegree = i;
      (*it).mCoeff = iRandom(a.mPrime);
    }
  }
}

void IFactorizationZassenhaus::repeatedSquaring(
    IAlgExtModularShortAux& a, 
    MP_INT& n, 
    IAlgExtModularShortAux& rem) {
  
  int i = mpz_sizeinbase(&n, 2) - 2;
  
  
  IAlgExtModularShortAux result(a);

  while (i >= 0) {
    IAlgExtModularShortAux tmp(result);
    result.mult(tmp);
    result.rem(rem);
    
    if (mpz_tstbit(&n, i)) {
      result.mult(a);
      result.rem(rem);
    }
    --i;
  }

  a.mHead.swap(result.mHead);
}

void IFactorizationZassenhaus::equalDegreeSplitting(
    IAlgExtModularShortAux& f, 
    IAlgExtModularShortAux& a, 
    unsigned d, 
    IAlgExtModularShortAux& g) {
  
  IASSERT(f.degree() > d);
  g.gcd(a, f);

  if (!g.isOne()) {
    g.mult(iInverse((*g.mHead.begin()).mCoeff, g.mPrime));
    return;
  }

  if (a.mPrime > 2) {
    MP_INT exp;
    mpz_init(&exp);
    mpz_ui_pow_ui(&exp, a.mPrime, d);
    mpz_sub_ui(&exp, &exp, 1ul);
    mpz_tdiv_q_2exp(&exp, &exp, 1);

    repeatedSquaring(a, exp, f);
    
    mpz_clear(&exp);
  } 
  else {
    // Use Trace polynomial
    IAlgExtModularShortAux tmp(a);

    for (int i = 1; i < d; i++) {
      IAlgExtModularShortAux tmp2(tmp);
      tmp.mult(tmp2);
      tmp.rem(f);
      a.add(tmp);
      a.rem(f);
    }
  }

  IAlgExtModularShortAux one(a.mPrime);
  one.setOne();
  a.sub(one);

  if (a.isZero()) {
    g.setZero();
    return;
  }

  g.gcd(a, f);
  g.mult(iInverse((*g.mHead.begin()).mCoeff, g.mPrime));

  if (g.degree() == 0 || g.degree() == f.degree())
    g.setZero();
}

void IFactorizationZassenhaus::equalDegreeFactorization(
    IAlgExtModularShortAux& a, 
    unsigned d, 
    IList<IAlgExtModularShortAux*>& factors) {
  
  if (a.degree() == d) {
    IList<IAlgExtModularShortAux*>::Iterator it = factors.begin();
    it.insert();
    (*it) = new IAlgExtModularShortAux(a);
    return;
  }

  IAlgExtModularShortAux g(a.mPrime);
  g.setZero();

  while (g.isZero()) {
    IAlgExtModularShortAux r(a.mPrime);
    randomPolynomial(r, a.degree());
    equalDegreeSplitting(a, r, d, g);
  }

  a.div(g);

  equalDegreeFactorization(g, d, factors);
  equalDegreeFactorization(a, d, factors);
}

void IFactorizationZassenhaus::degrees(IAlgExtModularShortAux& a, 
    IList<unsigned>& degrees) {

  IAlgExtModularShortAux x(a.mPrime);
  x.setOne();
  x.mult(1, 1);

  IAlgExtModularShortAux h(x);

  IAlgExtModularShortAux v(a);
  v.mult(iInverse((*v.mHead.begin()).mCoeff, v.mPrime));

  degrees.clear();

  IList<unsigned>::Iterator degit = degrees.begin();

  int d = 0;
  MP_INT p;
  mpz_init_set_ui(&p, v.mPrime);
  while (!v.isOne()) {
    d++;
    repeatedSquaring(h, p, a);

    IAlgExtModularShortAux g(v.mPrime);
    IAlgExtModularShortAux tmp(h);
    tmp.sub(x);
    g.gcd(tmp, v);
    g.mult(iInverse((*g.mHead.begin()).mCoeff, g.mPrime));

    if (!g.isOne()) {
      v.div(g);
      for (int i = 0; i < g.degree()/d; i++) {
        degit.insert();
        (*degit) = d;
        ++degit;
      }
    }
  }
  
}


void IFactorizationZassenhaus::factor(
    IAlgExtModularShortAux& a, 
    IList<IAlgExtModularShortAux*>& factors, 
    IList<unsigned>& multiplicities) {
  
  IAlgExtModularShortAux x(a.mPrime);
  x.setOne();
  x.mult(1, 1);

  IAlgExtModularShortAux h(x);

  IAlgExtModularShortAux v(a);
  v.mult(iInverse((*v.mHead.begin()).mCoeff, v.mPrime));

  factors.clear();
  multiplicities.clear();

  IList<IAlgExtModularShortAux*>::Iterator fit = factors.begin();
  IList<unsigned>::Iterator mit = multiplicities.begin();

  int d = 0;
  MP_INT p;
  mpz_init_set_ui(&p, v.mPrime);
  while (!v.isOne()) {
    d++;
    repeatedSquaring(h, p, a);

    IAlgExtModularShortAux g(v.mPrime);
    IAlgExtModularShortAux tmp(h);
    tmp.sub(x);
    g.gcd(tmp, v);
    g.mult(iInverse((*g.mHead.begin()).mCoeff, g.mPrime));

    if (!g.isOne()) {
      IList<IAlgExtModularShortAux*> gfactors;
      equalDegreeFactorization(g, d, gfactors);

      IList<IAlgExtModularShortAux*>::Iterator git = gfactors.begin();

      while (git) {
        fit.insert();
        mit.insert();

        (*fit) = (*git);
        (*mit) = 0;

        do {
          (*mit)++;
          v.div(*(*fit));

          tmp = v;
          tmp.rem(*(*fit));
        } while (tmp.isZero());

        ++git;
        ++fit;
        ++mit;
      }
    }
  }
  mpz_clear(&p);
}
