/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./igcdalgorithm.h"
#include "./igcdbrown.h"
#include "./igcdzippel.h"
#include "./igcdonevar.h"
#include "./igcdheu.h"
//#include "./igcddivbro.h"
#include "./inparametergmpzaux.h"

#include "../util/iprime.h"

IGcdAlgorithm::AbstractIterator* IGcdAlgorithm::ShiftIterator::clone() const {
  return new ShiftIterator(ma->clone(), mShift);
}

bool IGcdAlgorithm::ShiftIterator::isValid() const {
  return (ma->isValid() && ((int) ma->deg()) >= mShift);
}

void IGcdAlgorithm::ShiftIterator::next() {
  ma->next();
}

ModularShort IGcdAlgorithm::ShiftIterator::coeff() {
  return ma->coeff();
}

unsigned IGcdAlgorithm::ShiftIterator::deg() {
  return ma->deg() - mShift;
}

IGcdAlgorithm::AbstractIterator* IGcdAlgorithm::DelegationIterator::clone() const {
  return new DelegationIterator(ma);
}

bool IGcdAlgorithm::DelegationIterator::isValid() const {
  return (ma);
}
void IGcdAlgorithm::DelegationIterator::next() {
  ++ma;
}
ModularShort IGcdAlgorithm::DelegationIterator::coeff() {
  return (*ma).mCoeff;
}
unsigned IGcdAlgorithm::DelegationIterator::deg() {
  return (*ma).mMonom[0];
}

IGcdAlgorithm::AbstractIterator* IGcdAlgorithm::SumIterator::clone() const {
  return new SumIterator(ma->clone(), mb->clone(), mPrimeShort);
}

bool IGcdAlgorithm::SumIterator::isValid() const {
  return ma->isValid() || mb->isValid();
}
void IGcdAlgorithm::SumIterator::next() {
  if (!ma->isValid()) {
    mb->next();
    return;
  }
  if (!mb->isValid()) {
    ma->next();
    return;
  }
  unsigned dega = ma->deg();
  unsigned degb = mb->deg();
  if (dega > degb) {
    ma->next();
  } else if (dega < degb) {
    mb->next();
  } else {
    ma->next();
    mb->next();
  }
}

ModularShort IGcdAlgorithm::SumIterator::coeff() {
  if (!ma->isValid())
    return mb->coeff();
  if (!mb->isValid())
    return ma->coeff();
  unsigned dega = ma->deg();
  unsigned degb = mb->deg();
  if (dega > degb) {
    return ma->coeff();
  } else if (dega < degb) {
    return mb->coeff();
  } else {
    return (ma->coeff() + mb->coeff()) % mPrimeShort;
  }
}

unsigned IGcdAlgorithm::SumIterator::deg() {
  if (!ma->isValid())
    return mb->deg();
  if (!mb->isValid())
    return ma->deg();
  unsigned dega = ma->deg();
  unsigned degb = mb->deg();
  return (dega > degb) ? dega : degb;
}

IGcdAlgorithm* IGcdAlgorithm::create(IGcdAlgorithm::Algorithm alg) {
  IGcdAlgorithm* r= NULL;
  switch (alg) {
  case Brown:
    r = new IGcdBrown();
    break;
  case Zippel:
    r = new IGcdZippel();
    break;
  case Onevar:
    r = new IGcdOnevar();
    break;
  case Heu:
    r = new IGcdHeu();
    break;
    //  case Divbro:
    //    r = new IGcdDivbro();
    //    break;
  }

  return r;
}

unsigned* IGcdAlgorithm::sortPoly(const INParameterGmpZAux& a,
    const INParameterGmpZAux& b,
    INParameterGmpZAux& asorted,
    INParameterGmpZAux& bsorted) {

  IMonomAuxInterface::MonomIterator d = a.interface().create();
  degreeBounds(a, b, d);

  unsigned* permutation = a.interface().sortDecreasing(d);

  a.sort(asorted, permutation);
  b.sort(bsorted, permutation);

  delete[] d;
  
  return permutation;
}

ModularShort* IGcdAlgorithm::getRandomPoints(int n, ModularShort p) {
  ModularShort* result = new ModularShort[n];
  for (int i = 0; i < n; i++) {
    result[i] = iRandom(p);
    if (result[i] == 0)
      result[i]++;
  }

  return result;
}

void IGcdAlgorithm::degreeBounds(const INParameterGmpZAux& a,
    const INParameterGmpZAux& b,
    IMonomAuxInterface::MonomIterator d) {
  MP_INT la, lb, gamma;
  a.lc(la);
  b.lc(lb);
  mpz_init(&gamma);
  mpz_gcd(&gamma, &la, &lb);

  IPrime prime;
  ModularShort p = prime.prime();
  while (mpz_divisible_ui_p(&gamma, p)) {
    prime.next();
    p = prime.prime();
  }

  INParameterModularShortAux a_p(a, p);
  INParameterModularShortAux b_p(b, p);

  for (int i = 0; i < a.interface().getEndVar(); i++) {
    IOneParameterModularShortAux evala(p);
    IOneParameterModularShortAux evalb(p);

    do {
      ModularShort* random = getRandomPoints(a.interface().getEndVar() - 1, p);

      a_p.eval(evala, random, i);
      b_p.eval(evalb, random, i);
      delete[] random;
    } while (evala.isZero() || evalb.isZero() || evala.degree() != a_p.deg(i)
        || evalb.degree() != b_p.deg(i));

    evala.gcd(evalb);

    d[i] = evala.degree();
  }

  mpz_clear(&la);
  mpz_clear(&lb);
  mpz_clear(&gamma);
}

/* creates the polynomial x_n - c
 */
void IGcdAlgorithm::createPoly(IOneParameterModularShortAux& f, ModularShort c) {
  f.mHead.clear();
  IOneParameterModularShortAux::Iterator it(f.mHead.begin());

  it.insert();
  (*it).mCoeff = 1;
  (*it).mDegree = 1;

  ++it;

  it.insert();

  (*it).mCoeff = f.primeShort() - c;
  (*it).mCoeff %= f.primeShort();
  if ((*it).mCoeff == 0) {
    it.del();
  } else {
    (*it).mDegree = 0;
  }
}

/* for each even i compute the unique polynomial f such that $f \equiv pols[i] pmod m_1$ 
 * and $f \equiv pols[i+1] pmod m_2$, where $m_2 := x - alpha$
 * The result is stored in pols[i]
 *
 * pols must contain 6 polynomials.
 */
void IGcdAlgorithm::cra(IOneParameterModularShortAux& m1,
    ModularShort alpha,
    INParameterModularShortAux* pols[]) {
  IMonomAuxInterface interface = pols[1]->interface();
  ModularShort p = pols[1]->primeShort();
  IOneParameterModularShortAux c(p);
  IOneParameterModularShortAux d(p);

  IOneParameterModularShortAux m2(p);
  createPoly(m2, alpha);

  eea(m1, m2, c, d);

  for (int i = 0; i < 6; i+= 2) {
    INParameterModularShortAux backup(*pols[i]);
    pols[i]->minus();
    pols[i]->add(*pols[i+1]);
    pols[i]->mult(c);
    // mod m2
    INParameterModularShortAux e(interface, p);
    pols[i]->eval(e, alpha);
    pols[i]->swap(e);
    pols[i]->mult(m1);
    pols[i]->add(backup);
  }
}

/* cf. Labahn, Algorithm 5.2
 */
void IGcdAlgorithm::newton(INParameterModularShortAux& g,
    std::vector<INParameterModularShortAux*>& u,
    std::vector<ModularShort>& alpha) {
  ModularShort p = g.primeShort();
  std::vector<ModularShort> gamma;
  gamma.reserve(alpha.size());

  gamma.push_back(0);
  for (int k = 1; k < alpha.size(); k++) {
    ModularShort product = alpha[k] - alpha[0];
    if (product >= p)
      product -= p;
    else if (product < 0)
      product += p;

    for (int i = 1; i <= k-1; i++) {
      ModularShort tempcoeff = alpha[k] - alpha[i];
      if (tempcoeff >= p)
        tempcoeff -= p;
      else if (tempcoeff < 0)
        tempcoeff += p;
      product *= tempcoeff;
      product %= p;
    }

    product = iInverse(product, p);
    gamma.push_back(product);
  }

  for (int k = 1; k < u.size(); k++) {
    INParameterModularShortAux temp(*(u[k-1]));

    for (int j = k-2; j >= 0; j--) {
      ModularShort tempcoeff = alpha[k] - alpha[j];
      if (tempcoeff >= p)
        tempcoeff-= p;
      else if (tempcoeff < 0)
        tempcoeff+= p;

      temp.mult(tempcoeff);
      temp.add(*(u[j]));
    }

    u[k]->sub(temp);
    u[k]->mult(gamma[k]);
  }

  g.set(*(u[u.size() - 1]));

  for (int k = u.size() - 2; k >= 0; k--) {
    IOneParameterModularShortAux temp(p);
    createPoly(temp, alpha[k]);
    g.mult(temp);
    g.add(*(u[k]));
  }
}

/* integer cra for polynomials.
 * m1 \in Z,
 * m2 \in Z,
 * u1 \in Z[x_1, ..., x_n],
 * u2 \in Z_m2[x_1, ..., x_n]
 *
 * stores the result in u1
 */
void IGcdAlgorithm::cra(MP_INT& m1,
    ModularShort m2,
    INParameterGmpZAux& u1,
    INParameterGmpZAux& u2) {
  Iterator i(u1.mHead.begin());
  ConstIterator ia(u2.mHead.begin());

  while (i && ia) {
    switch (u1.interface().compare((*i).mMonom, (*ia).mMonom)) {
    case 1:
      // the monomial is in u1 but not in u2 -> nothing to do
      MP_INT tmp;
      mpz_init_set_ui(&tmp, 0ul);
      integerCra(m1, m2, (*i).mCoeff, tmp);
      mpz_clear(&tmp);
      ++i;
      break;
    case -1:
      // the monomial is in u2 but not in u1
      i.insert();
      (*i).mMonom = u1.interface().copy((*ia).mMonom);
      mpz_set_ui(&(*i).mCoeff, 0ul);
      integerCra(m1, m2, (*i).mCoeff, (*ia).mCoeff);
      ++i;
      ++ia;
      break;
    default:
      // both polynomials contain this monomial
      integerCra(m1, m2, (*i).mCoeff, (*ia).mCoeff);
      ++i;
      ++ia;
    }
  }

  while (ia) {
    i.insert();
    (*i).mMonom = u1.interface().copy((*ia).mMonom);
    mpz_set_ui(&(*i).mCoeff, 0ul);
    integerCra(m1, m2, (*i).mCoeff, (*ia).mCoeff);
    ++i;
    ++ia;
  }
}

/* integer cra for polynomials. Checks if the polynomial changes
 * m1 \in Z,
 * m2 \in Z,
 * u1 \in Z[x_1, ..., x_n],
 * u2 \in Z_m2[x_1, ..., x_n]
 *
 * stores the result in u1
 */
bool IGcdAlgorithm::cracheck(MP_INT& m1,
    ModularShort m2,
    INParameterGmpZAux& u1,
    INParameterModularShortAux& u2) {
  Iterator i(u1.mHead.begin());
  INParameterModularShortAux::ConstIterator ia(u2.mHead.begin());

  bool changed = false;

  MP_INT tmp;
  mpz_init(&tmp);
  MP_INT oldvalue;
  mpz_init(&oldvalue);

  while (i && ia) {
    switch (u1.interface().compare((*i).mMonom, (*ia).mMonom)) {
    case 1:
      // the monomial is in u1 but not in u2
      mpz_set_ui(&tmp, 0ul);
      mpz_set(&oldvalue, &(*i).mCoeff);
      integerCra(m1, m2, (*i).mCoeff, tmp);
      changed |= (mpz_cmp(&oldvalue, &(*i).mCoeff) != 0);
      ++i;
      break;
    case -1:
      // the monomial is in u2 but not in u1
      i.insert();
      (*i).mMonom = u1.interface().copy((*ia).mMonom);
      mpz_set_ui(&(*i).mCoeff, 0ul);
      mpz_set_ui(&tmp, (*ia).mCoeff);
      integerCra(m1, m2, (*i).mCoeff, tmp);
      changed = true;
      ++i;
      ++ia;
      break;
    default:
      // both polynomials contain this monomial
      mpz_set_ui(&tmp, (*ia).mCoeff);
      mpz_set(&oldvalue, &(*i).mCoeff);
      integerCra(m1, m2, (*i).mCoeff, tmp);
      changed |= (mpz_cmp(&oldvalue, &(*i).mCoeff) != 0);
      ++i;
      ++ia;
    }
  }

  while (ia) {
    i.insert();
    (*i).mMonom = u1.interface().copy((*ia).mMonom);
    mpz_set_ui(&(*i).mCoeff, 0ul);
    mpz_set_ui(&tmp, (*ia).mCoeff);
    integerCra(m1, m2, (*i).mCoeff, tmp);
    ++i;
    ++ia;
  }

  mpz_clear(&tmp);
  mpz_clear(&oldvalue);

  return changed;
}

/* cra for integers
 *
 * Result is stored in u1
 */
void IGcdAlgorithm::integerCra(MP_INT& m1,
    ModularShort m2,
    MP_INT& u1,
    MP_INT& u2) {
  MP_INT m;
  mpz_init_set_ui(&m, m2);

  MP_INT c;
  mpz_init_set(&c, &m1);
  mpz_invert(&c, &c, &m);

  MP_INT v;
  mpz_init_set(&v, &u2);
  mpz_sub(&v, &v, &u1);
  mpz_mul(&v, &v, &c);
  mpz_mod(&v, &v, &m);

  mpz_set(&c, &u1);
  mpz_addmul(&c, &m1, &v);

  mpz_set(&u1, &c);

  // u1 should be in the symmetric range of Z_{m1*m2} (i.e. <= m_1m_2/2)
  mpz_mul(&c, &m1, &m);
  mpz_fdiv_q_2exp(&v, &c, 1);
  if (mpz_cmp(&u1, &v) >= 0)
    mpz_sub(&u1, &u1, &c);

  mpz_clear(&c);
  mpz_clear(&v);
}

void IGcdAlgorithm::gcd(INParameterModularShortAux& a,
    const INParameterModularShortAux& b) {
  IASSERT(a.mPrimeShort == b.mPrimeShort);

  if (a.isZero()) {
    a.set(b);
    return;
  }

  if (b.isZero())
    return;

  if (a.fulldeg() == 0 || b.fulldeg() == 0)
    a.setOne();
  else {
    INParameterModularShortAux tmp(b);
    if (tmp.fulldeg() > a.fulldeg())
      a.mHead.swap(tmp.mHead);

    do {
      a.mHead.swap(tmp.mHead);
      tmp.rem(a);
    } while (!tmp.isZero() && tmp.fulldeg() > 0);

    if (!tmp.isZero())
      a.setOne();
    else
      a.ppcoeff();
  }
}

void IGcdAlgorithm::gcdByFastgcd(INParameterModularShortAux& a,
    const INParameterModularShortAux& b) {
  IASSERT(a.mPrimeShort == b.mPrimeShort);

  INParameterModularShortAux tmp(b);

  a.ppcoeff();
  tmp.ppcoeff();

  if (a.lm()[0] < tmp.lm()[0])
    a.swap(tmp);
  else if (a.lm()[0] == tmp.lm()[0]) {
    tmp.sub(a);
    if (tmp.isZero()) {
      a.ppcoeff();
      return;
    }
  }

  INParameterModularShortAux R[4] = { INParameterModularShortAux(a.interface(),
      a.primeShort()),
      INParameterModularShortAux(a.interface(), a.primeShort()),
      INParameterModularShortAux(a.interface(), a.primeShort()),
      INParameterModularShortAux(a.interface(), a.primeShort()) };

  fastgcd(a, tmp, a.lm()[0], tmp.lm()[0], a.lm()[0], R);
  //std::cout << "we are here!" << std::endl;

  a.mult(R[0]);
  a.addmul(R[1], tmp);
  a.ppcoeff();
}

int IGcdAlgorithm::fastgcd(INParameterModularShortAux& a,
    const INParameterModularShortAux& b,
    int n,
    int m,
    int k,
    INParameterModularShortAux R[]) {

  IASSERT(n >= m);
  //std::cout << "n = " << n << ", m = " << m << ", k = " << k << std::endl;

  if (b.isZero() || m < 0 || k < n - m) {
    R[0].setOne();
    R[1].setZero();
    R[2].setZero();
    R[3].setOne();
    //std::cout << "returning" << std::endl;
    return 0;
  }

  int d = k/2;

  int j = fastgcd(a, b, 2*d, 2*d - n + m, d, R);

  //std::cout << "multipliying(1) ... " << std::endl;
  INParameterModularShortAux rj(R[2]);
  //std::cout << "r_j = " << rj << std::endl;
  //std::cout << "a = " << a << std::endl;
  //std::cout << "n = " << n << std::endl;
  multTruncated(rj, a, n);
  //std::cout << "r_j = " << rj << std::endl;
  //std::cout << "R[3] = " << R[3] << std::endl;
  //std::cout << "b = " << b << std::endl;
  //std::cout << "m = " << m << std::endl;
  addmulTruncated(rj, R[3], b, m);
  //std::cout << "r_j = " << rj << std::endl;
  //std::cout << "finished" << std::endl;

  int nj = (rj.isZero()) ? -1 : rj.lm()[0];

  if (rj.isZero() || k < n - nj)
    return j-1;

  //std::cout << "multipliying(2) ... " << std::endl;
  INParameterModularShortAux rjm(R[0]);
  //std::cout << "r_{j-1} = " << rjm << std::endl;
  //std::cout << "a = " << a << std::endl;
  //std::cout << "n = " << n << std::endl;
  multTruncated(rjm, a, n);
  //std::cout << "r_{j-1} = " << rjm << std::endl;
  //std::cout << "R[1] = " << R[1] << std::endl;
  //std::cout << "b = " << b << std::endl;
  //std::cout << "m = " << m << std::endl;
  addmulTruncated(rjm, R[1], b, m);
  //std::cout << "r_{j-1} = " << rjm << std::endl;
  //std::cout << "finished" << std::endl;

  INParameterModularShortAux qj(a.interface(), a.primeShort());
  INParameterModularShortAux rjp(a.interface(), a.primeShort());

  //std::cout << "polydiv ... " << std::flush;
  polydiv(rjm, rj, qj, rjp);
  //std::cout << "finished" << std::endl;

  ModularShort rho = (rjp.isZero()) ? 1 : rjp.lc();

  if (rho != 1)
    rjp.div(rho);

  int njp = (rjp.isZero()) ? -1 : rjp.lm()[0];

  int ds = k - n + nj;

  INParameterModularShortAux S[4] = { INParameterModularShortAux(a.interface(),
      a.primeShort()),
      INParameterModularShortAux(a.interface(), a.primeShort()),
      INParameterModularShortAux(a.interface(), a.primeShort()),
      INParameterModularShortAux(a.interface(), a.primeShort()) };
  int h = fastgcd(rj, rjp, 2*ds, 2*ds - (nj - njp), ds, S) + j;
  //std::cout << "step 9" << std::endl;

  ModularShort rhoinv = iInverse(rho, a.primeShort());
  //std::cout << "got inverse" << std::endl;

  INParameterModularShortAux Q[4] = { INParameterModularShortAux(a.interface(),
      a.primeShort()),
      INParameterModularShortAux(a.interface(), a.primeShort()),
      INParameterModularShortAux(a.interface(), a.primeShort()),
      INParameterModularShortAux(a.interface(), a.primeShort()) };
  Q[0].setZero();
  Q[1].setOne();
  Q[2].setOne();
  Q[2].mult(rhoinv);
  Q[3].swap(qj);
  Q[3].mult(rhoinv);
  Q[3].minus();

  //std::cout << "matrix mult ... " << std::flush;
  matrixMult(Q, R);
  matrixMult(S, R);
  //std::cout << "finished" << std::endl;

  return h;
}

/* stores the result in b
 */
void IGcdAlgorithm::matrixMult(const INParameterModularShortAux a[],
    INParameterModularShortAux b[]) {
  INParameterModularShortAux tmp(b[0]);

  mult(b[0], a[0]);
  b[0].addmul(a[1], b[2]);

  mult(b[2], a[3]);
  b[2].addmul(a[2], tmp);

  tmp.set(b[1]);

  mult(b[1], a[0]);
  b[1].addmul(a[1], b[3]);
  mult(b[3], a[3]);
  b[3].addmul(a[2], tmp);
}

//void IGcdAlgorithm::invert(INParameterGmpZAux a, const INParameterGmpZAux b, unsigned l) {
//  int r;
//  if (l == 1)
//    r = 0;
//  else {
//    r = 1;
//    l -= 1; // otherwise the result is too big for powers of 2
//    while ((l >>= 1) != 0)
//      ++r;
//  }
//
//  a.setOne();
//
//  MP_INT two;
//  mpz_init_set_ui(&two, 2ul);
//
//  unsigned k = 2;
//
//  for (int i = 0; i < r; i++) {
//    INParameterGmpZAux tmp(a);
//    tmp.mult(a);
//    tmp.mult(b);
//
//    rem(tmp, k);
//
//    a.mult(two);
//    a.sub(tmp);
//    k <<= 1;
//  }
//
//  mpz_clear(&two);
//}
//
//void IGcdAlgorithm::rem(INParameterGmpZAux a, unsigned k) {
//  INParameterGmpZAux::Iterator it = a.begin();
//
//  while (it) {
//    if ((*it).mMonom[0] >= k)
//      it.del();
//    else
//      return;
//  }
//}
//
//void IGcdAlgorithm::rev(INParameterGmpZAux a, const INParameterGmpZAux& b) {
//  a.setZero();
//
//  INParameterGmpZAux::Iterator ia = a.begin();
//  INParameterGmpZAux::ConstIterator ib = b.begin();
//
//  unsigned deg = (*ib).mMonom[0];
//
//  while (ib) {
//    ia.insert();
//    (*ia).mMonom = a.interface().create();
//    (*ia).mMonom[0] = deg - (*ib).mMonom[0];
//    mpz_set(&(*ia).mCoeff, &(*ib).mCoeff);
//
//    ++ib;
//  }
//}
//
//void IGcdAlgorithm::polydiv(INParameterGmpZAux& q, INParameterGmpZAux& r, const INParameterGmpZAux a, const INParameterGmpZAux& b) {
//  unsigned dega = a.lm()[0];
//  unsigned degb = a.lm()[0];
//
//  if (dega < degb) {
//    q.setZero();
//    r.set(a);
//    return;
//  }
//
//  unsigned m = dega - degb;
//
//  INParameterGmpZAux revb(b.interface());
//  INParameterGmpZAux invb(b.interface());
//  rev(revb, b);
//  invert(invb, revb, m+1);
//
//  INParameterGmpZAux qs(a.interface());
//  rev(qs, a);
//  qs.mult(invb);
//  rem(qs, m+1);
//
//  rev(q, qs);
//
//  INParameterGmpZAux tmp(q);
//  tmp.mult(b);
//  r.set(a);
//  r.sub(tmp);
//}


void IGcdAlgorithm::mult(INParameterModularShortAux& a,
    const INParameterModularShortAux& b) {
  if (a.isZero() || b.isZero()) {
    a.setZero();
    return;
  }
  INParameterModularShortAux tmp(a);
  unsigned dega = a.lm()[0];
  unsigned degb = b.lm()[0];
  unsigned max = (dega > degb) ? dega : degb;

  unsigned n = 1;
  while (max != 0) {
    n <<= 1;
    max >>= 1;
  }

  DelegationIterator* ita = new DelegationIterator(tmp.begin());
  DelegationIterator* itb = new DelegationIterator(b.begin());
  karatsuba(a, ita, itb, n);

  delete ita;
  delete itb;
}

/* multiplicates a with b truncated by k. Here b truncated by k is the polynomial of degree k,
 * which coefficients are the first k+1 coefficients of b.
 */
void IGcdAlgorithm::multTruncated(INParameterModularShortAux &a,
    const INParameterModularShortAux &b,
    int k) {
  if (k < 0 || a.isZero()) {
    a.setZero();
    return;
  }

  INParameterModularShortAux tmp(a);
  unsigned dega = a.lm()[0];
  unsigned degb = b.lm()[0];
  int shift = degb - k;
  degb -= shift;
  unsigned max = (dega > degb) ? dega : degb;

  unsigned n = 1;
  while (max != 0) {
    n <<= 1;
    max >>= 1;
  }

  DelegationIterator* ita = new DelegationIterator(tmp.begin());
  ShiftIterator* itb = new ShiftIterator(new DelegationIterator(b.begin()), shift);
  karatsuba(a, ita, itb, n);

  delete ita;
  delete itb;
}

void IGcdAlgorithm::karatsuba(INParameterModularShortAux& result,
    AbstractIterator* f,
    AbstractIterator* g,
    unsigned n) {
  if (n == 1) {
    if (f->isValid() && f->coeff() != 0 && g->isValid() && g->coeff() != 0) {
      result.setOne();
      result.mult(f->coeff());
      result.mult(g->coeff());
    } else {
      result.setZero();
    }

    return;
  }

  unsigned k = n >> 1;

  AbstractIterator* F1 = new ShiftIterator(f->clone(), k);
  AbstractIterator* F0 = f->clone();
  while (F0->isValid() && F0->deg() >= k)
    F0->next();
  AbstractIterator* sumF = new SumIterator(F1->clone(), F0->clone(), result.primeShort());

  AbstractIterator* G1 = new ShiftIterator(g->clone(), k);
  AbstractIterator* G0 = g->clone();
  while (G0->isValid() && G0->deg() >= k)
    G0->next();
  AbstractIterator* sumG = new SumIterator(G1->clone(), G0->clone(), result.primeShort());

  INParameterModularShortAux prod0(result.interface(), result.primeShort());
  INParameterModularShortAux prod1(result.interface(), result.primeShort());
  INParameterModularShortAux sum(result.interface(), result.primeShort());

  karatsuba(prod0, F0, G0, k);
  karatsuba(prod1, F1, G1, k);
  karatsuba(sum, sumF, sumG, k);

  result.set(prod0);
  addmul(result, sum, k);
  submul(result, prod0, k);
  submul(result, prod1, k);
  addmul(result, prod1, n);

  delete F1;
  delete F0;
  delete sumF;
  delete G1;
  delete G0;
  delete sumG;
}

void IGcdAlgorithm::addmul(INParameterModularShortAux& a,
    const INParameterModularShortAux& b,
    unsigned deg) {
  IASSERT(a.primeShort() == b.primeShort());
  It ia = a.begin();
  ConstIt ib = b.begin();
  while (ia && ib) {
    int tmp = ((int) (*ia).mMonom[0]) - ((*ib).mMonom[0] + deg);
    if (tmp > 0) {
      ++ia;
    } else if (tmp < 0) {
      ia.insert();
      (*ia).mCoeff = (*ib).mCoeff;
      (*ia).mMonom = a.interface().copy((*ib).mMonom);
      (*ia).mMonom[0] += deg;
      ++ia;
      ++ib;
    } else {
      (*ia).mCoeff += (*ib).mCoeff;
      if ((*ia).mCoeff >= a.primeShort())
        (*ia).mCoeff -= a.primeShort();

      if ((*ia).mCoeff == 0)
        ia.del();
      else
        ++ia;
      ++ib;
    }
  }
  while (ib) {
    ia.insert();
    (*ia).mCoeff = (*ib).mCoeff;
    (*ia).mMonom = a.interface().copy((*ib).mMonom);
    (*ia).mMonom[0] += deg;
    ++ia;
    ++ib;
  }
  IASSERTVALID(a);
}

void IGcdAlgorithm::submul(INParameterModularShortAux& a,
    const INParameterModularShortAux& b,
    unsigned deg) {
  IASSERT(a.primeShort() == b.primeShort());
  It ia = a.begin();
  ConstIt ib = b.begin();
  while (ia && ib) {
    int tmp = ((int) (*ia).mMonom[0]) - ((*ib).mMonom[0] + deg);
    if (tmp > 0) {
      ++ia;
    } else if (tmp < 0) {
      ia.insert();
      (*ia).mCoeff = a.primeShort() - (*ib).mCoeff;
      (*ia).mMonom = a.interface().copy((*ib).mMonom);
      (*ia).mMonom[0] += deg;
      ++ia;
      ++ib;
    } else {
      (*ia).mCoeff -= (*ib).mCoeff;
      if ((*ia).mCoeff < 0)
        (*ia).mCoeff += a.primeShort();

      if ((*ia).mCoeff == 0)
        ia.del();
      else
        ++ia;
      ++ib;
    }
  }
  while (ib) {
    ia.insert();
    (*ia).mCoeff = a.primeShort() - (*ib).mCoeff;
    (*ia).mMonom = a.interface().copy((*ib).mMonom);
    (*ia).mMonom[0] += deg;
    ++ia;
    ++ib;
  }

  IASSERTVALID(a);
}

//XXX not efficient
void IGcdAlgorithm::addmulTruncated(INParameterModularShortAux &a,
    const INParameterModularShortAux &b,
    const INParameterModularShortAux &c,
    int k) {

  INParameterModularShortAux tmp(b);
  multTruncated(tmp, c, k);
  a.add(tmp);
}
