/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IGCDALGORITHM_H
#define IGCDALGORITHM_H

#include <vector>

#include "inparametergmpzaux.h"
#include "inparametermodularshortaux.h"
#include "inparameterutil.h"

#ifdef TEST
class IGcdAlgorithmTest;
#endif

class IGcdAlgorithm {
#ifdef TEST
  friend class IGcdAlgorithmTest;
#endif
protected:
  class AbstractIterator {
  public:
    virtual ~AbstractIterator() {}

    virtual AbstractIterator* clone() const = 0;

    virtual bool isValid () const = 0;
    virtual void next() = 0;
    virtual ModularShort coeff() = 0;
    virtual unsigned deg() = 0;
  };

  class ShiftIterator : public AbstractIterator {
    AbstractIterator* ma;
    int mShift;

  public:
    ShiftIterator(AbstractIterator *a, unsigned shift) : ma(a), mShift(shift) {}

    ~ShiftIterator() {
      delete ma;
    }

    IGcdAlgorithm::AbstractIterator* clone() const;

    bool isValid() const;
    void next();
    ModularShort coeff();
    unsigned deg();
  };

  class DelegationIterator: public AbstractIterator {
    INParameterModularShortAux::ConstIterator ma;
    
  public:
    DelegationIterator(INParameterModularShortAux::ConstIterator a): ma(a) {}

    AbstractIterator* clone() const;

    bool isValid() const;
    void next();
    ModularShort coeff();
    unsigned deg();
  };

  class SumIterator: public AbstractIterator {
    AbstractIterator *ma;
    AbstractIterator *mb;
    ModularShort mPrimeShort;
    
  public:
    SumIterator(AbstractIterator *a, AbstractIterator *b, ModularShort p): ma(a), mb(b), mPrimeShort(p) {}

    ~SumIterator() {
      delete ma;
      delete mb;
    }

    AbstractIterator* clone() const;

    bool isValid() const;
    void next();
    ModularShort coeff();
    unsigned deg();
  };

  void createPoly(IOneParameterModularShortAux& f, ModularShort c);

  unsigned* sortPoly(
      const INParameterGmpZAux& a, 
      const INParameterGmpZAux& b,
      INParameterGmpZAux& asorted,
      INParameterGmpZAux& bsorted);

  ModularShort* getRandomPoints(int n, ModularShort p);

  void degreeBounds(const INParameterGmpZAux& a, const INParameterGmpZAux& b, IMonomAuxInterface::MonomIterator d);

  void cra(IOneParameterModularShortAux& m1, const ModularShort alpha, INParameterModularShortAux* pols[]);

  void cra(MP_INT& m1, ModularShort m2, INParameterGmpZAux& u1, INParameterGmpZAux& u2);
  bool cracheck(MP_INT& m1, ModularShort m2, INParameterGmpZAux& u1, INParameterModularShortAux& u2);

  void integerCra(MP_INT& m1, ModularShort m2, MP_INT& u1, MP_INT& u2);

  void newton(INParameterModularShortAux& g, std::vector<INParameterModularShortAux*>& u, std::vector<ModularShort>& alpha);


  typedef INParameterGmpZAux::Iterator  Iterator;
  typedef INParameterGmpZAux::ConstIterator  ConstIterator;

  typedef INParameterModularShortAux::Iterator  It;
  typedef INParameterModularShortAux::ConstIterator  ConstIt;

  static void karatsuba(INParameterModularShortAux& result, AbstractIterator* f, AbstractIterator* g, unsigned n);

  static void mult(INParameterModularShortAux& a, const INParameterModularShortAux& b);

  static void multTruncated(INParameterModularShortAux &a, const INParameterModularShortAux &b, int k);
  static void addmulTruncated(INParameterModularShortAux &a, const INParameterModularShortAux &b, const INParameterModularShortAux &c, int k);

  static void addmul(INParameterModularShortAux& a, const INParameterModularShortAux& b, unsigned deg);
  static void submul(INParameterModularShortAux& a, const INParameterModularShortAux& b, unsigned deg);

  static void matrixMult(const INParameterModularShortAux a[], INParameterModularShortAux b[]);

//  static void invert(INParameterGmpZAux a, const INParameterGmpZAux b, unsigned l);
//
//  static void rem(INParameterGmpZAux a, unsigned k);
//
//  static void rev(INParameterGmpZAux a, const INParameterGmpZAux& b);
//
//  static void polydiv(INParameterGmpZAux& q, INParameterGmpZAux& r, const INParameterGmpZAux a, const INParameterGmpZAux& b);

public:
  enum Algorithm {
      Brown,
      Divbro,
      Zippel,
      Heu,
      Onevar
  };

  static IGcdAlgorithm* create(Algorithm alg);
  
  virtual void gcd(INParameterGmpZAux& a, INParameterGmpZAux& b, INParameterGmpZAux& g) = 0;

  /* a = gcd(a, b). Assumes that a and b are polynomials in \F_p[x_mEndVar]
   */
  void gcd(INParameterModularShortAux& a, const INParameterModularShortAux& b);

  void gcdByFastgcd(INParameterModularShortAux& a, const INParameterModularShortAux& b);

  int fastgcd(
      INParameterModularShortAux& a,
      const INParameterModularShortAux& b,
      int n,
      int m,
      int k,
      INParameterModularShortAux R[]);
};

#endif // IGCDALGORITHM_H
