/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <fstream>
#include <sstream>
#include <iomanip>

#include "./igcdbrown.h" 
#include "./inparameterutil.h" 
#include "../util/iprime.h"

void IGcdBrown::gcd(INParameterGmpZAux& a,
    INParameterGmpZAux& b,
    INParameterGmpZAux& g) {
  if (a.isZero()) {
    g = b;
    return;
  }

  if (b.isZero()) {
    g = a;
    return;
  }

  IMonomAuxInterface::MonomIterator dega = a.degrees();
  IMonomAuxInterface::MonomIterator degb = b.degrees();
  IMonomAuxInterface::MonomIterator deggcd = a.interface().copy(dega);
  a.interface().lcm(deggcd, degb);
  unsigned* permutation = a.interface().sortDecreasing(deggcd);

  INParameterGmpZAux tmpa(a.interface());
  a.sort(tmpa, permutation);
  INParameterGmpZAux tmpb(b.interface());
  b.sort(tmpb, permutation);

  brownM(tmpa, tmpb, g);

  unsigned* inverse = a.interface().inversePermutation(permutation);
  INParameterGmpZAux tmpg(g.interface());
  g.sort(tmpg, inverse);
  g.swap(tmpg);

  delete[] inverse;
  delete[] permutation;
  delete[] deggcd;
  delete[] dega;
  delete[] degb;
}

void IGcdBrown::brownP(INParameterModularShortAux& a,
    INParameterModularShortAux& b,
    INParameterModularShortAux& g,
    INParameterModularShortAux& a_,
    INParameterModularShortAux& b_) {

  g.setZero();
  a_.setZero();
  b_.setZero();

  // STEP 1: If n=1 then compute the GCD using Euclid's algorithm
  // and the cofactors by division and return.
  if (a.interface().getEndVar() == 1) {

    INParameterModularShortAux tmpa(a);
    INParameterModularShortAux tmpb(b);

    g = a;

    IGcdAlgorithm::gcd(g, b); // 0 ??

    a_ = a;
    a_.div(g);

    b_ = b;
    b_.div(g);
    return;
  }

  // otherwise compute the contents in Z_p[x_n] and their GCD.
  IOneParameterModularShortAux ca(a.primeShort()), cb(a.primeShort());
  a.pppoly(ca);
  b.pppoly(cb);
  IOneParameterModularShortAux cg(ca);
  cg.gcd(cb);

  // STEP 2: compute the correction coefficients
  //std::cout << "P.2" << std::endl;
  IOneParameterModularShortAux la(a.primeShort()), lb(a.primeShort());
  a.lp(la);
  b.lp(lb);
  IOneParameterModularShortAux gamma(la);
  gamma.gcd(lb);

  // STEP 3: bound the degree in x_n of gamma x g, gamma x a_, gamma x b_
  //std::cout << "P.3" << std::endl;
  unsigned B = a.deg();
  unsigned temp = b.deg();
  if (temp > B)
    B = temp;

  B += gamma.degree();

  // STEP 4:
  //std::cout << "P.4" << std::endl;
  IOneParameterModularShortAux m(a.primeShort());
  m.setOne();

  IMonomAuxInterface::MonomIterator e = a.interface().copy(a.lm());
  if (a.interface().compare(e, b.lm()) > 0) {
    a.interface().set(e, b.lm());
  }

  // LOOP
  INParameterModularShortAux g_m(a.interface(), a.primeShort()), a_m_(
      a.interface(), a.primeShort()), b_m_(a.interface(), a.primeShort());

  for (ModularShort alpha = 1; alpha < a.primeShort(); alpha++) {
    //std::cout << "alpha = " << alpha << std::endl;
    // STEP 6: Choose a new evaluation point alpha in Z_p such that
    // x_n - alpha does not divide la x lb. If no such evaluation
    // point exists then p is to small and the algorithm fails.
    //std::cout << "P.6" << std::endl;

    if (la.eval(alpha) == 0 || lb.eval(alpha) == 0)
      continue;

    // STEP 7: compute a_alpha = a mod <x_n - alpha>, b_alpha.
    //std::cout << "P.7" << std::endl;
    INParameterModularShortAux a_alpha(a.interface(), a.primeShort()), b_alpha(
        a.interface(), a.primeShort());
    a.eval(a_alpha, alpha);
    b.eval(b_alpha, alpha);

    // STEP 8: compute g_alpha, a_alpha_, b_alpha_ by recursion
    //std::cout << "P.8" << std::endl;
    INParameterModularShortAux g_alpha(a_alpha.interface(), a.primeShort()),
        a_alpha_(a_alpha.interface(), a.primeShort()), b_alpha_(
            a_alpha.interface(), a.primeShort());

    brownP(a_alpha, b_alpha, g_alpha, a_alpha_, b_alpha_);

    // STEP 9: Test for a, b relatively prime. If so, output cg, 
    // (ca/cg)a, (cb/cg)b
    //std::cout << "P.9" << std::endl;
    if (g_alpha.isUnit()) {
      a_.set(a);
      ca.div(cg);
      a_.mult(ca);

      b_.set(b);
      cb.div(cg);
      b_.mult(cb);

      g.setOne();
      g.mult(cg);

      delete[] e;

      return;
    }

    // STEP 10: If deg(g_alpha) > e then goto Loop (unlucky evaluation)
    //std::cout << "P.10" << std::endl;
    if (a.interface().compare1(g_alpha.lm(), e) > 0)
      continue;

    // STEP 11: Leading coefficient correction in Z_p[x_n]
    //std::cout << "P.11" << std::endl;
    ModularShort gam = gamma.eval(alpha);
    g_alpha.mult(gam);

    // TODO 8
    //correction of a_alpha, b_alpha!!

    // STEP 12: First Image?
    // If m=1 then set g_m = g_alpha, a_m_ = a_alpha_, b_m_ = g_alpha_, 
    // m = x_n - alpha, e = deg(g_alpha) and goto Loop
    // STEP 13: If deg(g_alpha) < e then all previous evaluations were
    // unlucky. Restart the algorithm keeping only the current evaluation.
    //std::cout << "P.12" << std::endl;
    if (m.isOne() || a.interface().compare1(g_alpha.lm(), e) < 0) {
      createPoly(m, alpha);

      a.interface().set(e, g_alpha.lm());

      g_m.swap(g_alpha);
      a_m_.swap(a_alpha_);
      b_m_.swap(b_alpha_);

      continue;
    }

    // STEP 14: Combine the new image with the old using the CRA
    //std::cout << "P.14" << std::endl;
    INParameterModularShortAux* pols[] = { &a_m_, &a_alpha_, &b_m_, &b_alpha_,
        &g_m, &g_alpha };
    cra(m, alpha, pols);

    // STEP 15: Set m = m x (x_n - alpha). If deg_{x_n}(m) <= B
    // then goto Loop.
    //std::cout << "P.15" << std::endl;
    IOneParameterModularShortAux mult(m.primeShort());
    createPoly(mult, alpha);
    m.mult(mult);

    if (m.degree() <= B)
      continue;

    // STEP 16: Test for termination.
    //std::cout << "P.16" << std::endl;
    if (g_m.deg() + a_m_.deg() == gamma.degree() + a.deg() && g_m.deg()
        + b_m_.deg() == gamma.degree() + b.deg()) {
      g.swap(g_m);
      g.pppoly();
      IOneParameterModularShortAux delta(g.primeShort());
      g.lp(delta);

      a_m_.div(delta);
      b_m_.div(delta);

      g.mult(cg);

      a_.set(ca);
      a_.div(cg);
      a_.mult(a_m_);

      b_.set(cb);
      b_.div(cg);
      b_.mult(b_m_);

      delete[] e;

      return;
    }

    // STEP 17: Goto Loop. All alpha's were unlucky.
    //std::cout << "STEP 17" << std::endl;
  }

  //FIXME: handle the case that p was to small
}

void IGcdBrown::brownM(INParameterGmpZAux& a,
    INParameterGmpZAux& b,
    INParameterGmpZAux& g) {
  g.setZero();

  // STEP 1: Compute integer contents and their GCD
  MP_INT ca, cb, cg;
  a.cont(ca);
  b.cont(cb);
  mpz_init(&cg);
  mpz_gcd(&cg, &ca, &cb);

  if (a.degree() == 0 || b.degree() == 0) {
    g.setOne();
    g.mult(cg);
    mpz_clear(&ca);
    mpz_clear(&cb);
    mpz_clear(&cg);
    return;
  }

  //Achtung: Seiteneffekt (a,b)! Wird vor dem return behoben.
  a.div(ca);
  b.div(cb);

  // STEP 2: compute the correction coefficients
  MP_INT la, lb, gamma;
  a.lc(la);
  b.lc(lb);
  mpz_init(&gamma);
  mpz_gcd(&gamma, &la, &lb);

  // STEP 3: estimate bound for twice the height of gamma x g, 
  // gamma x a_, gamma x b_
  MP_INT B, height_b;
  a.height(B);
  b.height(height_b);
  if (mpz_cmp(&B, &height_b) < 0) {
    mpz_set(&B, &height_b);
  }
  mpz_clear(&height_b);

  //B *= 2*gamma
  mpz_add(&B, &B, &B);
  mpz_mul(&B, &B, &gamma);

  // STEP 4:
  MP_INT m;
  mpz_init_set_ui(&m, 1ul);

  INParameterGmpZAux::MonomIterator e = a.interface().copy(a.lm());
  if (a.interface().compare(e, b.lm()) < 0)
    a.interface().set(e, b.lm());

  //LOOP
  INParameterGmpZAux g_m(a.interface());
  INParameterGmpZAux a_m_(a.interface());
  INParameterGmpZAux b_m_(a.interface());

  IPrime prime;

  while (true) {
    // STEP 6: choose a new prime p that does not divide la or lb

    if (!prime.next()) {
      std::cout << "ERROR! Not enough primes." << std::endl;
    }

    ModularShort p = prime.prime(); //mpz_get_ui(&prime);

    if (mpz_divisible_ui_p(&la, p) || mpz_divisible_ui_p(&lb, p)) {
      continue;
    }

    // STEP 7: compute a_p = a mod p, b_b = b mod p
    //std::cout << "M.7" << std::endl;
    INParameterModularShortAux a_p(a, p);
    INParameterModularShortAux b_p(b, p);

    // STEP 8: compute g_p, a_p_, b_p_ using algorithm p
    //std::cout << "M.8" << std::endl;
    INParameterModularShortAux g_p(a_p.interface(), p);
    INParameterModularShortAux a_p_(a_p.interface(), p);
    INParameterModularShortAux b_p_(a_p.interface(), p);

    brownP(a_p, b_p, g_p, a_p_, b_p_);

    // STEP 9: Test for a, b relatively prime. If so, output cg, 
    // (ca/cg)a, (cb/cg)b
    //std::cout << "M.9" << std::endl;
    if (g_p.fulldeg() == 0) {
      // revert change of a and b
      a.mult(ca);
      b.mult(cb);

      g.setOne();
      g.mult(cg);

      delete[] e;

      mpz_clear(&m);
      mpz_clear(&B);
      mpz_clear(&ca);
      mpz_clear(&cb);
      mpz_clear(&cg);
      mpz_clear(&gamma);
      mpz_clear(&la);
      mpz_clear(&lb);
      return;
    }

    // STEP 10: If deg(g_p) > e then goto Loop (unlucky prime)
    //std::cout << "M.10" << std::endl;
    if (a.interface().compare(g_p.lm(), e) > 0) {
      continue;
    }

    // STEP 11: Leading coefficient correction in Z
    //std::cout << "M.11" << std::endl;
    INParameterGmpZAux g_z(g_p.interface());
    INParameterGmpZAux a_z_(a_p_.interface());
    INParameterGmpZAux b_z_(b_p_.interface());
    g_p.convert(g_z);
    a_p_.convert(a_z_);
    b_p_.convert(b_z_);
    g_z.mult(gamma);
    //correction of a_p_, b_p_

    // STEP 12: First Image?
    // If m=1 then set g_m = g_p, a_m_ = a_p_, b_m_ = g_p_, 
    // m = p, e = deg(g_p) and goto Loop
    // STEP 13: If deg(g_alpha) < e then all previous evaluations were
    // unlucky. Restart the algorithm keeping only the current evaluation.
    //std::cout << "M.12" << std::endl;
    if ((mpz_cmp_ui(&m, 1ul) == 0) || g_z.interface().compare(g_z.lm(), e) < 0) {
      g_m.swap(g_z);
      a_m_.swap(a_z_);
      b_m_.swap(b_z_);

      mpz_set_ui(&m, p);

      a.interface().set(e, g_m.lm());

      continue;
    }

    // STEP 14: Combine the new image with the old using the CRA
    //std::cout << "M.14" << std::endl;
    cra(m, p, a_m_, a_z_);
    cra(m, p, b_m_, b_z_);
    cra(m, p, g_m, g_z);

    // STEP 15: Set m = m x p. If m <= B
    // then goto Loop.
    //std::cout << "M.15" << std::endl;
    mpz_mul_ui(&m, &m, p);

    if (mpz_cmp(&m, &B) <= 0) {
      continue;
    }

    MP_INT norm_ag;
    g_m.norm(a_m_, norm_ag);
    mpz_mul_2exp(&norm_ag, &norm_ag, 1ul);
    MP_INT norm_bg;
    g_m.norm(b_m_, norm_bg);
    mpz_mul_2exp(&norm_bg, &norm_bg, 1ul);

    // STEP 16: Test for termination.
    // Without explicitly computing g_m x a_m_ and g_m x b_m_ determine
    // if |g_m x a_m_| < m/2 and |g_m x b_m_| < m/2
    //std::cout << "M.16" << std::endl;
    if ((mpz_cmp(&m, &norm_ag) >=0) && (mpz_cmp(&m, &norm_bg) >= 0)) {
      // revert change of a and b
      a.mult(ca);
      b.mult(cb);

      //g = g_m;
      g.swap(g_m);
      g.pp();
      g.mult(cg);

      delete[] e;

      mpz_clear(&m);
      mpz_clear(&B);
      mpz_clear(&ca);
      mpz_clear(&cb);
      mpz_clear(&cg);
      mpz_clear(&gamma);
      mpz_clear(&la);
      mpz_clear(&lb);
      return;
    }

    // STEP 17: Goto Loop. All primes were unlucky.
  }
}

