/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <fstream>
#include <sstream>
#include <iomanip>

#include "./igcdheu.h" 
#include "./inparameterutil.h" 
#include "../util/iprime.h"

static inline unsigned max(unsigned a, unsigned b) {
  return (a > b) ? a: b;
}

void IGcdHeu::getXi(MP_INT& xi, INParameterGmpZAux& a, INParameterGmpZAux& b) {
  MP_INT norma;
  MP_INT normb;
  a.height(norma);
  b.height(normb);

  MP_INT beta;
  if (mpz_cmp(&norma, &normb) > 0) 
    mpz_init_set(&beta, &norma);
  else
    mpz_init_set(&beta, &normb);
  mpz_mul_2exp(&beta, &beta, 1);
  mpz_add_ui(&beta, &beta, 29);

  mpz_init(&xi);
  mpz_sqrt(&xi, &beta);
  mpz_cdiv_q_ui(&xi, &xi, 101);
  mpz_mul_ui(&xi, &xi, 10000);

  if (mpz_cmp(&xi, &beta) > 0)
    mpz_set(&xi, &beta);

  MP_INT tmpa;
  MP_INT tmpb;
  a.lc(tmpa);
  b.lc(tmpb);

  mpz_cdiv_q(&tmpa, &norma, &tmpa);
  mpz_cdiv_q(&tmpb, &normb, &tmpb);

  MP_INT tmp;
  if (mpz_cmp(&tmpa, &tmpb) < 0)
    mpz_init_set(&tmp, &tmpa);
  else
    mpz_init_set(&tmp, &tmpb);
  mpz_mul_2exp(&tmp, &tmp, 1);
  mpz_add_ui(&tmp, &tmp, 2);

  if (mpz_cmp(&xi, &tmp) < 0)
    mpz_set(&xi, &tmp);

  mpz_clear(&tmp);
  mpz_clear(&tmpa);
  mpz_clear(&tmpb);
  mpz_clear(&beta);
  mpz_clear(&norma);
  mpz_clear(&normb);
}

/* Horner's scheme
 */
void IGcdHeu::evaluate(MP_INT& result, INParameterGmpZAux& a, MP_INT& xi) {
  if (a.isZero()) {
    mpz_init_set_ui(&result, 0ul);
    return;
  }

  INParameterGmpZAux::ConstIterator it = a.begin();

  mpz_init_set(&result, &(*it).mCoeff);
  unsigned deg = (*it).mMonom[0];
  ++it;

  MP_INT tmp;
  mpz_init(&tmp);
  while (it) {
    int k = (*it).mMonom[0];
    mpz_pow_ui(&tmp, &xi, deg - k);
    mpz_mul(&result, &result, &tmp);
    mpz_add(&result, &result, &(*it).mCoeff);

    deg = k;

    ++it;
  }

  if (deg > 0) {
    mpz_pow_ui(&tmp, &xi, deg);
    mpz_mul(&result, &result, &tmp);
  }

  mpz_clear(&tmp);
}

void IGcdHeu::interpolate(INParameterGmpZAux& g, MP_INT& gamma, MP_INT& xi) {
  g.setZero();
  MP_INT e;
  mpz_init_set(&e, &gamma);

  MP_INT xi2;
  mpz_init(&xi2);
  mpz_cdiv_q_2exp(&xi2, &xi, 1);

  int i = 0;
  INParameterGmpZAux::Iterator it = g.begin();
  while (mpz_cmp_ui(&e, 0ul) != 0) {
    it.insert();
    if (mpz_cmp_ui(&e, 0ul) > 0) {
      mpz_fdiv_qr(&e, &(*it).mCoeff, &e, &xi);
      if (mpz_cmp(&(*it).mCoeff, &xi2) > 0) {
        mpz_sub(&(*it).mCoeff, &(*it).mCoeff, &xi);
        mpz_add_ui(&e, &e, 1ul);
      }
    }
    else {
      mpz_cdiv_qr(&e, &(*it).mCoeff, &e, &xi);
      if (mpz_cmpabs(&(*it).mCoeff, &xi2) > 0) {
        mpz_add(&(*it).mCoeff, &(*it).mCoeff, &xi);
        mpz_sub_ui(&e, &e, 1ul);
      }
    }
    (*it).mMonom = g.interface().create();
    g.interface().setZero((*it).mMonom);
    (*it).mMonom[0] = i;

    if (mpz_cmp_ui(&(*it).mCoeff, 0ul) == 0)
      it.del();
    
    ++i;
  }
  
  mpz_clear(&xi2);
  mpz_clear(&e);
}

void IGcdHeu::gcd(INParameterGmpZAux& a, 
    INParameterGmpZAux& b, 
    INParameterGmpZAux& g) {
  
  MP_INT conta;
  a.cont(conta);
  MP_INT contb;
  b.cont(contb);

  MP_INT cont;
  mpz_init(&cont);
  mpz_gcd(&cont, &conta, &contb);

  a.div(cont);
  b.div(cont);

  MP_INT xi;
  getXi(xi, a, b);

  int i;
  for (i = 0; i < 5; i++) {
    MP_INT evala;
    evaluate(evala, a, xi);
    MP_INT evalb;
    evaluate(evalb, b, xi);

    MP_INT gamma;
    mpz_init(&gamma);
    mpz_gcd(&gamma, &evala, &evalb);

    interpolate(g, gamma, xi);
    g.pp();

//std::cout << "testing divisibility" << std::endl;
    if (a.isdivisible(g) && b.isdivisible(g)) {
      mpz_clear(&evala);
      mpz_clear(&evalb);
      mpz_clear(&gamma);
      
      break;
    }

    INParameterGmpZAux tmpa(a.interface());
    mpz_divexact(&evala, &evala, &gamma);
    interpolate(tmpa, evala, xi);

//std::cout << "testing divisibility (a)" << std::endl;
    // XXX: replace by try
    if (a.isdivisible(tmpa)) {
      g.set(a);
      g.div(tmpa);
      if (b.isdivisible(g)) {
        mpz_clear(&evala);
        mpz_clear(&evalb);
        mpz_clear(&gamma);
        break;
      }
    }
      
    INParameterGmpZAux tmpb(b.interface());
    mpz_divexact(&evalb, &evalb, &gamma);
    interpolate(tmpb, evalb, xi);

//std::cout << "testing divisibility (b)" << std::endl;
    // XXX: replace by try
    if (b.isdivisible(tmpb)) {
      g.set(b);
      g.div(tmpb);
      if (a.isdivisible(g)) {
        mpz_clear(&evala);
        mpz_clear(&evalb);
        mpz_clear(&gamma);
        break;
      }
    }

    MP_INT tmp;
    mpz_init(&tmp);
    mpz_sqrt(&tmp, &xi);
    mpz_sqrt(&tmp, &tmp);
    mpz_cdiv_q_ui(&tmp, &tmp, 27011);
    mpz_mul(&xi, &xi, &tmp);
    mpz_mul_ui(&xi, &xi, 73794);

    mpz_clear(&gamma);
    mpz_clear(&evalb);
    mpz_clear(&evala);
    mpz_clear(&tmp);
  }

  a.mult(cont);
  b.mult(cont);
  g.mult(cont);

  mpz_clear(&xi);
  mpz_clear(&cont);
  mpz_clear(&contb);
  mpz_clear(&conta);

  if (i == 5)
    IERROR("gcdheu failed");
}
