/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "igcdzippel.h" 
#include "imonomaux.h" 
#include "inparameterutil.h" 
#include "../util/iprime.h"
#include "../util/itimer.h"


void IGcdZippel::gcd(INParameterGmpZAux& a, 
    INParameterGmpZAux& b, 
    INParameterGmpZAux& g) {
  
//std::cout << "gcd (n = " << a.interface().getEndVar() << ")" << std::endl;
//std::cout << "a = " << a << std::endl;
//std::cout << "b = " << b << std::endl;

  srand(31013);
  
  if (a.isZero()) {
    g = b;
    return;
  }

  if (b.isZero()) {
    g = a;
    return;
  }

  if (a.degree() == 0) {
    MP_INT la;
    a.lc(la);
    MP_INT cont;
    b.cont(cont);

    mpz_gcd(&cont, &cont, &la);
    g.setOne();
    g.mult(cont);

    mpz_clear(&cont);
    mpz_clear(&la);

//std::cout << "g = " << g << std::endl;
//std::cout << "leaving gcd (n = " << a.interface().getEndVar() << ")" << std::endl;
    return;
  }

  if (b.degree() == 0) {
    MP_INT lb;
    b.lc(lb);
    MP_INT cont;
    a.cont(cont);

    mpz_gcd(&cont, &cont, &lb);
    g.setOne();
    g.mult(cont);

    mpz_clear(&cont);
    mpz_clear(&lb);

//std::cout << "g = " << g << std::endl;
//std::cout << "leaving gcd (n = " << a.interface().getEndVar() << ")" << std::endl;
    return;
  }

  INParameterGmpZAux tmpa(a.interface());
  INParameterGmpZAux tmpb(b.interface());

  unsigned* permutation = sortPoly(a, b, tmpa, tmpb);

  INParameterGmpZAux content(a.interface());
  content.setOne();

  if (a.interface().getEndVar() > 1) {
    removeContent(tmpa, tmpb, content);
  } 
  else {
    MP_INT conta;
    MP_INT contb;
    tmpa.cont(conta);
    tmpa.cont(contb);
    mpz_gcd(&conta, &conta, &contb);

    content.mult(conta);
    
    mpz_clear(&conta);
    mpz_clear(&contb);
  }

//std::cout << "removed content (n = " << a.interface().getEndVar() << ")" << std::endl;

  IMonomAuxInterface::MonomIterator d = g.interface().create();

  degreeBounds(tmpa, tmpb, d);

//std::cout << "got degree bounds (n = " << a.interface().getEndVar() << ")" << std::endl;

  // check whether degree bounds are 0. then the gcd is an integer
  bool isInt = true;
  for (int i = 0; i < tmpa.interface().getEndVar(); i++)
    isInt &= (d[i] == 0);
  
  if (isInt) 
    g.setOne();
  else
    zippelM(tmpa, tmpb, g, d);

  g.mult(content);
  //a.mult(content);
  //b.mult(content);

  unsigned* inverse = a.interface().inversePermutation(permutation);
  INParameterGmpZAux gtmp(g.interface());
  g.sort(gtmp, inverse);
  g.swap(gtmp);

  delete[] d;
  delete[] inverse;
  delete[] permutation;
//std::cout << "g = " << g << std::endl;
//std::cout << "leaving gcd (n = " << a.interface().getEndVar() << ")" << std::endl;
}


void IGcdZippel::removeContent(INParameterGmpZAux& a, 
    INParameterGmpZAux& b, 
    INParameterGmpZAux& content) {
  
  IList<INParameterGmpZAux*> coeffs;

  getCoeffs(a, coeffs);
  getCoeffs(b, coeffs);

  INParameterGmpZAux minimum((*coeffs.begin())->interface());
  getMinimum(coeffs, minimum);

  gmp_randstate_t state;
  gmp_randinit_default(state);
  // use a custom seed to make the choice of values deterministic, i.e.
  // each run of the algorithm with the same parameters is identical
  gmp_randseed_ui(state, 1000000000ul);

  do {
    INParameterGmpZAux linearcombination((*coeffs.begin())->interface());
    getRandomLinearCombination(coeffs, linearcombination, state);

    INParameterGmpZAux candidate((*coeffs.begin())->interface());
    //candidate.gcd(minimum, linearcombination);
    gcd(minimum, linearcombination, candidate);

    convert(candidate, content);
    content.pp();
  } while (!a.isdivisible(content) || !b.isdivisible(content));
  
  gmp_randclear(state);

  MP_INT ca;
  MP_INT cb;
  a.cont(ca);
  b.cont(cb);

  mpz_gcd(&ca, &ca, &cb);

  content.mult(ca);

  a.div(content);
  b.div(content);

  IList<INParameterGmpZAux*>::Iterator it = coeffs.begin();
  while (it) {
    delete (*it);
    ++it;
  }

  mpz_clear(&ca);
  mpz_clear(&cb);
}

void IGcdZippel::getMinimum(IList<INParameterGmpZAux*>& list, 
    INParameterGmpZAux& minimum) {
  
  IList<INParameterGmpZAux*>::Iterator it = list.begin();
  INParameterGmpZAux* result = (*it);
  it.del();

  while (it) {
    if (result->degree() > (*it)->degree()) {
      result->swap(*(*it));
    }
    ++it;
  }

  minimum.swap(*result);
  delete result;
}

void IGcdZippel::getRandomLinearCombination(
    IList<INParameterGmpZAux*>& list, 
    INParameterGmpZAux& linearcombination, 
    gmp_randstate_t state) {
  
  linearcombination.setZero();

  MP_INT max;
  mpz_init_set_ui(&max, 300);

  IList<INParameterGmpZAux*>::Iterator it = list.begin();

  MP_INT rand;
  mpz_init(&rand);
  while (it) {
    mpz_urandomm(&rand, state, &max);
    mpz_add_ui(&rand, &rand, 1ul); //avoid 0
    if (iRandom(2))
      mpz_neg(&rand, &rand);
    linearcombination.addmul(*(*it), rand);
    ++it;
  }
  mpz_clear(&rand);
  mpz_clear(&max);
}

void IGcdZippel::convert(const INParameterGmpZAux& a, 
    INParameterGmpZAux& result) {
  
  result.setZero();

  INParameterGmpZAux::Iterator i = result.begin();
  INParameterGmpZAux::ConstIterator ia = a.begin();

  while (ia) {
    i.insert();
    mpz_set(&(*i).mCoeff, &(*ia).mCoeff);
    (*i).mMonom = result.interface().shiftup((*ia).mMonom);

    ++ia;
    ++i;
  }
}

void IGcdZippel::zippelP(
    INParameterModularShortAux& a,
    INParameterModularShortAux& b,
    INParameterModularShortAux& g,
    IMonomAuxInterface::MonomIterator d,
    bool deterministic) {

//std::cout << "zippelP" << std::endl;
  ModularShort p = a.primeShort();
  g.setZero();

  // STEP -1: If n=1 then compute the GCD using Euclid's algorithm
  // and the cofactors by division and return.
  if (a.interface().getEndVar() == 1) {
    g.interface().setEndVar(1);
    g = a;
    IGcdAlgorithm::gcd(g, b);  

    return;
  }

  // STEP 0: If the gcd of the inputs has content in x_n return fail (g = 0)
  IOneParameterModularShortAux conta(p);
  a.cont(conta);
  if (!conta.isOne()) {
    IOneParameterModularShortAux contb(p);
    b.cont(contb);
    contb.gcd(conta);
    if (!contb.isOne()) {
      return;
    }
  }

  // STEP 1: Compute the scaling factor gamma = gcd(lp(a), lp(b)) \in F_p[x_n]
  IOneParameterModularShortAux gamma(p), tmplp(p);
  a.lp(gamma);
  b.lp(tmplp);
  gamma.gcd(tmplp);

  // STEP 2: Choose v \in F_p - {0} at random such that gamma mod (x_n - v) \neq 0.
  // Set a_v = a mod (x_n - v), b_v = b mod (x_n - v) and compute the gcd g_v
  // with a recursive call.
  // If for n > 2 the algorithm returns fail or for n = 2 we have deg_{x_1}(g_v) > d_{x_1}
  // return fail. Otherwise set d_{x_1} = deg_{x_1}(g_v) and continue

  ModularShort startvalue = iRandom(p);
  ModularShort v = startvalue;

  while (true) {
    try {
      v++;
      v %= p;
      while ((v == 0 || gamma.eval(v) == 0) && v != startvalue) {
        v++;
        v %= p;
      }

      if (v == startvalue)
        IERROR("Not enough primes");


      INParameterModularShortAux a_v(a.interface(), p);
      INParameterModularShortAux b_v(b.interface(), p);
      a.eval(a_v, v);
      b.eval(b_v, v);

      INParameterModularShortAux* g_v = 
        new INParameterModularShortAux(a_v.interface(), p);
      zippelP(a_v, b_v, *g_v, d);

      if (g_v->isZero() 
          || (a.interface().getEndVar() == 2 
              && (*g_v->begin()).mMonom[0] > d[0])) {

        g.setZero();

        delete g_v;
        return;
      }

      d[0] = (*g_v->begin()).mMonom[0];

      // STEP 3
      // TODO
      bool multiple_scaling = true;

      int imagesNeeded = 
        a.interface().deg(d, a.interface().getEndVar() - 1) 
        + gamma.degree() 
        + 1;

      std::vector<INParameterModularShortAux*> gcds;
      std::vector<ModularShort> alpha;
      gcds.reserve(imagesNeeded);
      alpha.reserve(imagesNeeded);

      addToList(g_v, v, gamma, gcds, alpha);

      int images = 0;

      while (images < imagesNeeded && v != startvalue ) {
        v++;
        v %= p;
        while ((v == 0 || gamma.eval(v) == 0) && v != startvalue) {
          v++;
          v %= p;
        }

        if (v == startvalue)
          break;

        a.eval(a_v, v);
        b.eval(b_v, v);

        try {
          INParameterModularShortAux* gn = getGcdByEvaluation(*g_v, a_v, b_v, d);
          gn->ppcoeff();
          addToList(gn, v, gamma, gcds, alpha);
          images++;
        } 
        catch (IUnluckyError) { 
          /* unlucky evaluation. Use another evaluation point v */ 
          //std::cout << "caught it" << std::endl;
        }
      }

      if (v == startvalue)
        IERROR("Not enough primes");

      IGcdAlgorithm::newton(g, gcds, alpha);

      g.pppoly();
      g.ppcoeff();

      if (true) {
        // no probabilistic division test
        if (!a.isdivisible(g) || !b.isdivisible(g)) {
          std::vector<INParameterModularShortAux*>::iterator it = gcds.begin();
          while (it != gcds.end()) {
            delete (*it);
            ++it;
          }
          
          continue;
        }
      }
      else {
        //probabilistic division test
        INParameterModularShortAux g_1(a.interface(), p);
        INParameterModularShortAux a_1(a.interface(), p);
        INParameterModularShortAux b_1(a.interface(), p);

        ModularShort *points = NULL;
        do {
          if (points != NULL)
            delete[] points;
          points = getRandomPoints(a.interface().getEndVar() - 1, p);
          g.eval(g_1, points);
        } while ((*g.begin()).mMonom[0] != (*g_1.begin()).mMonom[0]);
        // XXX handle the case that p is too small

        a.eval(a_1, points);
        b.eval(b_1, points);

        delete[] points;

        if (!a_1.isdivisible(g_1) || !b_1.isdivisible(g_1)) {
          std::vector<INParameterModularShortAux*>::iterator it = gcds.begin();
          while (it != gcds.end()) {
            delete (*it);
            ++it;
          }
          
          continue;
        }
      }

      std::vector<INParameterModularShortAux*>::iterator it = gcds.begin();
      while (it != gcds.end()) {
        delete (*it);
        ++it;
      }
      return;

    } 
    catch (IBadFormError) {
      //std::cout << "bad form (in zippelP)" << std::endl;
    }
    catch (IInconsistentSystemError) {
      //std::cout << "inconsistent (in zippelP)" << std::endl;
    }
  }
}

void IGcdZippel::addToList(
    INParameterModularShortAux* g, 
    ModularShort v, 
    IOneParameterModularShortAux& gamma, 
    std::vector<INParameterModularShortAux*>& gcds, 
    std::vector<ModularShort>& alpha) {

  ModularShort scaling = gamma.eval(v);
  scaling *= iInverse(g->lc(), g->primeShort());
  scaling %= g->primeShort();
  g->mult(scaling);

  gcds.push_back(g);
  alpha.push_back(v);
}

int IGcdZippel::checkForMonomial(INParameterModularShortAux& a) {
  IList<INParameterModularShortAux::Term>::ConstIterator it = a.begin();

  int d = (*it).mMonom[0];
  int terms = 1;

  ++it;

  while (it) {
    if (d == (*it).mMonom[0]) {
      terms++;
    } 
    else {
      if (terms == 1)
        return d;

      terms = 1;
      d = (*it).mMonom[0];
    }

    ++it;
  }

  if (terms == 1)
    return d;

  return -1;
}

int IGcdZippel::numberOfTerms(INParameterModularShortAux& a) {
  IList<INParameterModularShortAux::Term>::ConstIterator it = a.begin();

  int d = (*it).mMonom[0];
  int terms = 1;

  int result = 1;

  ++it;

  while (it) {
    if (d == (*it).mMonom[0]) {
      terms++;
    } 
    else {
      if (terms > result)
        result = terms;

      terms = 1;
      d = (*it).mMonom[0];
    }

    ++it;
  }

  if (terms > result)
    return terms;

  return result;
}

void IGcdZippel::zippelM(INParameterGmpZAux& a, 
    INParameterGmpZAux& b, 
    INParameterGmpZAux& g, 
    IMonomAuxInterface::MonomIterator d) {
  
//std::cout << "zippelM" << std::endl;
  g.setZero();

  // STEP 1: Compute the scaling factor gamma = gcd(lc_x(a), lc_x(b)) \in \Z
  MP_INT la, lb, gamma;
  a.lc(la);
  b.lc(lb);
  mpz_init(&gamma);
  mpz_gcd(&gamma, &la, &lb);

  IPrime prime;
  while (true) {
    try {
      // STEP 2: Choose a random prime p such that gamma_p = gamma mod p \neq 0.
      // Set a_p = a mod p, b_p = b mod p, then compute from these a modular gcd image
      // g_p \in F_p[x_1, ..., x_n] with a call to zippelP.
      // If the algorithm fails, repeat, otherwise set d_x1 = deg_x1(g_p)
      // and continue

      ModularShort p;

      INParameterModularShortAux* g_f = NULL;

      while (g_f == NULL) {
        p = prime.prime();

        if (mpz_divisible_ui_p(&gamma, p)) {
          prime.next();
          // XXX: Handle the case that there are not enough primes
          continue;
        }

        INParameterModularShortAux a_p(a, p);
        INParameterModularShortAux b_p(b, p);

        INParameterModularShortAux g_p(a_p.interface(), p);

//std::cout << "zippelP with" << std::endl;
//std::cout << "p := " << p << ":" << std::endl;
//std::cout << "a := " << a << ":" << std::endl;
//std::cout << "b := " << b << ":" << std::endl;
//std::cout << "a_p := " << a_p << ":" << std::endl;
//std::cout << "b_p := " << b_p << ":" << std::endl;

        zippelP(a_p, b_p, g_p, d);

//std::cout << "zippel returned" << std::endl;
//std::cout << "g_p := " << g_p << ":" << std::endl;

        if (!g_p.isZero()) {
          d[0] = (*g_p.begin()).mMonom[0];
          g_f = new INParameterModularShortAux(a_p.interface(), p);
          g_f->swap(g_p);
        }
        else {
          prime.next();
        }
      }

      // STEP 3: Assume that g_p has no missing terms, and that the prime is not unlucky.
      // We call the assumed form g_f. There are two cases here.
      // 1. If there exists a coefficient of x1 in g_f that is a monomial, then we will
      //    normalize by setting the integer coefficient of that monomial to 1.
      //    Count the largest number of terms in any coefficient of x1 in g_f, calling this nx
      // 2. If there is no such coefficient, then multiple scaling must be used.
      //    Compute the minimum number of images needed to determine g_f for a new prime
      //    with multiple scaling, caling this nx

      // STEP 4: Set g_m = gamma_p/lc(g_p) x g_p mod g_p and m = p
      ModularShort gamma_p = mpz_fdiv_ui(&gamma, p);
      g_f->mult(gamma_p);

      INParameterGmpZAux g_m(a.interface());
      g_f->convert(g_m);
      MP_INT m;
      mpz_init_set_ui(&m, p);

      bool gm_changed = true;
      // STEP 5: Repeat until g_m has stopped changing for one iteration
      while (gm_changed) {
        // STEP 5.1: Choose a new random prime p such that gamma_p = gamma mod p \neq 0
        // and set a_p = a mod p, b_p = b mod p
        prime.next();
        p = prime.prime();
//std::cout << "using prime p = " << p << std::endl;

        if (mpz_divisible_ui_p(&gamma, p)) {
          // XXX: Handle the case that there are not enough primes
          continue;
        }

        ModularShort gamma_p = mpz_fdiv_ui(&gamma, p);

        INParameterModularShortAux a_p(a, p);
        INParameterModularShortAux b_p(b, p);

//        std::cout << "calculating gcd of " << std::endl;
//        std::cout << "*g_f = " << *g_f << std::endl;
//        std::cout << "a_p = " << a_p << std::endl;
//        std::cout << "b_p = " << b_p << std::endl;
//        std::cout << "d = {";
//        for (int i = 0; i < a_p.interface().getEndVar(); i++) {
//          std::cout << d[i] << ", ";
//        }
//        std::cout << "}" << std::endl;
//
        INParameterModularShortAux* gn = getGcdByEvaluation(*g_f, a_p, b_p, d);
        gn->ppcoeff();
        gn->mult(gamma_p);
        gm_changed = cracheck(m, p, g_m, *gn);
        mpz_mul_ui(&m, &m, p);

        delete gn;
      }

      g.set(g_m);
      g.pp();

      if (a.isdivisible(g) && b.isdivisible(g)) {
        mpz_clear(&la);
        mpz_clear(&lb);
        mpz_clear(&gamma);
        mpz_clear(&m);
        delete g_f;
        return;
      }

      //std::cout << "not divisible" << std::endl;
    }
    catch (IBadFormError e) { 
      //std::cout << "bad form" << std::endl;
    }
    catch (IInconsistentSystemError) { }
  }
}

INParameterModularShortAux* IGcdZippel::getGcdByEvaluation(
    INParameterModularShortAux& form, 
    INParameterModularShortAux& a_p, 
    INParameterModularShortAux& b_p, 
    IMonomAuxInterface::MonomIterator d) {

//std::cout << "p := " << a_p.primeShort() << ":" << std::endl;
//std::cout << "a_p := " << a_p << ":" << std::endl;
//std::cout << "b_p := " << b_p << ":" << std::endl;
//std::cout << "form := " << form << ":" << std::endl;

  ModularShort p = a_p.primeShort();


  if (a_p.interface().getEndVar() == 1) {
    INParameterModularShortAux* result = new INParameterModularShortAux(a_p);
    IGcdAlgorithm::gcd(*result, b_p);

    if ((*result->begin()).mMonom[0] < d[0]) {
      d[0] = (*result->begin()).mMonom[0];
      delete result;
      throw IBadFormError();
    }

    if ((*result->begin()).mMonom[0] > d[0]) {
      //std::cout << "throwing unlucky" << std::endl;
      delete result;
      throw IUnluckyError();
    }

    return result;
  } 

  int scalingdeg = checkForMonomial(form);

  IZippelSystem* system;
  
  if (scalingdeg == -1) 
    system = new IMultipleScalingSystem(form, p);
  else
    system = new ISingleScalingSystem(form, scalingdeg, p);

//std::cout << "scaling degree = " << scalingdeg << std::endl;
  bool failedOnce = false;
  while (system->degreeOfFreedom() > 0 
      && system->degreeOfFreedomNotChanged() < 3) {
    
    INParameterModularShortAux g_1(a_p.interface(), p);
    INParameterModularShortAux b_1(a_p.interface(), p);

    ModularShort *points = NULL;
    do {
      if (points != NULL)
        delete[] points;
      points = getRandomPoints(a_p.interface().getEndVar() - 1, p);
      a_p.eval(g_1, points);
      b_p.eval(b_1, points);
    } while (// there must be no degree drop. The cases, where one of the 
              // polynomials can be zero demands a special treatment
      (g_1.isZero() && !a_p.isZero())
      || (!g_1.isZero() && a_p.isZero())
      || (b_1.isZero() && !b_p.isZero())
      || (!b_1.isZero() && b_p.isZero())
      || (*g_1.begin()).mMonom[0] != (*a_p.begin()).mMonom[0] 
      || (*b_1.begin()).mMonom[0] != (*b_p.begin()).mMonom[0]);
    // TODO handle the case that p is too small
    
    IGcdAlgorithm::gcd(g_1, b_1);
    if ((*g_1.begin()).mMonom[0] < d[0]) {
      d[0] = (*g_1.begin()).mMonom[0];
      delete[] points;
      delete system;
      throw IBadFormError();
    }

    if ((*g_1.begin()).mMonom[0] > d[0]) {
      delete[] points;
      
      if (failedOnce) {
        //std::cout << "throwing unluck (failed once)" << std::endl;
        delete system;
        throw IUnluckyError();
      }

      failedOnce = true;
      continue;
    }

    try {
      system->add(g_1, points);
    }
    catch (IInconsistentSystemError) {
      delete[] points;
      delete system;
      throw;
    }

    failedOnce = false;
    delete[] points;
  }

  if (system->degreeOfFreedomNotChanged() >= 3) {
    //std::cout << "throwing unlucky (degree of freedom not changed >= 3)" << std::endl;
    delete system;
    throw IUnluckyError();
  }

  INParameterModularShortAux* result = 
    new INParameterModularShortAux(a_p.interface(), p);
  system->getSolution(*result);

//std::cout << "result = " << *result << std::endl;

  result->ppcoeff();

  delete system;

  IASSERTVALID(*result);
  return result;
}

void IGcdZippel::getCoeffs(const INParameterGmpZAux& a, 
    IList<INParameterGmpZAux*>& coeffs) {
  
  if (a.isZero())
    return;

  IList<INParameterGmpZAux*>::Iterator it = coeffs.begin();
  it.insert();
  (*it) = new INParameterGmpZAux(a.interface().getEndVar() - 1);
  (*it)->setZero();

  const IMonomAuxInterface& interface = (*it)->interface();

  INParameterGmpZAux::ConstIterator i = a.begin();
  IMonomAuxInterface::MonomIterator m = interface.copy(a.lm());

  INParameterGmpZAux::Term t;
  mpz_set(&t.mCoeff, &(*i).mCoeff);
  t.mMonom = interface.shiftdown((*i).mMonom);
  (*it)->add(t);

  ++i;

  while (i) {
    if (interface.deg(m, 0) != interface.deg((*i).mMonom, 0)) {
      ++it;
      it.insert();
      (*it) = new INParameterGmpZAux(interface);
      (*it)->setZero();

      interface.set(m, (*i).mMonom);
    }

    delete[] t.mMonom;

    mpz_set(&t.mCoeff, &(*i).mCoeff);
    t.mMonom = interface.shiftdown((*i).mMonom);
    (*it)->add(t);

    ++i;
  }

  delete[] m;

  // INParameterGmpZAux::~Term() takes care of deleting the monomial.
}

template <>
IAllocator IList<INParameterModularShortAux*>::Node::sAllocator(sizeof(IList<INParameterModularShortAux*>::Node));

template <>
IAllocator IList<INParameterGmpZAux*>::Node::sAllocator(sizeof(IList<INParameterGmpZAux*>::Node));

template <>
IAllocator IList<ModularShort>::Node::sAllocator(sizeof(IList<ModularShort>::Node));
