/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./igmpq.h"


IGmpQInterface::IGmpQInterface():
    ICoeffInterface() {
  mType = ICoeffInterface::GmpQ;
}

ICoeff* IGmpQInterface::create() {
  return new IGmpQ(this);
}

ICoeff* IGmpQInterface::copy(const ICoeff &a) {
  IASSERT(a.type() == ICoeffInterface::GmpQ);
  return new IGmpQ(a);
}

IGmpQ::IGmpQ(IGmpQInterface* r):
   ICoeff(r) {
  mpq_init(&mMP_RAT);
}

IGmpQ::IGmpQ(const ICoeff &a):
   ICoeff(a.coeffInterface()) {
  mpq_init(&mMP_RAT);
  mpq_set(&mMP_RAT, &((const IGmpQ&)a).mMP_RAT);
}

IGmpQ::~IGmpQ() {
  mpq_clear(&mMP_RAT);
}

void IGmpQ::set(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpQ);
  mpq_set(&mMP_RAT, &((const IGmpQ&)a).mMP_RAT);
}

void IGmpQ::setZero() {
  mpq_set_ui(&mMP_RAT, 0ul, 1ul);
}

void IGmpQ::setOne() {
  mpq_set_ui(&mMP_RAT, 1ul, 1ul);
}

bool IGmpQ::isNegative() const {
  return mpq_sgn(&mMP_RAT) < 0;
}

bool IGmpQ::isZero() const {
  return mpq_sgn(&mMP_RAT) == 0;
}

bool IGmpQ::isOne() const {
  return mpq_cmp_ui(&mMP_RAT, 1ul, 1ul) == 0;
}

bool IGmpQ::isComposite() const {
  return mpq_sgn(&mMP_RAT) < 0;
}

bool IGmpQ::isUnit() const {
  return !isZero();
}

int IGmpQ::norm() const {
//  return mpq_sizeinbase(mpq_numref(&mMP_RAT, 2) + mpq_sizeinbase(mpq_denref(&mMP_RAT, 2);
  return (int)(mpz_size(mpq_numref(&mMP_RAT)) + mpz_size(mpq_denref(&mMP_RAT)));
}

void IGmpQ::minus() {
  mpq_neg(&mMP_RAT, &mMP_RAT);
}

void IGmpQ::abs() {
  mpq_abs(&mMP_RAT, &mMP_RAT);
}

void IGmpQ::add(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpQ);
  mpq_add(&mMP_RAT, &mMP_RAT, &((const IGmpQ&)a).mMP_RAT);
}

void IGmpQ::sub(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpQ);
  mpq_sub(&mMP_RAT, &mMP_RAT, &((const IGmpQ&)a).mMP_RAT);
}

void IGmpQ::mult(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpQ);
  mpq_mul(&mMP_RAT, &mMP_RAT, &((const IGmpQ&)a).mMP_RAT);
}

void IGmpQ::diff(int , unsigned ) {
  mpq_set_ui(&mMP_RAT, 0ul, 1ul);
}

void IGmpQ::shift(int , unsigned ) {
}

void IGmpQ::pow(unsigned deg) {
  mpz_pow_ui(mpq_numref(&mMP_RAT), mpq_numref(&mMP_RAT), deg);
  mpz_pow_ui(mpq_denref(&mMP_RAT), mpq_denref(&mMP_RAT), deg);
}

void IGmpQ::inverse() {
  mpq_inv(&mMP_RAT, &mMP_RAT);
}

void IGmpQ::div(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpQ);
  mpq_div(&mMP_RAT, &mMP_RAT, &((const IGmpQ&)a).mMP_RAT);
}

void IGmpQ::div(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::GmpQ);
  IASSERT(b.type() == ICoeffInterface::GmpQ);
  mpq_div(&mMP_RAT, &((const IGmpQ&)a).mMP_RAT, &((const IGmpQ&)b).mMP_RAT);
}

void IGmpQ::gcd(const ICoeff& ) {
  IERROR("IGmpQ::gcd");
}

void IGmpQ::gcd(const ICoeff& , const ICoeff& ) {
  IERROR("IGmpQ::gcd");
}

void IGmpQ::addmul(const ICoeff& a, const ICoeff& b) {
  MP_RAT mp;
  mpq_init(&mp);
  mpq_mul(&mp, &((const IGmpQ&)a).mMP_RAT, &((const IGmpQ&)b).mMP_RAT);
  mpq_add(&mMP_RAT, &mMP_RAT, &mp);
  mpq_clear(&mp);
}

void IGmpQ::read(std::istream& in) {
  in >> (&mMP_RAT);
  mpq_canonicalize(&mMP_RAT);
}

void IGmpQ::write(std::ostream& out) const {
  out << (&mMP_RAT);
}

void IGmpQ::input(std::istream& in, bool expanded) {
  in >> (&mMP_RAT);
  mpq_canonicalize(&mMP_RAT);
}

void IGmpQ::output(std::ostream& out) const {
  out << (&mMP_RAT);
}

IAllocator IGmpQ::sAllocator(sizeof(IGmpQ));
