/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./igmpz.h"

IGmpZInterface::IGmpZInterface():
    ICoeffInterface() {
  mType = ICoeffInterface::GmpZ;
}

ICoeff* IGmpZInterface::create() {
  return new IGmpZ(this);
}

ICoeff* IGmpZInterface::copy(const ICoeff &a) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  return new IGmpZ(a);
}

IGmpZ::IGmpZ(IGmpZInterface* r):
   ICoeff(r) {
  mpz_init(&mMP_INT);
}

IGmpZ::IGmpZ(const ICoeff &a):
   ICoeff(a.coeffInterface()) {
  mpz_init_set(&mMP_INT, &((const IGmpZ&)a).mMP_INT);
}

IGmpZ::~IGmpZ() {
  mpz_clear(&mMP_INT);
}

void IGmpZ::set(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  mpz_set(&mMP_INT, &((const IGmpZ&)a).mMP_INT);
}

void IGmpZ::setZero() {
  mpz_set_ui(&mMP_INT, 0ul);
}

void IGmpZ::setOne() {
  mpz_set_ui(&mMP_INT, 1ul);
}

bool IGmpZ::isNegative() const {
  return mpz_sgn(&mMP_INT) < 0;
}

bool IGmpZ::isZero() const {
  return mpz_sgn(&mMP_INT) == 0;
}

bool IGmpZ::isOne() const {
  return mpz_cmp_ui(&mMP_INT, 1ul) == 0;
}

bool IGmpZ::isComposite() const {
  return mpz_sgn(&mMP_INT) < 0;
}

bool IGmpZ::isUnit() const {
  return !isZero();
}

int IGmpZ::norm() const {
 return mpz_sizeinbase(&mMP_INT, 2);
//   return (int)mpz_size(&mMP_INT);
}

void IGmpZ::minus() {
  mpz_neg(&mMP_INT, &mMP_INT);
}

void IGmpZ::abs() {
  mpz_abs(&mMP_INT, &mMP_INT);
}

void IGmpZ::add(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  mpz_add(&mMP_INT, &mMP_INT, &((const IGmpZ&)a).mMP_INT);
}

void IGmpZ::sub(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  mpz_sub(&mMP_INT, &mMP_INT, &((const IGmpZ&)a).mMP_INT);
}

void IGmpZ::mult(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  mpz_mul(&mMP_INT, &mMP_INT, &((const IGmpZ&)a).mMP_INT);
}

void IGmpZ::diff(int , unsigned ) {
  mpz_set_ui(&mMP_INT, 0ul);
}

void IGmpZ::shift(int , unsigned ) {
}

void IGmpZ::pow(unsigned deg) {
  mpz_pow_ui(&mMP_INT, &mMP_INT, deg);
}

void IGmpZ::inverse() {
  IERROR("IGmpZ::inverse");
}

bool IGmpZ::divisibility(const ICoeff& a) const {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  return mpz_divisible_p(&mMP_INT, &((const IGmpZ&)a).mMP_INT);
}

void IGmpZ::div(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  IASSERT(mpz_divisible_p(&mMP_INT, &((const IGmpZ&)a).mMP_INT));
  mpz_divexact(&mMP_INT, &mMP_INT, &((const IGmpZ&)a).mMP_INT);
}

void IGmpZ::div(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  IASSERT(b.type() == ICoeffInterface::GmpZ);
  IASSERT(mpz_divisible_p(&((const IGmpZ&)a).mMP_INT, &((const IGmpZ&)b).mMP_INT));
  mpz_divexact(&mMP_INT, &((const IGmpZ&)a).mMP_INT, &((const IGmpZ&)b).mMP_INT);
}

void IGmpZ::gcd(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  mpz_gcd(&mMP_INT, &mMP_INT, &((const IGmpZ&)a).mMP_INT);
}

void IGmpZ::gcd(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  IASSERT(b.type() == ICoeffInterface::GmpZ);
  mpz_gcd(&mMP_INT, &((const IGmpZ&)a).mMP_INT, &((const IGmpZ&)b).mMP_INT);
}

void IGmpZ::addmul(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  IASSERT(b.type() == ICoeffInterface::GmpZ);
  mpz_addmul(&mMP_INT, &((const IGmpZ&)a).mMP_INT, &((const IGmpZ&)b).mMP_INT);
}

void IGmpZ::addmul(const ICoeff& a, unsigned long b) {
  IASSERT(a.type() == ICoeffInterface::GmpZ);
  mpz_addmul_ui(&mMP_INT, &((const IGmpZ&)a).mMP_INT, b);
}

void IGmpZ::read(std::istream& in) {
  in >> (&mMP_INT);
}

void IGmpZ::write(std::ostream& out) const {
  out << (&mMP_INT);
}

void IGmpZ::input(std::istream& in, bool expanded) {
  in >> (&mMP_INT);
}

void IGmpZ::output(std::ostream& out) const {
  out << (&mMP_INT);
}

IAllocator IGmpZ::sAllocator(sizeof(IGmpZ));
