/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./igmpzz.h"

IGmpZZInterface::IGmpZZInterface():
    ICoeffInterface() {
  mType = ICoeffInterface::GmpZZ;
}

ICoeff* IGmpZZInterface::create() {
  return new IGmpZZ(this);
}

ICoeff* IGmpZZInterface::copy(const ICoeff &a) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  return new IGmpZZ(a);
}

IGmpZZ::IGmpZZ(IGmpZZInterface* r):
   ICoeff(r) {
  mpz_init(&mMP_INT);
}

IGmpZZ::IGmpZZ(const ICoeff &a):
   ICoeff(a.coeffInterface()) {
  mpz_init_set(&mMP_INT, &((const IGmpZZ&)a).mMP_INT);
}

IGmpZZ::~IGmpZZ() {
  mpz_clear(&mMP_INT);
}

void IGmpZZ::set(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  mpz_set(&mMP_INT, &((const IGmpZZ&)a).mMP_INT);
}

void IGmpZZ::setZero() {
  mpz_set_ui(&mMP_INT, 0ul);
}

void IGmpZZ::setOne() {
  mpz_set_ui(&mMP_INT, 1ul);
}

bool IGmpZZ::isNegative() const {
  return mpz_sgn(&mMP_INT) < 0;
}

bool IGmpZZ::isZero() const {
  return mpz_sgn(&mMP_INT) == 0;
}

bool IGmpZZ::isOne() const {
  return mpz_cmp_ui(&mMP_INT, 1ul) == 0;
}

bool IGmpZZ::isComposite() const {
  return mpz_sgn(&mMP_INT) < 0;
}

bool IGmpZZ::isUnit() const {
  return mpz_cmp_ui(&mMP_INT, 1ul) == 0 ||
         mpz_cmp_si(&mMP_INT, -1l) == 0;
}

int IGmpZZ::norm() const {
 return mpz_sizeinbase(&mMP_INT, 2);
//   return (int)mpz_size(&mMP_INT);
}

int IGmpZZ::cmpNorm(const ICoeff& a) const {
  return mpz_cmpabs(&mMP_INT, &((const IGmpZZ&)a).mMP_INT);
}

void IGmpZZ::minus() {
  mpz_neg(&mMP_INT, &mMP_INT);
}

void IGmpZZ::abs() {
  mpz_abs(&mMP_INT, &mMP_INT);
}

void IGmpZZ::add(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  mpz_add(&mMP_INT, &mMP_INT, &((const IGmpZZ&)a).mMP_INT);
}

void IGmpZZ::sub(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  mpz_sub(&mMP_INT, &mMP_INT, &((const IGmpZZ&)a).mMP_INT);
}

void IGmpZZ::mult(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  mpz_mul(&mMP_INT, &mMP_INT, &((const IGmpZZ&)a).mMP_INT);
}

void IGmpZZ::diff(int , unsigned ) {
  mpz_set_ui(&mMP_INT, 0ul);
}

void IGmpZZ::shift(int , unsigned ) {
}

void IGmpZZ::pow(unsigned deg) {
  mpz_pow_ui(&mMP_INT, &mMP_INT, deg);
}

void IGmpZZ::inverse() {
  IERROR("IGmpZZ::inverse");
}

bool IGmpZZ::divisibility(const ICoeff& a) const {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  return mpz_divisible_p(&mMP_INT, &((const IGmpZZ&)a).mMP_INT);
}

bool IGmpZZ::divisibilityProper(const ICoeff& a) const {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  return mpz_divisible_p(&mMP_INT, &((const IGmpZZ&)a).mMP_INT) &&
         mpz_cmpabs(&mMP_INT, &((const IGmpZZ&)a).mMP_INT) != 0;
}

void IGmpZZ::div(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  IASSERT(mpz_divisible_p(&mMP_INT, &((const IGmpZZ&)a).mMP_INT));
  mpz_divexact(&mMP_INT, &mMP_INT, &((const IGmpZZ&)a).mMP_INT);
}

void IGmpZZ::div(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  IASSERT(b.type() == ICoeffInterface::GmpZZ);
  IASSERT(mpz_divisible_p(&((const IGmpZZ&)a).mMP_INT, &((const IGmpZZ&)b).mMP_INT));
  mpz_divexact(&mMP_INT, &((const IGmpZZ&)a).mMP_INT, &((const IGmpZZ&)b).mMP_INT);
}

void IGmpZZ::quotRem(const ICoeff& a, const ICoeff& b, ICoeff& r) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  IASSERT(r.type() == ICoeffInterface::GmpZZ);
  mpz_tdiv_qr(&mMP_INT, &((IGmpZZ&)r).mMP_INT,  &((const IGmpZZ&)a).mMP_INT, &((const IGmpZZ&)b).mMP_INT);
}

void IGmpZZ::gcd(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  mpz_gcd(&mMP_INT, &mMP_INT, &((const IGmpZZ&)a).mMP_INT);
}

void IGmpZZ::gcd(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  IASSERT(b.type() == ICoeffInterface::GmpZZ);
  mpz_gcd(&mMP_INT, &((const IGmpZZ&)a).mMP_INT, &((const IGmpZZ&)b).mMP_INT);
}

void IGmpZZ::gcdExt(const ICoeff& a, const ICoeff& b, ICoeff& s, ICoeff& t) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  IASSERT(b.type() == ICoeffInterface::GmpZZ);
  IASSERT(s.type() == ICoeffInterface::GmpZZ);
  IASSERT(t.type() == ICoeffInterface::GmpZZ);
  mpz_gcdext(&mMP_INT, &((IGmpZZ&)s).mMP_INT,&((IGmpZZ&)t).mMP_INT, &((const IGmpZZ&)a).mMP_INT, &((const IGmpZZ&)b).mMP_INT);
  //std::cout << *this << " = " << a << " * " << s << " + " << b << " * " << t << std::endl;
}

void IGmpZZ::addmul(const ICoeff& a, const ICoeff& b) {
  IASSERT(a.type() == ICoeffInterface::GmpZZ);
  IASSERT(b.type() == ICoeffInterface::GmpZZ);
  mpz_addmul(&mMP_INT, &((const IGmpZZ&)a).mMP_INT, &((const IGmpZZ&)b).mMP_INT);
}

void IGmpZZ::read(std::istream& in) {
  in >> (&mMP_INT);
}

void IGmpZZ::write(std::ostream& out) const {
  out << (&mMP_INT);
}

void IGmpZZ::input(std::istream& in, bool expanded) {
  in >> (&mMP_INT);
}

void IGmpZZ::output(std::ostream& out) const {
  out << (&mMP_INT);
}

IAllocator IGmpZZ::sAllocator(sizeof(IGmpZZ));
