/***************************************************************************
 *   Copyright (C) 2008 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./inparameterfield.h"

INParameterFieldInterface::INParameterFieldInterface(IMonomInterface* interface) :
  ICoeffInterface(), mi(interface) {

  mType = ICoeffInterface::NParameterField;
}

ICoeff* INParameterFieldInterface::create() {
  return new INParameterField(this, mi);
}

ICoeff* INParameterFieldInterface::copy(const ICoeff& a) {
  IASSERT(a.type() == ICoeffInterface::NParameterField);
  return new INParameterField(a);
}

INParameterField::INParameterField(INParameterFieldInterface* r,
    IMonomInterface* interface) :
  ICoeff(r), mi(interface), mNumer(interface->dimIndepend()),
      mDenom(interface->dimIndepend()) {
}

INParameterField::INParameterField(const ICoeff &a) :
  ICoeff(a.coeffInterface()), mi(((const INParameterField&)a).mi), mNumer(((const INParameterField&)a).mNumer), mDenom(((const INParameterField&)a).mDenom) {
  IASSERT(a.type() == ICoeffInterface::NParameterField);
  IASSERTVALID(*this);
}

INParameterField::~INParameterField() {
}

void INParameterField::set(const ICoeff& a) {
  //std::cout << "INParameterField::set (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::NParameterField);
  mi = ((const INParameterField&)a).mi;
  mNumer = ((const INParameterField&)a).mNumer;
  mDenom = ((const INParameterField&)a).mDenom;
  IASSERTVALID(*this);
  //std::cout << "INParameterField::set (out)" << std::endl;
}

void INParameterField::setZero() {
  //std::cout << "INParameterField::setZero (in)" << std::endl;
  mNumer.setZero();
  mDenom.setOne();
  //std::cout << "INParameterField::setZero (out)" << std::endl;
}

void INParameterField::setOne() {
  //std::cout << "INParameterField::setOne (in)" << std::endl;
  mNumer.setOne();
  mDenom.setOne();
  //std::cout << "INParameterField::setOne (out)" << std::endl;
}

bool INParameterField::isNegative() const {
  //std::cout << "INParameterField::isNegative (in)" << std::endl;
  return (mNumer.isNegative() && !mDenom.isNegative()) || (!mNumer.isNegative()
      && mDenom.isNegative());
  //std::cout << "INParameterField::isNegative (out)" << std::endl;
}

bool INParameterField::isZero() const {
  //std::cout << "INParameterField::isZero (in)" << std::endl;
  return mNumer.isZero();
  //std::cout << "INParameterField::isZero (out)" << std::endl;
}

bool INParameterField::isOne() const {
  //std::cout << "INParameterField::isOne (in)" << std::endl;
  return mNumer.isOne() && mDenom.isOne();
  //std::cout << "INParameterField::isOne (out)" << std::endl;
}

bool INParameterField::isComposite() const {
  //std::cout << "INParameterField::isComposite (in)" << std::endl;
  return mNumer.isComposite() || !mDenom.isOne();
  //std::cout << "INParameterField::isComposite (out)" << std::endl;
}

bool INParameterField::isUnit() const {
  //std::cout << "INParameterField::isUnit (in)" << std::endl;
  // expecting mMinPoly to be irreducible
  return !mNumer.isZero();
  //std::cout << "INParameterField::isUnit (out)" << std::endl;
}

int INParameterField::norm() const {
  //std::cout << "INParameterField::norm (in)" << std::endl;
  IERROR("norm() not implemented in INParameterField");
  //std::cout << "INParameterField::norm (out)" << std::endl;
}

void INParameterField::minus() {
  //std::cout << "INParameterField::minus (in)" << std::endl;
  mNumer.minus();
  //std::cout << "INParameterField::minus (out)" << std::endl;
}

void INParameterField::abs() {
}

void INParameterField::add(const ICoeff& a) {
  //std::cout << "INParameterField::add (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::NParameterField);

  INParameterGmpZAux g(mDenom.interface());
  g.gcdConst(mDenom, ((INParameterField&)a).mDenom);
  
  if (g.isNegative()) {
    g.minus();
  }

  INParameterGmpZAux tmp(mDenom);
  tmp.div(g);
  INParameterGmpZAux tmpa(((INParameterField&)a).mDenom);
  tmpa.div(g);

  mNumer.mult(tmpa);
  mDenom.mult(tmpa);
  tmp.mult(((INParameterField&)a).mNumer);
  mNumer.add(tmp);
  
  g.gcdConst(mNumer, mDenom);
  if (!g.isOne()) {
    mNumer.div(g);
    mDenom.div(g);
  }

  if (mDenom.isNegative()) {
    mDenom.minus();
    mNumer.minus();
  }

  IASSERTVALID(*this);
  //std::cout << "INParameterField::add (out)" << std::endl;
}

void INParameterField::sub(const ICoeff& a) {
  //std::cout << "INParameterField::sub (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::NParameterField);

  INParameterGmpZAux g(mDenom.interface());
  g.gcdConst(mDenom, ((INParameterField&)a).mDenom);
  
  if (g.isNegative()) {
    g.minus();
  }

  INParameterGmpZAux tmp(mDenom);
  tmp.div(g);
  INParameterGmpZAux tmpa(((INParameterField&)a).mDenom);
  tmpa.div(g);

  mNumer.mult(tmpa);
  mDenom.mult(tmpa);
  tmp.mult(((INParameterField&)a).mNumer);
  mNumer.sub(tmp);
  
  g.gcdConst(mNumer, mDenom);
  if (!g.isOne()) {
    mNumer.div(g);
    mDenom.div(g);
  }

  if (mDenom.isNegative()) {
    mDenom.minus();
    mNumer.minus();
  }

  IASSERTVALID(*this);
  //std::cout << "INParameterField::sub (out)" << std::endl;
}

void INParameterField::mult(const ICoeff& a) {
  //std::cout << "INParameterField::mult (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::NParameterField);

  INParameterGmpZAux g1(mNumer.interface());
  g1.gcdConst(mNumer, ((INParameterField&)a).mDenom);
  
  if (g1.isNegative()) {
    g1.minus();
  }

  INParameterGmpZAux g2(mDenom.interface());
  g2.gcdConst(mDenom, ((INParameterField&)a).mNumer);

  if (g2.isNegative()) {
    g2.minus();
  }

  mNumer.div(g1);
  mDenom.div(g2);

  if (!g1.isZero()) {
    INParameterGmpZAux tmp(((INParameterField&)a).mDenom);
    tmp.div(g1);
    mDenom.mult(tmp);
  } else {
    mDenom.mult(((INParameterField&)a).mDenom);
  }

  if (!g2.isZero()) {
    INParameterGmpZAux tmp(((INParameterField&)a).mNumer);
    tmp.div(g2);
    mNumer.mult(tmp);
  } else {
    mNumer.mult(((INParameterField&)a).mNumer);
  }
  
  if (mDenom.isNegative()) {
    mNumer.minus();
    mDenom.minus();
  }

  IASSERTVALID(*this);
  //std::cout << "INParameterField::mult (out)" << std::endl;
}

void INParameterField::diff(int, unsigned) {
  IERROR("diff not implemented in INParameterField");
}

void INParameterField::shift(int, unsigned) {
  IERROR("shift not implemented in INParameterField");
}

void INParameterField::pow(unsigned deg) {
  //std::cout << "INParameterField::pow (in)" << std::endl;
  mNumer.pow(deg);
  mDenom.pow(deg);
  IASSERTVALID(*this);
  //std::cout << "INParameterField::pow (out)" << std::endl;
}

void INParameterField::inverse() {
  //std::cout << "INParameterField::inverse (in)" << std::endl;
  IASSERT(!isZero());

  mNumer.swap(mDenom);

  if (mDenom.isNegative()) {
    mNumer.minus();
    mDenom.minus();
  }
  //std::cout << "INParameterField::inverse (out)" << std::endl;
}

void INParameterField::div(const ICoeff& a) {
  //std::cout << "INParameterField::div (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::NParameterField);

  inverse();
  mult(a);
  inverse();

  IASSERTVALID(*this);
  //std::cout << "INParameterField::div (out)" << std::endl;
}

void INParameterField::div(const ICoeff& a, const ICoeff& b) {
  //std::cout << "INParameterField::div (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::NParameterField);
  IASSERT(b.type() == ICoeffInterface::NParameterField);

  set(a);
  div(b);

  IASSERTVALID(*this);
  //std::cout << "INParameterField::div (out)" << std::endl;
}

void INParameterField::gcd(const ICoeff& a) {
  //std::cout << "INParameterField::gcd (in)" << std::endl;

  IERROR("gcd not implemented in INParameterField");

  //std::cout << "INParameterField::gcd (out)" << std::endl;
}

void INParameterField::gcd(const ICoeff& a, const ICoeff& b) {
  //std::cout << "INParameterField::gcd (in)" << std::endl;
  IERROR("gcd not implemented in INParameterField");
  //std::cout << "INParameterField::gcd (out)" << std::endl;
}

void INParameterField::addmul(const ICoeff& a, const ICoeff& b) {
  //std::cout << "INParameterField::addmul (in)" << std::endl;
  IASSERT(a.type() == ICoeffInterface::NParameterField);
  IASSERT(b.type() == ICoeffInterface::NParameterField);

  INParameterField tmp(a);
  tmp.mult(b);
  add(tmp);

  IASSERTVALID(*this);
  //std::cout << "INParameterField::addmul (out)" << std::endl;
}

void INParameterField::read(std::istream &in) {
  mNumer.read(in, mi);
  mDenom.read(in, mi);
  IASSERTVALID(*this);
}

void INParameterField::write(std::ostream &out) const {
  mNumer.write(out, mi);
  mDenom.write(out, mi);
}

void INParameterField::input(std::istream& in, bool expanded) {
  mNumer.input(in, expanded, mi);
  mDenom.setOne();
}

void INParameterField::output(std::ostream& out) const {
  if (mNumer.isComposite() && !mDenom.isOne()) {
    out << "(";
    mNumer.output(mi, out);
    out << ")";
  } else
    mNumer.output(mi, out);

  if (!mDenom.isOne()) {
    out << "/(";
    mDenom.output(mi, out);
    out << ")";
  }
}

void INParameterField::assertValid(const char* fileName, int fileLine) const {
  mNumer.assertValid(fileName, fileLine);
  mDenom.assertValid(fileName, fileLine);

  IASSERT2(fileName, fileLine, !mDenom.isNegative());
  
  INParameterGmpZAux tmpnumer(mNumer);
  INParameterGmpZAux tmpdenom(mDenom);
  
  
  INParameterGmpZAux tmp(mNumer.interface());
  tmp.gcd(tmpnumer, tmpdenom);
  if (tmp.isNegative())
    tmp.minus();
  
  if (!tmp.isOne()) {
    std::cout << "numer = " << mNumer << std::endl;
    std::cout << "denom = " << mDenom << std::endl;
    std::cout << "gcd = " << tmp << std::endl;
  }
  
  IASSERT2(fileName, fileLine, tmp.isOne());
}

template<>
IAllocator INParameterField::sAllocator(sizeof(INParameterField));
