/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <vector>
#include <algorithm>
#include "inparametergmpz.h"
#include "inparametergmpzaux.h"
#include "inparameterutil.h"
#include "igcdalgorithm.h"

INParameterGmpZAux::INParameterGmpZAux(const INParameterGmpZAux& a) :
  mInterface(a.mInterface), mHead() {
  Iterator i(mHead.begin());
  ConstIterator ia(a.mHead.begin());
  while (ia) {
    i.insert();
    (*i).mMonom = interface().copy((*ia).mMonom);
    mpz_set(&(*i).mCoeff, &(*ia).mCoeff);
    ++i;
    ++ia;
  }
  IASSERTVALID(*this);
}

INParameterGmpZAux::INParameterGmpZAux(const IOneParameterGmpZAux& a) :
  mInterface(1), mHead() {

  Iterator i(mHead.begin());
  IOneParameterGmpZAux::ConstIterator ia(a.begin());
  while (ia) {
    i.insert();
    (*i).mMonom = interface().create();
    (*i).mMonom[0] = (*ia).mDegree;
    mpz_set(&(*i).mCoeff, &(*ia).mCoeff);
    ++i;
    ++ia;
  }
  IASSERTVALID(*this);
}

INParameterGmpZAux::INParameterGmpZAux(const INParameterGmpZAux& a,
    MonomConstIterator m,
    const MP_INT& c) :
  mInterface(a.interface()), mHead() {

  Iterator i(mHead.begin());
  ConstIterator ia(a.mHead.begin());
  while (ia) {
    i.insert();
    (*i).mMonom = interface().copy((*ia).mMonom);
    interface().mult((*i).mMonom, m);
    mpz_mul(&(*i).mCoeff, &(*ia).mCoeff, &c);
    ++i;
    ++ia;
  }
  IASSERTVALID(*this);
}

INParameterGmpZAux::INParameterGmpZAux(const IMonomAuxInterface& interface,
    std::vector<INParameterGmpZAux::Term*> terms) :
  mInterface(interface), mHead() {

  std::sort(terms.begin(), terms.end(), TermCompare(interface));

  INParameterGmpZAux::Iterator ip = mHead.begin();
  for (std::vector<INParameterGmpZAux::Term*>::iterator it = terms.begin(); it
      != terms.end(); it++) {

    // TODO: just insert the list into the polynomial instead of copying
    ip.insert();
    (*ip).mMonom = mInterface.copy((*it)->mMonom);
    mpz_set(&(*ip).mCoeff, &(*it)->mCoeff);
    ++ip;
  }
}

void INParameterGmpZAux::set(const INParameterGmpZAux& a) {
  mHead.clear();
  Iterator i = mHead.begin();
  ConstIterator ia(a.mHead.begin());
  while (ia) {
    i.insert();
    (*i).mMonom = interface().copy((*ia).mMonom);
    mpz_set(&(*i).mCoeff, &(*ia).mCoeff);
    ++i;
    ++ia;
  }
  IASSERTVALID(*this);
}

void INParameterGmpZAux::setOne() {
  mHead.clear();
  Iterator i(mHead.begin());
  i.insert();
  (*i).mMonom = interface().create();
  interface().setZero((*i).mMonom);
  mpz_set_ui(&(*i).mCoeff, 1ul);
  IASSERTVALID(*this);
}

void INParameterGmpZAux::add(const INParameterGmpZAux& a) {
  Iterator i1(mHead.begin());
  ConstIterator i2(a.mHead.begin());
  while (i1 && i2) {
    switch (interface().compare((*i1).mMonom, (*i2).mMonom)) {
    case 1:
      ++i1;
      break;
    case -1:
      i1.insert();
      (*i1).mMonom = interface().copy((*i2).mMonom);
      mpz_set(&(*i1).mCoeff, &(*i2).mCoeff);
      ++i1;
      ++i2;

      break;
    default:
      mpz_add(&(*i1).mCoeff, &(*i1).mCoeff, &(*i2).mCoeff);
      if (mpz_sgn(&(*i1).mCoeff) == 0)
        i1.del();
      else
        ++i1;
      ++i2;
    }
  }

  while (i2) {
    i1.insert();
    (*i1).mMonom = interface().copy((*i2).mMonom);
    mpz_set(&(*i1).mCoeff, &(*i2).mCoeff);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void INParameterGmpZAux::addmul(const INParameterGmpZAux& a, const MP_INT& b) {
  Iterator i1(mHead.begin());
  ConstIterator i2(a.mHead.begin());
  while (i1 && i2) {
    switch (interface().compare((*i1).mMonom, (*i2).mMonom)) {
    case 1:
      ++i1;
      break;
    case -1:
      i1.insert();
      (*i1).mMonom = interface().copy((*i2).mMonom);
      mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &b);
      ++i1;
      ++i2;

      break;
    default:
      mpz_addmul(&(*i1).mCoeff, &(*i2).mCoeff, &b);
      if (mpz_sgn(&(*i1).mCoeff) == 0)
        i1.del();
      else
        ++i1;
      ++i2;
    }
  }

  while (i2) {
    i1.insert();
    (*i1).mMonom = interface().copy((*i2).mMonom);
    mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &b);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void INParameterGmpZAux::add(const Term& t) {
  Iterator i(mHead.begin());

  if (!i) {
    i.insert();
    (*i).mMonom = interface().copy(t.mMonom);
    mpz_set(&(*i).mCoeff, &t.mCoeff);
    IASSERTVALID(*this);
    return;
  }

  while (i)
    switch (interface().compare((*i).mMonom, t.mMonom)) {
    case 1:
      ++i;
      break;
    case -1:
      i.insert();
      (*i).mMonom = interface().copy(t.mMonom);
      mpz_set(&(*i).mCoeff, &t.mCoeff);
      IASSERTVALID(*this);
      return;
    default:
      mpz_add(&(*i).mCoeff, &(*i).mCoeff, &t.mCoeff);
      if (mpz_sgn(&(*i).mCoeff) == 0)
        i.del();
      IASSERTVALID(*this);
      return;
    }

  i.insert();
  (*i).mMonom = interface().copy(t.mMonom);
  mpz_set(&(*i).mCoeff, &t.mCoeff);
  IASSERTVALID(*this);
}

void INParameterGmpZAux::sub(const INParameterGmpZAux& a) {
  Iterator i1(mHead.begin());
  ConstIterator i2(a.mHead.begin());
  while (i1 && i2)
    switch (interface().compare((*i1).mMonom, (*i2).mMonom)) {
    case 1:
      ++i1;
      break;
    case -1:
      i1.insert();
      (*i1).mMonom = interface().copy((*i2).mMonom);
      mpz_neg(&(*i1).mCoeff, &(*i2).mCoeff);
      ++i1;
      ++i2;
      break;
    default:
      mpz_sub(&(*i1).mCoeff, &(*i1).mCoeff, &(*i2).mCoeff);
      if (mpz_sgn(&(*i1).mCoeff) == 0)
        i1.del();
      else
        ++i1;
      ++i2;
    }

  while (i2) {
    i1.insert();
    (*i1).mMonom = interface().copy((*i2).mMonom);
    mpz_neg(&(*i1).mCoeff, &(*i2).mCoeff);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void INParameterGmpZAux::sub(const INParameterGmpZAux& a,
    MonomConstIterator m,
    const MP_INT& coeff) {
  Iterator i1(mHead.begin());
  ConstIterator i2(a.mHead.begin());
  while (i1 && i2)
    switch (interface().compare((*i1).mMonom, (*i2).mMonom, m)) {
    case 1:
      ++i1;
      break;
    case -1:
      i1.insert();
      (*i1).mMonom = interface().copy((*i2).mMonom);
      interface().mult((*i1).mMonom, m);
      mpz_neg(&(*i1).mCoeff, &(*i2).mCoeff);
      mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &coeff);
      ++i1;
      ++i2;
      break;
    default:
      mpz_submul(&(*i1).mCoeff, &(*i2).mCoeff, &coeff);
      if (mpz_sgn(&(*i1).mCoeff) == 0)
        i1.del();
      else
        ++i1;
      ++i2;
    }

  while (i2) {
    i1.insert();
    (*i1).mMonom = interface().copy((*i2).mMonom);
    interface().mult((*i1).mMonom, m);
    mpz_neg(&(*i1).mCoeff, &(*i2).mCoeff);
    mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &coeff);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);

}

void INParameterGmpZAux::mult(const MP_INT& coeff) {
  IASSERT(mpz_sgn(&coeff) != 0);
  ConstIterator i(mHead.begin());
  while (i) {
    mpz_mul(&(*i).mCoeff, &(*i).mCoeff, &coeff);
    ++i;
  }
  IASSERTVALID(*this);
}

void INParameterGmpZAux::mult(MonomConstIterator m) {
  ConstIterator i(mHead.begin());
  while (i) {
    interface().mult((*i).mMonom, m);
    ++i;
  }
  IASSERTVALID(*this);
}

void INParameterGmpZAux::lc(MP_INT& a) const {
  if (isZero())
    mpz_init_set_ui(&a, 0l);
  else
    mpz_init_set(&a, &(*(mHead.begin())).mCoeff);
}

void INParameterGmpZAux::height(MP_INT& a) const {
  ConstIterator i(mHead.begin());

  mpz_init_set_ui(&a, 0ul);

  while (i) {
    if (mpz_cmpabs(&(*i).mCoeff, &a)> 0)
      mpz_set(&a, &(*i).mCoeff);
    ++i;
  }

  mpz_abs(&a, &a);
}

void INParameterGmpZAux::cont(MP_INT &a) const {
  if (isZero()) {
    mpz_init_set_ui(&a, 0l);
    return;
  }

  ConstIterator i(mHead.begin());

  mpz_init_set(&a, &(*i).mCoeff);
  ++i;

  while (i && mpz_cmp_ui(&a, 1ul)) {
    mpz_gcd(&a, &a, &(*i).mCoeff);
    ++i;
  }
}

void INParameterGmpZAux::cont(INParameterGmpZAux& a,
    unsigned* permutation,
    int var) const {

  a.mInterface.setEndVar(a.interface().getEndVar() - var);

  if (isZero()) {
    a.setZero();
    return;
  }

  IList<INParameterGmpZAux*> coeffs;
  getCoeffs(coeffs, permutation, var);

  IList<INParameterGmpZAux*>::ConstIterator it = coeffs.begin();
  a.set(*(*it));
  ++it;

  INParameterGmpZAux tmp(a.interface());

  // TODO: implementation of a more efficient way
  while (it && !a.isOne()) {
    tmp.gcd(a, *(*it));
    a.swap(tmp);
    ++it;
  }

  it = coeffs.begin();
  while (it) {
    delete (*it);
    ++it;
  }
}

INParameterGmpZAux::MonomIterator INParameterGmpZAux::removeTrivialFactors() {
  MonomIterator m;
  if (isZero()) {
    m = mInterface.create();
    mInterface.setZero(m);
    return m;
  }

  ConstIterator it = begin();
  m = mInterface.copy((*it).mMonom);

  ++it;
  while (it && !mInterface.isZero(m)) {
    mInterface.gcd(m, (*it).mMonom);
    ++it;
  }

  div(m);
  IASSERTVALID(*this);

  return m;
}

void INParameterGmpZAux::sort(INParameterGmpZAux& a, unsigned* permutation) const {

  std::vector<INParameterGmpZAux::Term*> terms;

  for (INParameterGmpZAux::ConstIterator it = begin(); it; ++it) {
    INParameterGmpZAux::Term* t = new INParameterGmpZAux::Term();
    t->mMonom = a.interface().copy((*it).mMonom, permutation);
    mpz_set(&t->mCoeff, &(*it).mCoeff);

    terms.push_back(t);
  }

  INParameterGmpZAux tmp(interface(), terms);

  for (std::vector<INParameterGmpZAux::Term*>::iterator it = terms.begin(); it
      != terms.end(); ++it) {

    delete (*it);
  }

  a.swap(tmp);
}

void INParameterGmpZAux::getCoeffs(IList<INParameterGmpZAux*>& coeffs,
    unsigned* permutation,
    int var) const {

  if (isZero())
    return;

  std::vector<INParameterGmpZAux::Term*> terms;

  for (INParameterGmpZAux::ConstIterator i = begin(); i; ++i) {
    INParameterGmpZAux::Term* t = new INParameterGmpZAux::Term();
    t->mMonom = interface().copy((*i).mMonom, permutation);
    mpz_set(&t->mCoeff, &(*i).mCoeff);

    terms.push_back(t);
  }
  std::sort(terms.begin(), terms.end(), TermCompare(interface()));

  IList<INParameterGmpZAux*>::Iterator it = coeffs.begin();
  //  it.insert();
  //  (*it) = new INParameterGmpZAux(interface());
  //  (*it)->mInterface.setEndVar(interface().getEndVar() - var);
  //  (*it)->setZero();

  IMonomAuxInterface interface(this->interface());
  interface.setEndVar(this->interface().getEndVar() - var);

  std::vector<INParameterGmpZAux::Term*>::iterator startPoly = terms.begin();
  IMonomAuxInterface::MonomIterator m = interface.copy((*startPoly)->mMonom);

  std::vector<INParameterGmpZAux::Term*>::iterator i;
  for (i = terms.begin(); i != terms.end(); ++i) {
    if (interface.compareToVar(m, (*i)->mMonom, var) != 0) {
      it.insert();
      (*it) = new INParameterGmpZAux(interface);
      Iterator ip = (*it)->begin();
      while (startPoly != i) {
        ip.insert();
        mpz_set(&(*ip).mCoeff, &(*startPoly)->mCoeff);
        (*ip).mMonom = interface.shiftdown((*startPoly)->mMonom, var);
        ++ip;
        ++startPoly;
      }

      interface.set(m, (*i)->mMonom);
      ++it;
    }
  }

  // TODO: remove code duplication
  if (startPoly != i) {
    it.insert();
    (*it) = new INParameterGmpZAux(interface);
    Iterator ip = (*it)->begin();
    while (startPoly != i) {
      ip.insert();
      mpz_set(&(*ip).mCoeff, &(*startPoly)->mCoeff);
      (*ip).mMonom = interface.shiftdown((*startPoly)->mMonom, var);
      ++ip;
      ++startPoly;
    }
  }

  delete[] m;

  for (i = terms.begin(); i != terms.end(); ++i) {
    delete (*i);
  }
}

void INParameterGmpZAux::mult(const INParameterGmpZAux& a) {
  INParameterGmpZAux prod(interface());
  ConstIterator ia(a.mHead.begin());
  Iterator begin(prod.mHead.begin());

  while (ia) {
    ConstIterator i(mHead.begin());
    Iterator ip = begin;

    // look for the starting point
    while (ip && i && interface().compare((*ip).mMonom, (*i).mMonom, (*ia).mMonom) == 1)
      ++ip;

    begin = ip;

    while (ip && i)
      switch (interface().compare((*ip).mMonom, (*i).mMonom, (*ia).mMonom)) {
      case 1:
        ++ip;
        break;
      case -1:
        ip.insert();
        (*ip).mMonom = interface().copy((*i).mMonom);
        interface().mult((*ip).mMonom, (*ia).mMonom);
        mpz_mul(&(*ip).mCoeff, &(*i).mCoeff, &(*ia).mCoeff);
        IASSERT(mpz_sgn(&(*ip).mCoeff) != 0);
        ++ip;
        ++i;
        break;
      default:
        mpz_addmul(&(*ip).mCoeff, &(*i).mCoeff, &(*ia).mCoeff);
        if (mpz_sgn(&(*ip).mCoeff) == 0)
          ip.del();
        else
          ++ip;
        ++i;
      }

    while (i) {
      ip.insert();
      (*ip).mMonom = interface().copy((*i).mMonom);
      interface().mult((*ip).mMonom, (*ia).mMonom);
      mpz_mul(&(*ip).mCoeff, &(*i).mCoeff, &(*ia).mCoeff);
      IASSERT(mpz_sgn(&(*ip).mCoeff) != 0);
      ++ip;
      ++i;
    }

    ++ia;
  }

  swap(prod);
  IASSERTVALID(*this);
}

void INParameterGmpZAux::div(const MP_INT& coeff) {
  IASSERT(mpz_sgn(&coeff) != 0);
  ConstIterator i(mHead.begin());
  while (i) {
    IASSERT(mpz_divisible_p(&(*i).mCoeff, &coeff));
    mpz_divexact(&(*i).mCoeff, &(*i).mCoeff, &coeff);
    ++i;
  }
  IASSERTVALID(*this);
}

void INParameterGmpZAux::divtry(const MP_INT& coeff) {
  IASSERT(mpz_sgn(&coeff) != 0);
  Iterator i(mHead.begin());
  while (i) {
    if (mpz_divisible_p(&(*i).mCoeff, &coeff)) {
      mpz_divexact(&(*i).mCoeff, &(*i).mCoeff, &coeff);
      ++i;
    } else {
      // revert changes
      Iterator j(mHead.begin());
      while (i != j) {
        mpz_mul(&(*j).mCoeff, &(*j).mCoeff, &coeff);
        ++j;
      }

      throw IDivError();
    }
  }
  IASSERTVALID(*this);
}

bool INParameterGmpZAux::isdivisible(const MP_INT& coeff) const {
  IASSERT(mpz_sgn(&coeff) != 0);
  ConstIterator i(mHead.begin());
  while (i) {
    if (!mpz_divisible_p(&(*i).mCoeff, &coeff))
      return false;
    ++i;
  }
  return true;
}

void INParameterGmpZAux::div(const INParameterGmpZAux& a) {
  if (a.degree() == 0) {
    MP_INT lca;
    a.lc(lca);
    div(lca);
    mpz_clear(&lca);
    IASSERTVALID(*this);
    return;
  }

  INParameterGmpZAux result(interface());
  MP_INT lcthis;
  MP_INT lca;

  Iterator it = result.mHead.begin();

  while (!mHead.isEmpty()) {
    it.insert();
    lc(lcthis);
    a.lc(lca);

    (*it).mMonom = interface().create();
    interface().divide((*it).mMonom, lm(), a.lm());
    mpz_tdiv_q(&(*it).mCoeff, &lcthis, &lca);

    sub(a, (*it).mMonom, (*it).mCoeff);

    ++it;
    mpz_clear(&lcthis);
    mpz_clear(&lca);
  }

  swap(result);
  IASSERTVALID(*this);
}

void INParameterGmpZAux::div(MonomConstIterator m) {
  if (mInterface.isZero(m))
    return;

  Iterator it = begin();

  while (it) {
    mInterface.divide((*it).mMonom, m);
    ++it;
  }

  IASSERTVALID(*this);
}

void INParameterGmpZAux::divtry(const INParameterGmpZAux& a) {
  if (a.degree() == 0) {
    MP_INT lca;
    a.lc(lca);
    divtry(lca);
    mpz_clear(&lca);
    return;
  }

  INParameterGmpZAux backup(*this);

  INParameterGmpZAux result(interface());
  MP_INT lcthis;
  MP_INT lca;

  Iterator it = result.mHead.begin();

  while (!mHead.isEmpty()) {
    it.insert();
    lc(lcthis);
    a.lc(lca);

    (*it).mMonom = interface().create();
    if (interface().divisibility(lm(), a.lm()) && mpz_divisible_p(&lcthis, &lca)) {
      interface().divide((*it).mMonom, lm(), a.lm());
      mpz_tdiv_q(&(*it).mCoeff, &lcthis, &lca);
    } else {
      mpz_clear(&lcthis);
      mpz_clear(&lca);
      swap(backup);
      throw IDivError();
    }

    // maybe implementation of a method sub(poly, monom, coeff)
    //INParameterGmpZAux temp(a, (*it).mMonom, (*it).mCoeff);
    sub(a, (*it).mMonom, (*it).mCoeff);

    ++it;
    mpz_clear(&lcthis);
    mpz_clear(&lca);
  }

  swap(result);
}

void INParameterGmpZAux::gcdConst(const INParameterGmpZAux& a,
    const INParameterGmpZAux& b) {
  INParameterGmpZAux tmpa(a);
  INParameterGmpZAux tmpb(b);

  gcd(tmpa, tmpb);
}

void INParameterGmpZAux::gcd(INParameterGmpZAux& a, INParameterGmpZAux& b) {
  if (a.isZero()) {
    set(b);
    return;
  }

  if (b.isZero()) {
    set(a);
    return;
  }

  MonomIterator ma = a.removeTrivialFactors();
  MonomIterator mb = b.removeTrivialFactors();

  if (a.equalsUpToSign(b)) {
    set(a);
  }
  else if (a.interface().getEndVar() == 1) {
    if (a.interface().isZero(a.lm()) || b.interface().isZero(b.lm()))
      integerGcd(a, b);
    else if (a.isSparse() && b.isSparse())
      IGcdAlgorithm::create(IGcdAlgorithm::Onevar)->gcd(a, b, *this);
    else
      IGcdAlgorithm::create(IGcdAlgorithm::Heu)->gcd(a, b, *this);
  } 
  else {
    MonomIterator dega = a.degrees();
    MonomIterator degb = b.degrees();
    MonomIterator deggcd = a.interface().copy(dega);
    MonomIterator deglcm = a.interface().copy(dega);
    a.interface().gcd(deggcd, degb);
    a.interface().lcm(deglcm, degb);

    unsigned* permutationlcm = a.interface().sortZeroesThenDecreasing(deglcm);
    unsigned* permutationgcd = a.interface().sortZeroesThenDecreasing(deggcd);

    // check, whether some variables only occur in one polynomial
    // if so,  compute the gcd of the contents of the two polynomials,
    // where the contents is in the polynomial ring of the variables,
    // which occur in both a and b
    if (deggcd[permutationgcd[a.interface().getEndVar() - 1]] == 0) {
      // the two polynomials don't share any variables
      integerGcd(a, b);
    } else if (deglcm[permutationlcm[0]] == 0) {
      // there are some unused variables
      int zeroes = 0;
      while (deglcm[permutationlcm[zeroes]] == 0)
        zeroes++;

      INParameterGmpZAux conta(a.interface());

      a.cont(conta, permutationlcm, zeroes);

      if (conta.isOne()) {
        setOne();
      } else if (conta.interface().isZero(conta.lm())) {
        // gcd is an integer
        integerGcd(conta, b);
      } else {
        INParameterGmpZAux contb(b.interface());
        b.cont(contb, permutationlcm, zeroes);

        INParameterGmpZAux tmp(interface());
        tmp.mInterface.setEndVar(conta.interface().getEndVar());
        tmp.gcd(conta, contb);
        tmp.shiftup(zeroes);

        unsigned* inverse = interface().inversePermutation(permutationlcm);
        tmp.sort(*this, inverse);
        delete[] inverse;
      }
    } else {
      //std::cout << "choosing algorithm ... " << std::flush;
      IGcdAlgorithm* alg;
      if (a.isSparse() || b.isSparse()) {
        //std::cout << "using Zippel..." << std::endl;

        //        std::cout << "Zippel with " << std::endl;
        //        std::cout << "a = " << a << std::endl;
        //        std::cout << "b = " << b << std::endl;
        alg = IGcdAlgorithm::create(IGcdAlgorithm::Zippel);
      } else {
        //std::cout << "using Brown..." << std::endl;
        alg = IGcdAlgorithm::create(IGcdAlgorithm::Brown);
      }

      alg->gcd(a, b, *this);
    }

    delete[] permutationlcm;
    delete[] permutationgcd;
    delete[] deglcm;
    delete[] deggcd;
    delete[] degb;
    delete[] dega;
  }

  MonomIterator mg = a.interface().copy(ma);
  a.interface().gcd(mg, mb);

  mult(mg);
  a.mult(ma);
  b.mult(mb);

  delete[] mg;
  delete[] mb;
  delete[] ma;
}

void INParameterGmpZAux::integerGcd(INParameterGmpZAux& a, INParameterGmpZAux& b) {

  MP_INT conta;
  a.cont(conta);
  MP_INT contb;
  b.cont(contb);

  setZero();
  Iterator it = mHead.begin();
  it.insert();
  (*it).mMonom = mInterface.create();
  mInterface.setZero((*it).mMonom);

  mpz_gcd(&(*it).mCoeff, &conta, &contb);

  mpz_clear(&contb);
  mpz_clear(&conta);
}

bool INParameterGmpZAux::isSparse() const {
  // TODO: a more sophisticated method to measure spareseness
  MonomIterator deg = degrees();

  unsigned degree = 0;
  for (int i = 0; i < mInterface.getEndVar(); i++) {
    degree += deg[i];
  }

  delete[] deg;

  MP_INT maxTerms;
  mpz_init(&maxTerms);
  mpz_bin_uiui(&maxTerms,
      degree + mInterface.getEndVar(),
      mInterface.getEndVar());

  // allow 20% of the maximal possible terms.
  mpz_fdiv_q_ui(&maxTerms, &maxTerms, 5ul);

  bool result = (mpz_cmp_ui(&maxTerms, mHead.length()) >= 0);
  mpz_clear(&maxTerms);

  return result;
}

void INParameterGmpZAux::shiftup(int var) {
  mInterface.setEndVar(mInterface.getEndVar() + var);
  ConstIterator it = begin();

  while (it) {
    interface().shiftupInplace((*it).mMonom, var);
    ++it;
  }
}

bool INParameterGmpZAux::isdivisible(const INParameterGmpZAux& a) const {
  if (a.degree() == 0) {
    MP_INT lca;
    a.lc(lca);
    bool result = isdivisible(lca);
    mpz_clear(&lca);
    return result;
  }

  INParameterGmpZAux copy(*this);

  MP_INT lcthis;
  MP_INT lca;

  MP_INT coeff;
  mpz_init(&coeff);
  IMonomAuxInterface::MonomIterator monom;

  while (!copy.mHead.isEmpty()) {
    copy.lc(lcthis);
    a.lc(lca);

    monom = interface().create();
    if (interface().divisibility(copy.lm(), a.lm()) && mpz_divisible_p(&lcthis, &lca)) {
      interface().divide(monom, copy.lm(), a.lm());
      mpz_tdiv_q(&coeff, &lcthis, &lca);
    } else {
      mpz_clear(&lcthis);
      mpz_clear(&lca);

      delete[] monom;
      mpz_clear(&coeff);
      return false;
    }

    copy.sub(a, monom, coeff);

    mpz_clear(&lcthis);
    mpz_clear(&lca);
    delete[] monom;
  }

  mpz_clear(&coeff);
  return true;
}

int INParameterGmpZAux::norm() const {
  ConstIterator i(mHead.begin());
  int max = -1;
  while (i) {
    int n= mpz_size(&(*i).mCoeff);
    if (max < 0 || n> max)
    max = n;
    ++i;
  }
  return max;
  /*
   int sum = 0;
   while(i) {
   sum += mpz_size(&(*i).mCoeff);
   ++i;
   }
   return sum;
   */
}

/* assumes the worst-case: each coefficients of the polynomials has
 * the same modulus.
 * 
 * is there a better way?
 */
void INParameterGmpZAux::norm(INParameterGmpZAux& a, MP_INT& norm) const {
  MP_INT temp;
  a.height(temp);
  height(norm);
  mpz_mul(&norm, &norm, &temp);
  mpz_clear(&temp);

  for (int i = 0; i < a.interface().getEndVar(); i++) {
    unsigned deg = degree(i);
    unsigned deg_a = a.degree(i);
    if (deg_a < deg)
      deg = deg_a;
    mpz_mul_ui(&norm, &norm, deg + 1);
  }
}

void INParameterGmpZAux::oneNorm(MP_INT& norm) const {
  mpz_init_set_ui(&norm, 0ul);

  ConstIterator it = mHead.begin();

  MP_INT tmp;
  mpz_init(&tmp);
  while (it) {
    mpz_abs(&tmp, &(*it).mCoeff);
    mpz_add(&norm, &norm, &tmp);
    ++it;
  }
  mpz_clear(&tmp);
}

/* returns the degree in the end variable
 */
unsigned INParameterGmpZAux::degree() const {
  IASSERT(!isZero());
  ConstIterator i(mHead.begin());
  unsigned result = 0;
  while (i) {
    if (interface().degree((*i).mMonom) > result)
      result = interface().degree((*i).mMonom);
    ++i;
  }
  return result;
}

/* returns the degree in variable var
 */
unsigned INParameterGmpZAux::degree(int var) const {
  unsigned max = 0;

  ConstIterator i(mHead.begin());
  while (i) {
    if (interface().deg((*i).mMonom, var) > max)
      max = interface().deg((*i).mMonom, var);
    ++i;
  }

  return max;
}

INParameterGmpZAux::MonomIterator INParameterGmpZAux::degrees() const {
  MonomIterator result = mInterface.create();
  mInterface.setZero(result);

  for (ConstIterator it = mHead.begin(); it; ++it) {
    mInterface.lcm(result, (*it).mMonom);
  }

  return result;
}

void INParameterGmpZAux::minus() {
  ConstIterator i(mHead.begin());
  while (i) {
    mpz_neg(&(*i).mCoeff, &(*i).mCoeff);
    ++i;
  }
  IASSERTVALID(*this);
}

void INParameterGmpZAux::diff() {
}

void INParameterGmpZAux::diff(unsigned deg) {
}

void INParameterGmpZAux::pow(unsigned deg) {
  IASSERT(deg> 0);
  if (deg > 1) {
    INParameterGmpZAux p(*this);
    while (--deg > 0) {
      p.mult(*this);
    }
    mHead.swap(p.mHead);
  }
  IASSERTVALID(*this);
}

void INParameterGmpZAux::lp(INParameterGmpZAux& a) const {
  ConstIterator i = mHead.begin();

  MonomIterator m = interface().copy((*i).mMonom);

  a.setZero();
  Iterator ia = a.mHead.begin();
  while (i && interface().compare1(m, (*i).mMonom) == 0) {
    ia.insert();
    (*ia).mMonom = interface().create();
    interface().setZero((*ia).mMonom);
    interface().prolong((*ia).mMonom, interface().mEndVar, interface().deg((*i).mMonom, interface().mEndVar));
    mpz_set(&(*ia).mCoeff, &(*i).mCoeff);
    ++i;
    ++ia;
  }

  delete[] m;
}

bool INParameterGmpZAux::equalsUpToSign(const INParameterGmpZAux& a) const {
  if (interface().getEndVar() != a.interface().getEndVar()) {
    return false;
  }

  if (isZero())
    return a.isZero();
  else if (a.isZero())
    return false;

  ConstIterator i = mHead.begin();
  ConstIterator ia = a.mHead.begin();

  bool differentSign = (mpz_sgn(&(*i).mCoeff) + mpz_sgn(&(*ia).mCoeff) == 0);

  if (differentSign) {
    while (i && ia) {
      if (mpz_cmpabs(&(*i).mCoeff, &(*ia).mCoeff) 
          || (mpz_sgn(&(*i).mCoeff) + mpz_sgn(&(*ia).mCoeff) != 0)
          || mInterface.compare((*i).mMonom, (*ia).mMonom))
        return false;
      ++i;
      ++ia;
    }
  } else {
    while (i && ia) {
      if (mpz_cmp(&(*i).mCoeff, &(*ia).mCoeff) || mInterface.compare((*i).mMonom, (*ia).mMonom))
        return false;
      ++i;
      ++ia;
    }
  }

  if (i || ia)
    return false;

  return true;
}

void INParameterGmpZAux::rem(const INParameterGmpZAux& a) {
  IASSERT(interface().getEndVar() == 1 && a.interface().getEndVar() == 1);

  if (isZero() || a.isZero())
    return;

  MP_INT k1, k2, tmp;
  mpz_init(&k1);
  mpz_init(&k2);
  mpz_init(&tmp);

  while (!isZero() && (*begin()).mMonom[0] >= (*a.begin()).mMonom[0]) {
    Iterator i1(mHead.begin());
    ConstIterator i2(a.mHead.begin());

    unsigned deg = (*begin()).mMonom[0] - (*a.begin()).mMonom[0];
    //std::cout << "(*i1).mCoeff.length = " << (*i1).mCoeff._mp_size;
    //std::cout << "\n(*i2).mCoeff.length = " << (*i2).mCoeff._mp_size << std::endl;
    mpz_gcd(&tmp, &(*i1).mCoeff, &(*i2).mCoeff);
    mpz_tdiv_q(&k1, &(*i2).mCoeff, &tmp);
    mpz_tdiv_q(&k2, &(*i1).mCoeff, &tmp);
    mpz_neg(&k2, &k2);

    i1.del();
    ++i2;

    while (i1 && i2) {
      unsigned degree2 = (*i2).mMonom[0] + deg;
      if ((*i1).mMonom[0] > degree2) {
        mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &k1);
        ++i1;
      } else if ((*i1).mMonom[0] < degree2) {
        i1.insert();
        (*i1).mMonom = mInterface.create();
        (*i1).mMonom[0] = degree2;
        mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &k2);
        ++i1;
        ++i2;
      } else {
        mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &k1);
        mpz_addmul(&(*i1).mCoeff, &(*i2).mCoeff, &k2);
        if (mpz_sgn(&(*i1).mCoeff) == 0)
          i1.del();
        else
          ++i1;
        ++i2;
      }
    }
    while (i2) {
      i1.insert();
      (*i1).mMonom = mInterface.create();
      (*i1).mMonom[0] = (*i2).mMonom[0] + deg;
      mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &k2);
      ++i1;
      ++i2;
    }
    while (i1) {
      mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &k1);
      ++i1;
    }
    IASSERTVALID(*this);
  }
  mpz_clear(&k1);
  mpz_clear(&k2);
  mpz_clear(&tmp);
}

void INParameterGmpZAux::assertValid(const char* fileName, int fileLine) const {
  if (!mHead.isEmpty()) {
    ConstIterator i(mHead.begin());
    IASSERT2(fileName, fileLine, mpz_sgn(&(*i).mCoeff) != 0);
    ConstIterator prev(i);
    ++i;
    while (i) {
      IASSERT2(fileName, fileLine, mpz_sgn(&(*i).mCoeff) != 0);
      prev = i;
      ++i;
      if (i) {
        IASSERT2(fileName, fileLine, interface().compare((*prev).mMonom, (*i).mMonom)> 0);
      }
    }
  }
}

void INParameterGmpZAux::read(std::istream &in, IMonomInterface* mi) {
  mHead.clear();
  int l;
  in.read((char*)(&l), sizeof(l));

  IMonom* temp = mi->create();
  Iterator i(mHead.begin());
  while (l) {
    i.insert();
    (*i).mMonom = interface().create();
    temp->read(in);
    in >> (&(*i).mCoeff);
    interface().set((*i).mMonom, *temp);
    ++i;
    --l;
  }
  delete temp;
  IASSERTVALID(*this);
}

void INParameterGmpZAux::write(std::ostream &out, IMonomInterface* mi) const {
  int l = mHead.length();
  out.write((char*)(&l), sizeof(l));
  IMonom* temp = mi->create();
  ConstIterator i(mHead.begin());
  while (i) {
    interface().set(*temp, (*i).mMonom);
    temp->write(out);
    out << (&(*i).mCoeff);
    ++i;
  }
  delete temp;
}

//void INParameterGmpZAux::input(std::istream& in) {
//  setZero();
//  Iterator i(begin());
//
//  if ((in >> std::ws).peek() == '(')
//    in.get();
//  while ((in >> std::ws).peek() != ')') {
//    i.insert();
//    in >> (&(*i).mCoeff);
//    (*i).mMonom = create();
//    if ((in >> std::ws).peek() == '*') {
//      in.get();
//      in >> std::ws;
//      in >> *(*i).mMonom;
//    } else {
//      (*i).mMonom->setZero();
//    }
//    
//    ++i;
//  }
//
//    IASSERTVALID(*this);
//}

void INParameterGmpZAux::input(std::istream& in,
    bool expanded,
    IMonomInterface* mi) {
  setZero();

  if (expanded) {
    IMonom* temp = mi->create();

    std::vector<INParameterGmpZAux::Term*> terms;

    bool neg;

    do {
      INParameterGmpZAux::Term* t = new INParameterGmpZAux::Term();
      t->mMonom = interface().create();

      neg = false;
      if ((in >> std::ws).peek() == '-') {
        neg = true;
        in.get();
      } else if (in.peek() == '+')
        in.get();

      std::streampos pos = in.tellg();
      in >> (&t->mCoeff);
      if (in.fail()) {
        in.clear();
        in.seekg(pos);
        in.setstate(std::ios::goodbit);
        mpz_set_ui(&t->mCoeff, 1ul);

        in >> *temp;
        if (!in.fail()) {
          interface().set(t->mMonom, *temp);
        } else
          IERROR("expected 'monomial' or 'coefficient'");
      } else if ((in >> std::ws).peek() == '*') {
        in.get();

        in >> *temp;
        if (!in.fail()) {
          interface().set(t->mMonom, *temp);
        } else {
          in.clear();
          interface().setZero(t->mMonom);
        }
      } else {
        interface().setZero(t->mMonom);
      }

      if (neg)
        mpz_neg(&t->mCoeff, &t->mCoeff);

      terms.push_back(t);
    } while ((in >> std::ws).peek() == '+' || in.peek() == '-');

    INParameterGmpZAux tmp(interface(), terms);
    swap(tmp);

    delete temp;
  } else {
    INParameterGmpZAux::Term t;
    t.mMonom = interface().create();
    std::streampos pos = in.tellg();
    in >> (&t.mCoeff);
    if (in.fail()) {
      in.clear();
      in.seekg(pos);
      in.setstate(std::ios::goodbit);
      mpz_set_ui(&t.mCoeff, 1ul);

      IMonom* temp = mi->create();
      in >> *temp;
      interface().set(t.mMonom, *temp);
      delete temp;
    } else {
      interface().setZero(t.mMonom);
    }

    if (mpz_sgn(&t.mCoeff) != 0)
      add(t);
  }

  IASSERTVALID(*this);
}

void INParameterGmpZAux::output(IMonomInterface* mi, std::ostream& out) const {
  if (isZero())
    out << '0';
  else {
    MP_INT mp;
    mpz_init(&mp);
    ConstIterator i(begin());
    if (mpz_cmp_ui(&(*i).mCoeff, 1ul) == 0) {
      if (interface().degree((*i).mMonom))
        interface().output((*i).mMonom, mi, out);
      else
        out << '1';
    } else {
      out << (&(*i).mCoeff);
      if (interface().degree((*i).mMonom)) {
        out << '*';
        interface().output((*i).mMonom, mi, out);
      }
    }

    ++i;
    while (i) {
      if (mpz_sgn(&(*i).mCoeff)> 0) {
        out << " + ";
        if (mpz_cmp_ui(&(*i).mCoeff, 1ul) == 0) {
          if (interface().degree((*i).mMonom))
            interface().output((*i).mMonom, mi, out);
          else
            out << '1';
        } else {
          out << &(*i).mCoeff;
          if (interface().degree((*i).mMonom)) {
            out << '*';
            interface().output((*i).mMonom, mi, out);
          }
        }
      } else {
        out << " - ";
        if (mpz_cmp_si(&(*i).mCoeff, -1) == 0) {
          if (interface().degree((*i).mMonom))
            interface().output((*i).mMonom, mi, out);
          else
            out << '1';
        } else {
          mpz_neg(&mp, &(*i).mCoeff);
          out << &mp;
          if (interface().degree((*i).mMonom)) {
            out << '*';
            interface().output((*i).mMonom, mi, out);
          }
        }
      }
      ++i;
    }
    mpz_clear(&mp);
  }
}

void INParameterGmpZAux::output(std::ostream& out) const {
  if (isZero())
    out << '0';
  else {
    MP_INT mp;
    mpz_init(&mp);
    ConstIterator i(begin());
    if (mpz_cmp_ui(&(*i).mCoeff, 1ul) == 0) {
      if (interface().degree((*i).mMonom))
        interface().output((*i).mMonom, out);
      else
        out << '1';
    } else {
      out << (&(*i).mCoeff);
      if (interface().degree((*i).mMonom)) {
        out << '*';
        interface().output((*i).mMonom, out);
      }
    }

    ++i;
    while (i) {
      if (mpz_sgn(&(*i).mCoeff)> 0) {
        out << " + ";
        if (mpz_cmp_ui(&(*i).mCoeff, 1ul) == 0) {
          if (interface().degree((*i).mMonom))
            interface().output((*i).mMonom, out);
          else
            out << '1';
        } else {
          out << &(*i).mCoeff;
          if (interface().degree((*i).mMonom)) {
            out << '*';
            interface().output((*i).mMonom, out);
          }
        }
      } else {
        out << " - ";
        if (mpz_cmp_si(&(*i).mCoeff, -1) == 0) {
          if (interface().degree((*i).mMonom))
            interface().output((*i).mMonom, out);
          else
            out << '1';
        } else {
          mpz_neg(&mp, &(*i).mCoeff);
          out << &mp;
          if (interface().degree((*i).mMonom)) {
            out << '*';
            interface().output((*i).mMonom, out);
          }
        }
      }
      ++i;
    }
    mpz_clear(&mp);
  }
}

template <> IAllocator
    IList<INParameterGmpZAux::Term>::Node::sAllocator(sizeof(IList<INParameterGmpZAux::Term>::Node));
