/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef INPARAMETERGMPZAUX_H
#define INPARAMETERGMPZAUX_H

#include <gmp.h>
#include <vector>

#include "../util/iutil.h"
#include "../util/imodular.h"
#include "../util/ilist.h"

#include "icoeff.h"
#include "idiverror.h"
#include "imonomaux.h"
#include "ioneparametergmpzaux.h"
#include "../monom/imonom.h"

class INParameterModularShortAux;
class INParameterGmpZ;

class INParameterGmpZAux {
  friend class IGcdAlgorithm;
  friend class IGcdZippel;
  friend class IGcdOnevar;
  friend class INParameterModularShortAux;
  friend class INParameterGmpZ;
private:
  struct Term {
    IMonomAuxInterface::MonomIterator  mMonom;
    MP_INT   mCoeff;

    Term(): mMonom(NULL) {
      mpz_init(&mCoeff);
    }
    ~Term() {
      delete[] mMonom;
      mpz_clear(&mCoeff);
    }
  };

  class TermCompare : public std::binary_function<INParameterGmpZAux::Term*, INParameterGmpZAux::Term*, bool> {
    IMonomAuxInterface mInterface;

  public:
    TermCompare(const IMonomAuxInterface& interface): mInterface(interface) { }

    bool operator()(INParameterGmpZAux::Term* x, INParameterGmpZAux::Term* y) { 
      return mInterface.compare(x->mMonom, y->mMonom) > 0;
    }
  };

  IList<Term> mHead;
  IMonomAuxInterface mInterface;

public:
  typedef IList<Term>::Iterator  Iterator;
  typedef IList<Term>::ConstIterator  ConstIterator;
  typedef IMonomAuxInterface::MonomIterator MonomIterator;
  typedef IMonomAuxInterface::MonomConstIterator MonomConstIterator;

  INParameterGmpZAux(const IMonomAuxInterface& interface):
      mInterface(interface),
      mHead() {
  }

  INParameterGmpZAux(int n):
      mInterface(n),
      mHead() {
  }

  const IMonomAuxInterface& interface() const {
    return mInterface;
  }

  INParameterGmpZAux(const INParameterGmpZAux& a);
  INParameterGmpZAux(const IOneParameterGmpZAux& a);
  INParameterGmpZAux(const INParameterGmpZAux& a, MonomConstIterator m, const MP_INT& c);
  //INParameterGmpZAux(const INParameterModularShortAux& a);

  INParameterGmpZAux(const IMonomAuxInterface& interface, std::vector<INParameterGmpZAux::Term*> terms);

  void add(const Term& t);
  void add(const INParameterGmpZAux& a);
  void sub(const INParameterGmpZAux& a);
  void sub(const INParameterGmpZAux& a, MonomConstIterator m, const MP_INT& coeff);
  void mult(const MP_INT& coeff);
  void mult(MonomConstIterator m);
  void mult(const INParameterGmpZAux& a);
  void div(const INParameterGmpZAux& a);
  void div(MonomConstIterator m);
  void divtry(const INParameterGmpZAux& a);
  bool isdivisible(const INParameterGmpZAux& a) const;
  void cont(MP_INT& a) const;
  void cont(INParameterGmpZAux& a, unsigned* permutation, int var) const;
  void pp() {
    MP_INT content;
    cont(content);
    div(content);
    mpz_clear(&content);
  }

  MonomIterator removeTrivialFactors();

  void sort(INParameterGmpZAux& a, unsigned* permutation) const;

  void getCoeffs(IList<INParameterGmpZAux*>& coeffs, 
      unsigned* permutation,
      int var) const;
  
  //void nextcoeff(IList<Term>::ConstIterator& i, INParameterGmpZAux& coeff, int var);
  
  void lc(MP_INT& a) const;
  MonomIterator lm() const {
    IASSERT(!isZero());
    return (*mHead.begin()).mMonom;
  }
  void lp(INParameterGmpZAux& a) const;
  void height(MP_INT& a) const;
  void mult(unsigned degree, const MP_INT& coeff);
  void div(const MP_INT& coeff);
  void divtry(const MP_INT& coeff);
  bool isdivisible(const MP_INT& coeff) const;
//  void gcdBrown(const INParameterGmpZAux& a);
public:

  ~INParameterGmpZAux() {
    mHead.clear();
  }

  void operator=(const INParameterGmpZAux& a) {
    set(a);
//    mHead = a.mHead;
  }

  void swap(INParameterGmpZAux& a) {
    mHead.swap(a.mHead);
  }

  virtual void set(const INParameterGmpZAux& a);

  void setZero() { mHead.clear(); }
  void setOne();

  bool isZero() const { return mHead.isEmpty(); }
  bool isOne() const {
    ConstIterator i(mHead.begin());
    return i && 
      interface().degree((*i).mMonom) == 0 
      && mpz_cmp_ui(&(*i).mCoeff, 1ul) == 0;
  }
  bool isNegative() const {
    ConstIterator i(mHead.begin());
    return i && mpz_sgn(&(*i).mCoeff) < 0;
  }
  bool isComposite() const {
    ConstIterator i(mHead.begin());
    return mHead.length() > 1 || (i && mpz_sgn(&(*i).mCoeff) < 0);
  }
  int norm() const;
  void norm(INParameterGmpZAux& a, MP_INT& norm) const;
  void oneNorm(MP_INT& norm) const;
  unsigned degree() const;
  unsigned degree(int var) const;
  MonomIterator degrees() const;

  void minus();
  void mult(const INParameterGmpZAux& a, const INParameterGmpZAux& b);
  void diff();
  void diff(unsigned deg);
  void pow(unsigned deg);
  void div(MP_INT& multiplier, const INParameterGmpZAux& a);
  void div(MP_INT& multiplier, const INParameterGmpZAux& a, const INParameterGmpZAux& b) {
    *this = a;
    div(multiplier, b);
  }

  void gcdConst(const INParameterGmpZAux& a, const INParameterGmpZAux& b);

  void gcd(INParameterGmpZAux& a, INParameterGmpZAux& b);
  void integerGcd(INParameterGmpZAux& a, INParameterGmpZAux& b);
  
  bool isSparse() const;
  
  void shiftup(int var);
  
  void addmul(const INParameterGmpZAux& a, const INParameterGmpZAux& b) {
    INParameterGmpZAux tmp(a);
    tmp.mult(b);
    add(tmp);
  }

  void addmul(const INParameterGmpZAux& a, const MP_INT& b);
  
  bool equalsUpToSign(const INParameterGmpZAux& a) const;
  
  void rem(const INParameterGmpZAux& a);

  ConstIterator begin() const { return mHead.begin(); }
  Iterator begin() { return mHead.begin(); }

  void assertValid(const char* fileName, int fileLine) const;

  void read(std::istream &in, IMonomInterface* mi);
  void write(std::ostream &out, IMonomInterface* mi) const;

  void input(std::istream& in, bool expanded, IMonomInterface* mi);
//  void input(std::istream& in);
  void output(IMonomInterface* mi, std::ostream& out) const;

  void output(std::ostream& out) const;

  friend std::ostream& operator<<(std::ostream& out, const INParameterGmpZAux& a) {
    a.output(out);
    return out;
  }
};

#endif // INPARAMETERGMPZAUX_H
