/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./ioneparametergmpzaux.h"
#include "./inparametergmpzaux.h"

IOneParameterGmpZAux::IOneParameterGmpZAux(const INParameterGmpZAux& a) {
  IASSERT(a.interface().getEndVar() == 1);
  
  INParameterGmpZAux::ConstIterator ia = a.begin();
  Iterator i = mHead.begin();
  
  while (ia) {
    i.insert();
    (*i).mDegree = (*ia).mMonom[0];
    mpz_set(&(*i).mCoeff, &(*ia).mCoeff);
    
    ++i;
    ++ia;
  }
  
  IASSERTVALID(*this);
}


void IOneParameterGmpZAux::add(IOneParameterGmpZAux& a) {
  Iterator i1(mHead.begin());
  Iterator i2(a.mHead.begin());
  while (i1 && i2) {
    if ((*i1).mDegree > (*i2).mDegree) {
      ++i1;
    } else if ((*i1).mDegree < (*i2).mDegree) {
      i1.insert(i2.get());
      ++i1;
    } else {
      mpz_add(&(*i1).mCoeff, &(*i1).mCoeff, &(*i2).mCoeff);
      if (mpz_sgn(&(*i1).mCoeff) == 0)
        i1.del();
      else
        ++i1;
      ++i2;
    }
  }
  while (i2) {
    i1.insert(i2.get());
    ++i1;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::sub(IOneParameterGmpZAux& a) {
  Iterator i1(mHead.begin());
  Iterator i2(a.mHead.begin());
  while (i1 && i2) {
    if ((*i1).mDegree > (*i2).mDegree) {
      ++i1;
    } else if ((*i1).mDegree < (*i2).mDegree) {
      i1.insert(i2.get());
      mpz_neg(&(*i1).mCoeff, &(*i1).mCoeff);
      ++i1;
    } else {
      mpz_sub(&(*i1).mCoeff, &(*i1).mCoeff, &(*i2).mCoeff);
      if (mpz_sgn(&(*i1).mCoeff) == 0)
        i1.del();
      else
        ++i1;
      ++i2;
    }
  }
  while (i2) {
    i1.insert(i2.get());
    mpz_neg(&(*i1).mCoeff, &(*i1).mCoeff);
    ++i1;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::mult(const MP_INT& coeff) {
  IASSERT(mpz_sgn(&coeff) != 0);
  ConstIterator i(mHead.begin());
  while (i) {
    mpz_mul(&(*i).mCoeff, &(*i).mCoeff, &coeff);
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::shiftMult(int coeff) {
  IASSERT(!isZero());
  IASSERT(0 < coeff);

  unsigned long c = coeff;
  MP_INT tmp;
  mpz_init(&tmp);

  mHead.begin().insert();

  Iterator i(mHead.begin());
  Iterator j(mHead.begin());
  ++i;
  (*j).mDegree = (*i).mDegree + 1;
  mpz_set_ui(&(*i).mCoeff, 1ul);
  ++j;
  do {
    ++i;
    if (i) {
      if ((*i).mDegree != (*j).mDegree - 1) {
        i.insert();
        (*i).mDegree = (*j).mDegree - 1;
      }
      mpz_mul_ui(&tmp, &(*j).mCoeff, c);
      mpz_sub(&(*j).mCoeff, &(*i).mCoeff, &tmp);
      if (mpz_sgn(&(*j).mCoeff) == 0)
        j.del();
      else
        ++j;
    } else {
      mpz_mul_ui(&(*j).mCoeff, &(*j).mCoeff, c);
      if (mpz_sgn(&(*j).mCoeff) == 0)
        j.del();
      else
        mpz_neg(&(*j).mCoeff, &(*j).mCoeff);
      break;
    }
  } while (true);
  mpz_clear(&tmp);
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::eval(MP_INT &sum, int t) const {
  IASSERT(0 < t);

  MP_INT tmp;
  mpz_init(&tmp);

  ConstIterator i(mHead.begin());
  mpz_set_ui(&sum, 0ul);
  while (i) {
    mpz_ui_pow_ui(&tmp, t, (*i).mDegree);
    mpz_addmul(&sum, &(*i).mCoeff, &tmp);
    ++i;
  }
  mpz_clear(&tmp);
}

void IOneParameterGmpZAux::cont(MP_INT &a) const {
  if (isZero()) {
    mpz_init_set(&a, 0ul);
    return;
  }

  ConstIterator i(mHead.begin());

  mpz_init_set(&a, &(*i).mCoeff);
  ++i;

  while (i && mpz_cmp_ui(&a, 1ul)) {
    mpz_gcd(&a, &a, &(*i).mCoeff);
    ++i;
  }
}

void IOneParameterGmpZAux::mult(unsigned degree, const MP_INT& coeff) {
  IASSERT(mpz_sgn(&coeff) != 0);
  ConstIterator i(mHead.begin());
  while (i) {
    (*i).mDegree += degree;
    mpz_mul(&(*i).mCoeff, &(*i).mCoeff, &coeff);
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::div(const MP_INT& coeff) {
  IASSERT(mpz_sgn(&coeff) != 0);
  ConstIterator i(mHead.begin());
  while (i) {
    IASSERT(mpz_divisible_p(&(*i).mCoeff, &coeff));
    mpz_divexact(&(*i).mCoeff, &(*i).mCoeff, &coeff);
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::div(unsigned degree, const MP_INT& coeff) {
  IASSERT(mpz_sgn(&coeff) != 0);
  ConstIterator i(mHead.begin());
  while (i) {
    IASSERT((*i).mDegree >= degree);
    (*i).mDegree += degree;
    IASSERT(mpz_divisible_p(&(*i).mCoeff, &coeff));
    mpz_divexact(&(*i).mCoeff, &(*i).mCoeff, &coeff);
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::rem(const IOneParameterGmpZAux& a) {
  if (isZero() || a.isZero())
    return;

  MP_INT k1, k2, tmp;
  mpz_init(&k1);
  mpz_init(&k2);
  mpz_init(&tmp);

  while (!isZero() && degree() >= a.degree()) {
    Iterator i1(mHead.begin());
    ConstIterator i2(a.mHead.begin());

    unsigned deg = degree() - a.degree();
    //std::cout << "(*i1).mCoeff.length = " << (*i1).mCoeff._mp_size;
    //std::cout << "\n(*i2).mCoeff.length = " << (*i2).mCoeff._mp_size << std::endl;
    mpz_gcd(&tmp, &(*i1).mCoeff, &(*i2).mCoeff);
    mpz_tdiv_q(&k1, &(*i2).mCoeff, &tmp);
    mpz_tdiv_q(&k2, &(*i1).mCoeff, &tmp);
    mpz_neg(&k2, &k2);

    i1.del();
    ++i2;

    while (i1 && i2) {
      unsigned degree2 = (*i2).mDegree + deg;
      if ((*i1).mDegree > degree2) {
        mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &k1);
        ++i1;
      } else if ((*i1).mDegree < degree2) {
        i1.insert();
        (*i1).mDegree = degree2;
        mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &k2);
        ++i1;
        ++i2;
      } else {
        mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &k1);
        mpz_addmul(&(*i1).mCoeff, &(*i2).mCoeff, &k2);
        if (mpz_sgn(&(*i1).mCoeff) == 0)
          i1.del();
        else
          ++i1;
        ++i2;
      }
    }
    while (i2) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree + deg;
      mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &k2);
      ++i1;
      ++i2;
    }
    while (i1) {
      mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &k1);
      ++i1;
    }
    IASSERTVALID(*this);
  }
  mpz_clear(&k1);
  mpz_clear(&k2);
  mpz_clear(&tmp);
}

void IOneParameterGmpZAux::pseudorem(const IOneParameterGmpZAux& a,
    const MP_INT& alpha) {
  IASSERT(mpz_sgn(&alpha) != 0);

  mult(alpha);
  rem(a);
}

void IOneParameterGmpZAux::mod(MP_INT& q) {
  Iterator i(mHead.begin());

  while (i) {
    mpz_mod(&(*i).mCoeff, &(*i).mCoeff, &q);
    if (mpz_cmp_ui(&(*i).mCoeff, 0ul) == 0)
      i.del();
    else
      ++i;
  }

  IASSERTVALID(*this);
}

int IOneParameterGmpZAux::norm() const {
  ConstIterator i(mHead.begin());
  int max = -1;
  while (i) {
    int n= mpz_size(&(*i).mCoeff);
    if (max < 0 || n> max)
    max = n;
    ++i;
  }
  return max;
  /*
   int sum = 0;
   while(i) {
   sum += mpz_size(&(*i).mCoeff);
   ++i;
   }
   return sum;
   */
}

void IOneParameterGmpZAux::minus() {
  ConstIterator i(mHead.begin());
  while (i) {
    mpz_neg(&(*i).mCoeff, &(*i).mCoeff);
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::add(const IOneParameterGmpZAux& a) {
  Iterator i1(mHead.begin());
  ConstIterator i2(a.mHead.begin());
  while (i1 && i2) {
    if ((*i1).mDegree > (*i2).mDegree) {
      ++i1;
    } else if ((*i1).mDegree < (*i2).mDegree) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree;
      mpz_set(&(*i1).mCoeff, &(*i2).mCoeff);
      ++i1;
      ++i2;
    } else {
      mpz_add(&(*i1).mCoeff, &(*i1).mCoeff, &(*i2).mCoeff);
      if (mpz_sgn(&(*i1).mCoeff) == 0)
        i1.del();
      else
        ++i1;
      ++i2;
    }
  }
  while (i2) {
    i1.insert();
    (*i1).mDegree = (*i2).mDegree;
    mpz_set(&(*i1).mCoeff, &(*i2).mCoeff);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::sub(const IOneParameterGmpZAux& a) {
  Iterator i1(mHead.begin());
  ConstIterator i2(a.mHead.begin());
  while (i1 && i2) {
    if ((*i1).mDegree > (*i2).mDegree) {
      ++i1;
    } else if ((*i1).mDegree < (*i2).mDegree) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree;
      mpz_set(&(*i1).mCoeff, &(*i2).mCoeff);
      mpz_neg(&(*i1).mCoeff, &(*i1).mCoeff);
      ++i1;
      ++i2;
    } else {
      mpz_sub(&(*i1).mCoeff, &(*i1).mCoeff, &(*i2).mCoeff);
      if (mpz_sgn(&(*i1).mCoeff) == 0)
        i1.del();
      else
        ++i1;
      ++i2;
    }
  }
  while (i2) {
    i1.insert();
    (*i1).mDegree = (*i2).mDegree;
    mpz_set(&(*i1).mCoeff, &(*i2).mCoeff);
    mpz_neg(&(*i1).mCoeff, &(*i1).mCoeff);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::sub(const IOneParameterGmpZAux& a,
    MP_INT& coeff,
    unsigned degree) {
  
  Iterator i1(mHead.begin());
  ConstIterator i2(a.mHead.begin());
  while (i1 && i2) {
    if ((*i1).mDegree > (*i2).mDegree + degree) {
      ++i1;
    } else if ((*i1).mDegree < (*i2).mDegree + degree) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree + degree;
      mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &coeff);
      mpz_neg(&(*i1).mCoeff, &(*i1).mCoeff);
      ++i1;
      ++i2;
    } else {
      mpz_submul(&(*i1).mCoeff, &(*i2).mCoeff, &coeff);
      if (mpz_sgn(&(*i1).mCoeff) == 0)
        i1.del();
      else
        ++i1;
      ++i2;
    }
  }
  while (i2) {
    i1.insert();
    (*i1).mDegree = (*i2).mDegree + degree;
    mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &coeff);
    mpz_neg(&(*i1).mCoeff, &(*i1).mCoeff);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::mult(const IOneParameterGmpZAux& a) {
  if (isZero())
    return;

  IOneParameterGmpZAux prod;
  ConstIterator ia(a.mHead.begin());
  Iterator begin(prod.mHead.begin());

  while (ia) {
    ConstIterator i(mHead.begin());
    Iterator ip = begin;
    unsigned degree2 = (*i).mDegree + (*ia).mDegree;

    // look for the starting point
    while (ip && (*ip).mDegree > degree2)
      ++ip;

    begin = ip;

    while (ip && i) {
      degree2 = (*i).mDegree + (*ia).mDegree;
      if ((*ip).mDegree > degree2)
        ++ip;
      else if ((*ip).mDegree < degree2) {
        ip.insert();
        (*ip).mDegree = degree2;
        mpz_mul(&(*ip).mCoeff, &(*i).mCoeff, &(*ia).mCoeff);
        IASSERT(mpz_sgn(&(*ip).mCoeff) != 0);
        //if (mpz_sgn(&(*ip).mCoeff) == 0)
        //ip.del();
        //else
        ++ip;
        ++i;
      } else {
        mpz_addmul(&(*ip).mCoeff, &(*i).mCoeff, &(*ia).mCoeff);
        if (mpz_sgn(&(*ip).mCoeff) == 0)
          ip.del();
        else
          ++ip;
        ++i;
      }
    }

    while (i) {
      ip.insert();
      (*ip).mDegree = (*i).mDegree + (*ia).mDegree;
      mpz_mul(&(*ip).mCoeff, &(*i).mCoeff, &(*ia).mCoeff);
      IASSERT(mpz_sgn(&(*ip).mCoeff) != 0);
      //if (mpz_sgn(&(*ip).mCoeff) == 0)
      //ip.del();
      //else
      ++ip;
      ++i;
    }

    ++ia;
  }
  mHead.swap(prod.mHead);
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::mult(const IOneParameterGmpZAux& a,
    const IOneParameterGmpZAux& b) {
  setZero();
  if (a.isZero() || b.isZero())
    return;

  ConstIterator ib(b.mHead.begin());
  Iterator begin(mHead.begin());

  while (ib) {
    ConstIterator ia(a.mHead.begin());
    Iterator ip = begin;
    unsigned degree2 = (*ia).mDegree + (*ib).mDegree;

    // look for the starting point
    while (ip && (*ip).mDegree > degree2)
      ++ip;

    begin = ip;

    while (ip && ia) {
      degree2 = (*ia).mDegree + (*ib).mDegree;
      if ((*ip).mDegree > degree2)
        ++ip;
      else if ((*ip).mDegree < degree2) {
        ip.insert();
        (*ip).mDegree = degree2;
        mpz_mul(&(*ip).mCoeff, &(*ia).mCoeff, &(*ib).mCoeff);
        IASSERT(mpz_sgn(&(*ip).mCoeff) != 0);
        //if (mpz_sgn(&(*ip).mCoeff) == 0)
        //ip.del();
        //else
        ++ip;
        ++ia;
      } else {
        mpz_addmul(&(*ip).mCoeff, &(*ia).mCoeff, &(*ib).mCoeff);
        if (mpz_sgn(&(*ip).mCoeff) == 0)
          ip.del();
        else
          ++ip;
        ++ia;
      }
    }

    while (ia) {
      ip.insert();
      (*ip).mDegree = (*ia).mDegree + (*ib).mDegree;
      mpz_mul(&(*ip).mCoeff, &(*ia).mCoeff, &(*ib).mCoeff);
      IASSERT(mpz_sgn(&(*ip).mCoeff) != 0);
      //if (mpz_sgn(&(*ip).mCoeff) == 0)
      //ip.del();
      //else
      ++ip;
      ++ia;
    }

    ++ib;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::diff() {
  Iterator i(mHead.begin());
  while (i) {
    if ((*i).mDegree == 0) {
      i.del();
      break;
    }
    mpz_mul_ui(&(*i).mCoeff, &(*i).mCoeff, (*i).mDegree);
    --(*i).mDegree;
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::diff(unsigned deg) {
  IASSERT(deg> 0);
  Iterator i(mHead.begin());
  while (i) {
    if ((*i).mDegree < deg) {
      i.clear();
      break;
    }
    unsigned n = (*i).mDegree;
    (*i).mDegree -= deg;
    do {
      mpz_mul_ui(&(*i).mCoeff, &(*i).mCoeff, (*i).mDegree);
      --n;
    } while (n != (*i).mDegree);
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::pow(unsigned deg) {
  IASSERT(deg> 0);
  if (deg > 1) {
    IOneParameterGmpZAux p(*this);
    while (--deg > 0) {
      p.mult(*this);
    }
    mHead.swap(p.mHead);
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::div(const IOneParameterGmpZAux& a) {
  IASSERT(!isZero());
  IASSERT(!a.isZero());
  IASSERT(degree() >= a.degree());

  if (a.degree() == 0) {
    div((*(a.mHead.begin())).mCoeff);
    IASSERTVALID(*this);
    return;
  }

  MP_INT k;
  mpz_init(&k);

  IOneParameterGmpZAux quo;
  Iterator i(quo.mHead.begin());

  while (!isZero()) {
    IASSERT(degree() >= a.degree());
    Iterator i1(mHead.begin());
    ConstIterator i2(a.mHead.begin());

    unsigned deg = degree() - a.degree();
    IASSERT(mpz_divisible_p(&(*i1).mCoeff, &(*i2).mCoeff));
    mpz_divexact(&k, &(*i1).mCoeff, &(*i2).mCoeff);

    i.insert();
    (*i).mDegree = deg;
    mpz_set(&(*i).mCoeff, &k);

    ++i;
    IASSERTVALID(quo);

    mpz_neg(&k, &k);

    i1.del();
    ++i2;

    while (i1 && i2) {
      unsigned degree2 = (*i2).mDegree + deg;
      if ((*i1).mDegree > degree2) {
        ++i1;
      } else if ((*i1).mDegree < degree2) {
        i1.insert();
        (*i1).mDegree = degree2;
        mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &k);
        ++i1;
        ++i2;
      } else {
        mpz_addmul(&(*i1).mCoeff, &(*i2).mCoeff, &k);
        if (mpz_sgn(&(*i1).mCoeff) == 0)
          i1.del();
        else
          ++i1;
        ++i2;
      }
    }
    while (i2) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree + deg;
      mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &k);
      ++i1;
      ++i2;
    }
    IASSERTVALID(*this);
  }
  IASSERT(isZero());
  mHead.swap(quo.mHead);
  mpz_clear(&k);
}

/* divides by a if possible. returns true if it is possible, false otherwise
 */
bool IOneParameterGmpZAux::divtry(const IOneParameterGmpZAux& a) {
  IASSERT(!isZero());
  IASSERT(!a.isZero());

  if (degree() < a.degree())
    return false;

  IOneParameterGmpZAux tmp(*this);

  if (a.degree() == 0) {
    MP_INT lca;
    a.lc(lca);
    Iterator it = tmp.begin();
    while (it) {
      if (mpz_divisible_p(&(*it).mCoeff, &lca))
        mpz_divexact(&(*it).mCoeff, &(*it).mCoeff, &lca);
      else {
        mpz_clear(&lca);
        return false;
      }

      ++it;
    }

    swap(tmp);
    mpz_clear(&lca);
    return true;
  }

  IOneParameterGmpZAux quo;
  Iterator i(quo.mHead.begin());

  while (!tmp.isZero()) {
    if (a.degree() > tmp.degree() || 
        !mpz_divisible_p(&(*tmp.mHead.begin()).mCoeff,
        &(*a.mHead.begin()).mCoeff))
      return false;
    i.insert();
    (*i).mDegree = tmp.degree() - a.degree();
    mpz_divexact(&(*i).mCoeff, 
        &(*tmp.mHead.begin()).mCoeff, 
        &(*a.mHead.begin()).mCoeff);

    tmp.sub(a, (*i).mCoeff, (*i).mDegree);
    
    ++i;
  }
  
  swap(quo);
  return true;
}

void IOneParameterGmpZAux::pseudo_div(MP_INT& multiplier,
    const IOneParameterGmpZAux& a) {
  IASSERT(!isZero());
  IASSERT(!a.isZero());
  IASSERT(degree() >= a.degree());

  MP_INT k1, k2, tmp;
  mpz_init(&k1);
  mpz_init(&k2);
  mpz_init(&tmp);

  mpz_set_ui(&multiplier, 1ul);
  IOneParameterGmpZAux quo;
  Iterator i(quo.mHead.begin());

  while (!isZero()) {
    IASSERT(degree() >= a.degree());
    Iterator i1(mHead.begin());
    ConstIterator i2(a.mHead.begin());

    unsigned deg = degree() - a.degree();
    mpz_gcd(&tmp, &(*i1).mCoeff, &(*i2).mCoeff);
    mpz_tdiv_q(&k1, &(*i2).mCoeff, &tmp);
    mpz_tdiv_q(&k2, &(*i1).mCoeff, &tmp);

    mpz_mul(&multiplier, &multiplier, &k1);
    quo.mult(k1);

    i.insert();
    (*i).mDegree = deg;
    mpz_set(&(*i).mCoeff, &k2);

    ++i;
    IASSERTVALID(quo);

    mpz_neg(&k2, &k2);

    i1.del();
    ++i2;

    while (i1 && i2) {
      unsigned degree2 = (*i2).mDegree + deg;
      if ((*i1).mDegree > degree2) {
        mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &k1);
        ++i1;
      } else if ((*i1).mDegree < degree2) {
        i1.insert();
        (*i1).mDegree = degree2;
        mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &k2);
        ++i1;
        ++i2;
      } else {
        mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &k1);
        mpz_addmul(&(*i1).mCoeff, &(*i2).mCoeff, &k2);
        if (mpz_sgn(&(*i1).mCoeff) == 0)
          i1.del();
        else
          ++i1;
        ++i2;
      }
    }
    while (i2) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree + deg;
      mpz_mul(&(*i1).mCoeff, &(*i2).mCoeff, &k2);
      ++i1;
      ++i2;
    }
    while (i1) {
      mpz_mul(&(*i1).mCoeff, &(*i1).mCoeff, &k1);
      ++i1;
    }
    IASSERTVALID(*this);
  }
  IASSERT(isZero());
  mHead.swap(quo.mHead);
  mpz_clear(&k1);
  mpz_clear(&k2);
  mpz_clear(&tmp);
}

void IOneParameterGmpZAux::gcd(const IOneParameterGmpZAux& a) {
  if (isZero()) {
    set(a);
    return;
  }
  if (a.isZero())
    return;

  if (degree() == 0)
    if (a.degree() == 0) {
      ConstIterator i(mHead.begin());
      mpz_gcd(&(*i).mCoeff, &(*i).mCoeff, &(*a.mHead.begin()).mCoeff);
    } else {
      MP_INT cont2;
      mpz_init(&cont2);
      a.cont(cont2);
      ConstIterator i(mHead.begin());
      mpz_gcd(&(*i).mCoeff, &(*i).mCoeff, &cont2);
      mpz_clear(&cont2);
    }
  else if (a.degree() == 0) {
    MP_INT cont1;
    mpz_init(&cont1);
    cont(cont1);
    setOne();
    ConstIterator i(mHead.begin());
    mpz_gcd(&(*i).mCoeff, &cont1, &(*a.mHead.begin()).mCoeff);
    mpz_clear(&cont1);
  } else {
    IOneParameterGmpZAux tmp(a);
    MP_INT cont1, cont2, g12, alpha, beta, psi, t;
    mpz_init(&cont1);
    mpz_init(&cont2);
    mpz_init(&g12);
    mpz_init(&alpha);
    mpz_init_set_ui(&beta, 1ul);
    mpz_init_set_si(&psi, -1);
    mpz_init(&t);

    cont(cont1);
    tmp.cont(cont2);

    if (mpz_cmp_ui(&cont1, 1ul) != 0) {
      div(cont1);
      if (mpz_cmp_ui(&cont2, 1ul) == 0)
        mpz_set_ui(&g12, 1ul);
      else {
        tmp.div(cont2);
        mpz_gcd(&g12, &cont1, &cont2);
      }
    } else {
      if (mpz_cmp_ui(&cont2, 1ul) != 0)
        tmp.div(cont2);
      mpz_set_ui(&g12, 1ul);
    }

    if (tmp.degree() > degree())
      mHead.swap(tmp.mHead);

    if (((degree() - tmp.degree()) & 1) == 0)
      mpz_neg(&beta, &beta);

    unsigned delta = degree() - tmp.degree();
    do {
      mHead.swap(tmp.mHead);
      mpz_pow_ui(&alpha, &(*mHead.begin()).mCoeff, delta + 1);
      tmp.pseudorem(*this, alpha);
      tmp.div(beta);

      if (!tmp.isZero()) {
        mpz_pow_ui(&t, &(*mHead.begin()).mCoeff, delta);
        mpz_pow_ui(&psi, &psi, delta - 1);
        mpz_divexact(&psi, &t, &psi);
        if ((delta & 1) == 1)
          mpz_neg(&psi, &psi);
        delta = degree() - tmp.degree();
        mpz_pow_ui(&beta, &psi, delta);
        mpz_mul(&beta, &beta, &(*mHead.begin()).mCoeff);
        mpz_neg(&beta, &beta);
      }
    } while (!tmp.isZero() && tmp.degree() > 0);

    if (!tmp.isZero()) {
      mHead.swap(tmp.mHead);
      mpz_set(&(*mHead.begin()).mCoeff, &g12);
    } else {
      cont(cont1);
      if (mpz_cmp_ui(&cont1, 1ul) != 0)
        div(cont1);
      mult(g12);
    }
    mpz_clear(&cont1);
    mpz_clear(&cont2);
    mpz_clear(&g12);
    mpz_clear(&alpha);
    mpz_clear(&beta);
    mpz_clear(&psi);
    mpz_clear(&t);
  }
}

void IOneParameterGmpZAux::gcd(const IOneParameterGmpZAux& a,
    const IOneParameterGmpZAux& b) {
  if (a.isZero()) {
    set(b);
    return;
  }
  if (b.isZero()) {
    set(a);
    return;
  }

  if (a.degree() == 0) {
    setOne();
    if (b.degree() == 0)
      mpz_gcd(&(*mHead.begin()).mCoeff, &(*a.mHead.begin()).mCoeff, &(*b.mHead.begin()).mCoeff);
    else {
      MP_INT cont2;
      mpz_init(&cont2);
      b.cont(cont2);
      mpz_gcd(&(*mHead.begin()).mCoeff, &(*a.mHead.begin()).mCoeff, &cont2);
      mpz_clear(&cont2);
    }
  } else if (b.degree() == 0) {
    setOne();
    MP_INT cont1;
    mpz_init(&cont1);
    a.cont(cont1);
    mpz_gcd(&(*mHead.begin()).mCoeff, &cont1, &(*b.mHead.begin()).mCoeff);
    mpz_clear(&cont1);
  } else {
    mHead = a.mHead;
    IOneParameterGmpZAux tmp(b);
    MP_INT cont1, cont2, g12;
    mpz_init(&cont1);
    mpz_init(&cont2);
    mpz_init(&g12);

    cont(cont1);
    tmp.cont(cont2);

    if (mpz_cmp_ui(&cont1, 1ul) != 0) {
      div(cont1);
      if (mpz_cmp_ui(&cont2, 1ul) == 0)
        mpz_set_ui(&g12, 1ul);
      else {
        tmp.div(cont2);
        mpz_gcd(&g12, &cont1, &cont2);
      }
    } else {
      if (mpz_cmp_ui(&cont2, 1ul) != 0)
        tmp.div(cont2);
      mpz_set_ui(&g12, 1ul);
    }

    if (tmp.degree() > degree())
      mHead.swap(tmp.mHead);

    do {
      mHead.swap(tmp.mHead);
      tmp.rem(*this);
    } while (!tmp.isZero() && tmp.degree() > 0);

    if (!tmp.isZero()) {
      mHead.swap(tmp.mHead);
      mpz_set(&(*mHead.begin()).mCoeff, &g12);
    } else {
      cont(cont1);
      if (mpz_cmp_ui(&cont1, 1ul) != 0)
        div(cont1);
      mult(g12);
    }
    mpz_clear(&cont1);
    mpz_clear(&cont2);
    mpz_clear(&g12);
  }
}

void IOneParameterGmpZAux::set(const IOneParameterGmpZAux& a) {
  mHead.clear();
  Iterator i(mHead.begin());
  ConstIterator ia(a.mHead.begin());
  while (ia) {
    i.insert();
    (*i).mDegree = (*ia).mDegree;
    mpz_set(&(*i).mCoeff, &(*ia).mCoeff);
    ++i;
    ++ia;
  }
  IASSERTVALID(*this);
}

void IOneParameterGmpZAux::input(std::istream& in, const IVariables &parameters) {
  additive(in, parameters);
}

void IOneParameterGmpZAux::additive(std::istream& in, const IVariables &parameters) {
  multiplicative(in, parameters);
  int op = (in >> std::ws).peek();
  IOneParameterGmpZAux tmp;
  while(op == '+' || op == '-') {
    in.get();
    tmp.multiplicative(in, parameters);
    if (op == '+')
      add(tmp);
    else {
      sub(tmp);
    }
    op = in.peek();
  }
}

void IOneParameterGmpZAux::multiplicative(std::istream& in, const IVariables &parameters) {
  unary(in, parameters);
  int op = (in >> std::ws).peek();
  IOneParameterGmpZAux tmp;
  while(op == '*') {
    in.get();
    tmp.unary(in, parameters);
    mult(tmp);
    op = in.peek();
  }
}

void IOneParameterGmpZAux::unary(std::istream& in, const IVariables &parameters) {
  int ch = (in >> std::ws).peek();
  if (ch != '+' && ch != '-')
    power(in, parameters);
  else {
    int k = 1;
    do {
      ch = in.get();
      if (ch == '-')
        k *= -1;
      ch = (in >> std::ws).peek();
    } while (ch == '+' || ch == '-');
    power(in, parameters);
    if (k < 0)
      minus();
  }
}

void IOneParameterGmpZAux::power(std::istream& in, const IVariables &parameters) {
  bracket(in, parameters);
  int op = (in >> std::ws).peek();
  if (op == '^') {
    in.get();
    int deg=0;
    in >> std::ws >> deg;
    if (in.fail() || deg <= 0)
      IMESSAGE("expected 'degree > 0'");
    pow(deg);
  }
}

void IOneParameterGmpZAux::bracket(std::istream& in, const IVariables &parameters) {
  int op = (in >> std::ws).peek();
  if (op == '(') {
    in.get();
    additive(in, parameters);
    if (in.get() != ')') {
      IMESSAGE("expected ')'");
    }
  }
  else {
    std::streampos posbeg = in.tellg();
    int var = parameters.read(in);
    if (var == 0) {
      setOne();
      (*begin()).mDegree = 1;
      IASSERTVALID(*this);
    }
    else {
      IOneParameterGmpZAux::Iterator i(begin());
      i.clear();
      i.insert();
      in >> &(*i).mCoeff;
      if (in.fail()) {
        in.clear();
        in.setstate(std::ios::failbit);
        in.seekg(posbeg);
      }
      else 
        IASSERTVALID(*this);
    }
  }
}

void IOneParameterGmpZAux::read(std::istream &in) {
  mHead.clear();
  int l;
  in.read((char*)(&l), sizeof(l));
  Iterator i(mHead.begin());
  while (l) {
    i.insert();
    in.read((char*)(&(*i).mDegree), sizeof((*i).mDegree));
    in >> (&(*i).mCoeff);
    ++i;
    --l;
  }
}

void IOneParameterGmpZAux::write(std::ostream &out) const {
  int l = mHead.length();
  out.write((char*)(&l), sizeof(l));
  ConstIterator i(mHead.begin());
  while (i) {
    out.write((char*)(&(*i).mDegree), sizeof((*i).mDegree));
    out << (&(*i).mCoeff);
    ++i;
  }
}

void IOneParameterGmpZAux::assertValid(const char* fileName, int fileLine) const {
  if (!mHead.isEmpty()) {
    ConstIterator i(mHead.begin());
    IASSERT2(fileName, fileLine, mpz_sgn(&(*i).mCoeff) != 0);
    ConstIterator prev(i);
    ++i;
    while (i) {
      IASSERT2(fileName, fileLine, mpz_sgn(&(*i).mCoeff) != 0);
      IASSERT2(fileName, fileLine, (*prev).mDegree> (*i).mDegree);
      prev = i;
      ++i;
    }
  }
}

std::ostream& operator<<(std::ostream& out, const IOneParameterGmpZAux& a) {
  IOneParameterGmpZAux::ConstIterator i(a.begin());
  if (!i)
    out << '0';
  else {
    out << (&(*i).mCoeff) << "*t^" << (*i).mDegree;
    ++i;
    while (i) {
      out << " + " << (&(*i).mCoeff) << "*t^" << (*i).mDegree;
      ++i;
    }
  }
  return out;
}

template <> IAllocator
    IList<IOneParameterGmpZAux::Term>::Node::sAllocator(sizeof(IList<IOneParameterGmpZAux::Term>::Node));
