/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IONEPARAMETERGMPZAUX_H
#define IONEPARAMETERGMPZAUX_H

#include <iostream>
#include <gmp.h>

#include "../util/iutil.h"
#include "../util/imodular.h"
#include "../util/ilist.h"
#include "../util/ivariables.h"


class ITwoParameterGmpZAux;
class INParameterGmpZAux;
class IFactorizationZ;
class IOneParameterGmpZAux {
  friend class ITwoParameterGmpZAux;
  friend class IFactorizationZ;

public:
  struct Term {
    unsigned  mDegree;
    MP_INT  mCoeff;

    Term(): mDegree(0) { mpz_init(&mCoeff); }
    Term(const Term& a): mDegree(a.mDegree) { mpz_init_set(&mCoeff, &a.mCoeff); }
    ~Term() { mpz_clear(&mCoeff); }

    void operator=(const Term& a) {
      mDegree = a.mDegree;
      mpz_set(&mCoeff, &a.mCoeff);
    }
  };

protected:
  IList<Term>  mHead;

  void gcdBrown(const IOneParameterGmpZAux& a);

  void additive(std::istream& in, const IVariables &parameters);
  void multiplicative(std::istream& in, const IVariables &parameters);
  void unary(std::istream& in, const IVariables &parameters);
  void power(std::istream& in, const IVariables &parameters);
  void bracket(std::istream& in, const IVariables &parameters);
  
  
public:
  void input(std::istream& in, const IVariables &parameters);

  void add(IOneParameterGmpZAux& a);
  void sub(IOneParameterGmpZAux& a);
  void mult(const MP_INT& coeff);
  void shiftMult(int coeff);
  void eval(MP_INT& sum, int t) const;
  void cont(MP_INT& a) const;
  
  void pp() {
    MP_INT c;
    cont(c);
    div(c);
    mpz_clear(&c);
  }
  
  void mult(unsigned degree, const MP_INT& coeff);
  void div(const MP_INT& coeff);
  void div(unsigned degree,  const MP_INT& coeff);
  void rem(const IOneParameterGmpZAux& a);
  void pseudorem(const IOneParameterGmpZAux& a, const MP_INT& alpha);
  typedef IList<Term>::Iterator  Iterator;
  typedef IList<Term>::ConstIterator  ConstIterator;

  IOneParameterGmpZAux():
      mHead() {
  }
  IOneParameterGmpZAux(const IOneParameterGmpZAux& a):
      mHead(a.mHead) {
  }
  
  IOneParameterGmpZAux(const INParameterGmpZAux& a);

  ~IOneParameterGmpZAux() {}

  void operator=(const IOneParameterGmpZAux& a) {
    mHead = a.mHead;
  }

  void swap(IOneParameterGmpZAux& a) {
    mHead.swap(a.mHead);
  }

  void set(const IOneParameterGmpZAux& a);

  void setZero() { mHead.clear(); }

  void setOne() {
    mHead.clear();
    Iterator i(mHead.begin());
    i.insert();
    mpz_set_ui(&(*i).mCoeff, 1ul);
  }

  bool isZero() const { return mHead.isEmpty(); }
  bool isOne() const {
    ConstIterator i(mHead.begin());
    return i && (*i).mDegree == 0 && mpz_cmp_ui(&(*i).mCoeff, 1ul) == 0;
  }
  bool isNegative() const {
    ConstIterator i(mHead.begin());
    return i && mpz_sgn(&(*i).mCoeff) < 0;
  }
  bool isComposite() const {
    ConstIterator i(mHead.begin());
    return mHead.length() > 1 || (i && mpz_sgn(&(*i).mCoeff) < 0);
  }
  int norm() const;
  unsigned degree() const {
    IASSERT(!isZero());
    return (*mHead.begin()).mDegree;
  }
  
  void lc(MP_INT& a) const {
    if (isZero())
      mpz_init_set_ui(&a, 0ul);
    else
      mpz_init_set(&a, &(*mHead.begin()).mCoeff);
  }
  
  // return constant coefficient
  void cc(MP_INT& a) const {
    if (isZero()) {
      mpz_init_set_ui(&a, 0ul);
      return;
    }
    
    ConstIterator it = mHead.begin();
    while (it.isNext())
      ++it;
    
    if ((*it).mDegree == 0)
      mpz_init_set(&a, &(*it).mCoeff);
    else
      mpz_init_set_ui(&a, 0ul);
  }

  void minus();
  void add(const IOneParameterGmpZAux& a);
  void sub(const IOneParameterGmpZAux& a);
  void sub(const IOneParameterGmpZAux& a, MP_INT& coeff, unsigned degree);
  void mult(const IOneParameterGmpZAux& a);
  void mult(const IOneParameterGmpZAux& a, const IOneParameterGmpZAux& b);
  void diff();
  void diff(unsigned deg);
  void pow(unsigned deg);
  void div(const IOneParameterGmpZAux& a);
  bool divtry(const IOneParameterGmpZAux& a);
  void div(const IOneParameterGmpZAux& a, const IOneParameterGmpZAux& b) {
    *this = a;
    div(b);
  }
  void pseudo_div(MP_INT& multiplier, const IOneParameterGmpZAux& a);
  void pseudo_div(MP_INT& multiplier, const IOneParameterGmpZAux& a, const IOneParameterGmpZAux& b) {
    *this = a;
    pseudo_div(multiplier, b);
  }
  void gcd(const IOneParameterGmpZAux& a);
  void gcd(const IOneParameterGmpZAux& a, const IOneParameterGmpZAux& b);
  void addmul(const IOneParameterGmpZAux& a, const IOneParameterGmpZAux& b) {
    IOneParameterGmpZAux tmp;
    tmp.mult(a, b);
    add(tmp);
  }
  
  void mod(MP_INT& q);

  ConstIterator begin() const { return mHead.begin(); }
  Iterator begin() { return mHead.begin(); }

  void read(std::istream &in);
  void write(std::ostream &out) const;

  void assertValid(const char* fileName, int fileLine) const;

  friend std::ostream& operator<<(std::ostream& out, const IOneParameterGmpZAux& a);
};

#endif // IONEPARAMETERGMPZAUX_H
