/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdarg.h>

#include "./idivision.h"
#include "./ijanet.h"
#include "./ijanetlike.h"
#include "../gauss/ipolygauss.h"


IDivisionInterface::IDivisionInterface():
    mType(Indefinitely) {
}

IDivisionInterface* IDivisionInterface::create(Type type,
                                               IWrapInterface *wrapInterface, ...) {
  if (type == IDivisionInterface::Indefinitely)
    IERROR("error in IDivisionInterface: type == IDivisionInterface::Indefinitely");
  if (wrapInterface == NULL || wrapInterface->type() == IWrapInterface::Indefinitely)
    IERROR("error in IDivisionInterface: wrapInterface == NULL || wrapInterface->type() == IWrapInterface::Indefinitely");

  va_list marker;
  va_start(marker, wrapInterface);

  IDivisionInterface* r = NULL;
  switch(type) {
  case Janet:
    r = new IJanetInterface;
    break;
  case JanetLike:
    r = new IJanetLikeInterface;
    break;
  }
  va_end(marker);

  if (r) {
    r->mWrapInterface = wrapInterface;
  }
  return r;
}

int IDivision::nf(ICoeff &multiplier, IPoly &poly, const IPoly &red, ISetQ &setQ) const {
  int r = 1;
  if (coeffInterface()->type() != ICoeffInterface::GmpZZ) {
    poly.reduction(multiplier, red);
    if (!poly.isZero()) {
      IWrap* wrap = find(poly.lm());
      while(wrap) {
        poly.reduction(multiplier, wrap->poly());
        ++r;
        if (poly.isZero())
          break;
        wrap = find(poly.lm());
      }
    }
  }
  else { // coeffInterface()->type() == ICoeffInterface::GmpZZ
    ICoeff *g = coeffInterface()->create(),
           *s = coeffInterface()->create(),
           *t = coeffInterface()->create();
    if (poly.lc().divisibility(red.lc()))
      poly.reduction(multiplier, red);
    else {
      g->gcdExt(poly.lc(), red.lc(), *s, *t);
      IASSERT(!t->isZero());
      IPoly *p = polyInterface()->copy(poly),
            *r = polyInterface()->copy(red);
      p->mult(*s);
      r->mult(*t);
      IMonom *m = monomInterface()->create();
      m->divide(poly.lm(), red.lm());
      r->mult(*m);
      p->add(*r);
      poly.reduction(multiplier, *p);
      setQ.insert(p);
      delete r;
      delete m;
    }
    if (!poly.isZero()) {
      IWrap* wrap = find(poly.lm());
      if (wrap &&
          wrap->poly().lc().divisibilityProper(poly.lc()))
        wrap = NULL;
      while(wrap) {
        IPoly &red(wrap->poly());
        if (poly.lc().divisibility(red.lc()))
          poly.reduction(multiplier, red);
        else {
          g->gcdExt(poly.lc(), red.lc(), *s, *t);
          IASSERT(!t->isZero());
          IPoly *p = polyInterface()->copy(poly),
                *r = polyInterface()->copy(red);
          p->mult(*s);
          r->mult(*t);
          IMonom *m = monomInterface()->create();
          m->divide(poly.lm(), red.lm());
          r->mult(*m);
          p->add(*r);
          poly.reduction(multiplier, *p);
          setQ.insert(p);
          delete r;
          delete m;
        }
        ++r;
        if (poly.isZero())
          break;
        wrap = find(poly.lm());
        if (wrap &&
            wrap->poly().lc().divisibilityProper(poly.lc()))
          wrap = NULL;
      }
    }
  }
  IASSERT(isNf(poly));
  return r;
}

int IDivision::nf(ICoeff &multiplier, IPoly &poly) const {
  int r=0;
  if (!poly.isZero()) {
    IWrap* wrap = find(poly.lm());
    if (wrap && coeffInterface()->type() == ICoeffInterface::GmpZZ) {
      if (poly.lc().cmpNorm(wrap->poly().lc()) < 0)
        wrap = NULL;
    }
    while(wrap) {
      poly.reduction(multiplier, wrap->poly());
      ++r;
      if (poly.isZero())
        break;
      wrap = find(poly.lm());
      if (wrap && coeffInterface()->type() == ICoeffInterface::GmpZZ) {
	if (poly.lc().cmpNorm(wrap->poly().lc()) < 0)
	  wrap = NULL;
      }
    }
  }
  IASSERT(isNf(poly));
  return r;
}

bool IDivision::minPoly(const IPoly& poly, IOneParameterPoly& mp, int maxDegree) const {
  //IASSERT(poly.coeffInterface()->isField());

  ICoeff* multiplier = poly.coeffInterface()->create();
  multiplier->setOne();

  IPoly* tmp = poly.polyInterface()->create();
  tmp->setOne();

  IPolyGauss alg(poly.coeffInterface());
  alg.add(*multiplier, *tmp);

  bool linearDependent = false;

  while (!linearDependent && alg.size() < maxDegree) {
    tmp->mult(poly);
    nf(*multiplier, *tmp);
    if (!tmp->isZero())
      nfTail(*multiplier, *tmp);
    linearDependent = alg.add(*multiplier, *tmp);
  }

  delete multiplier;
  delete tmp;

  if (linearDependent) {
    alg.getMinPoly(mp);
    return true;
  }

  return false;
}

bool IDivision::isNf(const IPoly &poly) const {
  return poly.isZero() || find(poly.lm()) == NULL || coeffInterface()->type() == ICoeffInterface::GmpZZ;
}

int IDivision::nfTail(ICoeff &multiplier, IPoly &poly) const {
  IASSERT(!poly.isZero());
  int r = 0;
  IWrap* wrap = NULL;
  IPoly::Iterator &i(*poly.begin());
  do {
    ++i;
    if (!i)
      break;
    wrap = find(i.monom());
    if (wrap && coeffInterface()->type() == ICoeffInterface::GmpZZ) {
      if (i.coeff().cmpNorm(wrap->poly().lc()) < 0)
        wrap = NULL;
    }
  }
  while(wrap == NULL);

  if (wrap) {
    do {
      poly.redTail(multiplier, i, wrap->poly());
      ++r;
      while(i) {
        wrap = find(i.monom());
        if (wrap && coeffInterface()->type() == ICoeffInterface::GmpZZ) {
          if (i.coeff().cmpNorm(wrap->poly().lc()) < 0)
            wrap = NULL;
        }
        if (wrap)
          break;
        ++i;
      }
    }
    while(i);
    //the next step should be done in IAlgorithm::build
    //because for reduction of a polynomial modulo a given basis
    //it is not desired:
    //poly.pp();
  }
  IASSERT(isNfTail(poly));
  //IASSERT(poly.isPp());
  return r;
}

int IDivision::nfTailCanonical(ICoeff &multiplier, IPoly &poly) const {
  IASSERT(!poly.isZero());
  int r = 0;
  IWrap* wrap = NULL;
  IPoly::Iterator &i(*poly.begin());
  do {
    ++i;
    if (!i)
      break;
    if (i.coeff().isNegative())
      wrap = find(i.monom());
  }
  while(wrap == NULL);

  if (wrap) {
    poly.redTailCanonical(i, wrap->poly());
    ++r;
    ++i;
    while(i) {
      wrap = find(i.monom());
      if (wrap && coeffInterface()->type() == ICoeffInterface::GmpZZ) {
	if (i.coeff().cmpNorm(wrap->poly().lc()) < 0)
	  wrap = NULL;
      }
      if (wrap) {
	poly.redTail(multiplier, i, wrap->poly());
	++r;
      }
      if (i.coeff().isNegative()) {
	wrap = find(i.monom());
	if (wrap) {
	  poly.redTailCanonical(i, wrap->poly());
	  ++r;
	}
      }
      ++i;
    }
    while(i);
  }

  return r;
}

bool IDivision::isNfTail(const IPoly &poly) const {
  IASSERT(!poly.isZero());
  IPoly::ConstIterator &i(*poly.begin());
  IWrap *wrap = NULL;
  ++i;
  while(i) {
    IWrap *wrap = find(i.monom());
    if (wrap && coeffInterface()->type() == ICoeffInterface::GmpZZ) {
      if (i.coeff().cmpNorm(wrap->poly().lc()) < 0)
        wrap = NULL;
    }
    if (wrap)
      break;
    ++i;
  }
  return !wrap;
}
