/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./ijanet.h"


IJanetInterface::IJanetInterface():
    IDivisionInterface() {
  mType = IDivisionInterface::Janet;
}

IDivision* IJanetInterface::create() {
  return new IJanet(this);
}

inline void IJanet::Iterator::del() {
  Link tmp = *i;
  IASSERT(tmp);
  *i = tmp->mNextDeg;
  delete tmp;
}

void IJanet::Iterator::build(unsigned d, int var, IWrap *wrap) {
  IASSERT(d >= wrap->lm()[var]);
  Link r =  new IJanet::Node(wrap->lm()[var]);
  Link j = r;
  while(d > wrap->lm()[var]) {
    d -= wrap->lm()[var];
    ++var;
    j->mNextVar = new IJanet::Node(wrap->lm()[var]);
    j = j->mNextVar;
  }
  j->mWrap = wrap;

  IASSERT(!*i || r->mDeg < (*i)->mDeg);
  r->mNextDeg = *i;
  *i = r;
}

void IJanet::ConstIterator::prolong(ISetQ &q, int var) {
  while(nextDeg()) {
    IASSERT(nextVar());
    nextVar().prolong(q, var);
    deg();
  }
  if (nextVar())
    nextVar().prolong(q, var);
  else {
    if (wrap()->notBuild(var))
      q.insert(wrap(), var);
  }
}

void IJanet::ConstIterator::assertValid(const char* fileName, int fileLine) {
  while(nextDeg()) {
    IASSERT2(fileName, fileLine, degree() < nextDeg().degree());
    IASSERT2(fileName, fileLine, nextVar());
    nextVar().assertValid(fileName, fileLine);
    deg();
  }
  if (nextVar())
    nextVar().assertValid(fileName, fileLine);
  else
    IASSERT2(fileName, fileLine, wrap());
}

void IJanet::Iterator::clear() {
  while (nextDeg()) {
    IASSERT(nextVar());
    IJanet::Iterator((*i)->mNextVar).clear();
    del();
  }
  del();
}

IJanet::IJanet(IDivisionInterface* divisionInterface):
    IDivision(divisionInterface),
    mRoot(NULL) {
  int dim = monomInterface()->dimDepend();
  if (dim == 0)
    dim = 1;

  mRoot = new Link[dim];
  for(int i=0; i < dim; i++)
    mRoot[i] = NULL;
}

IJanet::~IJanet() {
  clear();
  delete[] mRoot;
}

IWrap* IJanet::find(const IMonom &m) const {
  int depend = 0;
  if (monomInterface()->dimDepend())
    depend = m.dependVar();

  Link root = mRoot[depend];

  IWrap* wrap = NULL;

  if (root) {
    IJanet::ConstIterator j(root);
    unsigned d = m.degree();
    int var = 0;
    if (d == 0) {
      if (j.degree() == 0)
        wrap = j.wrap();
    }
    else {
      do {
        while(j.degree() < m[var] && j.nextDeg())
          j.deg();

        if (j.degree() > m[var])
          break;
        else if (j.nextVar()) {
          d -= m[var];
          if (d == 0)
            break;
          ++var;
          j.var();
        }
        else {
          wrap = j.wrap();
	  IASSERT(wrap != NULL);
          IASSERT(m.divisibility(wrap->lm()));
          break;
        }
      } while(true);
    }
  }

  return wrap;
}

void IJanet::insert(IWrap *wrap, ISetQ &setQ) {
  IASSERT(wrap != NULL);
  IASSERT(find(wrap->lm()) == NULL || coeffInterface()->type() == ICoeffInterface::GmpZZ);

  int depend = 0;
  if (monomInterface()->dimDepend())
    depend = wrap->lm().dependVar();

  Link &root = mRoot[depend];

  unsigned d = wrap->lm().degree();
  IJanet::Iterator j(root);
  if (root == NULL) {
    wrap->clearBuild();
    j.build(d, 0, wrap);
  }
  else {
    int var = 0;
    do {
      while(j.degree() < wrap->lm()[var] && j.nextDeg())
        j.deg();

      if (j.degree() > wrap->lm()[var]) {
        setQ.insert(wrap, var);
        j.build(d, var, wrap);
        for(++var; var < wrap->lm().dimIndepend(); var++)
          wrap->clearBuild(var);
        break;
      }
      else if (j.degree() == wrap->lm()[var]) {
        if (j.nextDeg())
          setQ.insert(wrap, var);
        else
          wrap->clearBuild(var);
        d -= wrap->lm()[var];
        ++var;
        j.var();
      }
      else {
	IASSERT(IJanet::ConstIterator(j).nextVar());
        IJanet::ConstIterator(j).nextVar().prolong(setQ, var);
        j.deg();
        j.build(d, var, wrap);
        for(; var < wrap->lm().dimIndepend(); var++)
          wrap->clearBuild(var);
        break;
      }
    } while(true);
  }
  IASSERT(find(wrap->lm()) != NULL);
  IASSERTVALID(IJanet::ConstIterator(root));
}

void IJanet::insert(IWrap *wrap) {
  IASSERT(wrap != NULL);
  IASSERT(find(wrap->lm()) == NULL || coeffInterface()->type() == ICoeffInterface::GmpZZ);

  int depend = 0;
  if (monomInterface()->dimDepend())
    depend = wrap->lm().dependVar();

  Link &root = mRoot[depend];

  unsigned d = wrap->lm().degree();
  IJanet::Iterator j(root);
  if (root == NULL)
    j.build(d, 0, wrap);
  else {
    int var = 0;
    do {
      while(j.degree() < wrap->lm()[var] && j.nextDeg())
        j.deg();

      if (j.degree() > wrap->lm()[var]) {
        j.build(d, var, wrap);
        break;
      }
      else if (j.degree() == wrap->lm()[var]) {
        d -= wrap->lm()[var];
        ++var;
        j.var();
      }
      else {
        j.deg();
        j.build(d, var, wrap);
        break;
      }
    } while(true);
  }
  IASSERT(find(wrap->lm()) != NULL);
  IASSERTVALID(IJanet::ConstIterator(root));
}

void IJanet::update(IWrap* wrap, ISetQ &setQ) {
  IASSERT(wrap != NULL);
  IASSERT(find(wrap->lm()) != NULL || coeffInterface()->type() == ICoeffInterface::GmpZZ);

  int depend = 0;
  if (monomInterface()->dimDepend())
    depend = wrap->lm().dependVar();

  Link root = mRoot[depend];

  IJanet::ConstIterator j(root);
  int var = 0;
  do {
    while(j.degree() < wrap->lm()[var]) {
      IASSERT(j.nextDeg());
      j.deg();
    }
    IASSERT(j.degree() == wrap->lm()[var]);
    if (!j.nextDeg())
      wrap->clearBuild(var);
    else if (wrap->notBuild(var))
      setQ.insert(wrap, var);
    ++var;
    if (j.nextVar())
      j.var();
    else {
      IASSERT(wrap == j.wrap());
      for(; var < wrap->lm().dimIndepend(); var++)
        wrap->clearBuild(var);
      break;
    }
  } while(true);
}

void IJanet::clear() {
  int dim = monomInterface()->dimDepend();
  if (dim == 0)
    dim = 1;

  for(int i=0; i < dim; i++) {
    if (mRoot[i]) {
      IJanet::Iterator j(mRoot[i]);
      j.clear();
    }
    IASSERT(mRoot[i] == NULL);
  }
}

IAllocator IJanet::Node::sAllocator(sizeof(IJanet::Node));
