/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "igauss.h"
#include "iswapoperation.h"
#include "imultoperation.h"
#include "iaddmultoperation.h"

IGauss::~IGauss() {
// delete coefficients?
}

bool IGauss::addVector(IList<ICoeff*>& list) {
  // apply Gaussian algorithm on the new vector 

  int i = 0;
  VIterator vit = mVectors.begin();

  while (vit) {
    CIterator tmpit = list.begin();
    CIterator cit = (*vit).begin();

    while (tmpit && (*tmpit)->isZero() && cit && (*cit)->isZero()) {
      ++tmpit;
      ++cit;
    }

    if (!tmpit) { // tmp is the zero vector
      break;
    }

    if ((*cit)->isZero()) {
      vit.insert();
      (*vit).swap(list);

      // insert tmp into the matrix and shift the other rows down
      for (int j = i; j < mSize; j++) 
        mOps.add(new ISwapOperation(j, mSize));

      ++mSize;
      return false;
    }
    else if (!(*tmpit)->isZero()) { // && !(*cit)->isZero() 
      if (mCoeffInterface->isField()) {
        ICoeff* c = (*cit)->copy();
        c->inverse();
        c->mult(*(*tmpit));
        c->minus();

        mOps.add(new IAddMultOperation(mSize, i, c));

        while (tmpit && cit) {
          (*tmpit)->addmul(*(*cit), *c);
          ++tmpit;
          ++cit;
        }
      } 
      else {
        ICoeff* c = (*tmpit)->copy();
        c->minus();

        ICoeff* d = (*cit)->copy();

        ICoeff* g = mCoeffInterface->create();
        g->gcd(*c, *d);

        c->div(*g);
        d->div(*g);

        delete g;

        mOps.add(new IMultOperation(mSize, d));
        mOps.add(new IAddMultOperation(mSize, i, c));

        while (tmpit && cit) {
          (*tmpit)->mult(*d);
          (*tmpit)->addmul(*(*cit), *c);
          ++tmpit;
          ++cit;
        }
      }
    }

    ++vit;
    ++i;
  }

  vit.insert();
  (*vit).swap(list);
  ++mSize;

  CIterator it = (*vit).begin();

  while (it) {
    if (!(*it)->isZero())
      return false;
    ++it;
  }

  return true;
}


ICoeff** IGauss::createUnitVector(int i) const {
  IASSERT(mSize > 0);
  IASSERT(i < mSize);
  ICoeff** result = new ICoeff*[mSize];

  for (int j = 0; j < mSize; j++) {
    result[j] = mCoeffInterface->create();
    if (i == j)
      result[j]->setOne();
    else
      result[j]->setZero();
  }

  return result;
}


void IGauss::getMinPoly(IOneParameterPoly& minPoly) const {
  IASSERT(!mVectors.isEmpty());

  if (mCoeffInterface->isField()) {
    minPoly.setOne();
    minPoly.mult(mSize - 1);
  }
  else {
    ICoeff** v = createUnitVector(mSize - 1);
    mOps.apply(v);

    minPoly.setZero();
    minPoly.add(mSize - 1, *v[mSize-1]);

    delete[] v;
  }

  for (int i = 0; i < mSize - 1; i++) {
    ICoeff** v = createUnitVector(i);
    mOps.apply(v);

    minPoly.add(i, *v[mSize-1]);

    delete[] v;
  }

  ICoeff** v = createUnitVector(mSize - 1);
  mOps.apply(v);

  delete[] v;

  minPoly.pp();
}


std::ostream& operator<<(std::ostream& out, const IGauss& a) {
  IList<IList<ICoeff*> >::ConstIterator vit = a.mVectors.begin();

  while (vit) {
    out << "[";
    IList<ICoeff*>::ConstIterator cit = (*vit).begin();
    out << *(*cit);
    ++cit;
    while (cit) {
      out << ", " << *(*cit);
      ++cit;
    }
    out << "]" << std::endl;
    ++vit;
  }

  return out;
}

template <>
IAllocator IList<IList<ICoeff*> >::Node::sAllocator(sizeof(IList<IList<ICoeff*> >::Node));
