/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iomanip>

#include "izippelgauss.h"
#include "iswapoperation.h"
#include "imultoperation.h"
#include "iaddmultoperation.h"

ISubsystem::~ISubsystem() {
  for (VIterator it = mTerms.begin(); it != mTerms.end(); it++) {
    delete (*it);
  }

  for (VIterator it = mScalings.begin(); it != mScalings.end(); it++) {
    delete (*it);
  }
}

bool ISubsystem::addVector(std::list<ModularShort>* terms,
    std::list<ModularShort>* scalings) {
  VIterator termvectors = mTerms.begin();
  VIterator scalingvectors = mScalings.begin();

  while (termvectors != mTerms.end()) {
    CIterator newterms = terms->begin();
    CIterator termcoeffs = (*termvectors)->begin();

    while (newterms != terms->end() && (*newterms) == 0 && termcoeffs != (*termvectors)->end() && (*termcoeffs) == 0) {
      ++newterms;
      ++termcoeffs;
    }

    if (newterms == terms->end()) { // terms is the zero vector
      return true;
    }

    if ((*termcoeffs) == 0) {
      // insert the vectors at this poscalingvectorsion
      mTerms.insert(termvectors, terms);

      mScalings.insert(scalingvectors, scalings);

      ++mSize;
      return false;
    } else if ((*newterms) != 0) { // && !(*termcoeffs)->isZero() 
      // perform elimination
      ModularShort c = (*termcoeffs);
      c = iInverse(c, mPrime);
      c *= (*newterms);
      c %= mPrime;
      c = mPrime - c;

      while (newterms != terms->end() && termcoeffs != (*termvectors)->end()) {
        (*newterms) += (*termcoeffs)*c;
        (*newterms) %= mPrime;

        ++newterms;
        ++termcoeffs;
      }

      CIterator newscalings = scalings->begin();
      CIterator scalingcoeffs = (*scalingvectors)->begin();

      while (newscalings != scalings->end() && scalingcoeffs != (*scalingvectors)->end()) {
        (*newscalings) += (*scalingcoeffs)*c;
        (*newscalings) %= mPrime;

        ++newscalings;
        ++scalingcoeffs;
      }
    }

    ++termvectors;
    ++scalingvectors;
  }

  CIterator newterms = terms->begin();
  bool is_zero = true;
  while (newterms != terms->end() && is_zero) {
    is_zero &= ((*newterms) == 0);
    ++newterms;
  }

  if (is_zero)
    return true;

  mTerms.push_back(terms);
  mScalings.push_back(scalings);

  ++mSize;
  return false;
}

void ISubsystem::addScalingComponent() {
  VIterator scalingvectors = mScalings.begin();
  while (scalingvectors != mScalings.end()) {
    (*scalingvectors)->push_front(0);
    ++scalingvectors;
  }

  ++mComponents;
}

ModularShort* ISubsystem::getSolution(ModularShort scalings[]) {
  ModularShort* result = new ModularShort[mSize];

  int i = mSize - 1;
  for (std::list<std::list<ModularShort>*>::reverse_iterator it =
      mScalings.rbegin(); it != mScalings.rend(); ++it) {
    std::list<ModularShort>::reverse_iterator jt = (*it)->rbegin();
    result[i] = (*jt);
    ++jt;

    int j = mComponents - 2;
    while (jt != (*it)->rend()) {
      result[i] += scalings[j] * (*jt);
      result[i] %= mPrime;

      ++jt;
      --j;
    }

    --i;
  }

  i = mSize - 1;
  for (std::list<std::list<ModularShort>*>::reverse_iterator it = mTerms.rbegin(); it
      != mTerms.rend(); ++it) {
    std::list<ModularShort>::reverse_iterator jt = (*it)->rbegin();
    for (int j = mSize - 1; j > i; j--) {
      ModularShort tmp = (*jt)*result[j];
      tmp %= mPrime;
      result[i] += mPrime - tmp;
      result[i] %= mPrime;

      ++jt;
    }

    ModularShort c = iInverse((*jt), mPrime);
    result[i] *= c;
    result[i] %= mPrime;

    --i;
  }

  return result;
}

std::ostream& operator<<(std::ostream& out, const ISubsystem& a) {
  std::list<std::list<ModularShort>*>::const_iterator termvectors =
      a.mTerms.begin();
  std::list<std::list<ModularShort>*>::const_iterator scalingvectors =
      a.mScalings.begin();
  while (termvectors != a.mTerms.end()) {
    for (std::list<ModularShort>::const_iterator coeffs = (*termvectors)->begin(); coeffs != (*termvectors)->end(); ++coeffs) {
      out << std::setw(5) << (*coeffs) << " ";
    }
    out << " |  ";

    for (std::list<ModularShort>::const_iterator scalings = (*scalingvectors)->begin(); scalings != (*scalingvectors)->end(); ++scalings) {
      out << std::setw(5) << (*scalings) << " ";
    }
    out << std::endl;
    ++termvectors;
    ++scalingvectors;
  }

  return out;
}

ISingleSystem::~ISingleSystem() {
  //std::cout << "ISingleSystem:" << std::endl;
  //std::cout << *this << std::endl;
  VIterator it = mVectors.begin();
  while (it != mVectors.end()) {
    delete (*it);
    ++it;
  }
}

bool ISingleSystem::addVector(std::list<ModularShort>* list) {
  VIterator vectors = mVectors.begin();

  while (vectors != mVectors.end()) {
    CIterator newcoeffs = list->begin();
    CIterator vectorcoeffs = (*vectors)->begin();

    while (newcoeffs != list->end() 
        && (*newcoeffs) == 0 
        && vectorcoeffs != (*vectors)->end() 
        && (*vectorcoeffs) == 0) {
      ++newcoeffs;
      ++vectorcoeffs;
    }

    if (newcoeffs == list->end()) { // list is the zero vector
      return true;
    }

    if ((*vectorcoeffs) == 0) {
      // insert the vectors at this position
      mVectors.insert(vectors, list);

      ++mSize;
      return false;
    } else if ((*newcoeffs) != 0) { // && !(*vectorcoeffs)->isZero() 
      // perform elimination
      ModularShort c = (*vectorcoeffs);
      c = iInverse(c, mPrime);
      c *= (*newcoeffs);
      c %= mPrime;
      c = mPrime - c;

      while (newcoeffs != list->end() && vectorcoeffs != (*vectors)->end()) {
        (*newcoeffs) += (*vectorcoeffs)*c;
        (*newcoeffs) %= mPrime;

        ++newcoeffs;
        ++vectorcoeffs;
      }
    }

    ++vectors;
  }

  CIterator newvector = list->begin();
  bool is_zero = true;
  while (newvector != list->end() && is_zero) {
    is_zero &= ((*newvector) == 0);
    ++newvector;
  }

  if (is_zero) {
    return true;
  }

  if (!is_zero && newvector == list->end())
    throw IInconsistentSystemError();

  mVectors.push_back(list);

  ++mSize;
  return false;
}

void ISingleSystem::addScalingComponent() {
  VIterator vectors = mVectors.begin();
  while (vectors != mVectors.end()) {
    (*vectors)->push_front(0);
    ++vectors;
  }

  ++mComponents;
}

int ISingleSystem::degreeOfFreedom() const {
  int result = mComponents - mSize - 1;
  if (result < 0)
    return 0;
  return result;
}

ModularShort* ISingleSystem::getSolution() {
  ModularShort* result = new ModularShort[mSize];

  int i = mSize - 1;
  for (std::list<std::list<ModularShort>*>::reverse_iterator it =
      mVectors.rbegin(); it != mVectors.rend(); ++it) {
    std::list<ModularShort>::reverse_iterator jt = (*it)->rbegin();
    result[i] = (*jt);
    ++jt;

    for (int j = mSize - 1; j > i; j--) {
      ModularShort tmp = (*jt)*result[j];
      tmp %= mPrime;
      result[i] += mPrime - tmp;
      result[i] %= mPrime;

      ++jt;
    }

    ModularShort c = iInverse((*jt), mPrime);
    result[i] *= c;
    result[i] %= mPrime;
    //result[i] = mPrime - result[i];

    --i;
  }

  return result;
}

std::ostream& operator<<(std::ostream& out, const ISingleSystem& a) {
  for (std::list<std::list<ModularShort>*>::const_iterator vectors =
      a.mVectors.begin(); vectors != a.mVectors.end(); ++vectors) {
    for (std::list<ModularShort>::const_iterator coeffs = (*vectors)->begin(); 
        coeffs != (*vectors)->end(); 
        ++coeffs) {
      
      out << std::setw(5) << (*coeffs) << " ";
    }
    out << std::endl;
  }

  return out;
}

IMultipleScalingSystem::IMultipleScalingSystem(INParameterModularShortAux& form,
    ModularShort p) :
  mForm(form), mPrime(p), mScalings(p), mVars(form.interface().getEndVar()), mDeg((*form.begin()).mMonom[0]), mDegreeOfFreedomNotChanged(0) {

  mSystems = new ISubsystem*[mDeg + 1];
  for (int i = 0; i <= mDeg; i++) {
    mSystems[i] = new ISubsystem(p);
  }

  mNumberOfTerms = new int[mDeg + 1];
  for (int i = 0; i <= mDeg; i++) {
    mNumberOfTerms[i] = 0;
  }
  INParameterModularShortAux::ConstIterator it = form.begin();
  while (it) {
    mNumberOfTerms[(*it).mMonom[0]]++;
    ++it;
  }
}

IMultipleScalingSystem::~IMultipleScalingSystem() {
  for (int i = 0; i <= mDeg; i++) {
    delete mSystems[i];
  }

  delete[] mSystems;
  delete[] mNumberOfTerms;
}

ModularShort IMultipleScalingSystem::eval(IMonomAuxInterface::MonomIterator monom,
    ModularShort alpha[]) {
  ModularShort result = 1;

  for (int i = 0; i < mForm.interface().getEndVar() - 1; i++) {
    result *= iPow0(alpha[i], monom[i + 1], mPrime);
    result %= mPrime;
  }

  return result;
}

std::list<ModularShort>* IMultipleScalingSystem::getScalingVector(int deg,
    int size,
    INParameterModularShortAux& poly) {
  std::list<ModularShort>* result = new std::list<ModularShort>;

  INParameterModularShortAux::ConstIterator polyit = poly.begin();
  while (polyit && (*polyit).mMonom[0] > deg) {
    ++polyit;
  }

  if (polyit && (*polyit).mMonom[0] == deg) {
    result->push_back((*polyit).mCoeff);
  } else
    result->push_back(0);

  for (int i = 1; i < size; i++) {
    result->push_back(0);
  }

  return result;
}

void IMultipleScalingSystem::add(INParameterModularShortAux& poly,
    ModularShort alpha[]) {
  int tmp = degreeOfFreedom();

  for (int i = 0; i <= mDeg; i++) {
    mSystems[i]->addScalingComponent();
  }
  mScalings.addScalingComponent();

  int deg = -1;

  std::list<ModularShort>* gcdcoeffs;

  //std::cout << "form = " << mForm << std::endl;

  for (INParameterModularShortAux::ConstIterator it = mForm.begin(); it; ++it) {
    if ((*it).mMonom[0] != deg) {
      if (deg != -1) {
        std::list<ModularShort>* scalings = getScalingVector(deg,
            mScalings.components(), poly);
        //std::cout << "deg = " << deg << std::endl;
        //std::list<ModularShort>::iterator test = gcdcoeffs->begin();
        //while (test != gcdcoeffs->end()) {
        //  std::cout << std::setw(5) << (*test) << " ";
        //  ++test;
        //}
        //std::cout << std::endl;
        bool linearly_dependent = mSystems[deg]->addVector(gcdcoeffs, scalings);
        if (linearly_dependent) {
          mScalings.addVector(scalings);
        }
      }

      gcdcoeffs = new std::list<ModularShort>;

      deg = (*it).mMonom[0];
    }

    ModularShort t = eval((*it).mMonom, alpha);
    //std::cout << "gcdcoeffs.size() = " << gcdcoeffs->size() << std::endl;
    gcdcoeffs->push_back(t);
  }

  std::list<ModularShort>* scalings = getScalingVector(deg,
      mScalings.components(), poly);
  //std::cout << "deg = " << deg << std::endl;
  //std::list<ModularShort>::iterator test = gcdcoeffs->begin();
  //while (test != gcdcoeffs->end()) {
  //  std::cout << std::setw(5) << (*test) << " ";
  //  ++test;
  //}
  //std::cout << std::endl;
  bool linearly_dependent = mSystems[deg]->addVector(gcdcoeffs, scalings);
  if (linearly_dependent) {
    mScalings.addVector(scalings);
  }

  if (degreeOfFreedom() != tmp) {
    mDegreeOfFreedomNotChanged = 0;
  } else {
    mDegreeOfFreedomNotChanged++;
  }
}

int IMultipleScalingSystem::degreeOfFreedom() {
  int result = 0;
  for (int i = 0; i <= mDeg; i++) {
    result += mNumberOfTerms[i] - mSystems[i]->size();
  }

  result += mScalings.components() - mScalings.size();

  if (mScalings.components() == 0)
    return result;
  else
    return result - 1;
}

int IMultipleScalingSystem::degreeOfFreedomNotChanged() {
  return mDegreeOfFreedomNotChanged;
}

void IMultipleScalingSystem::getSolution(INParameterModularShortAux& poly) {
  poly.setZero();

  ModularShort *scalings = mScalings.getSolution();
  for (int i = 0; i < mScalings.size(); i++) {
    scalings[i] = mPrime - scalings[i];
  }

  INParameterModularShortAux::ConstIterator formit = mForm.begin();
  INParameterModularShortAux::Iterator polyit = poly.begin();

  for (int i = mDeg; i >= 0; i--) {
    ModularShort *sol = mSystems[i]->getSolution(scalings);

    for (int j = 0; j < mSystems[i]->size(); j++) {
      if (sol[j] != 0) {
        polyit.insert();
        (*polyit).mMonom = mForm.interface().copy((*formit).mMonom);
        (*polyit).mCoeff = sol[j];
        ++polyit;
      }
      ++formit;
    }

    delete[] sol;
  }

  delete[] scalings;
}

void IMultipleScalingSystem::output(std::ostream& out) const {
  out << mForm << std::endl;
  for (int i = mDeg; i >= 0; i--) {
    out << "deg " << i << std::endl;
    out << *mSystems[i] << std::endl;
    out << "-------" << std::endl;
  }

  out << "scalings: " << std::endl;
  out << mScalings << std::endl;
}

ISingleScalingSystem::ISingleScalingSystem(INParameterModularShortAux& form,
    int scalingDegree,
    ModularShort p) :
  mForm(form), mPrime(p), mScalingDegree(scalingDegree), mVars(form.interface().getEndVar()), mDeg((*form.begin()).mMonom[0]), mDegreeOfFreedomNotChanged(0) {

  mSystems = new ISingleSystem*[mDeg + 1];
  for (int i = 0; i <= mDeg; i++) {
    mSystems[i] = new ISingleSystem(p);
  }

  mNumberOfTerms = new int[mDeg + 1];
  for (int i = 0; i <= mDeg; i++) {
    mNumberOfTerms[i] = 0;
  }
  INParameterModularShortAux::ConstIterator it = form.begin();
  while (it) {
    if ((*it).mMonom[0] == scalingDegree) {
      mScalingMonom = form.interface().copy((*it).mMonom);
    }

    mNumberOfTerms[(*it).mMonom[0]]++;
    ++it;
  }
}

ISingleScalingSystem::~ISingleScalingSystem() {
  for (int i = 0; i <= mDeg; i++) {
    delete mSystems[i];
  }

  delete[] mSystems;
  delete[] mNumberOfTerms;
  delete[] mScalingMonom;
}

ModularShort ISingleScalingSystem::eval(IMonomAuxInterface::MonomIterator monom,
    ModularShort alpha[]) {
  ModularShort result = 1;

  for (int i = 0; i < mForm.interface().getEndVar() - 1; i++) {
    result *= iPow0(alpha[i], monom[i + 1], mPrime);
    result %= mPrime;
  }

  return result;
}

ModularShort ISingleScalingSystem::scalingCoeff(const INParameterModularShortAux& poly,
    ModularShort alpha[]) {
  ModularShort result = eval(mScalingMonom, alpha);
  INParameterModularShortAux::ConstIterator it = poly.begin();

  while (it) {
    if ((*it).mMonom[0] == mScalingDegree) {
      result *= iInverse((*it).mCoeff, mPrime);
      result %= mPrime;
      break;
    }

    IASSERT((*it).mMonom[0] >= mScalingDegree);

    ++it;
  }

  return result;
}

void ISingleScalingSystem::add(INParameterModularShortAux& poly,
    ModularShort alpha[]) {

  int tmp = degreeOfFreedom();
  ModularShort scaling = scalingCoeff(poly, alpha);

  int deg = -1;

  std::list<ModularShort>* gcdcoeffs;

  INParameterModularShortAux::ConstIterator it = mForm.begin();
  INParameterModularShortAux::ConstIterator ip = poly.begin();
  while (it) {
    if ((*it).mMonom[0] != deg) {
      if (deg != -1 && deg != mScalingDegree) {
        while (ip && (*ip).mMonom[0] > deg)
          ++ip;
        ModularShort gcd = 0;
        // TODO: check this!
        if (ip && (*ip).mMonom[0] == deg) {
          gcd = (*ip).mCoeff;
          gcd *= scaling;
          gcd %= mPrime;
        }
        gcdcoeffs->push_back(gcd);
        mSystems[deg]->addVector(gcdcoeffs);
      }

      if ((*it).mMonom[0] != mScalingDegree)
        gcdcoeffs = new std::list<ModularShort>;

      deg = (*it).mMonom[0];
    }

    if (deg != mScalingDegree) {
      ModularShort t = eval((*it).mMonom, alpha);
      gcdcoeffs->push_back(t);
    }

    ++it;
  }

  if (deg != mScalingDegree) {
    while (ip && (*ip).mMonom[0] > deg)
      ++ip;
    ModularShort gcd = 0;
    if (ip && (*ip).mMonom[0] == deg) {
      gcd = (*ip).mCoeff;
      gcd *= scaling;
      gcd %= mPrime;
    }
    gcdcoeffs->push_back(gcd);
    mSystems[deg]->addVector(gcdcoeffs);
  }

  if (degreeOfFreedom() != tmp) {
    mDegreeOfFreedomNotChanged = 0;
  } else {
    mDegreeOfFreedomNotChanged++;
  }
}

int ISingleScalingSystem::degreeOfFreedom() {
  int result = 0;
  for (int i = 0; i <= mDeg; i++) {
    if (i != mScalingDegree)
      result += mNumberOfTerms[i] - mSystems[i]->size();
  }

  return result;
}

int ISingleScalingSystem::degreeOfFreedomNotChanged() {
  return mDegreeOfFreedomNotChanged;
}

void ISingleScalingSystem::getSolution(INParameterModularShortAux& poly) {
  poly.setZero();

  INParameterModularShortAux::ConstIterator formit = mForm.begin();
  INParameterModularShortAux::Iterator polyit = poly.begin();

  for (int i = mDeg; i >= 0; i--) {
    if (i != mScalingDegree) {
      ModularShort *sol = mSystems[i]->getSolution();

      for (int j = 0; j < mSystems[i]->size(); j++) {
        if (sol[j] != 0) {
          polyit.insert();
          (*polyit).mMonom = mForm.interface().copy((*formit).mMonom);
          (*polyit).mCoeff = sol[j];
          ++polyit;
        }
        ++formit;
      }

      delete[] sol;
    } else {
      polyit.insert();
      (*polyit).mCoeff = 1;
      (*polyit).mMonom = mForm.interface().copy(mScalingMonom);
      ++polyit;
      ++formit;
    }
  }
}

void ISingleScalingSystem::output(std::ostream& out) const {
  out << mForm << std::endl;
  for (int i = mDeg; i >= 0; i--) {
    out << "deg " << i << std::endl;
    out << *mSystems[i] << std::endl;
    out << "-------" << std::endl;
  }
}
