/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          * *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IZIPPELGAUSS_H
#define IZIPPELGAUSS_H

#include <iostream>
#include <list>

#include "ichainoperation.h"
#include "../coeff/icoeff.h"
#include "../coeff/imonomaux.h"
#include "../coeff/inparametermodularshortaux.h"
#include "../gauss/izippelgauss.h"
#include "../poly/ioneparameterpoly.h"
#include "../util/imodular.h"

class IInconsistentSystemError {
};

class ISubsystem {
protected:
  std::list<std::list<ModularShort>*> mTerms;
  std::list<std::list<ModularShort>*> mScalings;
  int mSize;
  int mComponents;
  ModularShort mPrime;

  typedef std::list<std::list<ModularShort>*>::iterator VIterator;
  typedef std::list<ModularShort>::iterator CIterator;

public:
  ISubsystem(ModularShort p): mPrime(p), mSize(0), mComponents(0) { }
  ~ISubsystem();

  bool addVector(std::list<ModularShort>* terms, std::list<ModularShort>* scalings);

  void addScalingComponent();

  int size() const {
    return mSize;
  }

  ModularShort* getSolution(ModularShort scalings[]);

  friend std::ostream& operator<<(std::ostream& out, const ISubsystem& a);
};

class ISingleSystem {
protected:
  std::list<std::list<ModularShort>*> mVectors;
  int mSize;
  int mComponents;
  ModularShort mPrime;

  typedef std::list<std::list<ModularShort>*>::iterator VIterator;
  typedef std::list<ModularShort>::iterator CIterator;

public:
  ISingleSystem(ModularShort p): mPrime(p), mSize(0), mComponents(0) { }
  ~ISingleSystem();

  bool addVector(std::list<ModularShort>* list);

  void addScalingComponent();

  int size() const {
    return mSize;
  }

  int components() const {
    return mComponents;
  }

  int degreeOfFreedom() const;

  ModularShort* getSolution();

  friend std::ostream& operator<<(std::ostream& out, const ISingleSystem& a);
};

class IZippelSystem {
public:
  virtual ~IZippelSystem() { }
  virtual void add(INParameterModularShortAux& poly, ModularShort alpha[]) = 0;

  virtual int degreeOfFreedom() = 0;

  virtual int degreeOfFreedomNotChanged() = 0;

  virtual void getSolution(INParameterModularShortAux& poly) = 0;

  virtual void output(std::ostream& out) const = 0 ;
  
  friend std::ostream& operator<<(std::ostream& out, const IZippelSystem& a) {
    a.output(out);
    return out;
  }
};

class IMultipleScalingSystem : public IZippelSystem {
protected:
  INParameterModularShortAux mForm;
  ModularShort mPrime;
  ISubsystem** mSystems;
  ISingleSystem mScalings;
  int mVars;
  int mDeg;
  int mDegreeOfFreedomNotChanged;
  int* mNumberOfTerms;

private:
  ModularShort eval(IMonomAuxInterface::MonomIterator monom, ModularShort alpha[]);
  std::list<ModularShort>* getScalingVector(int deg, int size, INParameterModularShortAux& poly);

public:
  IMultipleScalingSystem(INParameterModularShortAux& form, ModularShort p);
  ~IMultipleScalingSystem();

  void add(INParameterModularShortAux& poly, ModularShort alpha[]);

  int degreeOfFreedom();

  int degreeOfFreedomNotChanged();

  void getSolution(INParameterModularShortAux& poly);
  
  void output(std::ostream& out) const;
};

class ISingleScalingSystem : public IZippelSystem {
protected:
  INParameterModularShortAux mForm;
  ModularShort mPrime;
  int mScalingDegree;
  IMonomAuxInterface::MonomIterator mScalingMonom;

  ISingleSystem** mSystems;
  int mVars;
  int mDeg;
  int mDegreeOfFreedomNotChanged;
  int* mNumberOfTerms;

private:
  ModularShort eval(IMonomAuxInterface::MonomIterator monom, ModularShort alpha[]);
  ModularShort scalingCoeff(const INParameterModularShortAux& poly, ModularShort alpha[]);

public:
  ISingleScalingSystem(INParameterModularShortAux& form, int scalingDegree, ModularShort p);
  ~ISingleScalingSystem();

  void add(INParameterModularShortAux& poly, ModularShort alpha[]);

  int degreeOfFreedom();

  int degreeOfFreedomNotChanged();

  void getSolution(INParameterModularShortAux& poly);
  
  void output(std::ostream& out) const;
};

#endif // IZIPPELGAUSS_H
