/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../util/iutil.h"
#include "./ideglex.h"


IDegLexInterface::IDegLexInterface(const ISystemType *systemType, IVariables* independ):
    ILexInterface(systemType, independ) {
  mOrder = IMonomInterface::DegLex;
}

IMonom* IDegLexInterface::create() {
  return new IDegLex(this);
}

IMonom* IDegLexInterface::copy(const IMonom &a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  IDegLex* r;
  if (this == a.monomInterface())
    r = new IDegLex(a);
  else {
    r = new IDegLex(this);
    r->IDegLex::setZero();
    for(int var=0; var < dimDepend(); var++)
      r->IDegLex::prolong(var, a.deg(var));
  }
  return r;
}

IDegLex::IDegLex(IDegLexInterface* r):
    ILex(r) {
}

IDegLex::IDegLex(const IMonom& a):
    ILex(a) {
}

IDegLex::~IDegLex() {
}

unsigned IDegLex::deg(int var) const {
  IASSERT(0 <= var && var < dimIndepend());
  unsigned d;
  if (var == dimIndepend()-1)
    d = mBegin[var];
  else
    d = mBegin[var] - mBegin[var+1];
   return d;
}

unsigned IDegLex::degree() const {
  return *mBegin;
}

int IDegLex::compare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  ConstIterator i = mBegin,
                ia = ((const IDegLex&)a).mBegin;
  if (*i > *ia)
    return 1;
  else if (*i < *ia)
    return -1;
  ConstIterator end = i + dimIndepend();
  do {
    ++i;
    ++ia;
  } while(i < end && *i == *ia);
  if (i == end)
    return 0;
  else if (*i < *ia)
    return 1;
  else
    return -1;
}

int IDegLex::compare(const IMonom& a, int var) const {
  IASSERT(monomInterface() == a.monomInterface());
  ConstIterator i = mBegin,
                ia = ((const IDegLex&)a).mBegin;
  if (*i > *ia)
    return 1;
  else if (*i < *ia)
    return -1;
  ConstIterator end = i + var;
  do {
    ++i;
    ++ia;
  } while(i < end && *i == *ia);
  if (i == end)
    return 0;
  else if (*i < *ia)
    return 1;
  else
    return -1;
}

int IDegLex::compare(const IMonom& a, const IMonom& b) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(monomInterface() == b.monomInterface());
  ConstIterator i = mBegin,
                ia = ((const IDegLex&)a).mBegin,
                ib = ((const IDegLex&)b).mBegin;
  if (*i > *ia + *ib)
    return 1;
  else if (*i < *ia + *ib)
    return -1;
  ConstIterator end = i + dimIndepend();
  do {
    ++i;
    ++ia;
    ++ib;
  } while(i < end && *i == *ia + *ib);
  if (i == end)
    return 0;
  else if (*i < *ia + *ib)
    return 1;
  else
    return -1;
}

int IDegLex::setQcompare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  ConstIterator i = mBegin,
                ia = ((const IDegLex&)a).mBegin;
  if (*i > *ia)
    return 1;
  else if (*i < *ia)
    return -1;
  ConstIterator end = i + dimIndepend();
  do {
    ++i;
    ++ia;
  } while(i < end && *i == *ia);
  if (i == end)
    return 0;
  else if (*i < *ia)
    return 1;
  else
    return -1;
}

void IDegLex::prolong(int var) {
  IASSERT(0 <= var && var < dimIndepend());
  Iterator i = mBegin;
  ConstIterator end = i + var;
  do {
    ++(*i++);
  } while(i <= end);
  IASSERTVALID(*this);
}

void IDegLex::prolong(int var, unsigned deg) {
  IASSERT(0 <= var && var < dimIndepend());
  Iterator i = mBegin;
  ConstIterator end = i + var;
  do {
    *i++ += deg;
  } while(i <= end);
  IASSERTVALID(*this);
}

void IDegLex::div(int var) {
  IASSERT(0 <= var && var < dimIndepend());
  Iterator i = mBegin;
  ConstIterator end = i + var;
  do {
    --(*i++);
  } while(i <= end);
  IASSERTVALID(*this);
}

bool IDegLex::divisibility(const IMonom& a) const{
  IASSERT(monomInterface() == a.monomInterface());
  for(int var = 0; var < dimIndepend(); var++)
    if (IDegLex::deg(var) < ((const IDegLex&)a).IDegLex::deg(var))
      return false;
  return true;
}

static inline unsigned min(unsigned a, unsigned b) {
  return (a < b) ? a: b;
}

unsigned IDegLex::gcd(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  unsigned d = 0;
  for(int var = 0; var < dimIndepend(); var++)
    d += min(IDegLex::deg(var), ((const IDegLex&)a).IDegLex::deg(var));
  return d;
}

static inline unsigned max(unsigned a, unsigned b) {
  return (a > b) ? a: b;
}

unsigned IDegLex::lcm(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  unsigned d = 0;
  for(int var = 0; var < dimIndepend(); var++)
    d += max(IDegLex::deg(var), ((const IDegLex&)a).IDegLex::deg(var));
  return d;
}

void IDegLex::assertValid(const char* fileName, int fileLine) const {
  unsigned d=0;
  for(int var = 0; var < dimIndepend(); var++) {
    IASSERT2(fileName, fileLine, deg(var) >= 0);
    d += deg(var);
  }
  IASSERT2(fileName, fileLine, d == degree());
}

IAllocator IDegLex::sAllocator(sizeof(IDegLex));
