/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../util/iutil.h"
#include "./idegrevlex.h"


IDegRevLexInterface::IDegRevLexInterface(const ISystemType *systemType, IVariables* independ):
    ILexInterface(systemType, independ) {
  mOrder = IMonomInterface::DegRevLex;
}

IMonom* IDegRevLexInterface::create() {
  return new IDegRevLex(this);
}

IMonom* IDegRevLexInterface::copy(const IMonom &a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  IDegRevLex* r;
  if (this == a.monomInterface())
    r = new IDegRevLex(a);
  else {
    r = new IDegRevLex(this);
    r->IDegRevLex::setZero();
    for(int var=0; var < dimDepend(); var++)
      r->IDegRevLex::prolong(var, a.deg(var));
  }
  return r;
}

IDegRevLex::IDegRevLex(IDegRevLexInterface* r):
    ILex(r) {
}

IDegRevLex::IDegRevLex(const IMonom& a):
    ILex(a) {
}

IDegRevLex::~IDegRevLex() {
}

unsigned IDegRevLex::deg(int var) const {
  IASSERT(0 <= var && var < dimIndepend());
  ConstIterator end = mBegin + dimIndepend();
  unsigned d;
  if (var)
    d = end[-var-1] - end[-var];
  else
    d = end[-var-1];
   return d;
}

unsigned IDegRevLex::degree() const {
  return *mBegin;
}

void IDegRevLex::prolong(int var) {
  IASSERT(0 <= var && var < dimIndepend());
  Iterator i = mBegin;
  ConstIterator end = i + dimIndepend() - var;
  do {
    ++(*i++);
  } while(i < end);
  IASSERTVALID(*this);
}

void IDegRevLex::prolong(int var, unsigned deg) {
  IASSERT(0 <= var && var < dimIndepend());
  Iterator i = mBegin;
  ConstIterator end = i + dimIndepend() - var;
  do {
    *i++ += deg;
  } while(i < end);
  IASSERTVALID(*this);
}

void IDegRevLex::div(int var) {
  IASSERT(0 <= var && var < dimIndepend());
  Iterator i = mBegin;
  ConstIterator end = i + dimIndepend() - var;
  do {
    --(*i++);
  } while(i < end);
  IASSERTVALID(*this);
}

bool IDegRevLex::divisibility(const IMonom& a) const{
  IASSERT(monomInterface() == a.monomInterface());
  for(int var = 0; var < dimIndepend(); var++)
    if (IDegRevLex::deg(var) < ((const IDegRevLex&)a).IDegRevLex::deg(var))
      return false;
  return true;
}

static inline unsigned min(unsigned a, unsigned b) {
  return (a < b) ? a: b;
}

unsigned IDegRevLex::gcd(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  unsigned d = 0;
  for(int var = 0; var < dimIndepend(); var++)
    d += min(IDegRevLex::deg(var), ((const IDegRevLex&)a).IDegRevLex::deg(var));
  return d;
}

static inline unsigned max(unsigned a, unsigned b) {
  return (a > b) ? a: b;
}

unsigned IDegRevLex::lcm(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  unsigned d = 0;
  for(int var = 0; var < dimIndepend(); var++)
    d += max(IDegRevLex::deg(var), ((const IDegRevLex&)a).IDegRevLex::deg(var));
  return d;
}

void IDegRevLex::assertValid(const char* fileName, int fileLine) const {
  unsigned d=0;
  for(int var = 0; var < dimIndepend(); var++) {
    IASSERT2(fileName, fileLine, deg(var) >= 0);
    d += deg(var);
  }
  IASSERT2(fileName, fileLine, d == degree());
}

IAllocator IDegRevLex::sAllocator(sizeof(IDegRevLex));
