/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../util/iutil.h"
#include "./ielim.h"


IElimInterface::IElimInterface(const ISystemType *systemType, IVariables* independ, int elim):
    ILexInterface(systemType, independ),
    mElim(elim) {
  IASSERT(0 < mElim && mElim < mDimIndepend);
  mOrder = IMonomInterface::Elim;
}

int IElimInterface::elim() const {
  return mElim;
}

IMonom* IElimInterface::create() {
  return new IElim(this);
}

IMonom* IElimInterface::copy(const IMonom &a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  IElim* r;
  if (this == a.monomInterface())
    r = new IElim(a);
  else {
    r = new IElim(this);
    r->IElim::setZero();
    for(int var=0; var < dimDepend(); var++)
      r->IElim::prolong(var, a.deg(var));
  }
  return r;
}

IElim::IElim(IElimInterface* r):
  ILex(r) {
}

IElim::IElim(const IMonom& a):
  ILex(a) {
}

IElim::~IElim() {
}

unsigned IElim::deg(int var) const {
  IASSERT(0 <= var && var < dimIndepend());
  unsigned d;
  if (elim() <= var) {
    var -= elim();
    ConstIterator end = mBegin + dimIndepend();
    if (var)
      d = end[-var-1] - end[-var];
    else
      d = end[-var-1];
  }
  else {
    ConstIterator end = mBegin + elim();
    if (var)
      d = end[-var-1] - end[-var];
    else
      d = end[-var-1];
  }
  return d;
}

unsigned IElim::degree() const {
  return *mBegin + mBegin[elim()];
}

void IElim::prolong(int var) {
  IASSERT(0 <= var && var < dimIndepend());
  if (elim() <= var) {
    var -= elim();
    Iterator i = mBegin + elim();
    ConstIterator end = mBegin + dimIndepend() - var;
    do {
      ++(*i++);
    } while(i < end);
  }
  else {
    Iterator i = mBegin;
    ConstIterator end = mBegin + elim() - var;
    do {
      ++(*i++);
    } while(i < end);
  }
  IASSERTVALID(*this);
}

void IElim::prolong(int var, unsigned deg) {
  IASSERT(0 <= var && var < dimIndepend());
  if (elim() <= var) {
    var -= elim();
    Iterator i = mBegin + elim();
    ConstIterator end = mBegin + dimIndepend() - var;
    do {
      *i++ += deg;
    } while(i < end);
  }
  else {
    Iterator i = mBegin;
    ConstIterator end = mBegin + elim() - var;
    do {
      *i++ += deg;
    } while(i < end);
  }
  IASSERTVALID(*this);
}

void IElim::div(int var) {
  IASSERT(0 <= var && var < dimIndepend());
  if (elim() <= var) {
    var -= elim();
    Iterator i = mBegin + elim();
    ConstIterator end = mBegin + dimIndepend() - var;
    do {
      --(*i++);
    } while(i < end);
  }
  else {
    Iterator i = mBegin;
    ConstIterator end = mBegin + elim() - var;
    do {
      --(*i++);
    } while(i < end);
  }
  IASSERTVALID(*this);
}

bool IElim::divisibility(const IMonom& a) const{
  IASSERT(monomInterface() == a.monomInterface());
  int var;
  for(var = 0; var < dimIndepend(); var++)
    if (IElim::deg(var) < ((const IElim&)a).IElim::deg(var))
      return false;
  return true;
}

static inline unsigned min(unsigned a, unsigned b) {
  return (a < b) ? a: b;
}

unsigned IElim::gcd(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  unsigned d = 0;
  for(int var = 0; var < dimIndepend(); var++)
    d += min(IElim::deg(var), ((const IElim&)a).IElim::deg(var));
  return d;
}

static inline unsigned max(unsigned a, unsigned b) {
  return (a > b) ? a: b;
}

unsigned IElim::lcm(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  unsigned d = 0;
  for(int var = 0; var < dimIndepend(); var++)
    d += max(IElim::deg(var), ((const IElim&)a).IElim::deg(var));
  return d;
}

void IElim::assertValid(const char* fileName, int fileLine) const {
  ILex::assertValid(fileName, fileLine);
}

IAllocator IElim::sAllocator(sizeof(IElim));
