/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../util/iutil.h"
#include "./ilex.h"


ILexInterface::ILexInterface(const ISystemType *systemType, IVariables* independ):
    IMonomInterface(systemType),
    mAllocator(sizeof(ILex::Word)*independ->dim()) {
  mOrder = IMonomInterface::Lex;
  mDimIndepend = independ->dim();
  mIndepend = independ;
}

IMonom* ILexInterface::create() {
  return new ILex(this);
}

IMonom* ILexInterface::copy(const IMonom &a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  ILex* r;
  if (this == a.monomInterface())
    r = new ILex(a);
  else {
    r = new ILex(this);
    r->ILex::setZero();
    for(int var=0; var < dimDepend(); var++)
      r->ILex::prolong(var, a.deg(var));
  }
  return r;
}

ILex::ILex(ILexInterface* r):
    IMonom(r),
    mBegin((Iterator)r->mAllocator.allocate()) {
}

ILex::ILex(const IMonom& a):
    IMonom((ILexInterface*)a.monomInterface()),
    mBegin((Iterator)((ILexInterface*)a.monomInterface())->mAllocator.allocate()) {
  Iterator i = mBegin;
  ConstIterator end = i + dimIndepend(), ia = ((const ILex&)a).mBegin;
  do {
    *i++ = *ia++;
  } while(i < end);
  IASSERTVALID(*this);
}

ILex::~ILex() {
  ((ILexInterface*)monomInterface())->mAllocator.deallocate(mBegin);
}

unsigned ILex::deg(int var) const {
  IASSERT(0 <= var && var < dimIndepend());
  return mBegin[var];
}

unsigned ILex::degree() const {
  unsigned degree = 0;
  ConstIterator i = mBegin;
  ConstIterator end = i + dimIndepend();
  do {
    degree += *i++;
  } while(i < end);
  return degree;
}

int ILex::dependVar() const {
  return -1;
}

void ILex::set(const IMonom& a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  if (monomInterface() == a.monomInterface()) {
    Iterator i = mBegin;
    ConstIterator end = i + dimIndepend(), ia = ((const ILex&)a).mBegin;
    do {
      *i++ = *ia++;
    } while(i < end);
  }
  else {
    setZero();
    for(int var=0; var < dimDepend(); var++)
      prolong(var, a[var]);
  }
  IASSERTVALID(*this);
}

void ILex::setZero() {
  Iterator i = mBegin;
  ConstIterator end = i + dimIndepend();
  do {
    *i++ = 0;
  } while(i < end);
  IASSERTVALID(*this);
}

void ILex::setDependVar(int) {
  IERROR("setDependVar");
}

void ILex::swap(IMonom& a) {
  IASSERT(monomInterface() == a.monomInterface());
  Iterator tmp = mBegin;
  mBegin = ((ILex&)a).mBegin;
  ((ILex&)a).mBegin = tmp;
}

void ILex::prolong(int var) {
  IASSERT(0 <= var && var < dimIndepend());
  ++mBegin[var];
  IASSERTVALID(*this);
}

void ILex::prolong(int var, unsigned deg) {
  IASSERT(0 <= var && var < dimIndepend());
  mBegin[var] += deg;
  IASSERTVALID(*this);
}

void ILex::div(int var) {
  IASSERT(0 <= var && var < dimIndepend());
  --mBegin[var];
  IASSERTVALID(*this);
}

void ILex::mult(const IMonom& a) {
  IASSERT(monomInterface() == a.monomInterface());
  Iterator i = mBegin;
  ConstIterator end = i + dimIndepend(), ia = ((const ILex&)a).mBegin;
  do {
    *i++ += *ia++;
  } while(i < end);
  IASSERTVALID(*this);
}

void ILex::mult(const IMonom& a, const IMonom& b) {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(monomInterface() == b.monomInterface());
  Iterator i = mBegin;
  ConstIterator end = i + dimIndepend(),
                ia = ((const ILex&)a).mBegin,
                ib = ((const ILex&)b).mBegin;
  do {
    *i++ = *ia++ + *ib++;
  } while(i < end);
  IASSERTVALID(*this);
}

bool ILex::divisibility(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  int var;
  for(var = 0; var < dimIndepend(); var++)
    if (ILex::deg(var) < ((const ILex&)a).ILex::deg(var))
      return false;
  return true;
}

void ILex::divide(const IMonom& a, const IMonom& b) {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(monomInterface() == b.monomInterface());
  IASSERT(a.divisibility(b));
  Iterator i = mBegin;
  ConstIterator end = i + dimIndepend(),
                ia = ((const ILex&)a).mBegin,
                ib = ((const ILex&)b).mBegin;
  do {
    *i++ = *ia++ - *ib++;
  } while(i < end);
  IASSERTVALID(*this);
}

static inline unsigned min(unsigned a, unsigned b) {
  return (a < b) ? a: b;
}

unsigned ILex::gcd(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  unsigned d = 0;
  for(int var = 0; var < dimIndepend(); var++)
    d += min(ILex::deg(var), ((const ILex&)a).ILex::deg(var));
  return d;
}

static inline unsigned max(unsigned a, unsigned b) {
  return (a > b) ? a: b;
}

unsigned ILex::lcm(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  unsigned d = 0;
  for(int var = 0; var < dimIndepend(); var++)
    d += max(ILex::deg(var), ((const ILex&)a).ILex::deg(var));
  return d;
}

int ILex::compare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  ConstIterator i = mBegin,
                ia = ((const ILex&)a).mBegin;
  ConstIterator end = i + dimIndepend();
  while(i < end && *i == *ia) {
    ++i;
    ++ia;
  }
  if (i == end)
    return 0;
  else if (*i > *ia)
    return 1;
  else
    return -1;
}

int ILex::compare(const IMonom& a, int var) const {
  IASSERT(monomInterface() == a.monomInterface());
  ConstIterator i = mBegin,
                    end = i + var,
                    ia = ((const ILex&)a).mBegin;
  while(i < end && *i == *ia) {
    ++i;
    ++ia;
  }
  if (i == end)
    return 0;
  else if (*i > *ia)
    return 1;
  else
    return -1;
}

int ILex::compare(const IMonom& a, const IMonom& b) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(monomInterface() == b.monomInterface());
  ConstIterator i = mBegin,
                ia = ((const ILex&)a).mBegin,
                ib = ((const ILex&)b).mBegin;
  ConstIterator end = i + dimIndepend();
  while(i < end && *i == *ia + *ib) {
    ++i;
    ++ia;
    ++ib;
  }
  if (i == end)
    return 0;
  else if (*i > *ia + *ib)
    return 1;
  else
    return -1;
}

int ILex::setQcompare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  ConstIterator i = mBegin,
                ia = ((const ILex&)a).mBegin;
  ConstIterator end = i + dimIndepend();
  while(i < end && *i == *ia) {
    ++i;
    ++ia;
  }
  if (i == end)
    return 0;
  else if (*i > *ia)
    return 1;
  else
    return -1;
}

void ILex::read(std::istream &in) {
  in.read((char*)mBegin, sizeof(ILex::Word)*mRealization->dimIndepend());
  IASSERTVALID(*this);
}

void ILex::write(std::ostream &out) const {
  out.write((char*)mBegin, sizeof(ILex::Word)*mRealization->dimIndepend());
}

void ILex::assertValid(const char* fileName, int fileLine) const {
  for(int var = 0; var < dimIndepend(); var++)
    IASSERT2(fileName, fileLine, deg(var) >= 0);
}

IAllocator ILex::sAllocator(sizeof(ILex));
