/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <cstdarg>

#include "../util/iutil.h"
#include "./imonom.h"
#include "./idegrevlex.h"
#include "./idegrevlexbyte.h"
#include "./ideglex.h"
#include "./ilex.h"
#include "./idegrevlexdeg.h"
#include "./ielim.h"
#include "./itopdegrevlex.h"
#include "./itopdegrevlexbyte.h"
#include "./itopdeglex.h"
#include "./itoplex.h"
#include "./itopelim.h"
#include "./itopdegrevlexdeg.h"
#include "./ipotdegrevlex.h"
#include "./ipotdeglex.h"
#include "./ipotlex.h"
#include "./ipotelim.h"
#include "./ipotdegrevlexdeg.h"
#include "./iposelimdegrevlex.h"
#include "./iposelimdeglex.h"
#include "./iposelimlex.h"
#include "./iposelimelim.h"
#include "./iposelimdegrevlexdeg.h"

IMonomInterface::IMonomInterface(const ISystemType *systemType):
    mOrder(Indefinitely),
    mSystemType(systemType),
    mDimIndepend(0),
    mIndepend(NULL),
    mDimDepend(0),
    mDepend(NULL) {
}

IMonomInterface* IMonomInterface::create(IMonomInterface::Order order,
                                         const ISystemType *systemType,
                                         IVariables* independ,
                                         IVariables* depend,
                                         ...) {
  if (order == IMonomInterface::Indefinitely)
    IERROR("error in IMonomInterface: order == IMonomInterface::Indefinitely");
  if (systemType == NULL || systemType->type() == ISystemType::Indefinitely)
    IERROR("error in IMonomInterface: systemType == NULL || systemType->type() == ISystemType::Indefinitely");

  va_list marker;
  va_start(marker, depend);

  IMonomInterface* r = NULL;
  switch(order) {
  case DegRevLex:
    r = new IDegRevLexInterface(systemType, independ);
    break;
  case DegRevLexByte:
    r = new IDegRevLexByteInterface(systemType, independ);
    break;
  case DegLex:
    r = new IDegLexInterface(systemType, independ);
    break;
  case Lex:
    r = new ILexInterface(systemType, independ);
    break;
  case Elim:
    r = new IElimInterface(systemType, independ, va_arg(marker, int));
    break;
  case DegRevLexDeg:
    r = new IDegRevLexDegInterface(systemType, independ, va_arg(marker, unsigned*));
    break;
  case TopDegRevLex:
    r = new ITopDegRevLexInterface(systemType, independ, depend);
    break;
  case TopDegRevLexByte:
    r = new ITopDegRevLexByteInterface(systemType, independ, depend);
    break;
  case TopDegLex:
    r = new ITopDegLexInterface(systemType, independ, depend);
    break;
  case TopLex:
    r = new ITopLexInterface(systemType, independ, depend);
    break;
  case TopElim:
    r = new ITopElimInterface(systemType, independ, depend, va_arg(marker, int));
    break;
  case TopDegRevLexDeg:
    r = new ITopDegRevLexDegInterface(systemType, independ, depend, va_arg(marker, unsigned*));
    break;
  case PotDegRevLex:
    r = new IPotDegRevLexInterface(systemType, independ, depend);
    break;
  case PotDegLex:
    r = new IPotDegLexInterface(systemType, independ, depend);
    break;
  case PotLex:
    r = new IPotLexInterface(systemType, independ, depend);
    break;
  case PotElim:
    r = new IPotElimInterface(systemType, independ, depend, va_arg(marker, int));
    break;
  case PotDegRevLexDeg:
    r = new IPotDegRevLexDegInterface(systemType, independ, depend, va_arg(marker, unsigned*));
    break;
  case PosElimDegRevLex:
    r = new IPosElimDegRevLexInterface(systemType, independ, depend, va_arg(marker, int));
    break;
  case PosElimDegLex:
    r = new IPosElimDegLexInterface(systemType, independ, depend, va_arg(marker, int));
    break;
  case PosElimLex:
    r = new IPosElimLexInterface(systemType, independ, depend, va_arg(marker, int));
    break;
  case PosElimDegRevLexDeg:
    r = new IPosElimDegRevLexDegInterface(systemType, independ, depend, va_arg(marker, int), va_arg(marker, unsigned*));
    break;
  case PosElimElim:
    r = new IPosElimElimInterface(systemType, independ, depend, va_arg(marker, int), va_arg(marker, int));
    break;
  }
  va_end(marker);

  return r;
}

bool IMonom::divisibilityTrue(const IMonom& a) const {
  return divisibility(a) && degree() > a.degree();
}

static inline unsigned max(unsigned a, unsigned b) {
  return (a > b) ? a: b;
}

void IMonom::lcm(const IMonom& a, const IMonom& b) {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(monomInterface() == b.monomInterface());
  setZero();
  for(int var = 0; var < dimIndepend(); var++)
    prolong(var, max(a.deg(var), b.deg(var)));
}

bool IMonom::equality(const IMonom& a, int var, unsigned degree) const {
  IASSERT(mRealization == a.mRealization);
  bool r = (depend() == NULL || depend() == a.depend()) && dependVar() == a.dependVar();
  if (r) {
    int i;
    for(i=0; i < var; i++)
      if (deg(i) != a.deg(i))
        break;
    if (i != var || deg(i) != (a.deg(i) + degree))
      r = false;
    else {
      for(++i; i < dimIndepend(); i++)
        if (deg(i) != a.deg(i))
          break;
      r = i == dimIndepend();
    }
  }
  return r;
}

std::istream& operator>>(std::istream& in, IMonom& a) {
  switch(a.monomInterface()->systemType()->type()) {
  case ISystemType::Polynomial: {
      std::streampos posbeg = in.tellg();
      int var = a.independ()->read(in);
      if (var < 0) {
        in.clear();
        in.setstate(std::ios::failbit);
      }
      else {
        a.setZero();
        int deg;
        do {
          deg = 1;
          std::streampos posbeg = in.tellg();
          if (!in.eof() && (in >> std::ws).peek() == '^') {
            in.get();
            in >> std::ws >> deg;
            if (in.fail() || deg < 0) {
              in.setstate(std::ios::failbit);
              IMESSAGE("expected 'degree >= 0'");
            }
          }
          a.prolong(var, deg);
          posbeg = in.tellg();
          if (in.eof() || in.peek() != '*')
            var = -1;
          else {
            in.get();
            var = a.independ()->read(in);
            if (var < 0) {
              in.clear();
              in.seekg(posbeg);
            }
          }
        } while(var >= 0);
        if (in.eof() &&
            deg >= 0)
          in.clear();
      }
    }
    break;
  case ISystemType::DifferentialEquation: {
      std::streampos posbeg = in.tellg();
      int var = a.depend()->read(in);
      if (var >= 0) {
        a.setZero();
        a.setDependVar(var);
        if (in.eof())
          in.clear();
      }
      else {
        in.seekg(posbeg);
        in >> std::ws;
        if ('d' != in.get() || 'f' != in.get() || '(' != in.get()) {
          in.clear();
          in.setstate(std::ios::failbit);
        }
        else {
          var = a.depend()->read(in);
          if (var < 0) {
            in.clear();
            in.setstate(std::ios::failbit);
          }
          else {
            a.setZero();
            a.setDependVar(var);
            int deg;
            while((in >> std::ws).peek() == ',') {
              deg = 1;
              in.get();
              var = a.independ()->read(in);
              if (var < 0) {
                in.setstate(std::ios::failbit);
                IMESSAGE("expected 'variable'");
                break;
              }
              std::streampos posbeg = in.tellg();
              if ((in >> std::ws).peek() == ')')
                a.prolong(var, 1);
              else if ((in >> std::ws).peek() == ',') {
                in.get();
                in >> std::ws >> deg;
                if (in.fail()) {
                  in.clear();
                  in.seekg(posbeg);
                  a.prolong(var, 1);
                }
                else if (deg >= 0) {
                  a.prolong(var, deg);
                }
                else {
                  in.setstate(std::ios::failbit);
                  IMESSAGE("expected 'degree >= 0'");
                  break;
                }
              }
            }
            if (in.fail() || (in >> std::ws).get() != ')') {
              in.setstate(std::ios::failbit);
              IMESSAGE("expected ')'");
            }
            else if (in.eof())
              in.clear();
          }
        }
      }
    }
    break;
  case ISystemType::FiniteDifferenceScheme: {
      std::streampos posbeg = in.tellg();
      int var = a.depend()->read(in);
      if (var >= 0) {
        a.setZero();
        a.setDependVar(var);
        if (in.eof())
          in.clear();
      }
      else {
        in.seekg(posbeg);
        in >> std::ws;
        if ('T' != in.get() || '(' != in.get()) {
          in.clear();
          in.setstate(std::ios::failbit);
        }
        else {
          var = a.depend()->read(in);
          if (var < 0) {
            in.clear();
            in.setstate(std::ios::failbit);
          }
          else {
            a.setZero();
            a.setDependVar(var);
            int deg;
            while((in >> std::ws).peek() == ',') {
              deg = 1;
              in.get();
              var = a.independ()->read(in);
              if (var < 0) {
                in.setstate(std::ios::failbit);
                IMESSAGE("expected 'variable'");
                break;
              }
              std::streampos posbeg = in.tellg();
              if ((in >> std::ws).peek() == ')')
                a.prolong(var, 1);
              else if ((in >> std::ws).peek() == ',') {
                in.get();
                in >> std::ws >> deg;
                if (in.fail()) {
                  in.clear();
                  in.seekg(posbeg);
                  a.prolong(var, 1);
                }
                else if (deg >= 0) {
                  a.prolong(var, deg);
                }
                else {
                  in.setstate(std::ios::failbit);
                  IMESSAGE("expected 'degree >= 0'");
                  break;
                }
              }
            }
            if (in.fail() || (in >> std::ws).get() != ')') {
              in.setstate(std::ios::failbit);
              IMESSAGE("expected ')'");
            }
            else if (in.eof())
              in.clear();
          }
        }
      }
    }
    break;
  }
  return in;
}

std::ostream& operator<<(std::ostream& out, const IMonom& a) {
  switch(a.monomInterface()->systemType()->type()) {
  case ISystemType::Polynomial: {
      IASSERT(a.monomInterface()->systemType()->module() > 0 || a.depend() == NULL);
      if (a.degree() == 0)
        out << '1';
      else {
        int i = 0;
        IVariables::ConstIterator j(a.independ()->begin());
        while(a.deg(i) == 0) {
          ++i;
          ++j;
        }
        out << *j;
        if (a.deg(i) > 1)
          out << '^' << a[i];
        ++i;
        ++j;
        while(j) {
          if (a.deg(i)) {
            out << '*' << *j;
            if (a.deg(i) > 1)
              out << '^' << a[i];
          }
          ++i;
          ++j;
        }
      }
    }
    break;
  case ISystemType::DifferentialEquation: {
      IASSERT(a.depend() != NULL);
      if (a.degree() == 0)
        if (a.depend() < 0)
          out << '1';
        else
          out << a.depend()->variable(a.dependVar());
      else {
        if (a.depend() < 0)
          out << '(';
        else
          out << "df(" << a.depend()->variable(a.dependVar());
        int i = 0;
        IVariables::ConstIterator j(a.independ()->begin());
        while(j) {
          if (a.deg(i)) {
            out << ',' << *j;
            if (a.deg(i) > 1)
              out << ',' << a.deg(i);
          }
          ++i;
          ++j;
        }
        out << ')';
      }
    }
    break;
  case ISystemType::FiniteDifferenceScheme: {
      IASSERT(a.depend() != NULL);
      if (a.degree() == 0)
        if (a.depend() < 0)
          out << '1';
      else
        out << a.depend()->variable(a.dependVar());
      else {
        if (a.depend() < 0)
          out << '(';
        else
          out << "T(" << a.depend()->variable(a.dependVar());
        int i = 0;
        IVariables::ConstIterator j(a.independ()->begin());
        while(j) {
          if (a.deg(i)) {
            out << ',' << *j;
            if (a.deg(i) > 1)
              out << ',' << a.deg(i);
          }
          ++i;
          ++j;
        }
        out << ')';
      }
    }
    break;
  }
  return out;
}
