/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IMONOM_H
#define IMONOM_H

#include <iostream>

#include "../system/isystemtype.h"
#include "../util/ivariables.h"

class IMonom;
class IMonomInterface {
  friend class IMonom;

public:
  enum Order {Indefinitely         = 0x00000000,
              DegRevLex            = 0x00000001,
              DegLex               = 0x00000002,
              Lex                  = 0x00000004,
              Elim                 = 0x00000008,
              Top                  = 0x00000010,
              Pot                  = 0x00000020,
              PosElim              = 0x00000040,
              SizeByte             = 0x00000080,
              DegRevLexDeg         = 0x00000100,
              DegRevLexByte        = DegRevLex | SizeByte,
              DegLexByte           = DegLex | SizeByte,
              TopDegRevLex         = Top | DegRevLex,
              TopDegRevLexByte     = Top | DegRevLex | SizeByte,
              TopDegLex            = Top | DegLex,
              TopDegLexByte        = Top | DegLex | SizeByte,
              TopLex               = Top | Lex,
              TopElim              = Top | Elim,
              TopDegRevLexDeg      = Top | DegRevLexDeg,
              PotDegRevLex         = Pot | DegRevLex,
              PotDegLex            = Pot | DegLex,
              PotLex               = Pot | Lex,
              PotElim              = Pot | Elim,
              PotDegRevLexDeg      = Pot | DegRevLexDeg,
              PosElimDegRevLex     = PosElim | DegRevLex,
              PosElimDegLex        = PosElim | DegLex,
              PosElimLex           = PosElim | Lex,
              PosElimElim          = PosElim | Elim,
              PosElimDegRevLexDeg  = PosElim | DegRevLexDeg
  };

protected:
  Order              mOrder;
  const ISystemType *mSystemType;
  int                mDimIndepend;
  IVariables*        mIndepend;
  int                mDimDepend;
  IVariables*        mDepend;

  IMonomInterface(const ISystemType *systemType);

public:
  static IMonomInterface* create(Order order,
                                 const ISystemType *systemType,
                                 IVariables* independ,
                                 IVariables* depend = NULL,...);
  virtual ~IMonomInterface() {}

  Order order() const { return mOrder; }
  const ISystemType* systemType() const { return mSystemType; }
  int dimIndepend() const { return mDimIndepend; }
  const IVariables* independ() const { return mIndepend; }
  int dimDepend() const { return mDimDepend; }
  const IVariables* depend() const { return mDepend; }
  int elim() const { return 0; }

  virtual IMonom* create() = 0;
  virtual IMonom* copy(const IMonom& a) = 0;
};

class IMonom {
protected:
  IMonomInterface* const         mRealization;

  IMonom(IMonomInterface* r): mRealization(r) { IASSERT(mRealization); }
  IMonom(const IMonom& a): mRealization(a.mRealization) {}

public:
  typedef IMonomInterface::Order Order;

  virtual ~IMonom() {}
  struct hash : public std::unary_function< IMonom, size_t > {
      size_t operator() (IMonom* const& monom) const {
          return monom->degree();
      }
  };

  struct equal {
      bool operator() (IMonom* const& m1, IMonom* const& m2) {
          if (m1->monomInterface() != m2->monomInterface())
              return false;
  
          int dimIndepend = m1->dimIndepend();
  
          if (m2->dimIndepend() != dimIndepend)
              return false;
  
          for (int i = 0; i < dimIndepend; i++) {
              if (m1->deg(i) != m2->deg(i))
                  return false;
          }
  
          return true;
      }
  };

  IMonomInterface* monomInterface() const { return mRealization; }
  IMonom* copy() const { return mRealization->copy(*this); }
  int order() const { return mRealization->order(); }
  int dimIndepend() const { return mRealization->dimIndepend(); }
  const IVariables* independ() const { return mRealization->independ(); }
  int dimDepend() const { return mRealization->dimDepend(); }
  const IVariables* depend() const { return mRealization->depend(); }

  virtual unsigned deg(int var) const = 0;
  unsigned operator[](int var) const { return deg(var); }
  virtual unsigned degree() const = 0;
  virtual int dependVar() const = 0;
  virtual unsigned weighted_degree() const { return degree(); }

  virtual void set(const IMonom& a) = 0;
  void operator=(const IMonom& a) { set(a); }
  virtual void setZero() = 0;
  virtual void setDependVar(int depend) = 0;
  virtual void swap(IMonom& a) = 0;

  virtual void prolong(int var) = 0;
  virtual void prolong(int var, unsigned deg) = 0;
  virtual void div(int var) = 0;
  virtual void mult(const IMonom& a) = 0;
  virtual void mult(const IMonom& a, const IMonom& b) = 0;

  virtual bool divisibility(const IMonom& a) const = 0;
  bool divisibilityTrue(const IMonom& a) const;
  virtual void divide(const IMonom& a, const IMonom& b) = 0;

  virtual unsigned gcd(const IMonom& a) const = 0;
  virtual unsigned lcm(const IMonom& a) const = 0;
  void lcm(const IMonom& a, const IMonom& b);

  bool equality(const IMonom& a, int var, unsigned degree) const;
  virtual int compare(const IMonom& a) const = 0;
  virtual int compare(const IMonom& a, const IMonom& b) const = 0;
  virtual int compare(const IMonom& a, int var) const {
    IERROR("compare");
  }
  virtual int setQcompare(const IMonom& a) const = 0;

  virtual void read(std::istream &in) = 0;
  virtual void write(std::ostream &out) const = 0;

  friend std::istream& operator>>(std::istream& in, IMonom& a);
  friend std::ostream& operator<<(std::ostream& out, const IMonom& a);
};

#endif // IMONOM_H
