/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../util/iutil.h"
#include "./itopdeglex.h"

ITopDegLexInterface::ITopDegLexInterface(const ISystemType *systemType, IVariables* independ, IVariables* depend):
    IDegLexInterface(systemType, independ) {
  IASSERT(mSystemType->module() > 0 || depend);
  mOrder = IMonomInterface::TopDegLex;
  if (depend) {
    IASSERT(mSystemType->module() == 0);
    mDimDepend = depend->dim();
    mDepend = depend;
  }
  else {
    IASSERT(mSystemType->module() > 0);
    mDimDepend = mSystemType->module();
  }
}

IMonom* ITopDegLexInterface::create() {
  return new ITopDegLex(this);
}

IMonom* ITopDegLexInterface::copy(const IMonom &a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  ITopDegLex* r;
  if (this == a.monomInterface())
    r = new ITopDegLex(a);
  else {
    r = new ITopDegLex(this);
    r->setZero();
    for(int var=0; var < dimDepend(); var++)
      r->prolong(var, a.deg(var));
    r->setDependVar(a.dependVar());
  }
  return r;
}

ITopDegLex::ITopDegLex(ITopDegLexInterface* r):
    IDegLex(r),
    mDependVar(-1) {
}

ITopDegLex::ITopDegLex(const IMonom& a):
    IDegLex(a),
    mDependVar(((const ITopDegLex&)a).mDependVar) {
  IASSERTVALID(*this);
}

ITopDegLex::~ITopDegLex() {
}

int ITopDegLex::dependVar() const {
  return mDependVar;
}

void ITopDegLex::set(const IMonom& a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  IDegLex::set(a);
  mDependVar = a.dependVar();
  IASSERTVALID(*this);
}

void ITopDegLex::setZero() {
  IDegLex::setZero();
  mDependVar = -1;
}

void ITopDegLex::setDependVar(int depend) {
  IASSERT(0 <= depend && depend < dimDepend());
  mDependVar = depend;
}

void ITopDegLex::swap(IMonom& a) {
  IDegLex::swap(a);
  int tmp = mDependVar;
  mDependVar = ((ITopDegLex&)a).mDependVar;
  ((ITopDegLex&)a).mDependVar = tmp;
}

void ITopDegLex::mult(const IMonom& a, const IMonom& b) {
  IDegLex::mult(a, b);
  mDependVar = ((ITopDegLex&)a).mDependVar;
}

bool ITopDegLex::divisibility(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  return mDependVar == ((ITopDegLex&)a).mDependVar && IDegLex::divisibility(a);
}

unsigned ITopDegLex::gcd(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(mDependVar == ((ITopDegLex&)a).mDependVar);
  return IDegLex::gcd(a);
}

unsigned ITopDegLex::lcm(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(mDependVar == ((ITopDegLex&)a).mDependVar);
  return IDegLex::lcm(a);
}

int ITopDegLex::compare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  if (mDependVar < 0) {
    if (((ITopDegLex&)a).mDependVar >= 0)
      return -1;
  }
  else {
    if (((ITopDegLex&)a).mDependVar < 0)
      return 1;
  }

  if (mDependVar < monomInterface()->systemType()->rightPart()) {
    if (((ITopDegLex&)a).mDependVar >= monomInterface()->systemType()->rightPart())
      return 1;
  }
  else {
    if (((ITopDegLex&)a).mDependVar < monomInterface()->systemType()->rightPart())
      return -1;
  }

  int cmp = IDegLex::compare(a);
  if (cmp != 0)
    return cmp;
  else if (mDependVar > ((ITopDegLex&)a).mDependVar)
    return -1;
  else if (mDependVar < ((ITopDegLex&)a).mDependVar)
    return 1;
  else
    return 0;
}

int ITopDegLex::compare(const IMonom& a, const IMonom& b) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(monomInterface() == b.monomInterface());
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(((ITopDegLex&)b).mDependVar == -1);
  if (mDependVar < 0) {
    if (((ITopDegLex&)a).mDependVar >= 0)
      return -1;
  }
  else {
    if (((ITopDegLex&)a).mDependVar < 0)
      return 1;
  }

  if (mDependVar < monomInterface()->systemType()->rightPart()) {
    if (((ITopDegLex&)a).mDependVar >= monomInterface()->systemType()->rightPart())
      return 1;
  }
  else {
    if (((ITopDegLex&)a).mDependVar < monomInterface()->systemType()->rightPart())
      return -1;
  }

  int cmp = IDegLex::compare(a, b);
  if (cmp != 0)
    return cmp;
  else if (mDependVar > ((ITopDegLex&)a).mDependVar)
    return -1;
  else if (mDependVar < ((ITopDegLex&)a).mDependVar)
    return 1;
  else
    return 0;
}

int ITopDegLex::setQcompare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  if (mDependVar < 0) {
    if (((ITopDegLex&)a).mDependVar >= 0)
      return -1;
  }
  else {
    if (((ITopDegLex&)a).mDependVar < 0)
      return 1;
  }

  if (mDependVar < monomInterface()->systemType()->rightPart()) {
    if (((ITopDegLex&)a).mDependVar >= monomInterface()->systemType()->rightPart())
      return 1;
  }
  else {
    if (((ITopDegLex&)a).mDependVar < monomInterface()->systemType()->rightPart())
      return -1;
  }

  int cmp = IDegLex::compare(a);
  if (cmp != 0)
    return cmp;
  else if (mDependVar > ((ITopDegLex&)a).mDependVar)
    return -1;
  else if (mDependVar < ((ITopDegLex&)a).mDependVar)
    return 1;
  else
    return 0;
}

void ITopDegLex::read(std::istream &in) {
  IDegLex::read(in);
  in.read((char*)&mDependVar, sizeof(mDependVar));
  IASSERTVALID(*this);
}

void ITopDegLex::write(std::ostream &out) const {
  IDegLex::write(out);
  out.write((char*)&mDependVar, sizeof(mDependVar));
}

void ITopDegLex::assertValid(const char* fileName, int fileLine) const {
  IDegLex::assertValid(fileName, fileLine);
  IASSERT2(fileName, fileLine, -1 <= mDependVar && mDependVar < dimDepend());
}

IAllocator ITopDegLex::sAllocator(sizeof(ITopDegLex));
