/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../util/iutil.h"
#include "./itopelim.h"

ITopElimInterface::ITopElimInterface(const ISystemType *systemType, IVariables* independ, IVariables* depend, int elim):
    IElimInterface(systemType, independ, elim) {
  IASSERT(mSystemType->module() > 0 || depend);
  mOrder = IMonomInterface::TopElim;
  if (depend) {
    IASSERT(mSystemType->module() == 0);
    mDimDepend = depend->dim();
    mDepend = depend;
  }
  else {
    IASSERT(mSystemType->module() > 0);
    mDimDepend = mSystemType->module();
  }
}

IMonom* ITopElimInterface::create() {
  return new ITopElim(this);
}

IMonom* ITopElimInterface::copy(const IMonom &a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  ITopElim* r;
  if (this == a.monomInterface())
    r = new ITopElim(a);
  else {
    r = new ITopElim(this);
    r->setZero();
    for(int var=0; var < dimDepend(); var++)
      r->prolong(var, a.deg(var));
    r->setDependVar(a.dependVar());
  }
  return r;
}

ITopElim::ITopElim(ITopElimInterface* r):
    IElim(r),
    mDependVar(-1) {
}

ITopElim::ITopElim(const IMonom& a):
    IElim(a),
    mDependVar(((const ITopElim&)a).mDependVar) {
  IASSERTVALID(*this);
}

ITopElim::~ITopElim() {
}

int ITopElim::dependVar() const {
  return mDependVar;
}

void ITopElim::set(const IMonom& a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  IElim::set(a);
  mDependVar = a.dependVar();
  IASSERTVALID(*this);
}

void ITopElim::setZero() {
  IElim::setZero();
  mDependVar = -1;
}

void ITopElim::setDependVar(int depend) {
  IASSERT(0 <= depend && depend < dimDepend());
  mDependVar = depend;
}

void ITopElim::swap(IMonom& a) {
  IElim::swap(a);
  int tmp = mDependVar;
  mDependVar = ((ITopElim&)a).mDependVar;
  ((ITopElim&)a).mDependVar = tmp;
}

void ITopElim::mult(const IMonom& a, const IMonom& b) {
  IElim::mult(a, b);
  mDependVar = ((ITopElim&)a).mDependVar;
}

bool ITopElim::divisibility(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  return mDependVar == ((ITopElim&)a).mDependVar && IElim::divisibility(a);
}

unsigned ITopElim::gcd(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(mDependVar == ((ITopElim&)a).mDependVar);
  return IElim::gcd(a);
}

unsigned ITopElim::lcm(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(mDependVar == ((ITopElim&)a).mDependVar);
  return IElim::lcm(a);
}

int ITopElim::compare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  if (mDependVar < 0) {
    if (((ITopElim&)a).mDependVar >= 0)
      return -1;
  }
  else {
    if (((ITopElim&)a).mDependVar < 0)
      return 1;
  }

  if (mDependVar < monomInterface()->systemType()->rightPart()) {
    if (((ITopElim&)a).mDependVar >= monomInterface()->systemType()->rightPart())
      return 1;
  }
  else {
    if (((ITopElim&)a).mDependVar < monomInterface()->systemType()->rightPart())
      return -1;
  }

  int cmp = IElim::compare(a);
  if (cmp != 0)
    return cmp;
  else if (mDependVar > ((ITopElim&)a).mDependVar)
    return -1;
  else if (mDependVar < ((ITopElim&)a).mDependVar)
    return 1;
  else
    return 0;
}

int ITopElim::compare(const IMonom& a, const IMonom& b) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(monomInterface() == b.monomInterface());
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(((ITopElim&)b).mDependVar == -1);
  if (mDependVar < 0) {
    if (((ITopElim&)a).mDependVar >= 0)
      return -1;
  }
  else {
    if (((ITopElim&)a).mDependVar < 0)
      return 1;
  }

  if (mDependVar < monomInterface()->systemType()->rightPart()) {
    if (((ITopElim&)a).mDependVar >= monomInterface()->systemType()->rightPart())
      return 1;
  }
  else {
    if (((ITopElim&)a).mDependVar < monomInterface()->systemType()->rightPart())
      return -1;
  }

  int cmp = IElim::compare(a, b);
  if (cmp != 0)
    return cmp;
  else if (mDependVar > ((ITopElim&)a).mDependVar)
    return -1;
  else if (mDependVar < ((ITopElim&)a).mDependVar)
    return 1;
  else
    return 0;
}

int ITopElim::setQcompare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  if (mDependVar < 0) {
    if (((ITopElim&)a).mDependVar >= 0)
      return -1;
  }
  else {
    if (((ITopElim&)a).mDependVar < 0)
      return 1;
  }

  if (mDependVar < monomInterface()->systemType()->rightPart()) {
    if (((ITopElim&)a).mDependVar >= monomInterface()->systemType()->rightPart())
      return 1;
  }
  else {
    if (((ITopElim&)a).mDependVar < monomInterface()->systemType()->rightPart())
      return -1;
  }

  int cmp = IElim::compare(a);
  if (cmp != 0)
    return cmp;
  else if (mDependVar > ((ITopElim&)a).mDependVar)
    return -1;
  else if (mDependVar < ((ITopElim&)a).mDependVar)
    return 1;
  else
    return 0;
}

void ITopElim::read(std::istream &in) {
  IElim::read(in);
  in.read((char*)&mDependVar, sizeof(mDependVar));
  IASSERTVALID(*this);
}

void ITopElim::write(std::ostream &out) const {
  IElim::write(out);
  out.write((char*)&mDependVar, sizeof(mDependVar));
}

void ITopElim::assertValid(const char* fileName, int fileLine) const {
  IElim::assertValid(fileName, fileLine);
  IASSERT2(fileName, fileLine, -1 <= mDependVar && mDependVar < dimDepend());
}

IAllocator ITopElim::sAllocator(sizeof(ITopElim));
