/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../util/iutil.h"
#include "./itoplex.h"

ITopLexInterface::ITopLexInterface(const ISystemType *systemType, IVariables* independ, IVariables* depend):
    ILexInterface(systemType, independ) {
  IASSERT(mSystemType->module() > 0 || depend);
  mOrder = IMonomInterface::TopLex;
  if (depend) {
    IASSERT(mSystemType->module() == 0);
    mDimDepend = depend->dim();
    mDepend = depend;
  }
  else {
    IASSERT(mSystemType->module() > 0);
    mDimDepend = mSystemType->module();
  }
}

IMonom* ITopLexInterface::create() {
  return new ITopLex(this);
}

IMonom* ITopLexInterface::copy(const IMonom &a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  ITopLex* r;
  if (this == a.monomInterface())
    r = new ITopLex(a);
  else {
    r = new ITopLex(this);
    r->setZero();
    for(int var=0; var < dimDepend(); var++)
      r->prolong(var, a.deg(var));
    r->setDependVar(a.dependVar());
  }
  return r;
}

ITopLex::ITopLex(ITopLexInterface* r):
    ILex(r),
    mDependVar(-1) {
}

ITopLex::ITopLex(const IMonom& a):
    ILex(a),
    mDependVar(((const ITopLex&)a).mDependVar) {
  IASSERTVALID(*this);
}

ITopLex::~ITopLex() {
}

int ITopLex::dependVar() const {
  return mDependVar;
}

void ITopLex::set(const IMonom& a) {
  IASSERT(dimIndepend() == a.dimIndepend());
  IASSERT(independ() == a.independ());
  IASSERT(dimDepend() == a.dimDepend());
  IASSERT(depend() == a.depend());

  ILex::set(a);
  mDependVar = a.dependVar();
  IASSERTVALID(*this);
}

void ITopLex::setZero() {
  ILex::setZero();
  mDependVar = -1;
}

void ITopLex::setDependVar(int depend) {
  IASSERT(0 <= depend && depend < dimDepend());
  mDependVar = depend;
}

void ITopLex::swap(IMonom& a) {
  ILex::swap(a);
  int tmp = mDependVar;
  mDependVar = ((ITopLex&)a).mDependVar;
  ((ITopLex&)a).mDependVar = tmp;
}

void ITopLex::mult(const IMonom& a, const IMonom& b) {
  ILex::mult(a, b);
  mDependVar = ((ITopLex&)a).mDependVar;
}

bool ITopLex::divisibility(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  return mDependVar == ((ITopLex&)a).mDependVar && ILex::divisibility(a);
}

unsigned ITopLex::gcd(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(mDependVar == ((ITopLex&)a).mDependVar);
  return ILex::gcd(a);
}

unsigned ITopLex::lcm(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(mDependVar == ((ITopLex&)a).mDependVar);
  return ILex::lcm(a);
}

int ITopLex::compare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  if (mDependVar < 0) {
    if (((ITopLex&)a).mDependVar >= 0)
      return -1;
  }
  else {
    if (((ITopLex&)a).mDependVar < 0)
      return 1;
  }

  if (mDependVar < monomInterface()->systemType()->rightPart()) {
    if (((ITopLex&)a).mDependVar >= monomInterface()->systemType()->rightPart())
      return 1;
  }
  else {
    if (((ITopLex&)a).mDependVar < monomInterface()->systemType()->rightPart())
      return -1;
  }

  int cmp = ILex::compare(a);
  if (cmp != 0)
    return cmp;
  else if (mDependVar > ((ITopLex&)a).mDependVar)
    return -1;
  else if (mDependVar < ((ITopLex&)a).mDependVar)
    return 1;
  else
    return 0;
}

int ITopLex::compare(const IMonom& a, const IMonom& b) const {
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(monomInterface() == b.monomInterface());
  IASSERT(monomInterface() == a.monomInterface());
  IASSERT(((ITopLex&)b).mDependVar == -1);
  if (mDependVar < 0) {
    if (((ITopLex&)a).mDependVar >= 0)
      return -1;
  }
  else {
    if (((ITopLex&)a).mDependVar < 0)
      return 1;
  }

  if (mDependVar < monomInterface()->systemType()->rightPart()) {
    if (((ITopLex&)a).mDependVar >= monomInterface()->systemType()->rightPart())
      return 1;
  }
  else {
    if (((ITopLex&)a).mDependVar < monomInterface()->systemType()->rightPart())
      return -1;
  }

  int cmp = ILex::compare(a, b);
  if (cmp != 0)
    return cmp;
  else if (mDependVar > ((ITopLex&)a).mDependVar)
    return -1;
  else if (mDependVar < ((ITopLex&)a).mDependVar)
    return 1;
  else
    return 0;
}

int ITopLex::setQcompare(const IMonom& a) const {
  IASSERT(monomInterface() == a.monomInterface());
  if (mDependVar < 0) {
    if (((ITopLex&)a).mDependVar >= 0)
      return -1;
  }
  else {
    if (((ITopLex&)a).mDependVar < 0)
      return 1;
  }

  if (mDependVar < monomInterface()->systemType()->rightPart()) {
    if (((ITopLex&)a).mDependVar >= monomInterface()->systemType()->rightPart())
      return 1;
  }
  else {
    if (((ITopLex&)a).mDependVar < monomInterface()->systemType()->rightPart())
      return -1;
  }

  int cmp = ILex::compare(a);
  if (cmp != 0)
    return cmp;
  else if (mDependVar > ((ITopLex&)a).mDependVar)
    return -1;
  else if (mDependVar < ((ITopLex&)a).mDependVar)
    return 1;
  else
    return 0;
}

void ITopLex::read(std::istream &in) {
  ILex::read(in);
  in.read((char*)&mDependVar, sizeof(mDependVar));
  IASSERTVALID(*this);
}

void ITopLex::write(std::ostream &out) const {
  ILex::write(out);
  out.write((char*)&mDependVar, sizeof(mDependVar));
}

void ITopLex::assertValid(const char* fileName, int fileLine) const {
  ILex::assertValid(fileName, fileLine);
  IASSERT2(fileName, fileLine, -1 <= mDependVar && mDependVar < dimDepend());
}

IAllocator ITopLex::sAllocator(sizeof(ITopLex));
