/***************************************************************************
 *   Copyright (C) 2006 by Daniel Robertz                                  *
 *   daniel@momo.math.rwth-aachen.de                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./ioneparameterpoly.h"

IOneParameterPoly::IOneParameterPoly(const IOneParameterPoly& a, 
    const char* var):
    ci(a.ci),
    mHead(),
    mVar(var) {

  Iterator      i1(mHead.begin());
  ConstIterator i2(a.mHead.begin());
  while(i2) {
    i1.insert();
    (*i1).mDegree = (*i2).mDegree;
    (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void IOneParameterPoly::operator=(const IOneParameterPoly& a) {
  setZero();

  Iterator      i1(mHead.begin());
  ConstIterator i2(a.mHead.begin());
  while(i2) {
    i1.insert();
    (*i1).mDegree = (*i2).mDegree;
    (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void IOneParameterPoly::mult(const ICoeff& coeff) {
  IASSERT(!coeff.isZero());
  ConstIterator i(mHead.begin());
  while(i) {
    (*i).mCoeff->mult(coeff);
    ++i;
  }
  IASSERTVALID(*this);
}


void IOneParameterPoly::mult(unsigned degree, const ICoeff& coeff) {
  IASSERT(!coeff.isZero());
  ConstIterator i(mHead.begin());
  while(i) {
    (*i).mDegree += degree;
    (*i).mCoeff->mult(coeff);
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterPoly::mult(unsigned degree) {
  ConstIterator i(mHead.begin());
  while(i) {
    (*i).mDegree += degree;
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterPoly::div(const ICoeff& coeff) {
  IASSERT(!coeff.isZero());
  ICoeff* tmp = ci->copy(coeff);
  tmp->inverse();
  ConstIterator i(mHead.begin());
  while(i) {
    (*i).mCoeff->mult(*tmp);
    ++i;
  }

  delete tmp;
  IASSERTVALID(*this);
}

void IOneParameterPoly::div(unsigned degree, const ICoeff& coeff) {
  IASSERT(!coeff.isZero());

  ICoeff* tmp = ci->copy(coeff);
  tmp->inverse();

  ConstIterator i(mHead.begin());
  while(i) {
   IASSERT((*i).mDegree >= degree);
   (*i).mDegree += degree;
   (*i).mCoeff->mult(*tmp);
   ++i;
  }

  delete tmp;
  IASSERTVALID(*this);
}

void IOneParameterPoly::rem(const IOneParameterPoly& a) {
  IASSERT(!a.isZero());
IASSERTVALID(*this);
IASSERTVALID(a);

  ICoeff* k2 = ci->create();

  while(!isZero() && degree() >= a.degree()) {
    Iterator  i1(mHead.begin());
    ConstIterator  i2(a.mHead.begin());

    unsigned deg = degree() - a.degree();
    k2->div(*(*i1).mCoeff, *(*i2).mCoeff);
    k2->minus();

    i1.del();
    ++i2;

    while(i1 && i2) {
      unsigned degree2 = (*i2).mDegree + deg;
      if ((*i1).mDegree > degree2) {
        ++i1;
      }
      else if ((*i1).mDegree < degree2) {
        i1.insert();
        (*i1).mDegree = degree2;
        (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
        (*i1).mCoeff->mult(*k2);
        ++i1;
        ++i2;
      }
      else {
        (*i1).mCoeff->addmul(*(*i2).mCoeff, *k2);
        if ((*i1).mCoeff->isZero())
          i1.del();
        else
          ++i1;
        ++i2;
      }
    }
    while(i2) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree + deg;
      (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
      (*i1).mCoeff->mult(*k2);
      ++i1;
      ++i2;
    }
    IASSERTVALID(*this);
  }
  delete k2;
}

void IOneParameterPoly::rem_ext(const IOneParameterPoly& a, IOneParameterPoly& c1, IOneParameterPoly& c2) {
  IASSERT(!a.isZero());

  ICoeff* k2 = ci->create();

  while(!isZero() && degree() >= a.degree()) {
    Iterator  i1(mHead.begin());
    ConstIterator  i2(a.mHead.begin());
    Iterator  j1(c1.mHead.begin());
    ConstIterator  j2(c2.mHead.begin());

    unsigned deg = degree() - a.degree();
    k2->div(*(*i1).mCoeff, *(*i2).mCoeff);
    k2->minus();

    i1.del();
    ++i2;

    while(i1 && i2) {
      unsigned degree2 = (*i2).mDegree + deg;
      if ((*i1).mDegree > degree2) {
        ++i1;
      }
      else if ((*i1).mDegree < degree2) {
        i1.insert();
        (*i1).mDegree = degree2;
        (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
        (*i1).mCoeff->mult(*k2);
        ++i1;
        ++i2;
      }
      else {
        (*i1).mCoeff->addmul(*(*i2).mCoeff, *k2);
        if ((*i1).mCoeff->isZero())
          i1.del();
        else
          ++i1;
        ++i2;
      }
    }
    while(i2) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree + deg;
      (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
      (*i1).mCoeff->mult(*k2);
      ++i1;
      ++i2;
    }

    while(j1 && j2) {
      unsigned degree2 = (*j2).mDegree + deg;
      if ((*j1).mDegree > degree2) {
        ++j1;
      }
      else if ((*j1).mDegree < degree2) {
        j1.insert();
        (*j1).mDegree = degree2;
        (*j1).mCoeff = ci->copy(*(*j2).mCoeff);
        (*j1).mCoeff->mult(*k2);
        ++j1;
        ++j2;
      }
      else {
        (*j1).mCoeff->addmul(*(*j2).mCoeff, *k2);
        if ((*j1).mCoeff->isZero()) 
          j1.del();
        else
          ++j1;
        ++j2;
      }
    }
    while(j2) {
      j1.insert();
      (*j1).mDegree = (*j2).mDegree + deg;
      (*j1).mCoeff = ci->copy(*(*j2).mCoeff);
      (*j1).mCoeff->mult(*k2);
      ++j1;
      ++j2;
    }
  }
  delete k2;
  IASSERTVALID(*this);
}

void IOneParameterPoly::minus() {
  ConstIterator i(mHead.begin());
  while(i) {
    (*i).mCoeff->minus();
    ++i;
  }
  IASSERTVALID(*this);
}

void IOneParameterPoly::pp() {
  if (isZero())
    return;
  if (coeffInterface()->type() == ICoeffInterface::GmpZZ)
    ;
  else if (coeffInterface()->isField()) {
    ICoeff *c = (*mHead.begin()).mCoeff->copy();
    if (!c->isOne()) {
      c->inverse();
      mult(*c);
    }
    delete c;
    IASSERTVALID(*this);
  }
  else {
    ICoeff *c = (*mHead.begin()).mCoeff->copy();
    ConstIterator i(mHead.begin());
    ++i;
    int k = 0;
    while(i && !c->isOne()) {
      c->gcd(*(*i).mCoeff);
      ++i;
    }
    if (!c->isOne())
      div(*c);

    IASSERTVALID(*this);
  }
}

void IOneParameterPoly::add(const IOneParameterPoly& a) {
  Iterator  i1(mHead.begin());
  ConstIterator  i2(a.mHead.begin());
  while(i1 && i2) {
    if ((*i1).mDegree > (*i2).mDegree) {
      ++i1;
    }
    else if ((*i1).mDegree < (*i2).mDegree) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree;
      (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
      ++i1;
      ++i2;
    }
    else {
      (*i1).mCoeff->add(*(*i2).mCoeff);
      if ((*i1).mCoeff->isZero())
        i1.del();
      else
        ++i1;
      ++i2;
    }
  }
  while(i2) {
    i1.insert();
    (*i1).mDegree = (*i2).mDegree;
    (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void IOneParameterPoly::add(int deg, const ICoeff& c) {
  if (c.isZero())
    return;

  Iterator  i1(mHead.begin());
  while(i1 && ((*i1).mDegree > deg)) 
    ++i1;

  if (i1 && ((*i1).mDegree == deg)) {
    (*i1).mCoeff->add(c);
    if ((*i1).mCoeff->isZero())
      i1.del();
  }
  else {
    i1.insert();
    (*i1).mDegree = deg;
    (*i1).mCoeff = ci->copy(c);
  }

  IASSERTVALID(*this);
}

void IOneParameterPoly::sub(const IOneParameterPoly& a) {
  Iterator  i1(mHead.begin());
  ConstIterator  i2(a.mHead.begin());
  while(i1 && i2) {
    if ((*i1).mDegree > (*i2).mDegree) {
      ++i1;
    }
    else if ((*i1).mDegree < (*i2).mDegree) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree;
      (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
      (*i1).mCoeff->minus();
      ++i1;
      ++i2;
    }
    else {
      (*i1).mCoeff->sub(*(*i2).mCoeff);
      if ((*i1).mCoeff->isZero())
        i1.del();
      else
        ++i1;
      ++i2;
    }
  }
  while(i2) {
    i1.insert();
    (*i1).mDegree = (*i2).mDegree;
    (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
    (*i1).mCoeff->minus();
    ++i1;
    ++i2;
  }
  IASSERTVALID(*this);
}

void IOneParameterPoly::mult(const IOneParameterPoly& a) {
  if (isZero())
    return;

  IOneParameterPoly prod(ci);
  ConstIterator ia(a.mHead.begin());
  Iterator begin(prod.mHead.begin());

  while(ia) {
    ConstIterator i(mHead.begin());
    Iterator ip = begin;
    unsigned degree2 = (*i).mDegree + (*ia).mDegree;

    // look for the starting point
    while(ip && (*ip).mDegree > degree2)
      ++ip;

    begin = ip;

    while(ip && i) {
      degree2 = (*i).mDegree + (*ia).mDegree;
      if ((*ip).mDegree > degree2)
        ++ip;
      else if ((*ip).mDegree < degree2) {
        ip.insert();
        (*ip).mDegree = degree2;
        (*ip).mCoeff = ci->copy(*(*i).mCoeff);
        (*ip).mCoeff->mult(*(*ia).mCoeff);
        IASSERT(!(*ip).mCoeff->isZero());
          ++ip;
        ++i;
      }
      else {
        (*ip).mCoeff->addmul(*(*i).mCoeff, *(*ia).mCoeff);
        if ((*ip).mCoeff->isZero())
          ip.del();
        else
          ++ip;
        ++i;
      }
    }

    while(i) {
      ip.insert();
      (*ip).mDegree = (*i).mDegree + (*ia).mDegree;
      (*ip).mCoeff = ci->copy(*(*i).mCoeff);
      (*ip).mCoeff->mult(*(*ia).mCoeff);
      IASSERT(!(*ip).mCoeff->isZero());
      ++ip;
      ++i;
    }

    ++ia;
  }
  mHead.swap(prod.mHead);
  IASSERTVALID(*this);
}

void IOneParameterPoly::mult(const IOneParameterPoly& a, const IOneParameterPoly& b) {
  setZero();
  if (a.isZero() || b.isZero())
    return;

  ConstIterator ib(b.mHead.begin());
  Iterator begin(mHead.begin());

  while(ib) {
    ConstIterator ia(a.mHead.begin());
    Iterator ip = begin;
    unsigned degree2 = (*ia).mDegree + (*ib).mDegree;

    // look for the starting point
    while(ip && (*ip).mDegree > degree2)
      ++ip;

    begin = ip;

    while(ip && ia) {
      degree2 = (*ia).mDegree + (*ib).mDegree;
      if ((*ip).mDegree > degree2)
        ++ip;
      else if ((*ip).mDegree < degree2) {
        ip.insert();
        (*ip).mDegree = degree2;
        (*ip).mCoeff = ci->copy(*(*ia).mCoeff);
        (*ip).mCoeff->mult(*(*ib).mCoeff);
        IASSERT(!(*ip).mCoeff->isZero());
        ++ip;
        ++ia;
      }
      else {
        (*ip).mCoeff->addmul(*(*ia).mCoeff, *(*ib).mCoeff);
        if ((*ip).mCoeff->isZero())
          ip.del();
        else
          ++ip;
        ++ia;
      }
    }

    while(ia) {
      ip.insert();
      (*ip).mDegree = (*ia).mDegree + (*ib).mDegree;
      (*ip).mCoeff = ci->copy(*(*ia).mCoeff);
      (*ip).mCoeff->mult(*(*ib).mCoeff);
      IASSERT(!(*ip).mCoeff->isZero());
      ++ip;
      ++ia;
    }

    ++ib;
  }
  IASSERTVALID(*this);
}

void IOneParameterPoly::pow(unsigned deg) {
  IASSERT(deg > 0);
  if (deg > 1) {
    IOneParameterPoly p(*this);
    while(--deg > 0) {
      p.mult(*this);
    }
    mHead.swap(p.mHead);
  }
  IASSERTVALID(*this);
}

void IOneParameterPoly::div(const IOneParameterPoly& a) {
  //IASSERT(a.degree() > 0);
  IASSERT(degree() >= a.degree());
  
  if (a.degree() == 0) {
    div(*(*a.begin()).mCoeff);
    return;
  }
  
  ICoeff* k1 = ci->create();
  ICoeff* k2 = ci->create();

  IOneParameterPoly  quo(ci);
  Iterator  i(quo.mHead.begin());

  while(!isZero()) {
    IASSERT(degree() >= a.degree());
    Iterator  i1(mHead.begin());
    ConstIterator  i2(a.mHead.begin());

    unsigned deg = degree() - a.degree();
    k1->set(*(*i2).mCoeff);
    k2->set(*(*i1).mCoeff);

    i.insert();
    (*i).mDegree = deg;
    (*i).mCoeff = ci->copy(*k2);
    ++i;
    IASSERTVALID(quo);

    k2->minus();

    i1.del();
    ++i2;

    while(i1 && i2) {
      unsigned degree2 = (*i2).mDegree + deg;
      if ((*i1).mDegree > degree2) {
        ++i1;
        (*i1).mCoeff->mult(*k1);
      }
      else if ((*i1).mDegree < degree2) {
        i1.insert();
        (*i1).mDegree = degree2;
        (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
        (*i1).mCoeff->mult(*k2);
        ++i1;
        ++i2;
      }
      else {
        (*i1).mCoeff->mult(*k1);
        (*i1).mCoeff->addmul(*(*i2).mCoeff, *k2);
        if ((*i1).mCoeff->isZero())
          i1.del();
        else
          ++i1;
        ++i2;
      }
    }
    while(i2) {
      i1.insert();
      (*i1).mDegree = (*i2).mDegree + deg;
      (*i1).mCoeff = ci->copy(*(*i2).mCoeff);
      (*i1).mCoeff->mult(*k2);
      ++i1;
      ++i2;
    }
    IASSERTVALID(*this);
  }
  IASSERT(isZero());
  mHead.swap(quo.mHead);
  delete k1;
  delete k2;
}

void IOneParameterPoly::gcd(const IOneParameterPoly& a) {
  IASSERT(!isZero());
  IASSERT(!a.isZero());

  if (degree() == 0 || a.degree() == 0) {
    ConstIterator  i(mHead.begin());
    (*i).mCoeff->setOne();
  }
  else {
    IOneParameterPoly tmp(a);

    if (tmp.degree() > degree())
      mHead.swap(tmp.mHead);

    do {
      mHead.swap(tmp.mHead);
      tmp.rem(*this);
    } while (!tmp.isZero() && tmp.degree() > 0);

    if (!tmp.isZero()) {
      mHead.swap(tmp.mHead);
      (*mHead.begin()).mCoeff->setOne();
    }
  }
}

void IOneParameterPoly::assertValid(const char* fileName, int fileLine) const {
  if (!mHead.isEmpty()) {
    ConstIterator i(mHead.begin());
    IASSERT2(fileName, fileLine, !(*i).mCoeff->isZero());
    IASSERT2(fileName, fileLine, (*i).mCoeff->coeffInterface() == ci);
    ConstIterator prev(i);
    ++i;
    while(i) {
      IASSERT2(fileName, fileLine, !(*i).mCoeff->isZero());
      IASSERT2(fileName, fileLine, (*i).mCoeff->coeffInterface() == ci);
      IASSERT2(fileName, fileLine, (*prev).mDegree > (*i).mDegree);
      prev = i;
      ++i;
    }
  }
}

void IOneParameterPoly::input(std::istream& in, const IVariables &parameters) {
  additive(in, parameters);
}

void IOneParameterPoly::additive(std::istream& in, const IVariables &parameters) {
  multiplicative(in, parameters);
  int op = (in >> std::ws).peek();
  IOneParameterPoly tmp(ci);
  while(op == '+' || op == '-') {
    in.get();
    tmp.multiplicative(in, parameters);
    if (op == '+')
      add(tmp);
    else {
      sub(tmp);
    }
    op = in.peek();
  }
}

void IOneParameterPoly::multiplicative(std::istream& in, const IVariables &parameters) {
  unary(in, parameters);
  int op = (in >> std::ws).peek();
  IOneParameterPoly tmp(ci);
  while(op == '*' || op == '/') {
    in.get();
    tmp.unary(in, parameters);
    if (op == '*')
      mult(tmp);
    else
      div(tmp);
    op = in.peek();
  }
}

void IOneParameterPoly::unary(std::istream& in, const IVariables &parameters) {
  int ch = (in >> std::ws).peek();
  if (ch != '+' && ch != '-')
    power(in, parameters);
  else {
    int k = 1;
    do {
      ch = in.get();
      if (ch == '-')
        k *= -1;
      ch = (in >> std::ws).peek();
    } while (ch == '+' || ch == '-');
    power(in, parameters);
    if (k < 0)
      minus();
  }
}

void IOneParameterPoly::power(std::istream& in, const IVariables &parameters) {
  bracket(in, parameters);
  int op = (in >> std::ws).peek();
  if (op == '^') {
    in.get();
    int deg=0;
    in >> std::ws >> deg;
    if (in.fail() || deg <= 0)
      IMESSAGE("expected 'degree > 0'");
    pow(deg);
  }
}

void IOneParameterPoly::bracket(std::istream& in, const IVariables &parameters) {
  int op = (in >> std::ws).peek();
  if (op == '(') {
    in.get();
    additive(in, parameters);
    if (in.get() != ')') {
      IMESSAGE("expected ')'");
    }
  }
  else {
    std::streampos posbeg = in.tellg();
    int var = parameters.read(in);
    if (var == 0) {
      setOne();
      (*begin()).mDegree = 1;
      IASSERTVALID(*this);
    }
    else {
      ICoeff* mp = ci->create();
      in >> *mp;
      if (in.fail()) {
        in.clear();
        in.setstate(std::ios::failbit);
        in.seekg(posbeg);
      }
      else {
        IOneParameterPoly::Iterator i(begin());
        i.clear();
        if (!mp->isZero()) {
          i.insert();
          (*i).mCoeff = mp;
        }
        IASSERTVALID(*this);
      }
    }
  }
}

void IOneParameterPoly::read(std::istream &in) {
  mHead.clear();
  int l;
  in.read((char*)(&l), sizeof(l));
  Iterator i(mHead.begin());
  while(l) {
    i.insert();
    in.read((char*)(&(*i).mDegree), sizeof((*i).mDegree));
    (*i).mCoeff = ci->create();
    (*i).mCoeff->read(in);
    ++i;
    --l;
  }
}

void IOneParameterPoly::write(std::ostream &out) const {
  int l = mHead.length();
  out.write((char*)(&l), sizeof(l));
  ConstIterator i(mHead.begin());
  while(i) {
    out.write((char*)(&(*i).mDegree), sizeof((*i).mDegree));
    (*i).mCoeff->write(out);
    ++i;
  }
}

void IOneParameterPoly::output(std::ostream& out, const std::string& var) const {
  IOneParameterPoly::ConstIterator i(begin());
  if (!i)
    out << '0';
  else {
    if ((*i).mCoeff->isOne()) {
      if ((*i).mDegree == 0) {
        out << '1';
      } else if ((*i).mDegree == 1) { 
        out << var;
      } else {
        out << var << '^' << (*i).mDegree;
      }
    }
    else {
      if ((*i).mCoeff->isComposite())
        out << "(" << *(*i).mCoeff << ")";
      else
        out << *(*i).mCoeff;


      if ((*i).mDegree == 1)
        out << '*' << var;
      else if ((*i).mDegree != 0)
        out << '*' << var << '^' << (*i).mDegree;
    }
    ++i;
    while(i) {
      out << " + ";
      if ((*i).mCoeff->isOne()) {
        if ((*i).mDegree == 0)
          out << '1';
        else if ((*i).mDegree == 1)
          out << var;
        else
          out << var << '^' << (*i).mDegree;
      }
      else {
        if ((*i).mCoeff->isComposite()) 
          out << "(" << *(*i).mCoeff << ")";
        else 
          out << *(*i).mCoeff;

        if ((*i).mDegree == 1)
          out << '*' << var;
        else if ((*i).mDegree != 0)
          out << '*' << var << '^' << (*i).mDegree;
      }
      ++i;
    }
  }
}


std::ostream& operator<<(std::ostream& out, const IOneParameterPoly& a) {
  a.output(out, a.mVar);
  return out;
}

template <>
IAllocator IList<IOneParameterPoly::Term>::Node::sAllocator(sizeof(IList<IOneParameterPoly::Term>::Node));
