/***************************************************************************
 *   Copyright (C) 2006 by Daniel Robertz                                  *
 *   daniel@momo.math.rwth-aachen.de                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IONEPARAMETERPOLY_H
#define IONEPARAMETERPOLY_H

#include <iostream>
#include <string>

#include "../coeff/icoeff.h"
#include "../util/ilist.h"
#include "../util/iutil.h"
#include "../util/ivariables.h"

class IOneParameterPoly {
private:
  struct Term {
    unsigned mDegree;
    ICoeff* mCoeff;

    Term(): mDegree(0), mCoeff(NULL) { }
    ~Term() {
      delete mCoeff;
    }

    void operator=(const Term& a) {
      mDegree = a.mDegree;
      mCoeff->set(*a.mCoeff);
    }
  };

  ICoeffInterface* ci;
  IList<Term>  mHead;
  std::string mVar;

protected:
  void equality(std::istream& in, const IVariables &parameters);
  void additive(std::istream& in, const IVariables &parameters);
  void multiplicative(std::istream& in, const IVariables &parameters);
  void unary(std::istream& in, const IVariables &parameters);
  void power(std::istream& in, const IVariables &parameters);
  void bracket(std::istream& in, const IVariables &parameters);

public:
  typedef IList<Term>::Iterator  Iterator;
  typedef IList<Term>::ConstIterator  ConstIterator;
  IOneParameterPoly(ICoeffInterface* interface, const char* var="t"):
      mHead(),
      ci(interface),
      mVar(var) {
  }
  IOneParameterPoly(const IOneParameterPoly& a, const char* var="t");

  ~IOneParameterPoly() {}

  ICoeffInterface* coeffInterface() {
    return ci;
  }

  void operator=(const IOneParameterPoly& a);

  void setZero() { mHead.clear(); }
  void setOne() {
    mHead.clear();
    Iterator i(mHead.begin());
    i.insert();
    (*i).mCoeff = ci->create();
    (*i).mCoeff->setOne();
  }

  void swap(IOneParameterPoly& a) { mHead.swap(a.mHead); }

  bool isZero() const { return mHead.isEmpty(); }
  bool isOne() const {
    ConstIterator i(mHead.begin());
    return i && (*i).mDegree == 0 && (*i).mCoeff->isOne();
  }
  bool isNegative() const {
    ConstIterator i(mHead.begin());
    return i && (*i).mCoeff->isNegative();
  }
  bool isComposite() const {
    ConstIterator i(mHead.begin());
    return mHead.length() > 1 || isNegative();
  }
  unsigned degree() const {
    IASSERT(!isZero());
    return (*mHead.begin()).mDegree;
  }

  void minus();

  void pp();

  void add(const IOneParameterPoly& a);
  void add(int deg, const ICoeff& c);

  void sub(const IOneParameterPoly& a);

  void mult(const ICoeff& coeff);
  void mult(unsigned degree, const ICoeff& coeff);
  void mult(unsigned degree);
  void mult(const IOneParameterPoly& a);
  void mult(const IOneParameterPoly& a, const IOneParameterPoly& b);

  void addmul(const IOneParameterPoly& a, const IOneParameterPoly& b) {
    IOneParameterPoly tmp(ci);
    tmp.mult(a, b);
    add(tmp);
  }

  void pow(unsigned deg);

  void div(const ICoeff& coeff);
  void div(unsigned degree,  const ICoeff& coeff);
  void div(const IOneParameterPoly& a);
  void div(const IOneParameterPoly& a, const IOneParameterPoly& b) {
    *this = a;
    div(b);
  }

  void gcd(const IOneParameterPoly& a);
  void gcd(const IOneParameterPoly& a, const IOneParameterPoly& b) {
    *this = a;
    gcd(b);
  }

  void rem(const IOneParameterPoly& a);
  void rem_ext(const IOneParameterPoly& a, IOneParameterPoly& c1, IOneParameterPoly& c2);

  ConstIterator begin() const { return mHead.begin(); }
  Iterator begin() { return mHead.begin(); }

  void read(std::istream &in);
  void write(std::ostream &out) const;

  void input(std::istream& in, const IVariables &parameters);
  void output(std::ostream& out, const std::string& var) const;

  friend std::ostream& operator<<(std::ostream& out, const IOneParameterPoly& a);

  friend std::istream& operator>>(std::istream& in, IOneParameterPoly& a) {
    IVariables tmp;
    tmp.add("t");
    a.input(in, tmp);
    return in;
  }

  void assertValid(const char* fileName, int fileLine) const;
};

#endif // IONEPARAMETERPOLY_H
