/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IPOLY_H
#define IPOLY_H

#include <iostream>

#include "../monom/imonom.h"
#include "../coeff/icoeff.h"

class IPoly;
class IPolyInterface {
  friend class IPoly;

public:
  enum Type {Indefinitely                 = 0x00000000,
             ForwardIterator              = 0x00000001,
             BidirectionalIterator        = 0x00000002,
             PolyList                     = 0x00000004 | ForwardIterator,
             PolyArray                    = 0x00000008 | BidirectionalIterator
  };

protected:
  Type          mType;
  const ISystemType *mSystemType;
  IMonomInterface *mMonomInterface;
  ICoeffInterface *mCoeffInterface;

  IPolyInterface();

public:
  static IPolyInterface* create(Type type,
                                ISystemType *systemType,
                                IMonomInterface* monomInterface,
                                ICoeffInterface* coeffInterface, ...);
  virtual ~IPolyInterface() {}

  Type type() const { return mType; }
  const ISystemType* systemType() const { return mSystemType; }
  IMonomInterface* monomInterface() const { return mMonomInterface; }
  ICoeffInterface* coeffInterface() const { return mCoeffInterface; }

  virtual IPoly* create() = 0;
  virtual IPoly* copy(const IPoly& a) = 0;
  virtual IPoly* copy(const IPoly& a, int var) = 0;
  virtual IPoly* copy(const IPoly& a, int var, unsigned deg) = 0;
};


class IDivision;
class IPoly {
  friend class IPolyInterface;

protected:
  IPolyInterface* mRealization;

  IPoly(IPolyInterface* r): mRealization(r) { IASSERT(mRealization); }
  IPoly(const IPoly& a): mRealization(a.mRealization) { IASSERT(mRealization); }

public:
  typedef IPolyInterface::Type Type;

  Type type() const { return mRealization->mType; }
  const ISystemType* systemType() const { return mRealization->mSystemType; }
  IMonomInterface* monomInterface() const { return mRealization->mMonomInterface; }
  ICoeffInterface* coeffInterface() const { return mRealization->mCoeffInterface; }


  class ConstIterator {
    friend class IPoly;

  protected:

    ConstIterator() {}

  public:
    virtual ~ConstIterator() {}

    virtual bool isPointTo() const = 0;
   operator bool() const { return isPointTo(); }

    virtual const ICoeff& coeff() const = 0;

    virtual const IMonom& monom() const = 0;

    virtual void next() = 0;
    void operator++() { next(); }
    virtual void prev() = 0;
    void operator--() { prev(); }

    virtual bool equal(const ConstIterator& a) const = 0;
    bool operator==(const ConstIterator& a) const { return equal(a); }
    bool operator!=(const ConstIterator& a) const { return !equal(a); }
  };

  class Iterator {
    friend class IPoly;

  protected:
    Iterator() {}

  public:
    virtual ~Iterator() {}
    virtual bool isPointTo() const = 0;
   operator bool() const { return isPointTo(); }

    virtual const ICoeff& coeff() const = 0;


    virtual const IMonom& monom() const = 0;

    virtual void next() = 0;
    void operator++() { next(); }
    virtual void prev() = 0;
    void operator--() { prev(); }

    virtual bool equal(const Iterator& a) const = 0;
    bool operator==(const Iterator& a) const { return equal(a); }
    bool operator!=(const Iterator& a) const { return !equal(a); }
  };

  virtual ~IPoly() {}

  IPolyInterface* polyInterface() const { return mRealization; }

  virtual void setZero() = 0;
  virtual void setOne() = 0;
  virtual void setDependVar(int var) = 0;
  virtual void setMonom(IMonom *m) = 0;
  virtual void setCoeff(ICoeff *c) = 0;
  virtual void setPoly(const IPoly& a, int dependVar) = 0;

  virtual void set(const IPoly& a) = 0;
  void operator=(const IPoly& a) { set(a); }
  virtual void swap(IPoly& a) = 0;

  virtual bool isZero() const = 0;
  virtual int length() const = 0;
  virtual int degree() const = 0;
  virtual int norm() const = 0;
  virtual const IMonom& lm() const = 0;
  virtual const ICoeff& lc() const = 0;
  virtual void minus() = 0;
  virtual void add(const IPoly &a) = 0;
  virtual void addTerm(ICoeff *c, IMonom *m) = 0;
  virtual void sub(const IPoly &a) = 0;
  virtual void mult(int var) = 0;
  virtual void mult(int var, unsigned deg) = 0;

  virtual void mult(const IMonom& m) = 0;
  virtual void mult(const ICoeff& c) = 0;
  virtual void mult(const IPoly &a) = 0;
  virtual void pow(unsigned deg) = 0;
  virtual void div(const ICoeff& c) = 0;

  virtual void content(ICoeff& c) const = 0;
  virtual void pp() = 0;
  virtual bool isPp() const = 0;

  virtual void reduction(ICoeff& multiplier, const IPoly& a) = 0;
  virtual void redTail(ICoeff &multiplier, Iterator& i, const IPoly& a) = 0;
  virtual void redTailCanonical(Iterator& i, const IPoly& a) {};
  virtual ConstIterator* begin() const = 0;
  virtual ConstIterator* beginReverse() const = 0;
  virtual Iterator* begin() = 0;
  virtual Iterator* beginReverse() = 0;

  virtual void read(std::istream &in) = 0;
  virtual void write(std::ostream &out) const = 0;

  friend std::ostream& operator<<(std::ostream& out, const IPoly& a);
};

#endif // IPOLY_H

