/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "ifactorizationzassenhaustest.h"

#include <cppunit/extensions/HelperMacros.h>

CPPUNIT_TEST_SUITE_REGISTRATION(IFactorizationZassenhausTest);

IAlgExtModularShortAux* IFactorizationZassenhausTest::getPoly(std::string str, ModularShort prime) {
  IAlgExtModularShortAux* result = new IAlgExtModularShortAux(prime);
  std::istringstream ist(str);
  result->input(ist, *vars);

  return result;
}

void IFactorizationZassenhausTest::setUp() {
  vars = new IVariables();
  vars->add("t");

  alg = new IFactorizationZassenhaus();
}

void IFactorizationZassenhausTest::tearDown() {
  delete alg;
}

void IFactorizationZassenhausTest::testRepeatedSquaring() {
  // taken from [von zur Gathen, Example 14.5]
  ModularShort p = 3;

  IAlgExtModularShortAux* f = getPoly("t^8 + t^7 - t^6 + t^5 - t^3 - t^2 - t ", p);

  IAlgExtModularShortAux* h = getPoly("t ", p);
  IAlgExtModularShortAux* expected = getPoly("t^3 ", p);

  MP_INT drei;
  mpz_init_set_ui(&drei, 3ul);
  alg->repeatedSquaring(*h, drei, *f);
  h->sub(*expected);
  CPPUNIT_ASSERT(h->isZero());
  delete h;
  delete expected;

  h = getPoly("t^3 ", p);
  expected = getPoly("-t^7 + t^6 + t^5 + t^4 - t ", p);

  alg->repeatedSquaring(*h, drei, *f);
  h->sub(*expected);
  CPPUNIT_ASSERT(h->isZero());
  delete h;
  delete expected;

  mpz_clear(&drei);
  delete f;
}

void IFactorizationZassenhausTest::testEqualDegreeSplitting() {
  // [von zur Gathen, Example 14.5 continued]
  ModularShort p = 3;

  IAlgExtModularShortAux* f = getPoly("t^4 + t^3 + t - 1 ", p);
  unsigned d = 2;
  IAlgExtModularShortAux* g = new IAlgExtModularShortAux(p);

  IAlgExtModularShortAux* a = getPoly("t + 1 ", p);
  IAlgExtModularShortAux* expected = getPoly("0 ", p);

  alg->equalDegreeSplitting(*f, *a, d, *g);
  g->sub(*expected);
  CPPUNIT_ASSERT(g->isZero());
  delete a;
  delete expected;

  a = getPoly("t ", p);
  expected = getPoly("t^2 + 1 ", p);

  alg->equalDegreeSplitting(*f, *a, d, *g);
  g->sub(*expected);
  CPPUNIT_ASSERT(g->isZero());
  delete a;
  delete expected;
  delete g;
  delete f;
}

void IFactorizationZassenhausTest::testEqualDegreeSplittingMod2() {
  ModularShort p = 2;

  IAlgExtModularShortAux* f = getPoly("t^6 + t^5 + t^4 + t^3 + t^2 + t + 1 ", p);
  unsigned d = 3;
  IAlgExtModularShortAux* g = new IAlgExtModularShortAux(p);

  IAlgExtModularShortAux* a = getPoly("t + 1 ", p);
  IAlgExtModularShortAux* expected = getPoly("t^3 + t + 1 ", p);

  alg->equalDegreeSplitting(*f, *a, d, *g);
  g->sub(*expected);
  CPPUNIT_ASSERT(g->isZero());
  delete a;
  delete expected;
  delete g;
  delete f;
}

void IFactorizationZassenhausTest::testRandomPolynomial() {
  // [von zur Gathen, Example 14.5 continued]
  ModularShort p = 3;

  IAlgExtModularShortAux r(p);
  unsigned d = 5;

  for (int i = 0; i < 5; i++) {
    alg->randomPolynomial(r, d);

    CPPUNIT_ASSERT(!r.isZero());
    CPPUNIT_ASSERT(r.degree() > 0);
    CPPUNIT_ASSERT(r.degree() < d);
  }
}

void IFactorizationZassenhausTest::testEqualDegreeFactorization() {
  ModularShort p = 17;

  IAlgExtModularShortAux* f = getPoly("t^8 + 3*t^7 ", p);
  IAlgExtModularShortAux* g = getPoly("8*t^6 + 2*t^5 ", p);
  f->add(*g);
  delete g;
  g = getPoly("6*t^4 + 12*t^3 ", p);
  f->add(*g);
  delete g;
  g = getPoly("3*t^2 + 14*t ", p);
  f->add(*g);
  delete g;
  g = getPoly("5 ", p);
  f->add(*g);
  delete g;

  CPPUNIT_ASSERT(f->mHead.length() == 9);

  IAlgExtModularShortAux* f1 = getPoly("t^2 + 5*t + 5 ", p);
  IAlgExtModularShortAux* f2 = getPoly("t^2 + 7*t + 3 ", p);
  IAlgExtModularShortAux* f3 = getPoly("t^2 + 8*t + 2 ", p);
  IAlgExtModularShortAux* f4 = getPoly("t^2 + 3 ", p);

  IList<IAlgExtModularShortAux*> expected;
  IList<IAlgExtModularShortAux*>::Iterator it = expected.begin();

  it.insert();
  (*it) = f1;
  it.insert();
  (*it) = f2;
  it.insert();
  (*it) = f3;
  it.insert();
  (*it) = f4;

  IList<IAlgExtModularShortAux*> factors;

  alg->equalDegreeFactorization(*f, 2, factors);

  CPPUNIT_ASSERT(factors.length() == 4);

  it = factors.begin();

  while (it) {
    IList<IAlgExtModularShortAux*>::Iterator ex = expected.begin();
    while (ex) {
      IAlgExtModularShortAux tmp(*(*ex));
      tmp.sub(*(*it));
      if (tmp.isZero()) {
        ex.del();
        break;
      }
      ++ex;
    }
    
    ++it;
  }

  CPPUNIT_ASSERT(expected.isEmpty());
}

void IFactorizationZassenhausTest::testFactorization() {
  ModularShort p = 17;

  IAlgExtModularShortAux* f = getPoly("3+4*t^2+11*t^9+16*t^13+12*t^11+9*t^10+8*t^12+14*t^14+t^18+4*t^17+8*t^16+9*t^15+6*t^6+3*t^5+5*t^8+15*t^7+4*t+12*t^3+10*t^4 ", p);
  CPPUNIT_ASSERT(f->mHead.length() == 19);

  IAlgExtModularShortAux* f1 = getPoly("t^2+5*t+5 ", p);
  IAlgExtModularShortAux* f2 = getPoly("t^2+7*t+3 ", p);
  IAlgExtModularShortAux* f3 = getPoly("t^3+5*t+2 ", p);
  IAlgExtModularShortAux* f4 = getPoly("t+12 ", p);

  IList<IAlgExtModularShortAux*> expected;
  IList<IAlgExtModularShortAux*>::Iterator it = expected.begin();

  it.insert();
  (*it) = f1;
  it.insert();
  (*it) = f2;
  it.insert();
  (*it) = f3;
  it.insert();
  (*it) = f4;

  IList<unsigned> expected_mult;
  IList<unsigned>::Iterator mit = expected_mult.begin();
  mit.insert();
  (*mit) = 1;
  mit.insert();
  (*mit) = 1;
  mit.insert();
  (*mit) = 3;
  mit.insert();
  (*mit) = 5;


  IList<IAlgExtModularShortAux*> factors;
  IList<unsigned> multiplicities;

  alg->factor(*f, factors, multiplicities);

  CPPUNIT_ASSERT(factors.length() == 4);
  CPPUNIT_ASSERT(multiplicities.length() == 4);

  it = factors.begin();
  IList<unsigned>::Iterator it_mult = multiplicities.begin();

  while (it) {
    IList<IAlgExtModularShortAux*>::Iterator ex = expected.begin();
    IList<unsigned>::Iterator ex_mult = expected_mult.begin();
    while (ex) {
      IAlgExtModularShortAux tmp(*(*ex));
      tmp.sub(*(*it));
      if (tmp.isZero()) {
        ex.del();
        CPPUNIT_ASSERT((*it_mult) == (*ex_mult));
        ex_mult.del();
        break;
      }
      ++ex;
      ++ex_mult;
    }
    
    ++it;
    ++it_mult;
  }

  CPPUNIT_ASSERT(expected.isEmpty());
  CPPUNIT_ASSERT(expected_mult.isEmpty());
}

void IFactorizationZassenhausTest::testFactorizationMod2() {
  ModularShort p = 2;

  IAlgExtModularShortAux* f = getPoly("t^15 - t", p);
  CPPUNIT_ASSERT(f->mHead.length() == 2);

  IAlgExtModularShortAux* f1 = getPoly("t^3 + t^2 + 1 ", p);
  IAlgExtModularShortAux* f2 = getPoly("t + 1 ", p);
  IAlgExtModularShortAux* f3 = getPoly("t^3 + t + 1 ", p);
  IAlgExtModularShortAux* f4 = getPoly("t ", p);

  IList<IAlgExtModularShortAux*> expected;
  IList<IAlgExtModularShortAux*>::Iterator it = expected.begin();

  it.insert();
  (*it) = f1;
  it.insert();
  (*it) = f2;
  it.insert();
  (*it) = f3;
  it.insert();
  (*it) = f4;

  IList<unsigned> expected_mult;
  IList<unsigned>::Iterator mit = expected_mult.begin();
  mit.insert();
  (*mit) = 2;
  mit.insert();
  (*mit) = 2;
  mit.insert();
  (*mit) = 2;
  mit.insert();
  (*mit) = 1;


  IList<IAlgExtModularShortAux*> factors;
  IList<unsigned> multiplicities;

  alg->factor(*f, factors, multiplicities);

  CPPUNIT_ASSERT(factors.length() == 4);
  CPPUNIT_ASSERT(multiplicities.length() == 4);

  it = factors.begin();
  IList<unsigned>::Iterator it_mult = multiplicities.begin();

  while (it) {
    IList<IAlgExtModularShortAux*>::Iterator ex = expected.begin();
    IList<unsigned>::Iterator ex_mult = expected_mult.begin();
    while (ex) {
      IAlgExtModularShortAux tmp(*(*ex));
      tmp.sub(*(*it));
      if (tmp.isZero()) {
        ex.del();
        CPPUNIT_ASSERT((*it_mult) == (*ex_mult));
        ex_mult.del();
        break;
      }
      ++ex;
      ++ex_mult;
    }
    
    ++it;
    ++it_mult;
  }

  CPPUNIT_ASSERT(expected.isEmpty());
  CPPUNIT_ASSERT(expected_mult.isEmpty());
}
