/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "igcdalgorithmtest.h"

#include <vector>

#include <cppunit/extensions/HelperMacros.h>

CPPUNIT_TEST_SUITE_REGISTRATION(IGcdAlgorithmTest);

void IGcdAlgorithmTest::setUp() {
  IGinvTest::setUp();
}

void IGcdAlgorithmTest::tearDown() {
  IGinvTest::tearDown();
}

void IGcdAlgorithmTest::testNewton() {
  ModularShort p = 97;
  
  INParameterModularShortAux* u1 = getModularPoly("-9*x - 21", p, 1);
  INParameterModularShortAux* u2 = getModularPoly("-3*x + 20", p, 1);
  INParameterModularShortAux* u3 = getModularPoly("5*x - 36", p, 1);

  INParameterModularShortAux* expected = getModularPoly("x*y^2 + 5*x*y + 41*y - 9*x - 21", p, 2);
  INParameterModularShortAux* g = getModularPoly("0", p, 2);

  std::vector<INParameterModularShortAux*> gcds;
  gcds.push_back(u1);
  gcds.push_back(u2);
  gcds.push_back(u3);

  std::vector<ModularShort> alpha;
  alpha.push_back(0);
  alpha.push_back(1);
  alpha.push_back(2);

  IGcdZippel algorithm;

  algorithm.newton(*g, gcds, alpha);

  g->sub(*expected);

  CPPUNIT_ASSERT(g->isZero());
}

void IGcdAlgorithmTest::testNewton2() {
  ModularShort p = 97;
  
  INParameterModularShortAux* u1 = getModularPoly("3*x", p, 1);
  INParameterModularShortAux* u2 = getModularPoly("7*x", p, 1);
  INParameterModularShortAux* u3 = getModularPoly("13*x", p, 1);
  INParameterModularShortAux* u4 = getModularPoly("21*x", p, 1);
  INParameterModularShortAux* u5 = getModularPoly("31*x", p, 1);

  INParameterModularShortAux* expected = getModularPoly("x*(y^2 + y + 1)", p, 2);
  INParameterModularShortAux* g = getModularPoly("0", p, 2);

  std::vector<INParameterModularShortAux*> gcds;
  gcds.push_back(u1);
  gcds.push_back(u2);
  gcds.push_back(u3);
  gcds.push_back(u4);
  gcds.push_back(u5);

  std::vector<ModularShort> alpha;
  alpha.push_back(1);
  alpha.push_back(2);
  alpha.push_back(3);
  alpha.push_back(4);
  alpha.push_back(5);

  IGcdZippel algorithm;

  algorithm.newton(*g, gcds, alpha);

  g->sub(*expected);

  CPPUNIT_ASSERT(g->isZero());
}

void IGcdAlgorithmTest::testDelegationIterator() {
  INParameterModularShortAux* pol = getModularPoly("x^3 + 3*x^2 + 2", 5, 1);
  IGcdAlgorithm::DelegationIterator it(pol->begin());
  CPPUNIT_ASSERT(it.isValid());
  CPPUNIT_ASSERT(it.deg() == 3);
  CPPUNIT_ASSERT(it.coeff() == 1);
  it.next();
  CPPUNIT_ASSERT(it.isValid());
  CPPUNIT_ASSERT(it.deg() == 2);
  CPPUNIT_ASSERT(it.coeff() == 3);
  it.next();
  CPPUNIT_ASSERT(it.isValid());
  CPPUNIT_ASSERT(it.deg() == 0);
  CPPUNIT_ASSERT(it.coeff() == 2);

  it.next();
  CPPUNIT_ASSERT(!it.isValid());
}

void IGcdAlgorithmTest::testShiftIterator() {
  INParameterModularShortAux* pol = getModularPoly("x^3 + 3*x^2 + 2", 5, 1);
  IGcdAlgorithm::ShiftIterator it1(new IGcdAlgorithm::DelegationIterator(pol->begin()), 3);
  CPPUNIT_ASSERT(it1.isValid());
  CPPUNIT_ASSERT(it1.deg() == 0);
  CPPUNIT_ASSERT(it1.coeff() == 1);
  it1.next();
  CPPUNIT_ASSERT(!it1.isValid());

  IGcdAlgorithm::ShiftIterator it2(new IGcdAlgorithm::DelegationIterator(pol->begin()), -1);
  CPPUNIT_ASSERT(it2.isValid());
  CPPUNIT_ASSERT(it2.deg() == 4);
  CPPUNIT_ASSERT(it2.coeff() == 1);
  it2.next();
  it2.next();
  CPPUNIT_ASSERT(it2.isValid());
  CPPUNIT_ASSERT(it2.deg() == 1);
  CPPUNIT_ASSERT(it2.coeff() == 2);
  it2.next();
  CPPUNIT_ASSERT(!it2.isValid());
}

void IGcdAlgorithmTest::testSumIterator() {
  INParameterModularShortAux* pol1 = getModularPoly("x^3 + 3*x^2 + 2", 5, 1);
  INParameterModularShortAux* pol2 = getModularPoly("x^4 + 3*x^3 + 3", 5, 1);

  IGcdAlgorithm::SumIterator it(new IGcdAlgorithm::DelegationIterator(pol1->begin()), new IGcdAlgorithm::DelegationIterator(pol2->begin()), pol1->primeShort());
  CPPUNIT_ASSERT(it.isValid());
  CPPUNIT_ASSERT(it.deg() == 4);
  CPPUNIT_ASSERT(it.coeff() == 1);
  it.next();
  CPPUNIT_ASSERT(it.isValid());
  CPPUNIT_ASSERT(it.deg() == 3);
  CPPUNIT_ASSERT(it.coeff() == 4);
  it.next();
  CPPUNIT_ASSERT(it.isValid());
  CPPUNIT_ASSERT(it.deg() == 2);
  CPPUNIT_ASSERT(it.coeff() == 3);
  it.next();
  CPPUNIT_ASSERT(it.isValid());
  CPPUNIT_ASSERT(it.deg() == 0);
  CPPUNIT_ASSERT(it.coeff() == 0);
  it.next();
  CPPUNIT_ASSERT(!it.isValid());
}

void IGcdAlgorithmTest::testAddmul() {
  INParameterModularShortAux* pol1 = getModularPoly("x^4 + x^3 + 3", 5, 1);
  INParameterModularShortAux* pol2 = getModularPoly("x^3 + 3*x^2 + 2", 5, 1);
  INParameterModularShortAux* result = getModularPoly("2*x^4 + 4*x^3 + 2*x + 3", 5, 1);

  IGcdAlgorithm::addmul(*pol1, *pol2, 1);
  pol1->sub(*result);

  CPPUNIT_ASSERT(pol1->isZero());
}

void IGcdAlgorithmTest::testSubmul() {
  INParameterModularShortAux* pol1 = getModularPoly("x^4 + x^3 + 3", 5, 1);
  INParameterModularShortAux* pol2 = getModularPoly("x^3 + 3*x^2 + 2", 5, 1);
  INParameterModularShortAux* result = getModularPoly("3*x^3 + 3*x + 3", 5, 1);

  IGcdAlgorithm::submul(*pol1, *pol2, 1);
  pol1->sub(*result);

  CPPUNIT_ASSERT(pol1->isZero());
}

void IGcdAlgorithmTest::testMult() {
  INParameterModularShortAux* pol1 = getModularPoly("2", 5, 1);
  INParameterModularShortAux* pol2 = getModularPoly("3", 5, 1);
  INParameterModularShortAux* result = getModularPoly("1", 5, 1);

  IGcdAlgorithm::mult(*pol1, *pol2);
  pol1->sub(*result);
  CPPUNIT_ASSERT(pol1->isZero());


  pol1 = getModularPoly("x + 1", 5, 1);
  pol2 = getModularPoly("x + 1", 5, 1);
  result = getModularPoly("x^2 + 2*x + 1", 5, 1);

  IGcdAlgorithm::mult(*pol1, *pol2);
  pol1->sub(*result);
  CPPUNIT_ASSERT(pol1->isZero());


  pol1 = getModularPoly("x^3 + 3*x^2 + 2", 5, 1);
  pol2 = getModularPoly("x^4 + 3*x^3 + 3", 5, 1);
  result = getModularPoly("x^7+x^6+4*x^3+4*x^5+4*x^2+2*x^4+1", 5, 1);

  IGcdAlgorithm::mult(*pol1, *pol2);
  pol1->sub(*result);
  CPPUNIT_ASSERT(pol1->isZero());
}

void IGcdAlgorithmTest::testMultTruncated() {
  INParameterModularShortAux* pol1 = getModularPoly("x^2 + 1", 5, 1);
  INParameterModularShortAux* pol2 = getModularPoly("x + 1", 5, 1);
  INParameterModularShortAux* result = getModularPoly("x^4 + x^3 + x^2 + x", 5, 1);

  IGcdAlgorithm::multTruncated(*pol1, *pol2, 2);
  pol1->sub(*result);
  CPPUNIT_ASSERT(pol1->isZero());
}
