/***************************************************************************
 *   Copyright (C) 2007 by Sebastian Jambor                                *
 *   sebastian@momo.math.rwth-aachen.de                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "izippelgausstest.h"

#include <cppunit/extensions/HelperMacros.h>

CPPUNIT_TEST_SUITE_REGISTRATION(IZippelGaussTest);

void IZippelGaussTest::getList(std::list<ModularShort>& l, ModularShort c[], int n) {
  for (int i = 0; i < n; i++) {
    l.push_back(c[i]);
  }
}

void IZippelGaussTest::setUp() {
  IGinvTest::setUp();

  INParameterGmpZ* p = (INParameterGmpZ*) getCoeff(ip, "x^3*y^2 + 9*x^3 + 4*y + 3");
  form = new INParameterModularShortAux(*p->mPoly, 13);
  delete p;

  p = (INParameterGmpZ*) getCoeff(ip, "x^3 + 12");
  eval1 = new INParameterModularShortAux(*p->mPoly, 17);
  alpha1 = new ModularShort[1];
  alpha1[0] = 1;
  delete p;

  p = (INParameterGmpZ*) getCoeff(ip, "x^3 + 8");
  eval2 = new INParameterModularShortAux(*p->mPoly, 17);
  alpha2 = new ModularShort[1];
  alpha2[0] = 2;
  delete p;

  p = (INParameterGmpZ*) getCoeff(ip, "x^3");
  eval3 = new INParameterModularShortAux(*p->mPoly, 17);
  alpha3 = new ModularShort[1];
  alpha3[0] = 3;
  delete p;

  p = (INParameterGmpZ*) getCoeff(ip, "7*x^3*y^2 + 11*x^3 + 11*y + 1");
  g = new INParameterModularShortAux(*p->mPoly, 17);
  delete p;
}

void IZippelGaussTest::tearDown() {
  delete eval3;
  delete eval2;
  delete eval1;
  delete form;

  IGinvTest::tearDown();
}

void IZippelGaussTest::testSingleSystem() {
  ModularShort p = 17;
  ISingleSystem system(p);

  system.addScalingComponent();
  system.addScalingComponent();
  system.addScalingComponent();

  CPPUNIT_ASSERT(system.degreeOfFreedom() == 2);

  ModularShort c1[] = { 1, 0, 1 };
  std::list<ModularShort>* l1 = new std::list<ModularShort>;
  getList(*l1, c1, 3);

  system.addVector(l1);
  CPPUNIT_ASSERT(system.degreeOfFreedom() == 1);

  ModularShort c2[] = { 0, 1, 1 };
  std::list<ModularShort>* l2 = new std::list<ModularShort>;
  getList(*l2, c2, 3);

  system.addVector(l2);
  CPPUNIT_ASSERT(system.degreeOfFreedom() == 0);

  system.addScalingComponent();
  system.addScalingComponent();
  CPPUNIT_ASSERT(system.degreeOfFreedom() == 2);

  ModularShort c3[] = { 1, 1, 0, 0, 0 };
  std::list<ModularShort>* l3 = new std::list<ModularShort>;
  getList(*l3, c3, 5);
  system.addVector(l3);
  CPPUNIT_ASSERT(system.degreeOfFreedom() == 1);

  ModularShort c4[] = { 1, 2, 0, 0, 3 };
  std::list<ModularShort>* l4 = new std::list<ModularShort>;
  getList(*l4, c4, 5);

  system.addVector(l4);
  CPPUNIT_ASSERT(system.degreeOfFreedom() == 0);

  ModularShort expectedSolution[] = {14, 3, 1, 1};

  ModularShort *solution = system.getSolution();

  for (int i = 0; i < 4; i++) {
    CPPUNIT_ASSERT(expectedSolution[i] == solution[i]);
  }
}

void IZippelGaussTest::testSingleSystemBug() {
  ModularShort p = 3119;
  ISingleSystem system(p);

  system.addScalingComponent();
  system.addScalingComponent();
  system.addScalingComponent();

  ModularShort c1[] = { 2852, 2569, 2132 };
  std::list<ModularShort>* l1 = new std::list<ModularShort>;
  getList(*l1, c1, 3);
  system.addVector(l1);

  system.addScalingComponent();

  ModularShort c2[] = { 2614, 0, 665, 697 };
  std::list<ModularShort>* l2 = new std::list<ModularShort>;
  getList(*l2, c2, 4);
  system.addVector(l2);

  system.addScalingComponent();

  ModularShort c3[] = { 1, 2846, 2115, 147, 1129 };
  std::list<ModularShort>* l3 = new std::list<ModularShort>;
  getList(*l3, c3, 5);
  system.addVector(l3);

  ModularShort c4[] = { 1913, 0, 0, 1518, 3011 };
  std::list<ModularShort>* l4 = new std::list<ModularShort>;
  getList(*l4, c4, 5);
  system.addVector(l4);

  CPPUNIT_ASSERT(system.degreeOfFreedom() == 0);

  ModularShort expectedSolution[] = {1304, 1870, 777, 2392};

  ModularShort *solution = system.getSolution();

  for (int i = 0; i < 4; i++) {
    CPPUNIT_ASSERT(expectedSolution[i] == solution[i]);
  }
}

void IZippelGaussTest::testSubsystem() {
  ModularShort p = 17;
  ISubsystem system(p);

  ModularShort c1[] = {1, 1};
  ModularShort s1[] = {1};
  ModularShort c2[] = {4, 1};
  ModularShort s2[] = {1, 0};
  ModularShort c3[] = {9, 1};
  ModularShort s3[] = {1, 0, 0};

  std::list<ModularShort>* l1 = new std::list<ModularShort>;
  getList(*l1, c1, 2);
  std::list<ModularShort>* sl1 = new std::list<ModularShort>;
  getList(*sl1, s1, 1);
  std::list<ModularShort>* l2 = new std::list<ModularShort>;
  getList(*l2, c2, 2);
  std::list<ModularShort>* sl2 = new std::list<ModularShort>;
  getList(*sl2, s2, 2);
  std::list<ModularShort>* l3 = new std::list<ModularShort>;
  getList(*l3, c3, 2);
  std::list<ModularShort>* sl3 = new std::list<ModularShort>;
  getList(*sl3, s3, 3);

  bool lindep;
  
  system.addScalingComponent();
  lindep = system.addVector(l1, sl1);
  CPPUNIT_ASSERT(lindep == false);

  system.addScalingComponent();
  lindep = system.addVector(l2, sl2);
  CPPUNIT_ASSERT(lindep == false);

  system.addScalingComponent();
  lindep = system.addVector(l3, sl3);
  CPPUNIT_ASSERT(lindep == true);

  ModularShort expectedScalings[] = {1, 3, 13};

  std::list<ModularShort>::iterator it = sl3->begin();
  for (int i = 0; i < 3; i++) {
    CPPUNIT_ASSERT(expectedScalings[i] == (*it));
    ++it;
  }

  ModularShort scalingSolution[] = {6, 5};
  ModularShort expectedSolution[] = {7, 11};
  ModularShort *solution = system.getSolution(scalingSolution);

  for (int i = 0; i < 2; i++) {
    CPPUNIT_ASSERT(expectedSolution[i] == solution[i]);
  }
}

void IZippelGaussTest::testDegreeOfFreedom() {
  IMultipleScalingSystem system(*form, 17);
  CPPUNIT_ASSERT(system.degreeOfFreedom() == 4);

  system.add(*eval1, alpha1);
  CPPUNIT_ASSERT(system.degreeOfFreedom() == 2);

  system.add(*eval2, alpha2);
  CPPUNIT_ASSERT(system.degreeOfFreedom() == 1);

  system.add(*eval3, alpha3);
  CPPUNIT_ASSERT(system.degreeOfFreedom() == 0);

  INParameterModularShortAux solution(eval1->interface(), 17);
  system.getSolution(solution);
  solution.sub(*g);

  CPPUNIT_ASSERT(solution.isZero());
}
