/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "./ihilbertpoly.h"

IHilbertPoly::~IHilbertPoly() {
  if (mCoeff) {
    for(int k = 0; k < mDim; k++)
      mpq_clear(mCoeff + k);

    delete[] mCoeff;
  }
}

void IHilbertPoly::setDim(int dim) {
  mDim = dim + 1;
  mCoeff = new MP_RAT[mDim];
  for(int k = 0; k < mDim; k++)
    mpq_init(mCoeff + k);
}

void IHilbertPoly::add(const IHilbertPoly& a) {
  for(int k = 0; k < mDim; k++)
    mpq_add(mCoeff + k, mCoeff + k, a.mCoeff + k);
}

void IHilbertPoly::sub(const IHilbertPoly& a) {
  for(int k = 0; k < mDim; k++)
    mpq_sub(mCoeff + k, mCoeff + k, a.mCoeff + k);
}

void IHilbertPoly::mult(int a) {
  MP_RAT b;
  mpq_init(&b);
  mpq_set_si(&b, (long)a, 1ul);
  for(int k = 0; k < mDim; k++) {
    mpq_mul(mCoeff + k, mCoeff + k, &b);
  }
  mpq_clear(&b);
}

void IHilbertPoly::binomial(long int a, unsigned long b) {
  for(int k = 0; k < mDim; k++)
    mpq_set_ui(mCoeff + k, 0ul, 1ul);

  if (b == 0)
    mpq_set_ui(mCoeff + 0, 1ul, 1ul);
  else {
    MP_RAT *tmp = new MP_RAT[b];
    for(unsigned long k = 0; k < b; k++)
      mpq_init(tmp + k);

    mpq_set_si(mCoeff + 0, a, 1ul);
    mpq_set_ui(mCoeff + 1, 1ul, 1ul);

    MP_RAT q;
    mpq_init(&q);

    for(unsigned long p = 2; p <= b; p++) {
      mpq_set_ui(&q, 1ul, p);
      for(unsigned long i = 0; i < p; i++) {
        mpq_set_ui(&q, 1ul, p);
        mpq_set(tmp + i, mCoeff + i);
        mpq_mul(tmp + i, tmp + i, &q);
      }

      mpq_set_si(&q, a-(long int)(p-1), p);

      for(unsigned long i = 0; i < p; i++)
        mpq_mul(mCoeff + i, mCoeff + i, &q);

      for(unsigned long i = 1; i <= p; i++)
        mpq_add(mCoeff + i, mCoeff + i, tmp + i-1);
    }
    for(int k = 0; k < mDim; k++)
      mpq_canonicalize(mCoeff + k);

    for(unsigned long k = 0; k < b; k++)
      mpq_clear(tmp + k);

    delete[] tmp;
  }
}

std::ostream& operator<<(std::ostream& out, const IHilbertPoly& a) {
  if (a.mDim < 0 || a.mCoeff == NULL) {
    out << "0";
    return out;
  }

  int k = a.mDim-1;
  while(mpq_sgn(a.mCoeff + k) == 0)
    --k;
  if (k == 0) {
    IASSERT(mpz_cmp_ui(mpq_denref(a.mCoeff), 1ul) == 0);
    out << mpq_numref(a.mCoeff);
  }
  else {
    if (mpz_cmp_ui(mpq_denref(a.mCoeff + k), 1ul) == 0) {
      if (mpz_cmp_ui(mpq_numref(a.mCoeff + k), 1ul) != 0)
        out << mpq_numref(a.mCoeff + k);
    }
    else
      out << (a.mCoeff + k);
    if (k > 1)
      out << "s^" << k;
    else
      out << 's';

    for(--k; k > 1 ; k--) {
      if (mpq_sgn(a.mCoeff + k) < 0) {
        mpq_neg(a.mCoeff + k, a.mCoeff + k);
        out << " - ";
        if (mpz_cmp_ui(mpq_denref(a.mCoeff + k), 1ul) == 0) {
          if (mpz_cmp_ui(mpq_numref(a.mCoeff + k), 1ul) != 0)
            out << mpq_numref(a.mCoeff + k);
          out << "s^" << k;
        }
        else
          out << (a.mCoeff + k) << "s^" << k;
        mpq_neg(a.mCoeff + k, a.mCoeff + k);
      }
      else if (mpq_sgn(a.mCoeff + k) > 0) {
        out << " + ";
        if (mpz_cmp_ui(mpq_denref(a.mCoeff + k), 1ul) == 0) {
          if (mpz_cmp_ui(mpq_numref(a.mCoeff + k), 1ul) != 0)
            out << mpq_numref(a.mCoeff + k);
          out << "s^" << k;
        }
        else
          out << (a.mCoeff + k) << "s^" << k;
      }
    }
    if (k) {
      IASSERT(k == 1);
      if (mpq_sgn(a.mCoeff + 1) < 0) {
        mpq_neg(a.mCoeff + 1, a.mCoeff + 1);
        out << " - ";
        if (mpz_cmp_ui(mpq_denref(a.mCoeff + 1), 1ul) == 0) {
          if (mpz_cmp_ui(mpq_numref(a.mCoeff + 1), 1ul) != 0)
            out << mpq_numref(a.mCoeff + 1);
          out << 's';
        }
        else
          out << (a.mCoeff + 1) << "s";
        mpq_neg(a.mCoeff + 1, a.mCoeff + 1);
      }
      else {
        out << " + ";
        if (mpz_cmp_ui(mpq_denref(a.mCoeff + 1), 1ul) == 0) {
          if (mpz_cmp_ui(mpq_numref(a.mCoeff + 1), 1ul) != 0)
            out << mpq_numref(a.mCoeff + 1);
          out << 's';
        }
        else
          out << (a.mCoeff + 1) << "s";
      }
      --k;
    }
    IASSERT(k == 0);
    IASSERT(mpz_cmp_ui(mpq_denref(a.mCoeff), 1ul) == 0);
    if (mpq_sgn(a.mCoeff) > 0)
      out << " + " << mpq_numref(a.mCoeff);
    else {
      mpq_neg(a.mCoeff, a.mCoeff);
      out << " - " << mpq_numref(a.mCoeff);
      mpq_neg(a.mCoeff, a.mCoeff);
    }
  }
  return out;
}

std::ostream& IHilbertPoly::latex(std::ostream& out) const {
  if (mDim < 0 || mCoeff == NULL) {
    out << "0";
    return out;
  }

  int k = mDim-1;
  while(mpq_sgn(mCoeff + k) == 0)
    --k;
  if (k == 0) {
    IASSERT(mpz_cmp_ui(mpq_denref(mCoeff), 1ul) == 0);
    out << mpq_numref(mCoeff);
  }
  else {
    if (mpz_cmp_ui(mpq_denref(mCoeff + k), 1ul) == 0) {
      if (mpz_cmp_ui(mpq_numref(mCoeff + k), 1ul) != 0)
        out << mpq_numref(mCoeff + k);
    }
    else
      out << "\\frac{" << mpq_numref(mCoeff + k) << ", "
           << mpq_denref(mCoeff + k) << '}';
    if (k > 1)
      out << "s^" << k;
    else
      out << 's';

    for(--k; k > 1 ; k--) {
      if (mpq_sgn(mCoeff + k) < 0) {
        mpq_neg(mCoeff + k, mCoeff + k);
        out << " - ";
        if (mpz_cmp_ui(mpq_denref(mCoeff + k), 1ul) == 0) {
          if (mpz_cmp_ui(mpq_numref(mCoeff + k), 1ul) != 0)
            out << mpq_numref(mCoeff + k);
          out << "s^" << k;
        }
        else
          out << "\\frac{" << mpq_numref(mCoeff + k) << ", "
               << mpq_denref(mCoeff + k) << "}s^" << k;
        mpq_neg(mCoeff + k, mCoeff + k);
      }
      else if (mpq_sgn(mCoeff + k) > 0) {
        out << " + ";
        if (mpz_cmp_ui(mpq_denref(mCoeff + k), 1ul) == 0) {
          if (mpz_cmp_ui(mpq_numref(mCoeff + k), 1ul) != 0)
            out << mpq_numref(mCoeff + k);
          out << "s^" << k;
        }
        else
          out << "\\frac{" << mpq_numref(mCoeff + k) << ", "
               << mpq_denref(mCoeff + k) << "}s^" << k;
      }
    }
    if (k) {
      IASSERT(k == 1);
      if (mpq_sgn(mCoeff + 1) < 0) {
        mpq_neg(mCoeff + 1, mCoeff + 1);
        out << " - ";
        if (mpz_cmp_ui(mpq_denref(mCoeff + 1), 1ul) == 0) {
          if (mpz_cmp_ui(mpq_numref(mCoeff + 1), 1ul) != 0)
            out << mpq_numref(mCoeff + 1);
          out << 's';
        }
        else
          out << "\\frac{" << mpq_numref(mCoeff + 1) << ", "
               << mpq_denref(mCoeff + 1) << "}s";
        mpq_neg(mCoeff + 1, mCoeff + 1);
      }
      else {
        out << " + ";
        if (mpz_cmp_ui(mpq_denref(mCoeff + 1), 1ul) == 0) {
          if (mpz_cmp_ui(mpq_numref(mCoeff + 1), 1ul) != 0)
            out << mpq_numref(mCoeff + 1);
          out << 's';
        }
        else
          out << "\\frac{" << mpq_numref(mCoeff + 1) << ", "
               << mpq_denref(mCoeff + 1) << "}s";
      }
      --k;
    }
    IASSERT(k == 0);
    IASSERT(mpz_cmp_ui(mpq_denref(mCoeff), 1ul) == 0);
    if (mpq_sgn(mCoeff) > 0)
      out << " + " << mpq_numref(mCoeff);
    else {
      mpq_neg(mCoeff, mCoeff);
      out << " - " << mpq_numref(mCoeff);
      mpq_neg(mCoeff, mCoeff);
    }
  }
  return out;
}

void IHilbertPoly::read(std::istream& in) {
  in.read((char*)(&mDim), sizeof(mDim));
  if (mCoeff) delete[] mCoeff;
  if (mDim > 0) {
    mCoeff = new MP_RAT[mDim];
    for (int i = 0; i < mDim; i++) {
      mpq_init(&mCoeff[i]);
      in >> (&mCoeff[i]);
    }
  }
  else
    mCoeff = NULL;
}

void IHilbertPoly::write(std::ostream& out) const {
  out.write((char*)(&mDim), sizeof(mDim));
  for (int i = 0; i < mDim; i++)
    out << (&mCoeff[i]);
}
