/***************************************************************************
 *   Copyright (C) 2004 by Blinkov Yu.A.                                   *
 *   BlinkovUA@info.sgu.ru                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ILIST_H
#define ILIST_H

#include "./iallocator.h"
#include <iostream>

template <typename T>
class IList {
private:
  struct Node {
    T     mData;
    Node* mNext;

    static IAllocator sAllocator;

    void* operator new(size_t) { return sAllocator.allocate(); }
    void operator delete(void *ptr) { sAllocator.deallocate(ptr); }

    explicit Node(Node *n): mData(), mNext(n) {}
    Node(const T& a, Node *n): mData(a), mNext(n) {}
    ~Node() {}
  };

  typedef Node *Link;

public:
  class Iterator;
  class ConstIterator {
    friend class Iterator;
    Link mLink;

  public:
    ConstIterator(): mLink(NULL) {}
    ConstIterator(Link n): mLink(n) {}
    ConstIterator(const ConstIterator& a): mLink(a.mLink) {}
    ~ConstIterator() {}

    void operator=(const ConstIterator& a) { mLink = a.mLink; }

    operator bool() const { return (mLink != NULL); }
    bool isNext() const { IASSERT(mLink); return (mLink->mNext != NULL); }
    T& operator*() const { IASSERT(mLink); return mLink->mData; }
    void operator++() {  IASSERT(mLink); mLink = mLink->mNext; }
  };

  class Iterator {
    Link *mLink;

  public:
    Iterator(): mLink(NULL) {}
    Iterator(Link &n): mLink(&n) {}
    Iterator(const Iterator& a): mLink(a.mLink) {}
    ~Iterator() {}

    static size_t sizeNode() { return sizeof(Node); }

    operator ConstIterator() { IASSERT(mLink); return *mLink; }
    void operator=(const Iterator& a) { mLink = a.mLink; }

    operator bool() const { IASSERT(mLink); return (*mLink != NULL); }
    bool isNext() const { IASSERT(mLink); return ((*mLink)->mNext != NULL); }
    T& operator*() const { IASSERT(mLink); IASSERT(*mLink); return (*mLink)->mData; }
    void operator++() { IASSERT(mLink); IASSERT(*mLink); mLink = &(*mLink)->mNext; }

    void insert() {
      IASSERT(mLink);
      *mLink = new Node(*mLink);
    }
    void insert(const T &Data) {
      IASSERT(mLink);
      *mLink = new Node(Data, *mLink);
    }
    void insert(ConstIterator j) {
      IASSERT(mLink);
      j.mLink->mNext = *mLink;
      *mLink = j.mLink;
    }
    ConstIterator get() {
      IASSERT(mLink);
      ConstIterator j(*mLink);
      *mLink = (*mLink)->mNext;
      j.mLink->mNext = NULL;
      return j;
    }
    void del() {
      IASSERT(mLink);
      Link tmp = *mLink;
      *mLink = tmp->mNext;
      delete tmp;
    }
    void clear() {
      IASSERT(mLink);
      while (*mLink)
        del();
    }

    friend bool operator==(const Iterator &a, const Iterator &b) {
      return a.mLink == b.mLink;
    }
    friend bool operator!=(const Iterator &a, const Iterator &b) {
      return a.mLink != b.mLink;
    }
  };

private:
  Link mHead;

public:
  IList(): mHead(NULL) {}
  IList(const IList& a): mHead(NULL) {
    Iterator i(mHead);
    ConstIterator j(a.begin());
    while (j) {
      i.insert(*j);
      ++i;
      ++j;
    }
  }
  ~IList() { begin().clear(); }

  void operator=(const IList& a) {
    clear();
    Iterator i(mHead);
    ConstIterator j(a.begin());
    while (j) {
      i.insert(*j);
      ++i;
      ++j;
    }
  }
  void clear() {
    begin().clear();
    mHead = NULL;
  }

  bool isEmpty() const { return mHead == NULL; }
  int length() const {
    ConstIterator i(mHead);
    int len=0;
    while(i) {
      ++len;
      ++i;
    }
    return len;
  }

  void swap(IList &s) {
    Link tmp = mHead;
    mHead = s.mHead;
    s.mHead = tmp;
  }
  void merge(IList &s) {
    Iterator i1 = begin(),
             i2 = s.begin();
    while(i1 && i2) {
      int cmp = compare(*i1, *i2);
      if (cmp > 0) {
        ++i1;
      }
      else if (cmp < 0) {
        i1.insert(i2.get());
        ++i1;
      }
      else {
        ++i1;
        i1.insert(i2.get());
        ++i1;
      }
    }
    while(i2) {
      i1.insert(i2.get());
      ++i1;
    }
  }
  void sort() {
    if (mHead && mHead->mNext) {
      IList lst;
      Iterator i1 = begin(),
               i2 = lst.begin();
      while(i1) {
        ++i1;
        if (i1)
          i2.insert(i1.get());
      }
      sort();
      lst.sort();
      merge(lst);
    }
  }

  ConstIterator begin() const { return mHead; }
  Iterator begin() { return mHead; }

  void read(std::istream& in) {
    clear();
    int l;
    in.read((char*)(&l), sizeof(l));
    Iterator i(mHead);
    while(l) {
      i.insert();
      (*i).read(in);
      ++i;
      --l;
    }
  }

  void write(std::ostream& out) const {
    int l = length();
    out.write((char*)(&l), sizeof(l));
    ConstIterator i(mHead);
    while(i) {
      (*i).write(out);
      ++i;
    }
  }
};

#endif // ILIST_H
